/*
 * the num2str cache.
 *
 * all this should eventually go into net/sunrpc/gssd_clnt.c instead,
 * with calls to gssd_X_to_Y cached transparently.
 *
 * used code from uid_cache, kernel/fork.c and 
 * from the dcache, fs/dcache.c
 */

#include <linux/string.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/sunrpc/svc.h>
#include <linux/sunrpc/auth_gss.h>
#include <linux/nfs4_prot.h>
#include <linux/n2s_cache.h>
#include <asm/uaccess.h>

#define USE_SLAB                0

#if USE_SLAB
#  define n2s_malloc()          kmem_cache_alloc (n2s_cachep, SLAB_KERNEL)
#  define n2s_free(p)           kmem_cache_free (n2s_cachep, p)
#  define n2s_load()            do {                                                                    \
													  if (!n2s_cachep)                                                      \
													  n2s_cachep = kmem_cache_create ("n2s cache", sizeof(n2s_cache_ent), \
																	  0, SLAB_HWCACHE_ALIGN, 0, 0);       \
													  if (!n2s_cachep)                                                      \
													  panic ("cannot create n2s cache\n");                                \
												      } while (0)
#  define n2s_unload()          kmem_cache_shrink (n2s_cachep)
#else
#  define n2s_malloc()          kmalloc (sizeof(n2s_cache_ent), GFP_KERNEL)
#  define n2s_free(p)           kfree (p)
#  define n2s_load()            do { } while (0)
#  define n2s_unload()          do { } while (0)
#endif

/* set the size of the hash tables with NUM2STR_HASHBITS */
#define NUM2STR_HASHBITS    	5 
#define NUM2STR_HASHSIZE       (1UL << NUM2STR_HASHBITS)
#define NUM2STR_HASHMASK       (NUM2STR_HASHSIZE-1)

/*
 *  This is our basic cache data structure.  It holds a (type, num, name) triple,
 *  and is involved in two hash tables, one for lookup by name, one for lookup by
 *  num.  Therefore, it contains two sets of pointers, one for traversing each
 *  hash chain it is part of.
 */

typedef struct n2s_cache_ent {
	unsigned int           type;
	unsigned int           num;
	utf8string             name;
	struct n2s_cache_ent   *next_num;      // next entry in num-based hash chain
	struct n2s_cache_ent   **pprev_num;    // prev entry in num-based hash chain
	struct n2s_cache_ent   *next_name;     // next entry in name-based hash chain
	struct n2s_cache_ent   **pprev_name;   // prev entry in name-based hash chain
} n2s_cache_ent;

n2s_cache_ent            *n2s_name_hash[NUM2STR_HASHSIZE];
n2s_cache_ent            *n2s_num_hash[NUM2STR_HASHSIZE];
spinlock_t               n2s_lock = SPIN_LOCK_UNLOCKED;  // a single lock controls access to both caches

#if USE_SLAB
extern kmem_cache_t *n2s_cachep;
#endif

/*
 *  Hash functions:
 *    numhashfn()          - hashes an int
 *    n2s_full_name_hash() - hashes a utf8string
 */

#define numhashfn(num)     (((num >> 8) ^ num) & (NUM2STR_HASHSIZE - 1))
#define init_name_hash()   0

/* partial hash update function. Assume roughly 4 bits per character */
static __inline__ unsigned long n2s_partial_name_hash(unsigned long c, unsigned long prevhash)
{
	prevhash = (prevhash << 4) | (prevhash >> (8*sizeof(unsigned long)-4));
	return prevhash ^ c;
}

/* Finally: cut down the number of bits to a int value (and try to avoid losing bits) */
static __inline__ unsigned long n2s_end_name_hash(unsigned long hash)
{
	if (sizeof(hash) > sizeof(unsigned int))
		hash += hash >> 4*sizeof(hash);
	return (unsigned long) hash;
}
static __inline__ unsigned int n2s_name_hash_index(unsigned long hash)
{
	hash = hash ^ (hash >> NUM2STR_HASHBITS) ^ (hash >> NUM2STR_HASHBITS*2);
	return (hash & NUM2STR_HASHMASK);
}

/* Compute the hash for a name string. */
static __inline__ unsigned int n2s_full_name_hash(utf8string *str)
{
	const unsigned char *name = str->utf8string_val;
	unsigned int len = str->utf8string_len;
	unsigned long hash = init_name_hash();
	unsigned long hash2;
  
	while (len--)
		hash = n2s_partial_name_hash(*name++, hash);
	hash2 = n2s_end_name_hash(hash);
	return n2s_name_hash_index(hash2);
}

/*
 *  Cache lookup functions
 *  These must be called with the n2s lock held!
 */

static inline n2s_cache_ent *
n2s_cache_lookup_num (unsigned int type, unsigned int num, unsigned int numhashent)
{
	n2s_cache_ent *p;

	for (p = n2s_num_hash[numhashent]; p; p = p->next_num)
		if ((p->type == type) && (p->num == num))
			return p;
	return 0;
}

static inline n2s_cache_ent *
n2s_cache_lookup_name (unsigned int type, utf8string *name, unsigned int namehashent)
{
	n2s_cache_ent *p;

	for (p = n2s_name_hash[namehashent]; p; p = p->next_name)
		if ((p->type == type) && !utf8cmp(name, &p->name))
			return p;
	return 0;
}

/*
 *    n2s_cache_add() - inserts an entry into the n2s cache
 *    n2scache_init() - called when module is loaded to initialize cache
 *    n2s_cache_shutdown() - called when module is unloaded to flush cache
 *  These must be called without the n2s lock held!
 */

static inline void
n2s_cache_free (n2s_cache_ent *p)
{
	kfree (p->name.utf8string_val);     // utf8copy() allocates this on the free store
						       n2s_free (p);
}

static inline int
n2s_cache_add (unsigned int type, unsigned int num, utf8string *name)
{
	unsigned int        namehashent = n2s_full_name_hash(name);
	unsigned int        numhashent = numhashfn(num);
	n2s_cache_ent       *new;
	int                 err = 0;

	if ( !(new = n2s_malloc())) {
		err = -1;
		goto out;
	}
	new->type = type;
	new->num = num;
	utf8copy (&new->name, name);
	if (!new->name.utf8string_val) {
		n2s_free (new);
		err = -1;
		goto out;
	}

	spin_lock (&n2s_lock);

	if (n2s_cache_lookup_num(type,num,numhashent)) {   // this can happen as the result of a race condition
								      n2s_cache_free (new);
	spin_unlock (&n2s_lock);
	goto out;
	}

	nfsv4_printk (n2s, "N2S: adding type=%d, num=%d, name=\"%.*s\"\n",
		      type, num, (int) name->utf8string_len, name->utf8string_val);

	// link new struct into num hashtable
		   if ( (new->next_num = n2s_num_hash[numhashent]))
			   new->next_num->pprev_num = &new->next_num;
	n2s_num_hash[numhashent] = new;
	new->pprev_num = &n2s_num_hash[numhashent];

	// link new struct into name hashtable
		   if ( (new->next_name = n2s_name_hash[namehashent]))
			   new->next_name->pprev_name = &new->next_name;
	n2s_name_hash[namehashent] = new;
	new->pprev_name = &n2s_name_hash[namehashent];

	spin_unlock (&n2s_lock);

 out:
	return err;
}

void
n2scache_init (void)
{
	int i;

	spin_lock (&n2s_lock);
	nfsv4_printk (n2s, "N2S init\n");
	n2s_load();
	for (i = 0; i < NUM2STR_HASHSIZE; i++) {
		n2s_num_hash[i] = 0;
		n2s_name_hash[i] = 0;
	}
	spin_unlock (&n2s_lock);

	/* XXX - dugsong testing */
	{
		u32 uid;
		struct xdr_netobj name;
    
		/* Test GSS-authenticated RPC call to GSSD. */
		name.data = "dugsong";
		name.len = 7;
    
		if (gssd_client_test(&name, &uid) < 0)
			printk("N2S: gssd_client_test failed!\n");
		else
			printk("N2S: gssd_client_test: %.*s -> %d\n",
			       (int) name.len, name.data, uid);
    
		/* Test regular GSSD calls. */
		uid = 2;
    
		if (gssd_uid_to_name(&uid, &name) < 0)
			printk("N2S: gssd_uid_to_name failed!\n");
		else {
			printk("N2S: gssd_uid_to_name: %d -> %.*s\n",
			       uid, (int) name.len, name.data);
			rpc_free(name.data);
		}
		name.data = "operator";
		name.len = 8;
    
		if (gssd_name_to_uid(&name, &uid) < 0)
			printk("N2S: gssd_name_to_uid failed!\n");
		else
			printk("N2S: gssd_name_to_uid: %.*s -> %d\n",
			       (int) name.len, name.data, uid);
	}
}

void
n2s_cache_shutdown (void)
{
	int                 i;
	n2s_cache_ent       *p = 0;
	n2s_cache_ent       *next;

	spin_lock (&n2s_lock);
	nfsv4_printk (n2s, "N2S shutdown\n");
	for (i = 0; i < NUM2STR_HASHSIZE; i++)
		for (p = n2s_num_hash[i]; p; p = next) {
			next = p->next_num;
			n2s_cache_free (p);
		}
	for (i = 0; i < NUM2STR_HASHSIZE; i++) {
		n2s_num_hash[i] = 0;
		n2s_name_hash[i] = 0;
	}
	n2s_unload ();
	spin_unlock (&n2s_lock);
}


/*
 *  get_num(), get_name() - perform uid/gid <-> name translation
 */

int
get_num (u_int type, utf8string *str, u_int *num)
{
	n2s_cache_ent *p;
	unsigned int namehashent = n2s_full_name_hash (str);

	spin_lock (&n2s_lock);
	p = n2s_cache_lookup_name (type, str, namehashent);
	spin_unlock (&n2s_lock);

	nfsv4_printk (n2s, "N2S get_num() : type=%d, name=\"%.*s\" -> ",
		      type, (int) str->utf8string_len, str->utf8string_val);
	if (p)
		nfsv4_printk (n2s, "%d\n", p->num);
	else
		nfsv4_printk (n2s, "cache miss\n");

	if (p)
		*num = p->num;
	else if (type == NUM2STR_UID) {
		if (gssd_name_to_uid((struct xdr_netobj *)str, num) < 0)
			return -1;
	}
	else if (type == NUM2STR_GID) {
		if (gssd_name_to_gid((struct xdr_netobj *)str, num) < 0)
			return -1;
	}
	else return -1;
  
	n2s_cache_add (type, *num, str);
	return 0;
}

/* caller will need to free results returned in utf8string */

int
get_name (u_int type, u_int num, utf8string *str)
{        
	n2s_cache_ent *p;
	unsigned int numhashent = numhashfn(num);

	spin_lock (&n2s_lock);
	p = n2s_cache_lookup_num (type, num, numhashent);
	spin_unlock(&n2s_lock);

	nfsv4_printk (n2s, "N2S get_name() : type=%d, num=%d -> ", type, num);
	if (p)
		nfsv4_printk (n2s, "\"%.*s\"\n", (int) p->name.utf8string_len, p->name.utf8string_val);
	else
		nfsv4_printk (n2s, "cache miss\n");

	if (p)
		utf8copy (str, &p->name);
	else if (type == NUM2STR_UID) {
		if (gssd_uid_to_name(&num, (struct xdr_netobj *)str) < 0)
			return -1;
	}
	else if (type == NUM2STR_GID) {
		if (gssd_gid_to_name(&num, (struct xdr_netobj *)str) < 0)
			return -1;
	}
	else return -1;

	n2s_cache_add (type, num, str);
	return 0;
}
