/*
 * the num2str cache.
 *
 * all this should eventually go into net/sunrpc/gssd_clnt.c instead,
 * with calls to gssd_X_to_Y cached transparently.
 *
 * used code from uid_cache, kernel/fork.c and 
 * from the dcache, fs/dcache.c
 *
 * cache consists of the following:
 *	SLAB cache for struct num2str's
 *	hash table on num
 *	hash table on name
 * 
 * the two hash's are joined at the hip - they 
 * share a spin lock and are created and destroyed 
 * together. if one exists, so does the other.
 *
 */

#include <linux/string.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/sunrpc/svc.h>
#include <linux/sunrpc/auth_gss.h>
#include <linux/nfs4_prot.h>
#include <linux/n2s_cache.h>
#include <asm/uaccess.h>

/* set the size of the hash tables with NUM2STR_HASHBITS */
#define NUM2STR_HASHBITS    	5 
#define NUM2STR_HASHSIZE     (1UL << NUM2STR_HASHBITS)
#define NUM2STR_HASHMASK     (NUM2STR_HASHSIZE-1)


#define numhashfn(num)  (((num >> 8) ^ num) & (NUM2STR_HASHSIZE - 1))
#define init_name_hash()                0

/* in unloaded memory, so when nfsd is reloaded it knows that the cache is 
 * available.
 */

extern kmem_cache_t *n2s_cachep;

/* Globals */
struct num_struct *numhash[NUM2STR_HASHSIZE];
struct name_struct *namehash[NUM2STR_HASHSIZE];
spinlock_t hash_lock = SPIN_LOCK_UNLOCKED;

/* the name->hash routines are from dcache.c */

/* partial hash update function. Assume roughly 4 bits per character */
static __inline__ unsigned long n2s_partial_name_hash(unsigned long c, unsigned long prevhash)
{
	prevhash = (prevhash << 4) | (prevhash >> (8*sizeof(unsigned long)-4));
	return prevhash ^ c;
}

/* Finally: cut down the number of bits to a int value (and try to avoid losing bits) */
static __inline__ unsigned long n2s_end_name_hash(unsigned long hash)
{
	if (sizeof(hash) > sizeof(unsigned int))
		hash += hash >> 4*sizeof(hash);
	return (unsigned long) hash;
}
static __inline__ unsigned int n2s_name_hash_index(unsigned long hash)
{
	hash = hash ^ (hash >> NUM2STR_HASHBITS) ^ (hash >> NUM2STR_HASHBITS*2);
	return (hash & NUM2STR_HASHMASK);
}

/* Compute the hash for a name string. */
static __inline__ unsigned int n2s_full_name_hash(utf8string *str)
{
	const unsigned char *name = str->utf8string_val;
	unsigned int len = str->utf8string_len;
	unsigned long hash = init_name_hash();
	unsigned long hash2;

	while (len--)
		hash = n2s_partial_name_hash(*name++, hash);
	hash2 = n2s_end_name_hash(hash);
	return n2s_name_hash_index(hash2);
}


/*
 * insert,remove and find routines must be called with the spinlock held.
 * these hash management routines are from the uid_cache in kernel/fork.c
 */

static inline int n2s_hash_insert(struct num_struct *nump, 
				  struct name_struct *namep,
				  unsigned int numhashent,
				  unsigned int namehashent)
{
	nfsv4_printk(n2s,"N2S_CACHE n2s_hash_insert. numhashent %d namehashent %d\n",
		     numhashent,namehashent);
	if((numhashent > NUM2STR_HASHSIZE) || (namehashent > NUM2STR_HASHSIZE))
		return -1;
	if((!namep) || (!nump))
		return -1;
	if((nump->next = numhash[numhashent]) != NULL)
		numhash[numhashent]->pprev = &nump->next;
	nump->pprev = &numhash[numhashent];
	numhash[numhashent] = nump;

	if((namep->next = namehash[namehashent]) != NULL)
		namehash[namehashent]->pprev = &namep->next;
	namep->pprev = &namehash[namehashent];
	namehash[namehashent] = namep;
	return 0;
}

static inline void n2s_hash_remove(struct num_struct *nump, 
				   struct name_struct * namep)
{
	nfsv4_printk(n2s,"N2S_CACHE n2s_hash_remove nump 0x%x,namep 0x%x\n",
		     (unsigned int)nump,(unsigned int)namep);
	if(nump->next)
		nump->next->pprev = nump->pprev;
	*nump->pprev = nump->next;

	nfsv4_printk(n2s,"N2S_CACHE n2s_hash_remove namep\n");
	if(namep->next)
		namep->next->pprev = namep->pprev;
	*namep->pprev = namep->next;
}

static inline struct num_struct *num_hash_find(unsigned int type, 
					       unsigned int num,
					       unsigned int hashent)
{
	struct num_struct *nump, *next;

	nfsv4_printk(n2s,"N2S_CACHE num_hash_find\n");
	next = numhash[hashent];
	for (;;) {
		nump = next;
		if (next) {
			next = nump->next;
			/* if not found, continue to look */
			if (!((nump->data->num == num) && (nump->data->type == type))) 
				continue;
			atomic_inc(&nump->data->count);
		}
		break;
	}
	return nump;
}

static inline struct name_struct *name_hash_find(unsigned int type,
						 utf8string *name, 
						 unsigned int hashent)
{
	struct name_struct *namep, *next;

	next = namehash[hashent];
	nfsv4_printk(n2s,"N2S_CACHE name_hash_find next 0x%x\n",
		     (unsigned int)next);
	for (;;) {
		namep = next;
		if (next) {
			next = namep->next;
			/* if not found, continue to look */
			if (!(!(utf8cmp(&namep->data->str,name)) && (namep->data->type == type)))
				continue;
			atomic_inc(&namep->data->count);
		}
		break;
	}
	nfsv4_printk(n2s,"N2S_CACHE name_hash_find return namep 0x%x\n",
		     (unsigned int)namep);
	return namep;
}


/* for garbage collection */
void free_n2s(struct num2str *ptr)
{
	unsigned int numhashent = numhashfn(ptr->num);
	unsigned int namehashent;
	struct num_struct *nump = NULL;
	struct name_struct *namep = NULL;

	if (ptr) {
		if (atomic_dec_and_test(&ptr->count)) {
			namehashent = n2s_full_name_hash(&ptr->str);
			spin_lock(&hash_lock);
			namep = name_hash_find(ptr->type,&ptr->str, namehashent);
			nump = num_hash_find(ptr->type,ptr->num, numhashent);
			n2s_hash_remove(nump,namep);
			kmem_cache_free(n2s_cachep, ptr);
			kfree(nump);
			kfree(namep);
			spin_unlock(&hash_lock);
		}
	}
}

int 
alloc_n2s(u_int type,u_int num, utf8string *str)
{
	unsigned int numhashent = numhashfn(num);
	unsigned int namehashent;
	struct num_struct *nump = NULL;
	struct name_struct *namep = NULL;
	struct num2str *new;
	int ret = 0;
	
	nfsv4_printk(n2s,"N2S_CACHE alloc_n2s. str:\n");
	print_utf8string(str);
	namehashent = n2s_full_name_hash(str);
	spin_lock(&hash_lock);
	nump = num_hash_find(type,num, numhashent);
	spin_unlock(&hash_lock);

	if (!nump) {
		struct num_struct *nump1 = NULL;
		nfsv4_printk(n2s,"N2S_CACHE alloc_n2s - kmem_cache_alloc\n");
		new = kmem_cache_alloc(n2s_cachep, SLAB_KERNEL);
		if (!new)
			return -EAGAIN;
		new->type = type;
		new->num = num;
		nfsv4_printk(n2s,"NEW: %p, new->num %d new->type %d\n",new,new->num, new->type);
		nfsv4_printk(n2s,"N2S_CACHE kmalloc\n");
		nump = kmalloc(sizeof(struct num_struct),GFP_KERNEL);
		if(nump){
			nfsv4_printk(n2s,"N2S_CACHE alloc_hash_element. nump %p\n",nump);
			nump->next = NULL;
			nump->pprev = NULL;
			nump->data = new;
			nump->namehashent = namehashent;
		}else{
			printk("N2S_CACHE nump kmalloc failed\n");
			ret = -ENOMEM;
			goto err_out;
		}
		namep = kmalloc(sizeof(struct name_struct),GFP_KERNEL);
		if(namep){ 
			nfsv4_printk(n2s,"N2S_CACHE alloc_hash_element. namep %p\n",namep);
			namep->next = NULL;
			namep->pprev = NULL;
			namep->data = new;
		}else {
			printk("N2S_CACHE namep kmalloc failed\n");
			ret = -ENOMEM;
			goto err_out;
		}
		utf8copy(&new->str,str);
		nfsv4_printk(n2s,"AFTER UTF8copy str: \n");
		print_utf8string(&nump->data->str);
		atomic_set(&new->count, 1);
	
		/*
		 * Before adding this, check whether we raced
		 * on adding the same num2str already..
		 */
		spin_lock(&hash_lock);
		nump1 = num_hash_find(type,num, numhashent);
		if (nump1) {
			nfsv4_printk(n2s,"N2S_CACHE alloc_n2s - race, found nump\n");
			ret = 0;
			goto err_out;
		} else {
			nfsv4_printk(n2s,"N2S_CACHE alloc_n2s - cache miss.\n");
			ret = n2s_hash_insert(nump,namep,numhashent,namehashent);
			if(ret != 0){
				printk("N2S_CACHE alloc_n2s - n2s_hash_insert error\n");
				ret = -1;
				goto err_out;
			}
		}
	}
	spin_unlock(&hash_lock);
	return 0;
 err_out:
	spin_unlock(&hash_lock);
	kmem_cache_free(n2s_cachep, new);
	kfree(nump);
	kfree(namep);
	return -1;
}

void n2scache_init(void)
{
	int i;

	nfsv4_printk(n2s,"N2S_CACHE init\n");
	if(!n2s_cachep){
		nfsv4_printk(n2s,"N2S_CACHE init calling kmem_cache_create\n");
		n2s_cachep = kmem_cache_create("n2s_cache", sizeof(struct num2str), 0, SLAB_HWCACHE_ALIGN, NULL, NULL);
		if(!n2s_cachep)
			panic("Cannot create num2str SLAB cache\n");
	}else{
		nfsv4_printk(n2s,"N2S_CACHE init n2s_cache already created\n");
	}
	for(i = 0; i < NUM2STR_HASHSIZE; i++){
		numhash[i] = 0;
		namehash[i] = 0;
	}
	/* XXX - dugsong testing */
	{
		u32 uid;
		struct xdr_netobj name;

		/* Test GSS-authenticated RPC call to GSSD. */
		name.data = "dugsong";
		name.len = 7;

		if (gssd_client_test(&name, &uid) < 0)
			printk("N2S: gssd_client_test failed!\n");
		else
			printk("N2S: gssd_client_test: %.*s -> %d\n",
			       (int) name.len, name.data, uid);

		/* Test regular GSSD calls. */
		uid = 2;
	  
		if (gssd_uid_to_name(&uid, &name) < 0)
			printk("N2S: gssd_uid_to_name failed!\n");
		else {
			printk("N2S: gssd_uid_to_name: %d -> %.*s\n",
			       uid, (int) name.len, name.data);
			rpc_free(name.data);
		}
		name.data = "operator";
		name.len = 8;
	  
		if (gssd_name_to_uid(&name, &uid) < 0)
			printk("N2S: gssd_name_to_uid failed!\n");
		else
			printk("N2S: gssd_name_to_uid: %.*s -> %d\n",
			       (int) name.len, name.data, uid);
	}
}


/* functions called by nfsd */

/* free hash table elements, and shrink kmem_cache for nfsd shutdown.
 * 
 * search numhash[] for num_structs. search namehash[] for matching
 * name_struct. free name_struct, num_struct, num2str structs.
 */
void n2s_cache_shutdown(void) 
{
	int i;
	struct num_struct *np,*next;
	struct name_struct *namep;

	nfsv4_printk(n2s,"NFSDV4 n2s_cache_shutdown\n");
	spin_lock(&hash_lock);
	for(i=0;i<NUM2STR_HASHSIZE;i++){
		next = numhash[i];
		for(;;){
			np = next;
			printk("        next 0x%x     i %d\n",(unsigned int)next,i);
			if(next){
				nfsv4_printk(n2s,"next->data->type %d,next->data->num %d, next->namehashent %d\n", next->data->type,next->data->num,next->namehashent);
				next = np->next;
				namep = name_hash_find(np->data->type,
						       &np->data->str,np->namehashent);
				if(namep){
					n2s_hash_remove(np,namep);
					nfsv4_printk(n2s,"Calling KMEM_CACHE_FREE\n");
					kmem_cache_free(n2s_cachep,np->data);
					kfree(np);
					kfree(namep);
					continue;
				}else{ 
					nfsv4_printk(n2s,"N2S_CACHE shutdown ERROR\n");
					break;
				}
			}
			break;
		}
	}
	spin_unlock(&hash_lock);
	nfsv4_printk(n2s,"Calling KMEM_CACHE_SHRINK\n");
	kmem_cache_shrink(n2s_cachep);
}

int
get_num(u_int type, utf8string *str, u_int *num)
{
	struct name_struct *namep = NULL;
	unsigned int namehashent;

	/* look in num2str cache */
	namehashent = n2s_full_name_hash(str);
	spin_lock(&hash_lock);
	namep = name_hash_find(type,str,namehashent);
	spin_unlock(&hash_lock);

	if(namep) {
		*num = namep->data->num;
		return 0;
	}

	/* cache lookup failed. go get num from daemon and 
	 * stuff results in cache.
	 */
	nfsv4_printk(n2s,"N2S_CACHE get_num cache miss: looking for %.*s\n",
		     (int) str->utf8string_len, str->utf8string_val);

	if (type == NUM2STR_UID) {
		if (gssd_name_to_uid((struct xdr_netobj *)str, num) < 0)
			return -1;
	}
	else if (type == NUM2STR_GID) {
		if (gssd_name_to_gid((struct xdr_netobj *)str, num) < 0)
			return -1;
	}
	else return -1;
	
	nfsv4_printk(n2s,"N2S_CACHE get_num: %.*s -> %d\n",
		     (int) str->utf8string_len, str->utf8string_val, *num);
	
	/* place in cache */
	alloc_n2s(type,*num,str);

	return 0;
}
        
/* caller will need to free results returned in utf8string */

int
get_name(u_int type, u_int num, utf8string *str)
{        
	struct num_struct *nump = NULL;
	unsigned int numhashent = numhashfn(num);

	nfsv4_printk(n2s,"N2S_CACHE get_name type %d, num %d\n",type,num);
	/* look in num2str cache */
	spin_lock(&hash_lock);
	nump = num_hash_find(type,num, numhashent);
	spin_unlock(&hash_lock);

	if (nump) {
		nfsv4_printk(n2s,"N2S_CACHE get_name cache hit. str: \n");
		print_utf8string(&nump->data->str);
		str->utf8string_len = nump->data->str.utf8string_len;
		str->utf8string_val = kmalloc (str->utf8string_len, GFP_KERNEL);
		memcpy (str->utf8string_val, nump->data->str.utf8string_val, str->utf8string_len);
		print_utf8string(str);
		return 0;
	}

	/* cache lookup failed. go get num from daemon and 
	 * stuff results in cache.
	 */
	nfsv4_printk(n2s,"N2S_CACHE get_name cache miss: looking for %d\n", num);
	if (type == NUM2STR_UID) {
		if (gssd_uid_to_name(&num, (struct xdr_netobj *)str) < 0)
			return -1;
	}
	else if (type == NUM2STR_GID) {
		if (gssd_gid_to_name(&num, (struct xdr_netobj *)str) < 0)
			return -1;
	}
	else return -1;

	nfsv4_printk(n2s,"N2S_CACHE get_name: %d -> %.*s\n",
		     num, (int) str->utf8string_len, str->utf8string_val);
	  
	/* place in cache */
	alloc_n2s(type,num,str);
	
	return 0;
}

