/*
 *  fs/nfsd/nfsd4xdr.c
 *
 *  client-side xdr for nfsv4 callbacks
 *
 *  by Kendrick Smith <kmsmith@umich.edu>
 *  Copyright (C) 2000 The Regents of the University of Michigan
 */

#include <linux/types.h>
#include <linux/sched.h>

#include <linux/sunrpc/xdr.h>
#include <linux/sunrpc/svc.h>
#include <linux/nfs4_prot.h>
#include <linux/malloc.h>

/*
 *  stuff lifted from fs/nfsd/nfsd4xdr.c
 */

static inline int
xdr_overflow_check(struct svc_buf *buf, u32 *p, int arglen)
{
	return p + arglen - buf->base <= buf->buflen;
}

static inline u32 *
xdr_decode_64(u32 *p, u64 *valp)
{
	*valp  = ((u64) ntohl(*p++)) << 32;
	*valp |= ntohl(*p++);
	return p;
}

static u32 *
xdr_decode_utf8string(struct svc_rqst *rqstp, u32 *p, utf8string *utf)
{
	int len = sizeof(utf->utf8string_len) + ntohl(*p);

	if (!(xdr_overflow_check(&rqstp->rq_argbuf, p, len)))
		return 0;
	utf->utf8string_len = ntohl(*p++);
	utf->utf8string_val = (char *)p;
	return p + XDR_QUADLEN(utf->utf8string_len);
}

static inline u32 *
xdr_decode_fh4(struct svc_rqst *rqstp,u32 *p, nfs_fh4 *fhp)
{
	int len = sizeof(fhp->nfs_fh4_len) + XDR_QUADLEN(ntohl(*p));

	if (!(xdr_overflow_check(&rqstp->rq_argbuf, p, len)))
		return 0;
	fhp->nfs_fh4_len = ntohl(*p++);
	fhp->nfs_fh4_val = (char *)p;
	return p + XDR_QUADLEN(fhp->nfs_fh4_len);
}

static u32*
xdr_decode_bitmap4(struct svc_rqst *rqstp,u32 *p, bitmap4 *bm4)
{
/* ANDROS note: instead of kmalloc, decode BVAL's in place. */

	int i, len = sizeof(bm4->BLEN) + ntohl(*p)*sizeof(uint32_t);

	if (!xdr_overflow_check(&rqstp->rq_argbuf, p, len))
		return 0;
	bm4->BLEN = ntohl(*p++);
	bm4->BVAL = p;
	for (i = 0; i < bm4->BLEN; i++) {
		*p = ntohl(*p);
		p++;
	}
	return p;
}

static inline u32 *
xdr_encode_64(u32 *p, u64 val)
{
	*p++ = htonl(val >> 32);
	*p++ = htonl(val & 0xffffffff);
	return p;
}

static inline u32 *
xdr_encode_utf8string(struct svc_rqst *rqstp,u32 *p, utf8string  *utf)
{
	int len = sizeof(utf->utf8string_len) + XDR_QUADLEN(utf->utf8string_len); 

	if (!(xdr_overflow_check(&rqstp->rq_resbuf, p, len)))
		return 0;
	*p++=htonl(utf->utf8string_len);
	memcpy(p, utf->utf8string_val, utf->utf8string_len);
	return p + XDR_QUADLEN(utf->utf8string_len);
}

static u32 *
xdr_encode_bitmap4(struct svc_rqst *rqstp,u32 *p, bitmap4 *bm4)
{
	int i, len = sizeof(bm4->BLEN) + (bm4->BLEN * sizeof(uint32_t));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
		return 0;
	*p++ = htonl(bm4->BLEN);
	for (i = 0; i < bm4->BLEN; i++)
		*p++ = htonl(bm4->BVAL[i]);	
	return p;
}

static u32*
xdr_encode_fattr4(struct svc_rqst *rqstp,u32 *p,struct fattr4 *res)
{
	bitmap4 *bm4 = &res->attrmask;
	int i, j;
	char *cptr = res->FATTRSVAL;
	u32 *save, *start;
	
	if (!(p = xdr_encode_bitmap4(rqstp,p,bm4)))
		return 0;
	if (!xdr_overflow_check(&rqstp->rq_resbuf, p, 4))
		return 0;
	save = p++;
	start = p;
	
	for (i = 0; i < bm4->bitmap4_len; i++)
		for (j = 0; j < 32; j++)
			if (bm4->bitmap4_val[i] & (1 << j))
				switch (i*32 + j) {
				case FATTR4_CHANGE:
				case FATTR4_SIZE:
					if (!xdr_overflow_check(&rqstp->rq_resbuf, p, 8))
						return 0;
					p = xdr_encode_64(p,*(u64 *)cptr);
					p += 2;
					break;
				case FATTR4_TIME_ACCESS:
				case FATTR4_TIME_MODIFY:
					if (!xdr_overflow_check(&rqstp->rq_resbuf, p, 12))
						return 0;
					p = xdr_encode_64(p,*(u64 *)cptr);
					*(uint32_t *) p = htonl(*(uint32_t *) p);
					p += 1;
					break;
				}
	
	*save = htonl((p - start) * 4);
	return p;
}

int
xdr_decode_nullcb_args(struct svc_rqst *rqstp, u32 *p, void *data)
{
	return 1;       /* 1 indicates success on svc side */
}

int
xdr_encode_nullcb_res(struct svc_rqst *rqstp, u32 *p, void *data)
{
	rqstp->rq_resbuf.len = p - rqstp->rq_resbuf.base;
	return 1;       /* 1 indicates success on svc side */
}

int
xdr_free_nullcb_res(struct svc_rqst *rqstp, u32 *p, void *resp)
{
	return 1;          /* 1 indicates success */
}

static inline u32 *
xdr_decode_cb_getattr4args (struct svc_rqst *rqstp, u32 *p, CB_GETATTR4args *argp)
{
	if (!(p = xdr_decode_fh4(rqstp, p, &argp->fh)))
		return 0;
	p = xdr_decode_bitmap4(rqstp, p, &argp->attr_request);
	return p;
}

static inline u32 *
xdr_decode_cb_recall4args (struct svc_rqst *rqstp, u32 *p, CB_RECALL4args *argp)
{
	if (!xdr_overflow_check(&rqstp->rq_argbuf, p, 12))
		return 0;
	p = xdr_decode_64(p, &argp->stateid);
	argp->truncate = ntohl(*p++);
	p = xdr_decode_fh4(rqstp, p, &argp->fh);
	return p;
}

int
xdr_decode_compoundcb_args (struct svc_rqst *rqstp, u32 *p, CB_COMPOUND4args *argp)
{
	int i;
	int len;
	int op;

	if (!(p = xdr_decode_utf8string(rqstp, p, &argp->tag)))
		return 0;
	if (!xdr_overflow_check(&rqstp->rq_argbuf, p, 8))
		return 0;
	argp->minorversion = ntohl(*p++);
	len = ntohl(*p++);
	argp->argarray.argarray_len = len;
	argp->argarray.argarray_val = (nfs_cb_argop4 *)
		kmalloc(len * sizeof(nfs_cb_argop4), GFP_KERNEL);

	for (i = 0; i < len; i++) {
		if (!xdr_overflow_check(&rqstp->rq_argbuf, p, 4))
			return 0;
		op = ntohl(*p++);
		argp->argarray.argarray_val[i].argop = op;
		switch (op) {
		case OP_CB_GETATTR:
			if (!(p = xdr_decode_cb_getattr4args(rqstp, p, &argp->argarray.argarray_val[i].nfs_cb_argop4_u.opcbgetattr)))
				return 0;
			break;
		case OP_CB_RECALL:
			if (!(p = xdr_decode_cb_recall4args(rqstp, p, &argp->argarray.argarray_val[i].nfs_cb_argop4_u.opcbrecall)))
				return 0;
			break;
		default:
			printk ("!!! default reached when decoding cb args\n");
			return -EINVAL;
		}
	}

	return 1;      /* 1 indicates success on svc side */
}

static inline u32 *
xdr_encode_cb_getattr4res (struct svc_rqst *rqstp, u32 *p, CB_GETATTR4res *resp)
{
	if (!xdr_overflow_check(&rqstp->rq_resbuf, p, 8))
		return 0;
	*p++ = htonl(resp->status);
	if (!resp->status)
		p = xdr_encode_fattr4(rqstp, p, &resp->CB_GETATTR4res_u.resok4.obj_attributes);
	return p;
}

static inline u32 *
xdr_encode_cb_recall4res (struct svc_rqst *rqstp, u32 *p, CB_RECALL4res *resp)
{
	*p++ = htonl(resp->status);
	return p;
}

int
xdr_encode_compoundcb_res (struct svc_rqst *rqstp, u32 *p, CB_COMPOUND4res *resp)
{
	int i;
	int len;
	int op;

	if (!xdr_overflow_check(&rqstp->rq_resbuf, p, 4))
		return 0;
	*p++ = htonl(resp->status);
	if (!(p = xdr_encode_utf8string(rqstp, p, &resp->tag)))
		return 0;
	len = resp->resarray.resarray_len;
	if (!xdr_overflow_check(&rqstp->rq_resbuf, p, 4))
		return 0;
	*p++ = htonl (len);

	for (i = 0; i < len; i++) {
		if (!xdr_overflow_check(&rqstp->rq_resbuf, p, 4))
			return 0;
		op = resp->resarray.resarray_val[i].resop;
		*p++ = htonl (op);
		switch (op) {
		case OP_CB_GETATTR:
			if (!(p = xdr_encode_cb_getattr4res(rqstp, p, &resp->resarray.resarray_val[i].nfs_cb_resop4_u.opcbgetattr)))
				return 0;
			break;
		case OP_CB_RECALL:
			if (!(p = xdr_encode_cb_recall4res(rqstp, p, &resp->resarray.resarray_val[i].nfs_cb_resop4_u.opcbrecall)))
				return 0;
			break;
		default:
			printk("!!! default reached when encoding cb resp\n");
			return -EINVAL;
		}
	}

	rqstp->rq_resbuf.len = p - rqstp->rq_resbuf.base;
	return 1;           /* 1 indicates success on svc side */
}

void
xdr_free_compoundcb_args (CB_COMPOUND4args *argp)
{
	kfree(argp->argarray.argarray_val);
}

int
xdr_free_compoundcb_res(struct svc_rqst *rqstp, u32 *p, CB_COMPOUND4res *resp)
{
	int i;

	kfree(resp->tag.utf8string_val);
	
	for (i = 0; i < resp->resarray.resarray_len; i++)
		switch (resp->resarray.resarray_val[i].resop) {
		case OP_CB_GETATTR:
		{
			CB_GETATTR4res *rp = &resp->resarray.resarray_val[i].nfs_cb_resop4_u.opcbgetattr;
			if (!rp->status) {
				fattr4 *ap = &rp->CB_GETATTR4res_u.resok4.obj_attributes;
				kfree(ap->attrmask.bitmap4_val);
				kfree(ap->attr_vals.attrlist4_val);
			}
		}
		break;
		case OP_CB_RECALL:
			break;        /* nothing to do */
		default:
			printk("!!! default reached when freeing cb resp\n");
			return 0;
		}
	
	return 1;         /* 1 indicates success on svc side */
}
