/*
 *  linux/fs/nfs/nfs4deleg.c
 *
 *  client-side delegation support for NFSv4
 *
 *  by Kendrick Smith <kmsmith@umich.edu>
 *  Copyright (C) 2000 The Regents of the University of Michigan
 */

/*  list of things to fix/think about:
 *    - determining correct values for NFS4_CB_BUFSIZE, NFS4_CB_XDRSIZE (right
 *      now I am using huge buffers instead of worrying about calculating the
 *      proper sizes)
 *    - xdr_decode_bitmap4() currently requires preallocating the bitmap buffer,
 *      which could create buffer overflow
 *    - as in the rest of the xdr routines, buffer overflow checking is needed
 *    - xdr_decode_fh4() currently does an unnecessary kmalloc/memcpy/kfree
 *    - what are the r_netid and r_addr fields in cb_client4? (see setup_cb_client4)
 *    - right now, the response to a CB_RECALL and the ensuing DELEGRETURN
 *      can be done in any order.  is it possible to serialize things so that
 *      the CB_RECALL response is guaranteed to happen first?
 */

#define __NO_VERSION__

#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/nfs.h>
#include <linux/in.h>
#include <linux/uio.h>
#include <linux/version.h>
#include <linux/unistd.h>
#include <linux/malloc.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/sunrpc/auth.h>
#include <linux/sunrpc/types.h>
#include <linux/sunrpc/stats.h>
#include <linux/sunrpc/svc.h>
#include <linux/sunrpc/svcsock.h>
#include <linux/nfs.h>
#include <linux/nfs_fs.h>
#include <linux/nfs4_prot.h>

#define ENABLE_CALLBACKS          0
#define ACCEPT_READ_DELEGATIONS   0
#define ACCEPT_WRITE_DELEGATIONS  0
#define BOUNCE_DELEGATIONS        0
#define DELEG_PORT                23137
#define DELEG_NTHREADS            1

#define ALLOWED_SIGS	          (sigmask(SIGKILL))
#define SHUTDOWN_SIGS	          (sigmask(SIGKILL) | sigmask(SIGINT) | sigmask(SIGQUIT))

#define RC_NOCACHE                0
#define NFS4_CB_MAXTHREADS        1
#define NFS4_CB_BUFSIZE           1024
#define NFS4_CB_XDRSIZE           1024

#define DELEG_HASH_BITS           8
#define DELEG_HASH_SIZE           (1 << DELEG_HASH_BITS)
#define DELEG_HASH_MASK           (DELEG_HASH_SIZE - 1)
#define DELEG_HASH_MULT           33

static int proc_nullcb(struct svc_rqst *rqstp, void *argp, void *resp);
static int proc_compoundcb(struct svc_rqst *rqstp, CB_COMPOUND4args *argp, CB_COMPOUND4res *resp);
static int __nfs4_delegreturn(struct nfs_server *server, stateid4 stateid);
extern struct svc_program nfs4_cb_program;

static int cb_initialized = 0;
static struct nfs4_deleg_inode *deleg_hash[DELEG_HASH_SIZE];

static inline __u32
nfs_fh4_hashval(nfs_fh4 *fh)
{
	int i;
	__u32 hashval = 0;

	for (i = 0; i < fh->nfs_fh4_len; i++)
		hashval = (hashval * DELEG_HASH_MULT + fh->nfs_fh4_val[i]) & DELEG_HASH_MASK;
	return hashval;
}

static inline __u32
nfs_fh_hashval(struct nfs_fh *fh)
{
	int i;
	__u32 hashval = 0;

	for (i = 0; i < fh->size; i++)
		hashval = (hashval * DELEG_HASH_MULT + fh->data[i]) & DELEG_HASH_MASK;
	return hashval;
}

static inline struct nfs4_deleg_inode *
deleg_lookup(nfs_fh4 *fh)
{
	struct nfs4_deleg_inode *dinode;
	__u32 hashval = nfs_fh4_hashval(fh);

	for (dinode = deleg_hash[hashval]; dinode; dinode = dinode->hash_next)
		if ((dinode->fh.size == fh->nfs_fh4_len) &&
		    !memcmp(dinode->fh.data, fh->nfs_fh4_val, fh->nfs_fh4_len))
			break;
	return dinode;
}

/*
 *  Initialize delegation state
 */

static void
cb_thread (struct svc_rqst *rqstp)
	/* this is the main routine for kernel callback threads */
{
	struct svc_serv *serv = rqstp->rq_server;
	int err;

	MOD_INC_USE_COUNT;
	nfsv4_printk(callback, "callback kernel thread starting\n");

	/* set up kernel thread */
	lock_kernel();
	exit_mm(current);
	sprintf(current->comm, "nfs4_cb");
	current->session = 1;
	current->pgrp = 1;
	current->fs->umask = 0;

	/* let svc_process check server's authentication */
	rqstp->rq_auth = 1;

	for (;;) {            /* main request loop */
		
		/* block all but shutdown signals */
		spin_lock_irq(&current->sigmask_lock);
		siginitsetinv(&current->blocked, SHUTDOWN_SIGS);
		recalc_sigpending(current);
		spin_unlock_irq(&current->sigmask_lock);

		/* pull up data from socket */
		do {
			err = svc_recv (serv, rqstp, MAX_SCHEDULE_TIMEOUT);
		} while (err == -EAGAIN);
		if (err < 0)
			break;

		/* process request with signals blocked */
		spin_lock_irq(&current->sigmask_lock);
		siginitsetinv(&current->blocked, ALLOWED_SIGS);
		recalc_sigpending(current);
		spin_unlock_irq(&current->sigmask_lock);
		svc_process(serv, rqstp);
	}

	if (err != -EINTR)
		printk (KERN_WARNING "nfs4_cb thread: terminating on error %d\n", -err);
	else {
		unsigned int signo;
		for (signo = 1; signo <= _NSIG; signo++)
			if (sigismember(&current->signal, signo) &&
			    !sigismember(&current->blocked, signo))
				break;
		printk(KERN_WARNING "nfs4_cb thread: terminating on signal %d\n", signo);
	}

	svc_exit_thread (rqstp);       /* destroy thread's resources */
	nfsv4_printk (callback, "callback kernel thread exiting\n");
	MOD_DEC_USE_COUNT;
}

int
nfs4_deleg_init()
{
	struct svc_serv *serv;
	int i;
	int error = 0;

	if (!ENABLE_CALLBACKS) {
		nfsv4_printk(callback, "nfs4_deleg_init(): callbacks disabled\n");
		goto out;
	}
	if (cb_initialized) {
		nfsv4_printk(callback, "nfs4_deleg_init(): already initialized!\n");
		goto out;
	}
	cb_initialized = 1;
	nfsv4_printk(callback, "nfs4_deleg_init(): starting callback svc on port=%d (%d threads)\n",
		     DELEG_PORT, DELEG_NTHREADS);
	
	if (!(serv = svc_create(&nfs4_cb_program, NFS4_CB_BUFSIZE, NFS4_CB_XDRSIZE))) {
		error = -ENOMEM;
		goto out;
	}
     	if ((error = svc_makesock(serv, IPPROTO_UDP, DELEG_PORT)) < 0)
		goto destroy;
	for (i = 0; i < DELEG_NTHREADS; i++)
		if ((error = svc_create_thread(cb_thread, serv)) < 0)
			goto destroy;

	for (i = 0; i < DELEG_HASH_SIZE; i++)
		deleg_hash[i] = 0;
	
 destroy:
	svc_destroy(serv);

 out:
	nfsv4_printk(callback, "nfs4_deleg_init(): returning status %d\n", error);
	return error;
}


/*
 *  Callback svc
 */

static struct svc_stat nfs4_cb_stats = { &nfs4_cb_program,
					 0, 0, 0, 0, 0, 0, 0, 0 };

static struct svc_procedure nfs4_cb_procedures[] = {
	{					/* CB_NULL */
		(svc_procfunc) proc_nullcb,             /* process */
		(kxdrproc_t) xdr_decode_nullcb_args,    /* decode args */
		(kxdrproc_t) xdr_encode_nullcb_res,     /* encode res */
		(kxdrproc_t) xdr_free_nullcb_res,       /* free res */
		4,        /* would 0 be safe here? */   /* arg size */
		4,        /* would 0 be safe here? */   /* resp size */
		0,                                      /* call count */
		RC_NOCACHE                              /* ??? */
	},
	{					/* CB_COMPOUND */
		(svc_procfunc) proc_compoundcb,           /* process rqst */
		(kxdrproc_t) xdr_decode_compoundcb_args,  /* decode args */
		(kxdrproc_t) xdr_encode_compoundcb_res,   /* encode res */
		(kxdrproc_t) xdr_free_compoundcb_res,     /* free res */
		sizeof(struct CB_COMPOUND4args),          /* arg size */
		sizeof(struct CB_COMPOUND4res),           /* resp size */
		0,                                        /* call count */
		RC_NOCACHE                                /* ??? */
	}
};

static struct svc_version nfs4_cb_version1 = {
	1,					  /* version number */
	2,                                        /* number of procedures */
	nfs4_cb_procedures,                       /* procedure table */
	NULL                                      /* dispatcher (default) */
};

static struct svc_version *nfs4_cb_version[] = {
	NULL,                                     /* version 0 */
	&nfs4_cb_version1                         /* version 1 */
};

static struct svc_program nfs4_cb_program = {
	NFS4_CALLBACK,                            /* program number */
	1,                                        /* lowest supported version */
	1,                                        /* highest supported version */
	2,                                        /* # of entries in nfs4_cb_version */
	nfs4_cb_version,                          /* per-version virtual ops */
	"nfs4_cb",                                /* program name */
	&nfs4_cb_stats                            /* stats table */
};

static int
proc_nullcb(struct svc_rqst *rqstp, void *argp, void *resp)
{
	nfsv4_printk(callback, "received null callback\n");
	return 0;          /* 0 indicates success */
}

static int
proc_cb_getattr(CB_GETATTR4args *argp, CB_GETATTR4res *resp)
{
#if 0
	struct nfs4_deleg_inode   *statep;
	struct inode              *inode;
	struct iattr              iattr;
	int                       status = 0;
	uint32_t                  *bitmap_val = argp->attr_request.bitmap4_val;

	nfsv4_printk (deleg, "deleg: received CB_GETATTR\n");

	if (argp->fh.nfs_fh4_len > NFS_MAXFHSIZE)
		argp->fh.nfs_fh4_len = NFS_MAXFHSIZE;
  
	if ( !(statep = deleg_lookup(addr,&argp->fh))) {
		printk ("!!! bad filehandle in cb_getattr !!!\n");
		status = NFS4ERR_BADHANDLE;
		goto out;
	}
	inode = statep->inode;

	memset (&iattr, 0, sizeof(iattr));
	if (bitmap_val[0] & 0x08) {           // FATTR4_CHANGE
							 printk ("!!! change info requested on callback.. not yet supported !!!\n");
	}
	if (bitmap_val[0] & 0x010) {          // FATTR4_SIZE
							 iattr.ia_valid |= ATTR_SIZE;
	iattr.ia_size = inode->i_size;
	}
	if (bitmap_val[1] & 0x8000) {        // FATTR4_TIME_ACCESS
							iattr.ia_valid |= ATTR_ATIME;
	iattr.ia_atime = inode->i_atime;
	}
	if (bitmap_val[1] & 0x200000) {      // FATTR4_TIME_MODIFY
							iattr.ia_valid |= ATTR_MTIME;
	iattr.ia_mtime = inode->i_mtime;
	}

	if (iattr_to_fattr4 (&iattr, &resp->CB_GETATTR4res_u.resok4.obj_attributes)) {
		status = NFS4ERR_RESOURCE;
		goto out;
	}

 out:
	resp->status = status;
	if (nfsv4_debug_deleg) {
		printk ("deleg: CB_GETATTR returning status %d\n", status);
		if (status) {
			printk ("--------------- beginning of returned attributes -----------------");
			dump_fattr4 (&resp->CB_GETATTR4res_u.resok4.obj_attributes);
			printk ("--------------- end of returned attributes -----------------");
		}
	}
	return status;
#else
	return NFS4ERR_NOTSUPP;
#endif
}

static int
proc_cb_recall(CB_RECALL4args *argp, CB_RECALL4res *resp)
{
	struct nfs4_deleg_inode   *dinode;
	int                       status = 0;

	nfsv4_printk(deleg, "received CB_RECALL, stateid=%d fh_len=%d\n",
		     (int) argp->stateid, argp->fh.nfs_fh4_len);
	
	if (argp->fh.nfs_fh4_len > NFS_MAXFHSIZE)
		argp->fh.nfs_fh4_len = NFS_MAXFHSIZE;

	if (!(dinode = deleg_lookup(&argp->fh))) {
		nfsv4_printk(deleg, "bad filehandle in CB_RECALL\n");
		status = NFS4ERR_BADHANDLE;
		goto out;
	}
	if (dinode->stateid != argp->stateid) {
		nfsv4_printk(deleg, "stateid mismatch in CB_RECALL\n");
		status = NFS4ERR_BAD_STATEID;
		goto out;
	}
	nfsv4_printk(deleg, "found dinode on CB_RECALL\n");

	nfs4_delegreturn(dinode->inode);

 out:
	resp->status = status;
	nfsv4_printk(deleg, "CB_RECALL returning status %d\n", status);
	return status;
}

static int
proc_compoundcb(struct svc_rqst *rqstp, CB_COMPOUND4args *argp, CB_COMPOUND4res *resp)
{
	int i;
	int len;
	nfs_cb_argop4 *argop4;
	nfs_cb_resop4 *resop4;
	int status = 0;
    
	nfsv4_printk (callback, "callback: received compound callback; tag = \"%.*s\"\n",
		      (int)argp->tag.utf8string_len, argp->tag.utf8string_val);
	utf8copy(&resp->tag, &argp->tag);
  
	len = argp->argarray.argarray_len;
	resp->resarray.resarray_val = (nfs_cb_resop4 *)
		kmalloc(len * sizeof(nfs_cb_resop4), GFP_KERNEL);

	for (i = 0; i < len; i++) {
		if (status)
			break;
		argop4 = &argp->argarray.argarray_val[i];
		resop4 = &resp->resarray.resarray_val[i];
		resop4->resop = argop4->argop;
		switch (argop4->argop) {
		case OP_CB_GETATTR:
			status = proc_cb_getattr(&argop4->nfs_cb_argop4_u.opcbgetattr,
						 &resop4->nfs_cb_resop4_u.opcbgetattr);
			break;
		case OP_CB_RECALL:
			status = proc_cb_recall(&argop4->nfs_cb_argop4_u.opcbrecall,
						&resop4->nfs_cb_resop4_u.opcbrecall);
			break;
		default:
			printk ("!!! default case reached in proc_compoundcb(): panicking\n");
			return -EINVAL;
		}
	}
	resp->status = status;
	resp->resarray.resarray_len = i;

	xdr_free_compoundcb_args(argp);
	return 0;          /* 0 indicates success */
}

/*
 *  Called by setup_setclientid()
 */

void
nfs4_setup_cb_client4(cb_client4 *clp)
{
	if (ENABLE_CALLBACKS) {
		nfsv4_printk(deleg, "setup_cb_client4(): callbacks enabled\n");
		clp->cb_program = NFS4_CALLBACK;
		clp->cb_location.r_netid = "";
		clp->cb_location.r_addr = "";
	}
	else {
		nfsv4_printk(deleg, "setup_cb_client4(): callbacks disabled\n");
		clp->cb_program = 0;
		clp->cb_location.r_netid = "";
		clp->cb_location.r_addr = "";
	}
}

/*
 *  Accept a delegation
 */

void
nfs4_accept_delegation(struct dentry *dentry, open_delegation4 *deleg)
{
	struct inode *inode = dentry->d_inode;
	struct nfs4_deleg_inode *dinode;
	stateid4 stateid;
	struct nfs_fh *fh;
	__u32 hashval;

	if (!inode)
		goto out;
	
	nfsv4_printk(deleg, "nfs4_accept_delegation(): called with inode=%ld\n", inode->i_ino);
	down(&inode->i_sem);
	
	switch (deleg->delegation_type) {
	case OPEN_DELEGATE_NONE:
		nfsv4_printk(deleg, "nfs4_accept_delegation(): received OPEN_DELEGATE_NONE\n");
		break;

	case OPEN_DELEGATE_READ:
		stateid = deleg->open_delegation4_u.read.stateid;
		nfsv4_printk(deleg, "nfs4_accept_delegation(): received OPEN_DELEGATE_READ, stateid=%d\n",
			     (int) stateid);
		if (!ACCEPT_READ_DELEGATIONS) {
			nfsv4_printk(deleg, "nfs4_accept_delegation(): read delegations disabled\n");
			goto out_return;
		}
		if (NFS_DELEGSTATE(inode)) {
			nfsv4_printk(deleg, "nfs4_accept_delegation(): already have a delegation for this file!\n");
			goto out_return;
		}
		fh = &NFS_DENTRY(dentry)->fh;
		hashval = nfs_fh_hashval(fh);
		if (!(dinode = kmalloc(sizeof(*dinode), GFP_KERNEL)))
			goto out_return;
		dinode->inode = inode;
		dinode->type = OPEN_DELEGATE_READ;
		dinode->stateid = stateid;
		dinode->hashval = hashval;
		dinode->fh.size = fh->size;
		memcpy(dinode->fh.data, fh->data, fh->size);
		if ((dinode->hash_next = deleg_hash[hashval]))
			dinode->hash_next->hash_prev = dinode;
		deleg_hash[hashval] = dinode;
		dinode->hash_prev = NULL;
		NFS_DELEGSTATE(inode) = dinode;
		break;
		
	case OPEN_DELEGATE_WRITE:
		stateid = deleg->open_delegation4_u.read.stateid;
		if (!ACCEPT_WRITE_DELEGATIONS) {
			nfsv4_printk(deleg, "nfs4_accept_delegation(): write delegations disabled\n");
			goto out_return;
		}
		printk("!!! accepting write delegation; not implemented!\n");
		goto out_return;
		
	default:
		printk("!!! server returned corrupt delegation type\n");
		break;
	}

	up(&inode->i_sem);
	if (BOUNCE_DELEGATIONS) {
		nfsv4_printk(deleg, "nfs4_accept_delegation(): bouncing delegation\n");
		nfs4_delegreturn(inode);
	}
	
 out:
	nfsv4_printk(deleg, "nfs4_accept_delegation(): done\n");
	return;
 out_return:
	up(&inode->i_sem);
	__nfs4_delegreturn(NFS_SERVER(inode), stateid);
	goto out;
}

/*
 *  functions for delegation return
 */

struct dr_data {
	struct rpc_message msg;
	struct COMPOUND4args cargs;
	struct COMPOUND4res cres;
	nfs_argop4 aop4[1];
	nfs_resop4 rop4[1];
};

static void
delegreturn_done(struct rpc_task *task)
{
	struct dr_data *calldata = (struct dr_data *) task->tk_calldata;

#if 0
	rpc_release_task(task);
	kfree(calldata);
#endif
	nfsv4_printk(deleg, "async OP_DELEGRETURN returns status %d\n",
		     calldata->cres.status);
}

static int
__nfs4_delegreturn(struct nfs_server *server, stateid4 stateid)
{
	struct dr_data *calldata;
	char tag[] = "__nfs4_delegreturn() OP_DELEGRETURN";
	int status = 0;

	nfsv4_printk(deleg, "__nfs4_delegreturn starting with stateid %d\n", (int) stateid);

	if (!(calldata = kmalloc(sizeof(*calldata), GFP_KERNEL))) {
		status = -ENOMEM;
		goto out;
	}
	setup_compound(1, &calldata->cargs, &calldata->cres, tag,
		       calldata->aop4, calldata->rop4);
	calldata->cargs.OPVAL[0].argop = OP_DELEGRETURN;
	calldata->cargs.OPVAL[0].OPU.opdelegreturn.stateid = stateid;
	calldata->msg.rpc_proc = NFSPROC4_COMPOUND;
	calldata->msg.rpc_argp = &calldata->cargs;
	calldata->msg.rpc_resp = &calldata->cres;
	calldata->msg.rpc_cred = NULL;

	status = rpc_call_async(server->client, &calldata->msg, 0, delegreturn_done, calldata);
	
 out:
	nfsv4_printk(deleg, "__nfs4_delegreturn returning status %d\n", status);
	return status;
}

int
nfs4_delegreturn(struct inode *inode)
{
	struct nfs4_deleg_inode *dinode;
	int status = 0;

	nfsv4_printk(deleg, "nfs4_delegreturn() starting with inode %ld\n", inode->i_ino);
	
	down(&inode->i_sem);
	if (!(dinode = NFS_DELEGSTATE(inode))) {
		up(&inode->i_sem);
		nfsv4_printk(deleg, "nfs4_delegreturn(): inode is not delegated!\n");
		goto out;
	}
	if (dinode->hash_next)
		dinode->hash_next->hash_prev = dinode->hash_prev;
	if (dinode->hash_prev)
		dinode->hash_prev->hash_next = dinode->hash_next;
	else
		deleg_hash[dinode->hashval] = dinode->hash_next;
	NFS_DELEGSTATE(inode) = NULL;
	up(&inode->i_sem);

	status = __nfs4_delegreturn(NFS_SERVER(inode), dinode->stateid);
	kfree(dinode);

 out:
	nfsv4_printk(deleg, "nfs4_delegreturn returning status %d\n", status);
	return status;
}
