/*
 * linux/fs/nfsd/nfsdlock.c
 * 
 * based on:
 * linux/fs/lockd/svcsubs.c
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 *
 */                            
#include <linux/fs.h>
#include <linux/sunrpc/svc.h>
#include <linux/nfs_fs.h>
#include <linux/malloc.h> 



/*
 * Global hash tables
 */

/* NFS4_NRHASH = number of hash buckets 
*  for lockowners. - will need to store nfs4_files, nfs4share, etc 
*  when delegation is implemented, so #if 0 all that stuff..
*/

#define NFS4_NRHASH  64 
#define NFS4_HASH_BITS    5

#if 0
static struct nfs4_file *  nfs4_files[NFS4_NRHASH];
static struct semaphore   nfs4_file_sema = MUTEX;         
long	nfs4_filecnt;
#endif /*0*/

static struct nfs4_lock_owner *nfs4_lock_owners[NFS4_NRHASH];
static struct semaphore   nfs4_lowner_sema = MUTEX;         
long	nfs4_lock_ownercnt;

/* Global verifier */
extern char nfsd_verf[];

/* used for both nfs4_file and nfs4_lock_owner */
static unsigned int nfs4_hash(unsigned long element1,unsigned long element2)
{
	unsigned long tmp = (unsigned long) element2 | (unsigned long) element1;
	tmp = tmp + (tmp >> NFS4_HASH_BITS) + (tmp >> NFS4_HASH_BITS*2);
	return tmp & (NFS4_NRHASH - 1);
} 


/*
* Compare lockowner4's
*/
static inline int
nfs4_cmp_owner(pid_t pid,clientid4 clid, struct nfs4_lock_owner *owner)
{
	return ((clid == owner->lo_clientid) && (pid == owner->lo_pid)); 
}

/* 
* Initalize lock state
*/
void init_lock_state(void)
{
	int i;

	nfsv4_printk(lock,"NFSV4 Initalize Lock State\n");
	nfs4_lock_ownercnt = 0;
	for(i=0;i<NFS4_NRHASH;i++){
		nfs4_lock_owners[i] = NULL;	
#if 0
		nfs4_files[i] = NULL;	
#endif /*0*/
	}
}


/*
* Clean up
*/
void shutdown_lock_state(void)
{
	struct nfs4_lock_owner *owner,*o;
#if 0
	struct nfs4_file *file,*f;
#endif /*0*/
	int i;

	nfsv4_printk(lock,"NFSV4 Shutdown Lock State\n");
	for(i=0;i<NFS4_NRHASH;i++){
#if 0
		for(file = nfs4_files[i];file;){
			freefilecnt++;
			f = file->f_next;
			kfree(file);
			file = f;
		} 
#endif /*0*/
		for(owner = nfs4_lock_owners[i];owner;){
			o =owner->lo_next;
			kfree(owner);
			owner = o;
		}
	}
	nfsv4_printk(lock,"NFSV4 shutdown_lock_state done\n");
}

/*
 * Lockowner lookup/creation
 *
 * Called by OPEN  
 */
struct nfs4_lock_owner *
nfs4_get_lockowner(pid_t pid,clientid4 clid)
{
	struct nfs4_lock_owner	*lowner, **hp;
	int		hash;

	nfsv4_printk(lock,"NFSV4: nfs4_get_lockowner clientid: [0x%lx , %08lx] pid %d\n",
		     (unsigned long)(clid >> 32),
		     (unsigned long)(clid & 0xFFFFFFFFUL),pid);

	hash = nfs4_hash((unsigned long)(clid >> 32),
			 (unsigned long)(clid & 0xFFFFFFFFUL));

	/* Lock hash table */
	down(&nfs4_lowner_sema);

/* look at lockd's garbage collecting..
   if (time_after_eq(jiffies, next_gc))
   nlm_gc_hosts();
*/

	for (hp = &nfs4_lock_owners[hash]; (lowner = *hp); hp = &lowner->lo_next) {
		nfsv4_printk(lock,"NFSV4:get_lockowner: lowner->clid[0x%lx , %08lx],lowner->pid %d\n",
			     (unsigned long)(lowner->lo_clientid >> 32),
			     (unsigned long)(lowner->lo_clientid & 0xFFFFFFFFFFUL),lowner->lo_pid);
		if (nfs4_cmp_owner(pid, clid, lowner)) {
			if (hp != &nfs4_lock_owners[hash]) {
				/* move found lock_owner to front of hash linked list (MRU) */
				*hp = lowner->lo_next;
				lowner->lo_next = nfs4_lock_owners[hash];
				nfs4_lock_owners[hash] = lowner;
			}
			up(&nfs4_lowner_sema);
			nfsv4_printk(lock,"NFSV4: lockowner found lowner %p lo_seq_num %d\n",lowner,lowner->lo_seq_num);
			return lowner;
		}
	}

	/* No lockowner found, create it */
	nfsv4_printk(lock,"NFSV4: creating lockowner entry\n");

	if (!(lowner=(struct nfs4_lock_owner *)kmalloc(sizeof(*lowner),GFP_KERNEL)))
		goto out;
	nfs4_lock_ownercnt++;
	memset(lowner,0,sizeof(*lowner));

	lowner->lo_clientid = clid;
	lowner->lo_pid = current->pid;
	lowner->lo_seq_num = 0;


	lowner->lo_next   = nfs4_lock_owners[hash];
	nfs4_lock_owners[hash] = lowner;

/* look at lockd garage collecting
   if (++nrhosts > NLM_HOST_MAX)
   next_gc = 0;
*/

 out:
	up(&nfs4_lowner_sema);
	nfsv4_printk(lock,"NFSV4: nfs4_get_lockowner returning lowner %p\n",lowner);
	return lowner;
}

int
save_stateid(struct file *filp, stateid4 stateid)
{
	struct nfs4_file *f_data;

	nfsv4_printk(lock, "NFSV4: save_stateid: stateid [0x%lx , %08lx]\n",
		     (unsigned long)(stateid >> 32),
		     (unsigned long)(stateid & 0xFFFFFFFFUL));

	if (!filp)
		return -EINVAL;
	f_data = kmalloc(sizeof(struct nfs4_file), GFP_KERNEL);
	if(!f_data)
		return -ENOMEM;
	f_data->f_magic = NFS4_FILE_MAGIC;
	f_data->f_stateid = stateid;
	filp->private_data = f_data;
	nfsv4_printk(lock, "NFSV4: save_stateid: fdata->f_stateid [0x%lx , %08lx]\n",
		     (unsigned long)(f_data->f_stateid >> 32),
		     (unsigned long)(f_data->f_stateid & 0xFFFFFFFFUL));
	return 0;
}

void
release_stateid(struct file *filp)
{
	if(nfs4_check_file_stateid(filp)){
		kfree(filp->private_data);
		filp->private_data = NULL;
	}else
		nfsv4_printk(lock,KERN_ERR "NFSV4: file lock state corrupted!\n");
}

