/*
 *  linux/include/fs/nfs/nfs4util.c
 *
 *  nfs version 4 protocol utility funcitons
 *
 *  by Andy Adamson <andros@citi.umich.edu>
 *  Copyright (C) 1999 The Regents of the University of Michigan
 *
 */

#ifdef __linux__
#include <linux/param.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/utsname.h>
#include <linux/errno.h>
#include <net/sock.h>
#include <linux/string.h>
#include <linux/in.h>
#include <linux/pagemap.h>
#include <linux/sunrpc/xprt.h>
#include <linux/nfs.h>
#include <linux/nfs4_prot.h>
#include <linux/nfs_fs.h>
#include <linux/n2s_cache.h>
#include <linux/sunrpc/svc.h>
#include <linux/sunrpc/clnt.h>

#include <asm/segment.h>
#endif

#ifdef __OpenBSD__
#define USE_LINUX_FATTR
#include <sys/param.h>
#include <sys/conf.h>
#include <sys/ioctl.h>
#include <sys/signal.h>
#include <sys/proc.h>
#include <sys/namei.h>
#include <sys/vnode.h>
#include <sys/kernel.h>
#include <sys/mount.h>
#include <sys/buf.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/systm.h>
#include <sys/stat.h>

#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>

#include <nfs/rpcv2.h>
#include <nfs/nfsproto.h>
#include <nfs/nfsnode.h>
#include <nfs/nfs.h>
#include <nfs/nfsmount.h>
#include <nfs/xdr_subs.h>
#include <nfs/nfsm_subs.h>
#include <nfs/nfsdiskless.h>
#include <nfs/nqnfs.h>
#include <nfs/nfs_var.h>
#include <nfs/nfs4_prot.h>

extern char *strtok(char *s, const char *delim);
#endif

extern unsigned int bit4_mask[];
extern verifier4 nfs_verf;

struct getattr_ret {
	fattr4_type         type;
	fattr4_size         size;
	fattr4_fsid				  fsid;
	fattr4_fileid       fileid;
	fattr4_maxname      maxname;
	fattr4_maxread      maxread;
	fattr4_maxwrite     maxwrite;
	fattr4_mode         mode;
	fattr4_numlinks     numlinks;
	fattr4_rawdev       rawdev;
	fattr4_space_used   space_used;
	fattr4_time_access  atime;
	fattr4_time_create  ctime;
	fattr4_time_modify  mtime;
};

static struct {
	unsigned int    mode;
	nfs_ftype4      nfstype;
} nfs_type4fmt[] = {
	{ 0,              NF4NON   },
	{ S_IFREG,        NF4REG   },
	{ S_IFDIR,        NF4DIR   },
	{ S_IFBLK,        NF4BLK   },
	{ S_IFCHR,        NF4CHR   },
	{ S_IFLNK,        NF4LNK   },
	{ S_IFSOCK,       NF4SOCK  },
	{ S_IFIFO,        NF4FIFO  },
	{ 0,              NF4BAD   }
};


/* 
 * printk's for debugging.
 */


/* nfsv4_printk toggles*/
int nfsv4_debug_attributes      = 0;    /* fattr4, including xdr */
int nfsv4_debug_fh4             = 1;    /* file handle, including xdr */
int nfsv4_debug_utf8            = 1;    /* strings , including xdr*/
int nfsv4_debug_compound        = 1;    /* compound call */
int nfsv4_debug_xdr             = 1;    /* generic xdr: u64,bitmap4,specdata4 */
int nfsv4_debug_v4tov3          = 1;    /* stuff v3 structs from v4*/
int nfsv4_debug_inode           = 1;    /* generic inode  (inode.c)*/
int nfsv4_debug_svc             = 1;    /* generic svc  (nfssvc.c)*/
int nfsv4_debug_dir             = 1;    /* readdir  (dir.c)*/
int nfsv4_debug_vv3             = 1;    /* translation functions */
int nfsv4_debug_write           = 0;    /* write.c */
int nfsv4_debug_read            = 0;    /* read.c */
int nfsv4_debug_readdir         = 0;    /* READDIR */
int nfsv4_debug_free            = 0;    /* nfs4_util.c */
int nfsv4_debug_open            = 0;    /* OPEN call (file.c) */
int nfsv4_debug_create          = 0;    /* nfs_create  */
int nfsv4_debug_close           = 0;    /* CLOSE call (file.c) */
int nfsv4_debug_rename          = 0;    /* RENAME call (dir.c) */
int nfsv4_debug_n2s             = 0;    /* num2str driver */
int nfsv4_debug_secinfo         = 1;    /* SECINFO4 */
int nfsv4_debug_fileops         = 0;    /* uimplemented file.c operations */
int nfsv4_debug_bsize           = 0;    /* xdr bsize calculation */
int nfsv4_debug_setclntid       = 1;    /* SETCLIENTID */
int nfsv4_debug_setattr         = 0;    /* SETATTR */
int nfsv4_debug_lock            = 1;    
int nfsv4_debug_verifier        = 1;
int nfsv4_debug_andros          = 1;
int nfsv4_debug_kendrick        = 0;
int nfsv4_debug_callback        = 0;
int nfsv4_debug_deleg           = 0;

#ifndef __linux__
static u32
nfs4_time(void)
{
	struct timeval tv;

	microtime(&tv);
	return tv.tv_sec;
}
#endif

void
print_nfs_fh4(nfs_fh4 *fh4)
{
	char *kptr= fh4->nfs_fh4_val;
	int i;

	nfsv4_printk(fh4,"NFSV4 print_nfs_fh4 ");
	nfsv4_printk(fh4,"nfs_fh4_val: 0x%x\n",(unsigned int)kptr);
	if(fh4->nfs_fh4_len <= NFS4_FHSIZE){
		for(i=0;i<fh4->nfs_fh4_len;i++){
			nfsv4_printk(fh4,"%x",*kptr);
			kptr++;
		}
		nfsv4_printk(fh4,"\n");
	}else
		nfsv4_printk(fh4," nfs_fh4_len too big: %d\n",fh4->nfs_fh4_len);
}

#ifndef __OpenBSD__
void
print_nfs_fh(struct nfs_fh *fh)
{
	char *kptr= fh->data;
	int i;

	nfsv4_printk(fh4,"NFSV4 print_nfs_fh ");
	if(fh->size <= NFS4_FHSIZE){
		for(i=0;i<fh->size;i++){
			nfsv4_printk(fh4,"%x",*kptr);
			kptr++;
		}
		nfsv4_printk(fh4,"\n");
	}else
		nfsv4_printk(fh4," size too big: %d\n",fh->size);
}
void
print_fh(struct nfs_fh *fh)
{
	char *ptr = fh->data;
	int i;

	for(i=0;i<fh->size;i++){
		nfsv4_printk(fh4,"%x",*ptr);
		ptr++;
	}
	nfsv4_printk(fh4,"\n");
}

void
print_rpcsec_gss_info(rpcsec_gss_info *ptr)
{
	int i;
	char *p;

	nfsv4_printk(secinfo,"NFSV4: print_rpcsec_gss_info\n");
	/*secinfo4*/
	nfsv4_printk(secinfo,
		     "NFSV4: sec_info4 len: %d\n sec_info4 val:",ptr->oid.sec_oid4_len);
	p = ptr->oid.sec_oid4_val;
	for(i=0;i<ptr->oid.sec_oid4_len;i++)
		nfsv4_printk(secinfo,"%c",*p++);
	nfsv4_printk(secinfo,"NFSV4: qop: %d\n",ptr->qop);
	nfsv4_printk(secinfo,"NFSV4: service: %d\n",ptr->service);
}

u32 *
set_entry_attrs(struct nfs_entry *entry,u32 *p)
{
	int i,j;
	int fattrslen=0,error = 0;
	int bit;
	bitmap4 bm4;

	bm4.BLEN = *p++;
	bm4.BVAL = p;
	p += bm4.BLEN;
			
	fattrslen = *p++;	
	nfsv4_printk(readdir,"NFSV4:set_entry_attrs fattrslen %d\n",fattrslen);
	/* for each bit set in the bit mask */
	for(i=0;i<bm4.BLEN;i++){
		for(j=0;j<32;j++){
			/* if bit is set*/
			if(bm4.BVAL[i] & bit4_mask[j]){
				bit = (i*32)+j;
				nfsv4_printk(readdir,"NFSV4:set_entry_attrs bit %d is set\n",bit);
				switch(bit){
				case FATTR4_FILEID:
					entry->ino = *(fattr4_fileid *)p;
					p +=2;
					nfsv4_printk(readdir,"NFSV4:set_entry_attrs fileid %Ld\n",entry->ino);
					break;
				case FATTR4_TIME_MODIFY:
					entry->fattr.mtime = *(int64_t *)p;
					p += 2;
					p++; /* move past nseconds */
					nfsv4_printk(readdir,
						     "NFSV4:set_entry_attrs mtime %Ld p %p\n",entry->fattr.mtime,p);
					break;
				default:
 					printk("NFSV4: set_entry_attrs invalid case %d returning -NFS4ERR_EINVAL\n",bit);
					return ERR_PTR(-EINVAL);
					break;
				}
			}
		}
	}
	entry->fattr.valid = 1;
	nfsv4_printk(readdir,"NFSV4 set_entry_attrs returning %d \n",error);
	return p;
}
#endif /* OpenBSD */

int
utf8cmp(utf8string *s1,utf8string *s2)
{
	nfsv4_printk(utf8,"NFSDV4 utf8cmp\n");
	print_utf8string(s1);
	print_utf8string(s2);
	return strncmp(s1->utf8string_val,s2->utf8string_val,s1->utf8string_len);
}

void
utf8copy(utf8string *dest, utf8string *src)
{
	dest->utf8string_len = src->utf8string_len;
	if(dest->utf8string_len > 0){
		dest->utf8string_val = kmalloc(dest->utf8string_len,GFP_KERNEL);
		if(dest->utf8string_val != NULL)
			strncpy(dest->utf8string_val,src->utf8string_val,dest->utf8string_len);
	}
}

void
verf_init(void)
{
#ifdef __linux__
	time_t boot_time = CURRENT_TIME;

	/* stuff the 4 bytes of time into the first 4 bytes
	 * of the nfs4 verifier
	 */
	memset(nfs_verf,0,NFS4_VERIFIER_SIZE);
	*(time_t *)nfs_verf = boot_time;
#else
	microtime((struct timeval *) &nfs_verf);
#endif
	print_verifier4(nfs_verf);
}

void
set_verifier(verifier4 target,verifier4 src)
{
	int i;

	if(!src)
		for(i=0;i<NFS4_VERIFIER_SIZE;i++)
			target[i] = 0;
	else
		for(i=0;i<NFS4_VERIFIER_SIZE;i++)
			target[i] = src[i];
}
void
print_verifier4(verifier4 v1)
{
	nfsv4_printk(verifier,"NFSV4 nfs_verf:  ");
	nfsv4_printk(verifier,"%x\n",*(unsigned int *)&v1[0]);
	nfsv4_printk(verifier,"%x\n",*(unsigned int *)&v1[4]);
}
/* boolean compare */
int
cmp_verifier(verifier4 v1, verifier4 v2)
{
	int i;

	for(i=0;i<NFS4_VERIFIER_SIZE;i++)
		if(v1[i] != v2[i])
			return 0;
	return 1;
}

/* 
 * these calls kfree all kmalloc'ed storage used in the
 * compound call and operations. it is assumed unless otherwise
 * stated that if a pointer exists in a compound call and is 
 * not NULL, it has been kmalloc'ed and should therefore be freed.
 */
/* kfree non NULL bitmap4 values, and set pointer to NULL */
void
nfs4_kfree_bitmap4(bitmap4 *bm4)
{
	if(bm4->BVAL){
		kfree(bm4->BVAL);
		bm4->BVAL = NULL;
	}
}

/* kfree non NULL nfs_fh4 values, and set pointer to NULL */
void
nfs4_kfree_fh4(nfs_fh4 *fh4)
{
	if(fh4->nfs_fh4_val){
		kfree(fh4->nfs_fh4_val);
		fh4->nfs_fh4_val = NULL;
	}
}

/* kfree non NULL utf8string values, and set pointer to NULL */
void
nfs4_kfree_utf8(utf8string *utf8)
{
	if(utf8->utf8string_val){
		kfree(utf8->utf8string_val);
		utf8->utf8string_val = NULL;
	}
}

/* walk the compound4 args, kfree any pointers that are not NULL
 * then set them to NULL.
 */
void
nfs4_kfree_compound_args(COMPOUND4args *args)
{
	int i,j;

	nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_args\n");
	for(i=0;i<args->OPCNT;i++){
		switch(args->OPVAL[i].argop){ 
		case OP_ACCESS:
				/* nothing to do */
			break;
		case OP_CREATE:
			/* nothing to do */
			/* NOTE: in the case of a symlink, the symlink name is NOT allocated
			   on the free store, so no call to kfree() is needed.
			   (see nfs_symlink() in dir.c) */
			break;
		case OP_DELEGRETURN:
				/* nothing to do */
			break;
		case OP_GETATTR:
		{
			GETATTR4args *argp= &args->OPVAL[i].OPU.opgetattr;
	
			nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_args GETATTR\n");
			nfs4_kfree_bitmap4(&argp->attr_request);
		}
		break; 
		case OP_GETFH:
			nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_args GETFH\n");
				/* nothing to do */
			break;
		case OP_LINK:
		{
			LINK4args *argp = &args->OPVAL[i].OPU.oplink;

			nfsv4_printk(free, "NFSV4 nfs4_kfree_compound_args LINK\n");
			if (argp->newname.utf8string_val){
				kfree(argp->newname.utf8string_val);
				argp->newname.utf8string_val = NULL;
			}
		}
		break;
		case OP_LOOKUP:
		{
			LOOKUP4args *argp= &args->OPVAL[i].OPU.oplookup;

			nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_args LOOKUP\n");
				/* assume PLEN = # of kmalloc'ed PVAL's */
			for(j=0;j<argp->PLEN;j++)
				nfs4_kfree_utf8(&argp->PVAL[j]);
			if(argp->PVAL){	
				kfree(argp->PVAL);	
				argp->PVAL = NULL;
			}
		}
		break;
		case OP_LOOKUPP:
			/* nothing to do */
			break;
		case OP_OPEN:
		{
			OPEN4args *argp= &args->OPVAL[i].OPU.opopen;

			nfsv4_printk(free, "NFSV4 nfs4_kfree_compound_args OPEN\n");
			for(j=0;j<argp->claim.CLAIM_PATH_LEN;j++)
				nfs4_kfree_utf8(&argp->claim.CLAIM_PATH_VAL[j]);
			if(argp->claim.CLAIM_PATH_VAL){	
				kfree(argp->claim.CLAIM_PATH_VAL);	
				argp->claim.CLAIM_PATH_VAL = NULL;
			}
			switch(argp->openhow.opentype){
			case OPEN4_NOCREATE:
				break;
			case OPEN4_CREATE:
				if(argp->openhow.HOW.CREATE_ATTR.attrmask.BVAL)
					kfree(argp->openhow.HOW.CREATE_ATTR.attrmask.BVAL);
				if(argp->openhow.HOW.CREATE_ATTR.FATTRSVAL)
					kfree(argp->openhow.HOW.CREATE_ATTR.FATTRSVAL);
				break;
			}

			if (argp->OWNER_VAL){
				kfree(argp->OWNER_VAL);
				argp->OWNER_VAL = NULL;
			}
			nfsv4_printk(free, "NFSV4 nfs4_kree_compound_args OPEN done\n");
		}
		break;
		case OP_PUTFH:
		{
			PUTFH4args *argp= &args->OPVAL[i].OPU.opputfh;

			nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_args PUTFH\n");
			nfs4_kfree_fh4(&argp->object);
		}
		break;
		case OP_PUTROOTFH:
			nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_args GETFH - noop\n");
				/* nothing to do */
			break;
		case OP_READ:
			nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_args READ noop\n");
				/* nothing to do */
			break;
		case OP_READDIR:
		{
			READDIR4args *argp= &args->OPVAL[i].OPU.opreaddir;

			nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_args READDIR\n");
			nfs4_kfree_bitmap4(&argp->attr_request);
		}
		break;
		case OP_READLINK:
			// nothing to do
				break;
		case OP_RENAME:
		{
			RENAME4args *argp =&args->OPVAL[i].OPU.oprename;

			nfsv4_printk(free, "NFSV4 nfs4_kfree_compound_args RENAME\n");
			if (argp->OLDNAME_VAL){
				kfree(argp->OLDNAME_VAL);
				argp->OLDNAME_VAL = NULL;
			}
			if (argp->NEWNAME_VAL){
				kfree(argp->NEWNAME_VAL);
				argp->NEWNAME_VAL = NULL;
			}
		}
		break;
		case OP_RESTOREFH:
			nfsv4_printk (free, "NFSV4 nfs4_kfree_compound_args RESTOREFH\n");
			/* nothing to do */
			break;
		case OP_SAVEFH:
			nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_args SAVEFH\n");
			/* nothing to do */
			break;
		case OP_SETATTR:
		{
			fattr4 *fattrp = &args->OPVAL[i].OPU.opsetattr.obj_attributes;
			nfsv4_printk (free, "NFSV4 nfs4_kfree_compound_args SETATTR\n");
			kfree (fattrp->attrmask.bitmap4_val);
			kfree (fattrp->attr_vals.attrlist4_val);
		}
		break;
		case OP_WRITE:
			/* nothing to do, use sunrpc return iovec's 
			 * for argp->WRITEVAL.
			 */
			nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_args WRITE - noop\n");
			break;
		default:
			nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_args DEFAULT - invalid op %d\n",args->OPVAL[i].argop);
			break;
		} /*switch*/
	}
	nfsv4_printk(free, "NFSV4 nfs4_kree_compound_args done\n");
}


/* walk the compound4 res, kfree any pointers that are not NULL
 * then set them to NULL.
 */
void
nfs4_kfree_compound_res(COMPOUND4res *res)
{
	int i;
	nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_res\n");
	for(i=0;i<res->RESCNT;i++){
		switch(res->RESVAL[i].resop){ 
			case OP_ACCESS:
			/* nothing to do */
			break;
			case OP_DELEGRETURN:
			/* nothing to do */
			break;
			case OP_GETATTR:
			{
			GETATTR4res *resp= &res->RESVAL[i].RESU.opgetattr;

				nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_res GETATTR1\n");
				nfs4_kfree_bitmap4(&resp->GFATTR4.attrmask);
				nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_res GETATTR2\n");
				if(resp->GFATTR4.FATTRSVAL){
					kfree(resp->GFATTR4.FATTRSVAL);
					resp->GFATTR4.FATTRSVAL = NULL;
				}
			}
			break; 
			case OP_GETFH:
				/* nothing to do, now that kmalloc has been removed from xdr decode */
				break;
			case OP_LINK:
			/* nothing to do */
			break;
			case OP_LOOKUP:
			/* nothing to do */
			break;
		        case OP_LOOKUPP:
			  /* nothing to do */
			break;
			case OP_OPEN:
			/* nothing to do */
			break;
			case OP_PUTFH:
			/* nothing to do */
			break;
			case OP_PUTROOTFH:
			/* nothing to do */
			break;
			case OP_READ:
			/* nothing to do - use VFS read page cache and 
			* sunrpc return iovec's for resp->READVAL
			*/
			break;
			case OP_READDIR:
			{
			READDIR4res *resp= &res->RESVAL[i].RESU.opreaddir;
	
				nfsv4_printk(free,"NFSV4 nfs4_kfree_compound_res READDIR\n");
				if(resp->ENTRIES){
					kfree(resp->ENTRIES);
					resp->ENTRIES = NULL;
				}
			}
			break;
		        case OP_READLINK:
			  // nothing to do
			  // NOTE: the link buffer is not allocated on the free store, so it doesn't
			  // need a kfree()
			break;
			case OP_RENAME:
			/* nothing to do */
			break;
		        case OP_RESTOREFH:
			  /* nothing to do */
			break;
			case OP_SAVEFH:
			/* nothing to do */
			break;
		        case OP_SETATTR:
			  kfree (res->RESVAL[i].RESU.opsetattr.attrsset.bitmap4_val);
			break;
			case OP_WRITE:
			/* nothing to do */
			break;
			default:
			nfsv4_printk(free,
				"NFSV4 nfs4_kfree_compound_res DEFAULT - invalid op %d\n",
				res->RESVAL[i].resop);
			break;
		}
	}
	nfsv4_printk(free, "NFSV4 nfs4_kree_compound_res done\n");
}


/* 
 * these calls calculate the size of the xdr rpc arg buf in words
 */

int
get_compound_arg_size(char *str)
{
int size = 0;

	/* tag */
	/* utf8string_len */
	size += 1;
	/* utf8string_val */
	size += XDR_QUADLEN(strlen(str));

	/* minorversion */
	size += 1;
	/* argarray_len*/
	size += 1;
	nfsv4_printk(compound,"get_compound_size: size %d\n",size);
	return size;
}

int
get_pathname4_size(char *path)
{
				int size = 0;
				char *cp;
				int i;

				/* pathname4_len */
				size += 1;

				/* pathname4_val */
				i = 0;
				cp = strtok(path, "/");
				while (cp) {
								/* utf8string_len */
								size += 1; 
								/* utf8string_val */
								size += XDR_QUADLEN(strlen(cp));
								i++;
								cp = strtok(NULL, "/");
				}

				nfsv4_printk(compound,"get_pathname4_size: size %d\n",size);
				return size;                                                  
}

int
get_putrootfh_arg_size(void)
{
				int size = 0;
				/* nfs_opnum4 */	
				size += 1;
				return(size);
}

int
get_getfh_arg_size(void)
{
				int size = 0;
				/* nfs_opnum4 */	
				size += 1;
				return(size);
}

int
get_lookup_arg_size(char *path)
{
				int size = 0;
				/* nfs_opnum4 */	
				size += 1;

				size += get_pathname4_size(path);
				return(size);
}

int
get_getattr4_size(int len)
{
				int size = 0;
				/* nfs_opnum4 */	
				size += 1;

				/* bitmap4_len*/
				size += 1;
				/* bitmap4_val*/
				size += len;
				return size;
}


/* set pointers */

void
setup_compound(int opcnt,COMPOUND4args *cargs,COMPOUND4res *cres, char *tag,nfs_argop4 *aop4,nfs_resop4 *rop4)
{
	int size = 0;
	cargs->tag.utf8string_len = strlen(tag);
	cargs->tag.utf8string_val = tag;
	cargs->OPCNT = opcnt;
	cargs->minorversion = 0;

	if (nfsv4_debug_compound)
		print_utf8string(&cargs->tag);

	cargs->OPVAL = aop4;
	size = opcnt*sizeof(struct nfs_argop4);
	memset(cargs->OPVAL,0,size);

	/* don't care about res tag. */
	cres->tag.utf8string_len = 0;
	cres->tag.utf8string_val = NULL;
	cres->RESCNT = opcnt;

	cres->RESVAL = rop4;
	size = opcnt*sizeof(struct nfs_resop4);
	memset(cres->RESVAL,0,size);
}                                              

int
setup_bitmap4(bitmap4 *bm,int len)
{
	int err = 0;
	bm->BLEN = len;
	bm->BVAL = kmalloc(len * sizeof(__u32),GFP_KERNEL);
	if(bm->BVAL){
		memset(bm->BVAL,0,len * sizeof(__u32));
	} else
		err = NFS4ERR_RESOURCE;
	return err;
}

void
nfs4_setup_readdir (READDIR4args *args, READDIR4res *res, nfs_cookie4 cookie, bitmap4 *req, struct iovec *iov)
{
	int bmlen = req->bitmap4_len;

	memset (args, 0, sizeof(*args));
	args->cookie = cookie;
	args->dircount = 0;
	args->maxcount = PAGE_SIZE;
	args->attr_request.bitmap4_len = bmlen;
	args->attr_request.bitmap4_val = (uint32_t *) kmalloc (4 * bmlen, GFP_KERNEL);
	memcpy (args->attr_request.bitmap4_val, req->bitmap4_val, 4 * bmlen);

	res->ENTRIES = 0;
#ifdef __linux__
	memcpy (&args->iov[1], iov, sizeof(*iov));
#endif
}

void
setup_readdir_res(READDIR4args *args,READDIR4res *res)
{
	res->ENTRIES = kmalloc(args->dircount,GFP_KERNEL);
	memset(res->ENTRIES,0,args->dircount); 
}

void
stuffutf8(utf8string *utf8, char *str)
{
	utf8->utf8string_len = strlen(str);
	if(utf8->utf8string_len > 0){
		utf8->utf8string_val = kmalloc(utf8->utf8string_len,GFP_KERNEL);
		memcpy(utf8->utf8string_val,str,utf8->utf8string_len);
	}
}

/* given a string, stuff compound lookup arg */
void
str_to_lookup4(char *s, int len, LOOKUP4args *args)
{
	utf8string *str;

	/* Allocate component */
	args->PLEN = 1;
	args->PVAL = (component4 *) kmalloc(sizeof(component4), GFP_KERNEL);
	str = &args->PVAL[0];

	/* Copy string */
	str->utf8string_len = len;
	str->utf8string_val = kmalloc(len * sizeof(char), GFP_KERNEL);
	memcpy(str->utf8string_val, s, len);
}

#ifndef __OpenBSD__
/* given a dentry qstr, stuff the pathname4 struct */
void
qstr_to_utf8(struct qstr *dname, utf8string *str)
{
	str->utf8string_len = dname->len;
	str->utf8string_val = kmalloc(dname->len * sizeof(char),GFP_KERNEL);
	memcpy(str->utf8string_val,dname->name,dname->len);
}

/* given a dentry qstr, stuff compound lookup args. */
void
qstr_to_lookup4(struct qstr *dname, LOOKUP4args *args)
{
/*	str_to_lookup4(dname->name, dname->len, args); not tested but should work */
	args->PLEN = 1;
	args->PVAL = (component4 *) kmalloc(args->PLEN * sizeof(component4), GFP_KERNEL);
	qstr_to_utf8(dname, &args->PVAL[0]);
}
#endif

/*
 * given a filesystem path, stuff compound lookup args.
 * NOTE: not utf8 yet - assume ascii mnt_path.
 */
int
parse_mnt_path(char *mnt_path, LOOKUP4args *args)
{
	char *cp,*next;
	int i,len,flen=0;

	/* common case - mount (pseudo) root */
	len = strlen(mnt_path);
	if(!strcmp(mnt_path,"/")){
		args->PLEN = 1;
		args->PVAL =(component4 *)kmalloc(sizeof(component4),GFP_KERNEL);
		memset(args->PVAL,0,flen * sizeof(component4));
		args->PVAL[0].utf8string_val = kmalloc(sizeof(char),GFP_KERNEL);
		memset(args->PVAL[0].utf8string_val,0,len);
		memcpy(args->PVAL[0].utf8string_val,"/",strlen("/"));
		args->PVAL[0].utf8string_len = strlen("/");
		print_utf8string(&args->PVAL[0]);
		return 0;
	}
	/* count the path components */
	cp=mnt_path;
	for(i=0;i<len;i++){
		next = (cp + 1);
		if(*cp == '/') {
			if((*next == '/') || (*next == '\0'))
				continue;
			flen++;
		}
		cp++;
	}
	i = 0;
	args->PLEN = flen;
	args->PVAL =(component4 *)kmalloc(flen * sizeof(component4),GFP_KERNEL);
	memset(args->PVAL,0,flen * sizeof(component4));

	cp = strtok(mnt_path, "/");
	while (cp) {
		/* stuff the filename4 value*/
		len = strlen(cp);
		args->PVAL[i].utf8string_len = len;
		args->PVAL[i].utf8string_val = kmalloc(len * sizeof(char),GFP_KERNEL);
		memset(args->PVAL[i].utf8string_val,0,len);
		memcpy(args->PVAL[i].utf8string_val,cp,len);
		print_utf8string(&args->PVAL[i]);
		i++;
		cp = strtok(NULL, "/");
	}

	return 0;
}




int
set_bitmap4_bit(int b,bitmap4 *bm)
{
	int index,which_bit;

	index = (b/32);
	which_bit = b % 32;
	if(index >= bm->BLEN)
		return -1;
	bm->BVAL[index] |= bit4_mask[which_bit];
	return 0;
}


#ifndef __OpenBSD__
/* 
 * these calls use nfsv4 compound call results 
 * to stuff v3 result structures. 
 */

/* NOTE: assumed both v4 and v3_fh have storage */
int
nfs4_stuff_v3fh(nfs_fh4 *v4_fh,struct nfs_fh *v3_fh)
{
	memset(v3_fh->data,0,NFS_MAXFHSIZE);
	v3_fh->size = v4_fh->nfs_fh4_len;
	memcpy(v3_fh->data,v4_fh->nfs_fh4_val,v3_fh->size);
	return 0;
}

int
nfs4_stuff_v3dentry(nfs_fh4 *v4_fh,struct dentry *dentry)
{
	struct nfs_dentry       *d_fsdata = NULL;
	struct nfs_fh           *v3_fh;

	nfsv4_printk(fh4,"NFSV4 nfs4_stuff_v3fh for %s\n",dentry->d_name.name);
	if(v4_fh->nfs_fh4_len > NFS_MAXFHSIZE){
		nfsv4_printk(fh4,"NFSV4 nfs4_stuff_v3fh: v4 len %d too big.\n",v4_fh->nfs_fh4_len);
		return -ENOMEM;
	}
	d_fsdata = kmalloc(sizeof(*d_fsdata),GFP_KERNEL);
	if(!d_fsdata){
		nfsv4_printk(fh4,"NFSV4 nfs4_stuff_v3fh: kmalloc failed\n");
		return -ENOMEM;
	}
	memset(d_fsdata,0,sizeof(*d_fsdata));
	v3_fh = &d_fsdata->fh;
	v3_fh->size = v4_fh->nfs_fh4_len;
	memcpy(v3_fh->data,v4_fh->nfs_fh4_val,v3_fh->size);
	dentry->d_fsdata = d_fsdata;
	nfsv4_printk(fh4,"NFSV4 nfs4_stuff_v3fh: NFS_FH(dentry):\n");
	print_nfs_fh(NFS_FH(dentry));
	return 0;
}
#endif /* OpenBSD */

static void
init_nfsfsinfo(struct nfs_fsinfo *fsinfo)
{
/* default values in case server fails to return
 * requested GETATTR attributes
 * totally wild guess....
*/

	fsinfo->rtmax = 1024;
	fsinfo->wtmax = 1024;
	fsinfo->namelen = 32;
	fsinfo->bsize = 0;
}

void nfs4_stuff_nfsfsinfo(fattr4 *fptr,struct nfs_fsinfo *fsinfo)
{
	int i,j;
	int bit;
	bitmap4 *bm4 = &fptr->attrmask;
	char *cptr = fptr->FATTRSVAL;

	nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfsinfo. flen %d\n",fptr->FATTRSLEN);
	init_nfsfsinfo (fsinfo);

	if(!bm4)
		return;

	/* for each bit set in the bit mask */
	for(i=0;i<bm4->BLEN;i++){
		for(j=0;j<32;j++){
			/* if bit is set*/
			if(bm4->BVAL[i] & bit4_mask[j]){
				bit = (i*32)+j;
				switch (bit) {
				case FATTR4_SUPPORTED_ATTRS:
				{
					bitmap4 *b4p = (bitmap4 *) cptr;
					int size = sizeof(u_int) + (b4p->BLEN * sizeof(uint32_t));    
					cptr += size;
				}
				break;
				case FATTR4_TYPE:
					cptr += sizeof(fattr4_type);
					break;
				case FATTR4_FH_EXPIRE_TYPE:
					cptr += sizeof(fattr4_fh_expire_type); 
					break;
				case FATTR4_CHANGE:
					cptr += sizeof(fattr4_change);
					break;
				case FATTR4_SIZE:
					cptr += sizeof(fattr4_size);
					break;
				case FATTR4_LINK_SUPPORT:
					cptr += sizeof(fattr4_link_support);
					break;
				case FATTR4_SYMLINK_SUPPORT:
					cptr += sizeof(fattr4_symlink_support);
					break;
				case FATTR4_NAMED_ATTR:
					cptr += sizeof(fattr4_named_attr);
					break;
				case FATTR4_FSID:
					cptr += sizeof(fattr4_fsid);
					break;
				case FATTR4_UNIQUE_HANDLES:
					cptr += sizeof(fattr4_unique_handles);
					break;
				case FATTR4_LEASE_TIME:
					cptr += sizeof(fattr4_lease_time);
					break;
				case FATTR4_RDATTR_ERROR:
					cptr += sizeof(fattr4_rdattr_error);
					break;

					/*
					 *  need to code FATTR4_ACL
					 */

				case  FATTR4_ARCHIVE:
					cptr += sizeof(fattr4_archive);
					break;
				case  FATTR4_CANSETTIME:
					cptr += sizeof(fattr4_cansettime);
					break;
				case  FATTR4_CASE_INSENSITIVE:
					cptr += sizeof(fattr4_case_insensitive);
					break;
				case  FATTR4_CASE_PRESERVING:
					cptr += sizeof(fattr4_case_preserving);
					break;
				case  FATTR4_CHOWN_RESTRICTED:
					cptr += sizeof(fattr4_chown_restricted);
					break;
				case  FATTR4_FILEHANDLE:
				{
					int len = *(int *)cptr;
					cptr += sizeof(u_int) + len; 
				}
				break;
				case FATTR4_FILEID:
					cptr += sizeof(fattr4_fileid);
					break;
				case  FATTR4_FILES_AVAIL:
					cptr += sizeof(fattr4_files_avail);
					break;
				case  FATTR4_FILES_FREE:
					fsinfo->ffiles = *((fattr4_files_free *) cptr);
					cptr += sizeof(fattr4_files_free);
					break;
				case  FATTR4_FILES_TOTAL:
					fsinfo->tfiles = *((fattr4_files_total *) cptr);
					cptr += sizeof(fattr4_files_total);
					break;

					/*
					 *  need to code FATTR4_FS_LOCATIONS
					 */

				case FATTR4_HIDDEN:
					cptr += sizeof(fattr4_hidden);
					break;
				case FATTR4_HOMOGENEOUS:
					cptr += sizeof(fattr4_homogeneous);
					break;
				case FATTR4_MAXFILESIZE:
					cptr += sizeof(fattr4_maxfilesize);
					break;
				case FATTR4_MAXLINK:
					cptr += sizeof(fattr4_maxlink);
					break;
				case FATTR4_MAXNAME:
					fsinfo->namelen = *(fattr4_maxname *)cptr;
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfsinfo FATTR4_MAXNAME %d\n",fsinfo->namelen);
					cptr += sizeof(fattr4_maxname);
					break;
				case FATTR4_MAXREAD:
					fsinfo->rtmax = *(fattr4_maxread *)cptr;
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfsinfo FATTR4_MAXREAD %d\n",fsinfo->rtmax);
					cptr += sizeof(fattr4_maxread);
					break;
				case FATTR4_MAXWRITE:
					fsinfo->wtmax = *(fattr4_maxwrite *)cptr;
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfsinfo FATTR4_MAXWRITE %d\n",fsinfo->wtmax);
					cptr += sizeof(fattr4_maxwrite);
					break;

					/*
					 *  need to code FATTR4_MIMETYPE
					 */

				case FATTR4_MODE:
					cptr += sizeof(fattr4_mode);
					break;
				case  FATTR4_NO_TRUNC:
					cptr += sizeof(fattr4_no_trunc);
					break;
				case FATTR4_NUMLINKS:
					cptr += sizeof(fattr4_numlinks);
					break;
				case FATTR4_OWNER:
				{
					fattr4_owner *str = (fattr4_owner *) cptr;
					cptr += sizeof(fattr4_owner) + str->utf8string_len;
				}
				break;
				case FATTR4_OWNER_GROUP:
				{
					fattr4_owner_group *str = (fattr4_owner_group *) cptr;
					cptr += sizeof(fattr4_owner_group) + str->utf8string_len;
				}
				break;
				case  FATTR4_QUOTA_HARD:
					cptr += sizeof(fattr4_quota_hard);
					break;
				case  FATTR4_QUOTA_SOFT:
					cptr += sizeof(fattr4_quota_soft);
					break;
				case  FATTR4_QUOTA_USED:
					cptr += sizeof(fattr4_quota_used);
					break;
				case FATTR4_RAWDEV:
					cptr += sizeof(fattr4_rawdev);
					break;
				case  FATTR4_SPACE_AVAIL:
					fsinfo->abytes = *((fattr4_space_avail *) cptr);
					cptr += sizeof(fattr4_space_avail);
					break;
				case  FATTR4_SPACE_FREE:
					fsinfo->fbytes = *((fattr4_space_free *) cptr);
					cptr += sizeof(fattr4_space_free);
					break;
				case  FATTR4_SPACE_TOTAL:
					fsinfo->tbytes = *((fattr4_space_total *) cptr);
					cptr += sizeof(fattr4_space_total);
					break;
				case FATTR4_SPACE_USED:
					cptr += sizeof(fattr4_space_used);
					break;
				case  FATTR4_SYSTEM:
					cptr += sizeof(fattr4_system);
					break;
				case FATTR4_TIME_ACCESS:
					cptr += sizeof(fattr4_time_access);
					break;
				case FATTR4_TIME_ACCESS_SET:
					cptr += sizeof(fattr4_time_access_set);     // should never occur
											       break;
				case FATTR4_TIME_BACKUP:
					cptr += sizeof(fattr4_time_backup);
					break;
				case FATTR4_TIME_CREATE:
					cptr += sizeof(fattr4_time_create);
					break;
				case FATTR4_TIME_DELTA:
					cptr += sizeof(fattr4_time_delta);
					break;
				case FATTR4_TIME_METADATA:
					cptr += sizeof(fattr4_time_metadata);
					break;
				case FATTR4_TIME_MODIFY:
					cptr += sizeof(fattr4_time_modify);
					break;
				case FATTR4_TIME_MODIFY_SET:
					cptr += sizeof(fattr4_time_modify_set);     // this case should never occur
											       break;
				default:
					printk ("!!!! default reached in stuff_nfsfsinfo() !!!! -- panicking\n");
					return;
				}
			}
		}
	}
}

static void
init_nfsfattr(struct nfs_fattr *fattr)
{
/* set default values in case server doesn't return
 * needed values - really a wild guess...
*/
	memset(fattr, 0, sizeof *fattr);
	fattr->mode = 0x8080;
	fattr->nlink = 1;
#ifdef __linux__
	fattr->du.nfs3.used = 1;
	fattr->mtime =  fattr->ctime = fattr->atime = CURRENT_TIME;
#else
	fattr->used = 1;
	fattr->mtime.nfsv3_sec = fattr->ctime.nfsv3_sec = fattr->atime.nfsv3_sec = nfs4_time();
#endif
}

#ifdef __linux__
#define FATTR_SETTIME(a, r) ((a) = (r)->seconds)
#else
#define FATTR_SETTIME(a, r) ((a).nfsv3_sec = (r)->seconds, (a).nfsv3_nsec = (r)->nseconds)
#endif

void nfs4_stuff_nfsfattr (fattr4 *fptr, struct nfs_fattr *fattr)
{
	int i,j;
	int bit;
	bitmap4 *bm4 = &fptr->attrmask;
	char *cptr = fptr->FATTRSVAL;
	int flen = fptr->FATTRSLEN;
	int fmode = 0;
  
	nfsv4_printk(v4tov3,"ANDROS! NFSV4 nfs4_stuff_nfsfatr. flen %d\n",flen);
	init_nfsfattr(fattr);

	if(!bm4)
		return;

	/* for each bit set in the bit mask */

	for(i=0;i<bm4->BLEN;i++) {
		for(j=0;j<32;j++) {
			if(bm4->BVAL[i] & bit4_mask[j]) {
				bit = (i*32)+j;
				switch(bit) {
				case FATTR4_SUPPORTED_ATTRS:
				{
					bitmap4 *b4p = (bitmap4 *) cptr;
					int size = sizeof(u_int) + (b4p->BLEN * sizeof(uint32_t));    
					cptr += size;
				}
				break;
				case FATTR4_TYPE:
				{
					u32 type = *(fattr4_type *)cptr;
					if(type > NF4BAD)
						type=NF4BAD;
					fattr->mode = fmode = nfs_type4fmt[type].mode;
					fattr->type =nfs_type4fmt[type].nfstype;
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfatr FATTR4_TYPE. fmode 0x%x fattr->mode 0x%x fattr->type 0x%x \n",fmode,fattr->mode,fattr->type);
					cptr += sizeof(fattr4_type);
				}
				break;
				case FATTR4_FH_EXPIRE_TYPE:
					cptr += sizeof(fattr4_fh_expire_type); 
					break;
				case FATTR4_CHANGE:
					cptr += sizeof(fattr4_change);
					break;
				case FATTR4_SIZE:
				{
					fattr->size = *(fattr4_size *)cptr;
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfatr FATTR4_SIZE \n");
					cptr += sizeof(fattr4_size);
				}
				break;
				case FATTR4_LINK_SUPPORT:
					cptr += sizeof(fattr4_link_support);
					break;
				case FATTR4_SYMLINK_SUPPORT:
					cptr += sizeof(fattr4_symlink_support);
					break;
				case FATTR4_NAMED_ATTR:
					cptr += sizeof(fattr4_named_attr);
					break;
				case FATTR4_FSID:
				{
					fattr4_fsid *ptr = (fattr4_fsid *)cptr;
					u32 *lptr = (u32 *)ptr;

					fattr->fsid = ptr->major;
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfatr FATTR4_FSID \n");
					lptr++;
					*lptr = ptr->minor;
					cptr += sizeof(fattr4_fsid);
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfatrFSID [0x%lx , %08lx]\n",
						     (unsigned long)(fattr->fsid >> 32),
						     (unsigned long)(fattr->fsid & 0xFFFFFFFFUL));
				}
				break;
				case  FATTR4_UNIQUE_HANDLES:
					cptr += sizeof(fattr4_unique_handles);
					break;
				case  FATTR4_LEASE_TIME:
					cptr += sizeof(fattr4_lease_time);
					break;
				case  FATTR4_RDATTR_ERROR:
					cptr += sizeof(fattr4_rdattr_error);
					break;

					/*
					 *  need to code FATTR4_ACL
					 */

				case  FATTR4_ARCHIVE:
					cptr += sizeof(fattr4_archive);
					break;
				case  FATTR4_CANSETTIME:
					cptr += sizeof(fattr4_cansettime);
					break;
				case  FATTR4_CASE_INSENSITIVE:
					cptr += sizeof(fattr4_case_insensitive);
					break;
				case  FATTR4_CASE_PRESERVING:
					cptr += sizeof(fattr4_case_preserving);
					break;
				case  FATTR4_CHOWN_RESTRICTED:
					cptr += sizeof(fattr4_chown_restricted);
					break;
				case  FATTR4_FILEHANDLE:
				{
					int len = (int)*(int *)cptr;
					cptr += sizeof(u_int) + len; 
				}
				break;
				case FATTR4_FILEID:
				{
					fattr->fileid =*(fattr4_fileid *)cptr;
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfatr FATTR4_FILEID %Ld\n",fattr->fileid);
					cptr += sizeof(fattr4_fileid);
				}
				break;
				case  FATTR4_FILES_AVAIL:
					cptr += sizeof(fattr4_files_avail);
					break;
				case  FATTR4_FILES_FREE:
					cptr += sizeof(fattr4_files_free);
					break;
				case  FATTR4_FILES_TOTAL:
					cptr += sizeof(fattr4_files_total);
					break;

					/*
					 *  need to code FATTR4_FS_LOCATIONS
					 */

				case FATTR4_HIDDEN:
					cptr += sizeof(fattr4_hidden);
					break;
				case FATTR4_HOMOGENEOUS:
					cptr += sizeof(fattr4_homogeneous);
					break;
				case FATTR4_MAXFILESIZE:
					cptr += sizeof(fattr4_maxfilesize);
					break;
				case FATTR4_MAXLINK:
					cptr += sizeof(fattr4_maxlink);
					break;
				case FATTR4_MAXNAME:
					cptr += sizeof(fattr4_maxname);
					break;
				case  FATTR4_MAXREAD:
					cptr += sizeof(fattr4_maxread);
					break;
				case  FATTR4_MAXWRITE:
					cptr += sizeof(fattr4_maxwrite);
					break;

					/*
					 *  need to code FATTR4_MIMETYPE
					 */

				case FATTR4_MODE:
				{
					u32 tmp = *(fattr4_mode *)cptr;
					fattr->mode = (*(fattr4_mode *)cptr & ~S_IFMT) | fmode;
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfatr FATTR4_MODE. wire_mode 0x%x fmode 0x%x fattr->mode 0x%x \n",tmp,fmode,fattr->mode);
					cptr += sizeof(fattr4_mode);
				}
				break;
				case  FATTR4_NO_TRUNC:
					cptr += sizeof(fattr4_no_trunc);
					break;
				case FATTR4_NUMLINKS:
				{
					fattr->nlink = *(fattr4_numlinks *)cptr;
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfatr FATTR4_NUMLINKS \n");
					cptr += sizeof(fattr4_numlinks);
				}
				break;
				case FATTR4_OWNER:
				{
					fattr4_owner *str = (fattr4_owner *) cptr;
#ifndef __OpenBSD__
					int uid;
					int err;
					int i;

					nfsv4_printk (kendrick, "stuff_nfsfattr() calling get_num() with user name=\"");
					for (i = 0; i < str->utf8string_len; i++)
						nfsv4_printk (kendrick, "%c", str->utf8string_val[i]);
					nfsv4_printk (kendrick, "\"\n");
					err = get_num (NUM2STR_UID, str, &uid);
					if (err) {
						nfsv4_printk (kendrick, "get_num() returned error\n");
						uid = 0;             // assume root if there is no response from cache
												     }
					else
						nfsv4_printk (kendrick, "get_num() returned uid %d\n", uid);

					fattr->uid = uid;
#else
					fattr->uid = 32767;
#endif
					cptr += sizeof(fattr4_owner) + str->utf8string_len;
				}
				break;
				case FATTR4_OWNER_GROUP:
				{
					fattr4_owner_group *str = (fattr4_owner_group *) cptr;
#ifndef __OpenBSD__
					int gid;
					int err;
					int i;

					nfsv4_printk (kendrick, "stuff_nfsfattr() calling get_num() with group name=\"");
					for (i = 0; i < str->utf8string_len; i++)
						nfsv4_printk (kendrick, "%c", str->utf8string_val[i]);
					nfsv4_printk (kendrick, "\"\n");
					err = get_num (NUM2STR_GID, str, &gid);
					if (err) {
						nfsv4_printk (kendrick, "get_num() returned error\n");
						gid = 0;             // assume root if there is no response from cache
												     }
					else
						nfsv4_printk (kendrick, "get_num() returned gid %d\n", gid);

					fattr->gid = gid;
#else
					fattr->gid = 39;
#endif
					cptr += sizeof(fattr4_owner_group) + str->utf8string_len;
				}
				break;
				case  FATTR4_QUOTA_HARD:
					cptr += sizeof(fattr4_quota_hard);
					break;
				case  FATTR4_QUOTA_SOFT:
					cptr += sizeof(fattr4_quota_soft);
					break;
				case  FATTR4_QUOTA_USED:
					cptr += sizeof(fattr4_quota_used);
					break;
				case FATTR4_RAWDEV:
				{
					fattr4_rawdev *ptr = (fattr4_rawdev *)cptr;
#ifdef __linux__
					fattr->rdev = (ptr->specdata1 << MINORBITS) | (ptr->specdata2 & MINORMASK);
#else
					fattr->rdev.specdata1 = ptr->specdata1;
					fattr->rdev.specdata2 = ptr->specdata2;
#endif
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfatr FATTR4_RAWDEV \n");
					cptr += sizeof(fattr4_rawdev);
				}
				break;
				case  FATTR4_SPACE_AVAIL:
					cptr += sizeof(fattr4_space_avail);
					break;
				case  FATTR4_SPACE_FREE:
					cptr += sizeof(fattr4_space_free);
					break;
				case  FATTR4_SPACE_TOTAL:
					cptr += sizeof(fattr4_space_total);
					break;
				case FATTR4_SPACE_USED:
				{
#ifdef __linux__
					fattr->du.nfs3.used =  *(fattr4_space_used *)cptr;
#else
					fattr->used = *(fattr4_space_used *)cptr;
#endif
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfatr FATTR4_SPACE_USED \n");
					cptr += sizeof(fattr4_space_used);
				}
				break;
				case  FATTR4_SYSTEM:
					cptr += sizeof(fattr4_system);
					break;
				case FATTR4_TIME_ACCESS:
				{
					fattr4_time_access *ptr = (fattr4_time_access *)cptr;
					FATTR_SETTIME(fattr->atime, ptr);
#ifdef __linux__
					nfsv4_printk(v4tov3,
						     "NFSV4 nfs4_stuff_nfsfatr FATTR4_TIME_ACCESS fattr->atime: [0x%lx , %08lx]\n",
						     (unsigned long)(fattr->atime >> 32),
						     (unsigned long)(fattr->atime  & 0xFFFFFFFFUL));
#endif

					cptr += sizeof(fattr4_time_access);
				}
				break;
				case FATTR4_TIME_ACCESS_SET:
					cptr += sizeof(fattr4_time_access_set);     // this case should never occur
											       break;

				case FATTR4_TIME_BACKUP:
					cptr += sizeof(fattr4_time_backup);
					break;
				case FATTR4_TIME_CREATE:
				{
					fattr4_time_create *ptr = (fattr4_time_create *)cptr;
					FATTR_SETTIME(fattr->ctime, ptr);
#ifdef __linux__
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfatr FATTR4_TIME_CREATE. fattr->ctime: [0x%lx , %08lx]\n",
						     (unsigned long)(fattr->ctime >> 32),
						     (unsigned long)(fattr->ctime  & 0xFFFFFFFFUL));
#endif
					cptr += sizeof(fattr4_time_create);
				}
				break;
				case FATTR4_TIME_DELTA:
					cptr += sizeof(fattr4_time_delta);
					break;
				case FATTR4_TIME_METADATA:
					cptr += sizeof(fattr4_time_metadata);
					break;
				case FATTR4_TIME_MODIFY:
				{
					fattr4_time_modify *ptr = (fattr4_time_modify *)cptr;
					FATTR_SETTIME(fattr->mtime, ptr);
#ifdef __linux__
					nfsv4_printk(v4tov3,"NFSV4 nfs4_stuff_nfsfatr FATTR4_TIME_MODIFY fattr->mtime: [0x%lx , %08lx]\n",
						     (unsigned long)(fattr->mtime >> 32),
						     (unsigned long)(fattr->mtime  & 0xFFFFFFFFUL));
#endif
					cptr += sizeof(fattr4_time_modify);
				}
				break;
				case FATTR4_TIME_MODIFY_SET:
					cptr += sizeof(fattr4_time_modify_set);     // this case should never occur
											       break;
				default:
					printk ("!! SERIOUS PROBLEM !!  -- default reached in stuff_nfsfattr().  panicking\n");
					return;
				}
			}
		}
	}

#ifdef __linux__
	/* update mode bits a la v3 */
	fattr->valid |= (NFS_ATTR_FATTR | NFS_ATTR_FATTR_V3);
#endif
}

/* GETATTR
 * the idea here is to do a v4 call to stuff the v3 fattr,
 * which is the result of the v3 getattr call.
 * currently 50 v4 attributes which fit into a bitmap4 size of 2 __u32's.
 *
 * for v3 getattr i need:
 *      v3 fattr        v4 attributes
 * ----------------------------------------------------
 *      valid           nfs client flag
 *      type            fattr4_type
 *      fsid            fattr4_fsid
 *      fileid          fattr4_fileid
 *      mode            fattr4_mode
 *      rdev            fattr4_rawdev
 *      size            fattr4_size
 *      mtime           fattr4_time_modify
 *      atime           fattr4_time_access
 *      ctime           fattr4_time_create
 *      nlink           fattr4_numlinks
 *      uid             fattr4_owner (note: this is a string)
 *      gid             fattr4_owner_group  (note: this is a string)
 *      du.nfs3.used    fattr4_space_used
 *
 *      CACHE CONSISTENCY STUFF
 *      pre_size        fattr4_???
 *      pre_mtime       fattr4_???
 *
 */

static uint32_t  __standard_bm[]      = { 0x00100112, 0x0024a23a };
bitmap4          standard_bm          = { 2, __standard_bm };
static uint32_t  __no_owner_bm[]      = { 0x00100112, 0x0024a20a };
bitmap4          no_owner_bm          = { 2, __no_owner_bm };
static uint32_t  __only_owner_bm[]    = { 0x00000000, 0x00000030 };
bitmap4          only_owner_bm        = { 2, __only_owner_bm };
static uint32_t  __readdir_bm[]       = { 0x00100000, 0x00200000 };      // fileid, ctime
bitmap4          readdir_bm           = { 2, __readdir_bm };
static uint32_t  __statfs_bm[]        = { 0x20c00000, 0x00001c00 };      // files_free, files_total, maxname, space_avail,
	bitmap4          statfs_bm            = { 2, __statfs_bm };              // space_free, space_total

	int nfs4_setup_getattr (GETATTR4args *cargs, GETATTR4res *cres, bitmap4 *bm)
{
	int err;
	int bufsize;
	char *buf;

	// setup bitmap in arg

		   err = setup_bitmap4 (&cargs->attr_request, bm->bitmap4_len);
	if (err)
		return err;
	memcpy (cargs->attr_request.bitmap4_val, bm->bitmap4_val, bm->bitmap4_len * sizeof(uint32_t));

	// setup bitmap in result

		   err = setup_bitmap4 (&cres->GFATTR4.attrmask, bm->bitmap4_len);
	if (err)
		return err;

	// setup attribute buffer in result
		   // future optimization : store the buffer size in parallel with the bitmap to avoid
		   // an everytime call to get_fattr4_res_buf_size()

		   bufsize = get_fattr4_res_buf_size (bm);
	buf = (char *) kmalloc (bufsize, GFP_KERNEL);
	if (!buf) {
		cres->GFATTR4.attr_vals.attrlist4_len = 0;
		cres->GFATTR4.attr_vals.attrlist4_val = 0;
		return NFS4ERR_RESOURCE;
	}
	cres->GFATTR4.attr_vals.attrlist4_len = bufsize;
	cres->GFATTR4.attr_vals.attrlist4_val = buf;
	memset (buf, 0, bufsize);
	return 0;
}

int nfs4_setupv3_getattr(GETATTR4args *cargs,GETATTR4res *cres)
{
#if 0
	return nfs4_setup_getattr (cargs, cres, &standard_bm);
#endif 0
	return nfs4_setup_getattr (cargs, cres, &no_owner_bm);
}

void
nfs4_setup_fhputfh(PUTFH4args *cargs, void *fhp, int fhsize)
{
	cargs->object.nfs_fh4_val = kmalloc(fhsize, GFP_KERNEL);
	memcpy(cargs->object.nfs_fh4_val, fhp, fhsize);
	cargs->object.nfs_fh4_len = fhsize;
}

#ifndef __OpenBSD__
/* PUTFH
 * stuff the v3 filehandle into the v4 struct.
 */
void
nfs4_setup_putfh(PUTFH4args *cargs,struct dentry *dentry)
{
	nfsv4_printk(fh4,"NFSV4: nfs4_setup_putfh for %s fh len: %d\n",
		     dentry->d_name.name,(NFS_FH(dentry))->size);
	cargs->object.nfs_fh4_val = kmalloc((NFS_FH(dentry))->size,GFP_KERNEL);
	nfs4_setup_fhputfh(cargs, (void *) (NFS_FH(dentry))->data, (NFS_FH(dentry))->size);
	print_nfs_fh4(&cargs->object);
}

int
nfs4_setup_open (OPEN4args *cargs, struct dentry *dentry, int flags, int mode,
		 struct nfs_server *server)
{
	int status = 0;
	open_claim4 *claim4 = &cargs->claim;
	openflag4 *flag4 = &cargs->openhow;
	nfs_lockowner4 *owner4 = &cargs->owner;
	struct nfs4_lock_owner *lowner;
	
	/* setup open_claim4 */
	claim4->claim = CLAIM_NULL;
	claim4->CLAIM_PATH_LEN = 1;
	if (!(claim4->CLAIM_PATH_VAL = kmalloc(sizeof(component4), GFP_KERNEL))) {
		status = -ENOMEM;
		goto out;
	}
	qstr_to_utf8(&dentry->d_name, &claim4->CLAIM_PATH_VAL[0]);

	/* setup openflag4 */
	/* for now, if the O_EXCL bit is set, a GUARDED4 call is put on the wire.
	 * later, this will become an EXCLUSIVE4 call.
	 */
	if (flags & O_CREAT) {
		createhow4 *how = &flag4->openflag4_u.how;
		struct iattr attr;
		
		attr.ia_mode = mode;
		attr.ia_uid = current->fsuid;
		attr.ia_gid = current->fsgid;
		if (flags & O_TRUNC) {
			attr.ia_valid = ATTR_MODE | ATTR_UID | ATTR_GID | ATTR_SIZE;
			attr.ia_size = 0;
		}
		else
			attr.ia_valid = ATTR_MODE | ATTR_UID | ATTR_GID;
		
		flag4->opentype = OPEN4_CREATE;
		how->mode = (flags & O_EXCL) ? GUARDED4 : UNCHECKED4;
		if ((status = iattr_to_fattr4(&attr, &how->createhow4_u.createattrs)))
			goto out_free_claim;
	}
	else
		flag4->opentype = OPEN4_NOCREATE;

	/* setup lockowner, seqid, share bits */
	if (!(lowner = nfs4_get_lockowner(current->pid, server->nfs4_clientid))) {
		status = -ENOMEM;
		goto out_free_flag;
	}
	owner4->clientid = server->nfs4_clientid;
	owner4->owner.owner_len = sizeof(pid_t);
	if (!(owner4->owner.owner_val = kmalloc(sizeof(pid_t), GFP_KERNEL))) {
		status = -ENOMEM;
		goto out_free_flag;
	}
	*((pid_t *) owner4->owner.owner_val) = current->pid;
	cargs->seqid = ++lowner->lo_seq_num;
	switch (flags & 3) {
	case 0:
		cargs->share_access = OPEN4_SHARE_ACCESS_READ;
		break;
	case 1:
		cargs->share_access = OPEN4_SHARE_ACCESS_WRITE;
		break;
	case 2:
	case 3:
		cargs->share_access = OPEN4_SHARE_ACCESS_BOTH;
		break;
	}
	cargs->share_deny = OPEN4_SHARE_DENY_NONE;

 out:
	return status;
 out_free_flag:
	if (flag4->opentype == OPEN4_CREATE) {
		fattr4 *f4 = &flag4->openflag4_u.how.createhow4_u.createattrs;
		kfree(f4->attrmask.bitmap4_val);
		kfree(f4->attr_vals.attrlist4_val);
	}
 out_free_claim:
	kfree(claim4->CLAIM_PATH_VAL[0].utf8string_val);
	kfree(claim4->CLAIM_PATH_VAL);
	goto out;
}

void
nfs4_setup_rename(RENAME4args *args, struct qstr *old_name, struct qstr *new_name)
{
        nfsv4_printk(rename, "NFSV4: setup rename\n");
        nfsv4_printk(rename, "NFSV4: old name %s, new name %s\n", old_name->name, new_name->name);

        qstr_to_utf8(old_name, &args->oldname);
        qstr_to_utf8(new_name, &args->newname);
}

void
nfs4_setup_link(LINK4args *args, struct qstr *new_name)
{
	nfsv4_printk(compound, "NFSV4: setup link\n");
	nfsv4_printk(compound, "NFSV4: new name %s\n", new_name->name);

	qstr_to_utf8(new_name, &args->newname);

}

void
nfs4_setup_remove(REMOVE4args *args, struct qstr *target_name)
{
	nfsv4_printk(compound, "NFSV4: setup remove\n");
	nfsv4_printk(compound, "NFSV4: remove target name %s\n", target_name->name);

	qstr_to_utf8(target_name, &args->target);
}

void
nfs4_setup_mkdir(CREATE4args *args, struct qstr *dir_name)
{
	nfsv4_printk(compound, "NFSV4: setup CREATE, mkdir\n");
	nfsv4_printk(compound, "NFSV4: create %s\n", dir_name->name);
	qstr_to_utf8(dir_name, &args->objname);
	args->objtype.type = NF4DIR;
}

void
nfs4_setup_symlink(CREATE4args *args, struct qstr *dentry_name, struct qstr *sym_name)
{
	nfsv4_printk(compound, "NFSV4: setup CREATE symlink\n");

	qstr_to_utf8 (dentry_name, &args->objname);
	args->objtype.type = NF4LNK;
	qstr_to_utf8 (sym_name, &args->objtype.createtype4_u.linkdata);
}

int
iattr_to_fattr4 (struct iattr *iattr, fattr4 *fattr)
{
	int maxlen = 0;
	char *p;
	unsigned int ia_valid = iattr->ia_valid;
	int err = 0;

	/* set bitmap
	 */
	fattr->attrmask.bitmap4_len = 2;
	fattr->attrmask.bitmap4_val = (uint32_t *) kmalloc(2 * sizeof(uint32_t), GFP_KERNEL);
	if (!fattr->attrmask.bitmap4_val) {
		err = -ENOMEM;
		goto out;
	}
	fattr->attrmask.bitmap4_val[0] =  0;              /* zero bitmap initially */
	fattr->attrmask.bitmap4_val[1] =  0;
	if (ia_valid & ATTR_SIZE)
		fattr->attrmask.bitmap4_val[0] |= 0x010;  /* FATTR4_SIZE is 4 */
	if (ia_valid & ATTR_MODE)
		fattr->attrmask.bitmap4_val[1] |= 0x02;   /* FATTR4_MODE is 33 */
	if (ia_valid & ATTR_UID)
		fattr->attrmask.bitmap4_val[1] |= 0x10;   /* FATTR4_OWNER is 36 */
	if (ia_valid & ATTR_GID)
		fattr->attrmask.bitmap4_val[1] |= 0x20;   /* FATTR4_OWNER_GROUP is 37 */

	/* allocate and fill attribute buffer
	 */
	maxlen = get_fattr4_res_buf_size(&fattr->attrmask);
	if (maxlen <= 0)
		maxlen = 1;
	nfsv4_printk (setattr, "setup_setattr(): allocating fattr buffer %d bytes\n", maxlen);
	p = fattr->attr_vals.attrlist4_val = kmalloc(maxlen, GFP_KERNEL);
	if (!p) {
		err = -ENOMEM;
		goto out_free_bitmap;
	}

	if (ia_valid & ATTR_SIZE) {          /* FATTR4_SIZE */
		nfsv4_printk (setattr, "setup_setattr(): FATTR4_SIZE %ld\n", iattr->ia_size);
		*((fattr4_size *) p) = iattr->ia_size;
		p += sizeof(fattr4_size);
	}
	if (ia_valid & ATTR_MODE) {           /* FATTR4_MODE */
		nfsv4_printk (setattr, "setup_setattr(): FATTR4_MODE %d\n", iattr->ia_mode);
		*((fattr4_mode *) p) = iattr->ia_mode & (S_IFMT | S_IALLUGO);
		p += sizeof(fattr4_mode);
	}
	if (ia_valid & ATTR_UID) {            /* FATTR4_OWNER */
		utf8string s;
		if ((err = get_name(NUM2STR_UID, iattr->ia_uid, &s))) {
			printk ("get_name() returned error %d\n", err);
			goto out_free_attrs;
		}
		*((u_int *) p) = s.utf8string_len;
		p += sizeof(u_int);
		*((char **) p) = p + sizeof(char *);
		p += sizeof(char *);
		memcpy(p, s.utf8string_val, s.utf8string_len);
		p += s.utf8string_len;
		kfree(s.utf8string_val);
	}	
	if (ia_valid & ATTR_GID) {            /* FATTR4_OWNER_GROUP */
		utf8string s;
		if ((err = get_name (NUM2STR_GID, iattr->ia_gid, &s))) {
			printk ("get_name() returned error %d\n", err);
			goto out_free_attrs;
		}
		*((u_int *) p) = s.utf8string_len;
		p += sizeof(u_int);
		*((char **) p) = p + sizeof(char *);
		p += sizeof(char *);
		memcpy(p, s.utf8string_val, s.utf8string_len);
		p += s.utf8string_len;
		kfree(s.utf8string_val);
	}
	fattr->attr_vals.attrlist4_len = p - fattr->attr_vals.attrlist4_val;
	
	if (nfsv4_debug_setattr) {
		printk ("iattr_to_fattr4() created the following fattr --------\n");
		dump_fattr4 (fattr);
		printk ("  ----- end of created fattr\n");
	}
 out:
	return err;
 out_free_attrs:
	kfree (fattr->attr_vals.attrlist4_val);	
 out_free_bitmap:
	kfree (fattr->attrmask.bitmap4_val);
	goto out;
}


int nfs4_setup_setattr (SETATTR4args *args, SETATTR4res *res, struct iattr *attr)
{
	nfsv4_printk (setattr, "setup_setattr(): starting\n");

	// setup args
		   args->stateid = 0;         // indicates no change in file size
							 if (iattr_to_fattr4 (attr, &args->obj_attributes))
								 return -1;

	// setup res
		   res->attrsset.bitmap4_val = (uint32_t *) kmalloc (2 * sizeof(uint32_t), GFP_KERNEL);
	return 0;
}

void
nfs4_setup_secinfo(SECINFO4args *args, struct qstr *new_name)
{
	nfsv4_printk(compound, "NFSV4: nfs4_setup_secinfo \n");
	nfsv4_printk(compound, "NFSV4: new name %s\n", new_name->name);

	qstr_to_utf8(new_name, &args->name);
}

/* called when an error value of NFS4ERROR_WRONGSED is returned
 * from another NFSV4 operation. mark the static client secinfo 
 * array 
 * NOTE: called from dir.c:nfs_lookup() and uses directory inode
 * and dentry of name to be looked up.
 */

int
do_secinfo(struct inode *inode,struct dentry *dentry)
{
	char tag[] = "do_secinfo() PUTFH SECINFO";
	nfs_argop4 aop4[2];
	nfs_resop4 rop4[2]; 
	COMPOUND4args cargs;
	COMPOUND4res cres;
	int i,error = 0;
	SECINFO4res *sres = NULL;
	secinfo4 *ptr;
	u32 flavor = -1;
	struct rpc_auth *save_auth;
	
	nfsv4_printk(secinfo,"NFSV4: do_secinfo\n");
	/* kmalloc space for ops and results plus tags */
	setup_compound(2,&cargs,&cres,tag,aop4,rop4);
	/* nfsv4 PUTFH */
	cargs.OPVAL[0].argop = OP_PUTFH;
	nfs4_setup_putfh(&cargs.OPVAL[0].OPU.opputfh,dentry);
	
	/* nfsv4 SECINFO */
	cargs.OPVAL[1].argop = OP_SECINFO;
	nfs4_setup_secinfo(&cargs.OPVAL[1].OPU.opsecinfo,&dentry->d_name);
	nfsv4_printk(secinfo,"NFSV4: do_secinfo 1\n");
	
	/* Save old auth handle, replace with AUTH_GSS. */
	save_auth = NFS_CLIENT(inode)->cl_auth;
	rpcauth_create(RPC_AUTH_GSS, NFS_CLIENT(inode));
	
	nfsv4_printk(secinfo,"NFSV4: do_secinfo 2\n");
	error = nfs4_proc_compound(NFS_SERVER(inode),&cargs,&cres);
	if(error > 1000)
		error = -EINVAL;
	else error = -error;
	nfs4_kfree_compound_args(&cargs);
	nfsv4_printk(secinfo,"NFSV4: do_secinfo 3\n");
	
	/* Restore old auth handle. */
	NFS_CLIENT(inode)->cl_auth = save_auth;
	nfsv4_printk(secinfo,"NFSV4: do_secinfo 4\n");
	
	/* partial results - i.e. PUTFH succeeds, SECINFO fails,
	 * are meaningless.
	 */
	if(error){
		nfsv4_printk(secinfo,"NFSV4: do_secinfo ERROR %d\n",error);
		nfs4_kfree_compound_res(&cres);
		return error;
	}
  
	sres = &cres.RESVAL[1].RESU.opsecinfo;
	ptr = sres->SECRES.SECINFO4resok_val;
	
	nfsv4_printk(secinfo,"NFSV4: do_secinfo SUCCESS \n");
	for(i=0;i<sres->SECRES.SECINFO4resok_len;i++){
		flavor = ptr->flavor;
		
		switch(flavor){
		case RPC_AUTH_NULL:
			break;
		case RPC_AUTH_UNIX:
			break;	
		case RPC_AUTH_GSS:
			/* compare to the static client state struct that 
			 * contains the QOP supported by this client. 
			 * mark the matches.
			 */
			nfsv4_printk(secinfo,"NFSV4: do_secinfo rpcsec_gss_info: \n");
			/*
			  print_rpcsec_gss_info(ptr);
			*/
			break;
		default:
			nfsv4_printk(secinfo,"NFSV4: do_secinfo INVALID flavor %d\n",ptr->flavor);
			break;
		}
		ptr += sizeof(uint32_t);
		ptr += ptr->flavor_info.flavor_info_len;
	}
	if (NFS_CLIENT(inode) != NULL &&
	    NFS_CLIENT(inode)->cl_auth->au_ops->au_flavor != flavor) {
		nfsv4_printk(secinfo, "NFSV4: do_secinfo changing to "
			     "flavor %d\n", flavor);
		/* Set the mech, service, QOP, based on negotiated secinfo. */
		rpcauth_destroy(NFS_CLIENT(inode)->cl_auth);
		rpcauth_create(flavor, NFS_CLIENT(inode));
	}
	return 0;
}

int
nfs4_setup_close(CLOSE4args *args ,clientid4 clid, stateid4 stateid)
{
	struct nfs4_lock_owner *lowner;
	int err = 0;

	nfsv4_printk(close,"NFSV4 nfs4_setup_close\n");
	/* set the state id */
	args->stateid = stateid;

	/* get the lockowner */
	err = NFS4ERR_RESOURCE;
	lowner = nfs4_get_lockowner(current->pid,clid);
	if(!lowner)
		goto out_fail;
	args->seqid = ++lowner->lo_seq_num;
	return 0;

 out_fail:
	args->seqid = 0;
	return err;
}

void
setup_setclientid(SETCLIENTID4args *args, char *ipaddr)
{
	set_verifier(args->client.verifier,nfs_verf);
	args->client.id.id_len = strlen(ipaddr);
	args->client.id.id_val = kmalloc(strlen(ipaddr),GFP_KERNEL);
	memset(args->client.id.id_val,0,strlen(ipaddr));
	memcpy(args->client.id.id_val,ipaddr,strlen(ipaddr));
	nfsv4_printk(setclntid,
		     "NFSV4 setup_setclientid id_val %s verifier:\n", args->client.id.id_val);	
	print_verifier4(args->client.verifier);
	nfs4_setup_cb_client4(&args->callback);
}

int
do_setclientid(struct nfs_server *server)
{
	char tag[] = "do_setclientid() SETCLIENTID";
	char tag1[] = "do_setclientid() SETCLIENTID_CONFIRM";
	nfs_argop4 aop4[1];
	nfs_resop4 rop4[1];
	COMPOUND4args cargs;
	COMPOUND4res cres;
	SETCLIENTID4res *res;
	int error = 0;
	verifier4 srv_verf;
	struct rpc_auth *save_auth;
	clientid4 tmpclid = 0;
  
	nfsv4_printk(setclntid,"NFSV4: do_setclientid\n");
	setup_compound(1,&cargs,&cres,tag,aop4,rop4);
  
	/* nfsv4 SETCLIENTID */
	cargs.OPVAL[0].argop = OP_SETCLIENTID;
	setup_setclientid(&cargs.OPVAL[0].OPU.opsetclientid,server->ipaddr);

#if 0
	/* Save old auth handle, replace with AUTH_GSS. */
	save_auth = server->client->cl_auth;
	rpcauth_create(RPC_AUTH_GSS, server->client);
#endif /*0*/
	error = nfs4_proc_compound(server,&cargs,&cres);
	if(error > 1000)
		error = -EINVAL;
	else error = -error;

	/* Restore old auth handle. */

#if 0
	server->client->cl_auth = save_auth;
#endif /*0*/

	if(!error){
		/* call SETCLIENTID_CONFIRM */
		res = &cres.RESVAL[0].RESU.opsetclientid;
		nfsv4_printk(setclntid,"NFSV4: do_setclientid CONFIRM. clientid: %Ld\n",
			     res->CLIDOK.clientid);
		nfsv4_printk(setclntid,"NFSV4: do_setclientid CONFIRM. setclientid_confirm:\n");
		print_verifier4(res->CLIDOK.setclientid_confirm);

		memset(srv_verf,0,NFS4_VERIFIER_SIZE);
		set_verifier(srv_verf,res->CLIDOK.setclientid_confirm);
		print_verifier4(srv_verf);
		tmpclid = res->CLIDOK.clientid;
    
		/* free setclientid call parameters */
		nfs4_kfree_compound_args(&cargs);
		nfs4_kfree_compound_res(&cres);
    
		/* call setclientid_confirm */
		setup_compound(1,&cargs,&cres,tag1,aop4,rop4);
    
		cargs.OPVAL[0].argop = OP_SETCLIENTID_CONFIRM;
		set_verifier(cargs.OPVAL[0].OPU.opsetclientid_confirm.setclientid_confirm,srv_verf);
		nfsv4_printk(setclntid,"NFSV4: do_setclientid CONFIRM. sending verf:\n");
		print_verifier4(cargs.OPVAL[0].OPU.opsetclientid_confirm.setclientid_confirm);
    
		error = nfs4_proc_compound(server,&cargs,&cres);
		if(error > 1000)
			error = -EINVAL;
		else error = -error;
		if(error){
			nfsv4_printk(setclntid,
				     "NFSV4 SETCLIENTID_CONFIRM failed. error %d\n",error);
			goto out_err;
		}
		nfsv4_printk(setclntid,"NFSV4 SETCLIENTID_CONFIRM success!\n");
		/* save clientid returned and confirmed by server */
		server->nfs4_clientid = tmpclid;
	}
 out_err:
	nfs4_kfree_compound_args(&cargs);
  
	nfsv4_printk(setclntid, "NFSV4: do_setclientid: error %d\n",error);
  
	nfs4_kfree_compound_res(&cres);
	return error;
}

/* prepare an OPEN_CONFIRM call*/
u32
nfs4_setup_open_confirm(OPEN_CONFIRM4args *args, OPEN4res *openres,clientid4 clid)
{
	struct nfs4_lock_owner *lowner;
	u32 err = 0;

	nfsv4_printk(open,"NFSV4 nfs4_setup_open_confirm\n");

	/* get the lockowner */
	err = NFS4ERR_RESOURCE;
	lowner = nfs4_get_lockowner(current->pid,clid);
	if(!lowner)
		goto out_fail;
	args->seqid = ++lowner->lo_seq_num;
	set_verifier(args->open_confirm,openres->OPEN_RES.open_confirm);
	print_verifier4(args->open_confirm);
	return 0;

 out_fail:
	args->seqid = 0;
	return err;
}

/* see if server requested an OPEN_CONFIRM. if so send one.
 */
u32
check_open_confirm(OPEN4res *res, struct nfs_server *server, struct dentry *dentry)
{
	char tag[] = "check_open_confirm() PUTFH OPEN_CONFIRM";
	nfs_argop4 aop4[2];
	nfs_resop4 rop4[2];
	COMPOUND4args cargs;
	COMPOUND4res cres;
	u32 nfserr = 0;

	nfsv4_printk(open,"NFSV4:check_open_confirm clientid: %Ld\n",server->nfs4_clientid);
	if (!(res->OPEN_RES.rflags & OPEN4_RESULT_CONFIRM)){
		nfsv4_printk(open,"NFSV4:check_open_confirm: no open confirm requested\n");	
		goto out;
	}

	nfsv4_printk(setclntid,"NFSV4: check_open_confirm sending OPEN_CONFIRM\n");
	setup_compound(2, &cargs, &cres, tag, aop4, rop4);
  	/* ------------ PUTFH ------------ */
	cargs.OPVAL[0].argop = OP_PUTFH;
  	nfs4_setup_putfh(&cargs.OPVAL[0].OPU.opputfh, dentry);
	/* ------------ OPEN_CONFIRM ------------ */
	cargs.OPVAL[1].argop = OP_OPEN_CONFIRM;
	nfs4_setup_open_confirm(&cargs.OPVAL[1].OPU.opopen_confirm, res, server->nfs4_clientid);

	if ((nfserr = nfs4_proc_compound(server,&cargs,&cres))) {
		nfsv4_printk(open, "NFSV4 OPEN_CONFIRM failed. err %d\n",nfserr);
		if(nfserr > 1000)
			nfserr = -EINVAL;
		else nfserr = -nfserr;
		goto out;
	}
	/* replace old OPEN stateid by OPEN_CONFIRM stateid */
	res->OPEN_RES.stateid = cres.RESVAL[1].RESU.opopen_confirm.OPEN_CON_RES.stateid;

	nfsv4_printk(open,"NFSV4 OPEN_CONFIRM success! stateid [0x%lx , %08lx]\n",
		     (unsigned long)(res->OPEN_RES.stateid >> 32),
		     (unsigned long)(res->OPEN_RES.stateid & 0xFFFFFFFFUL));
	
 out:
	return nfserr;
}
#endif /* OpenBSD */

#ifdef __OpenBSD__
void *
kmalloc(unsigned long s, int t)
{
	void *p;

	MALLOC(p, void *, s, M_NFSREQ, M_WAITOK);
	return p;
}

void kfree(void *p)
{
	FREE(p, M_NFSREQ);
}
#endif

void
print_fattr4(fattr4 *fptr)
{
#if 0
	if (nfsv4_debug_attributes)
		dump_fattr4(fptr);
#endif 
}

void
hex_dump (char *p, int no_bytes)
{
	int i;

	for (i = 0; i < no_bytes; i++) {
		printk ("%x", (unsigned char) p[i]);

		if ((i % 56) == 55)
			printk ("\n\n");
		else if ((i % 8) == 7)
			printk ("\n");
		else if ((i % 4) == 3)
			printk ("   ");
		else
			printk (" ");
	}
	printk ("\n");
}

// prints a fattr4 unconditionally (i.e., regardless of which debug flags are set)
// based on the old version of print_fattr4(), but a couple of bugs have been fixed

void
dump_fattr4 (fattr4 *fptr)
{
	int i,j;
	int bit;
	bitmap4 *bm4 = &fptr->attrmask;
	char *cptr = fptr->FATTRSVAL;

	if(!bm4)
		return;
	/* for each bit set in the bit mask */
	printk ("NFSV4:print_fattr4 bitmap len %d\n",bm4->BLEN);
	printk ("NFSV4:print_fattr4 attrlist len %d\n",fptr->FATTRSLEN);

	for(i=0;i<bm4->BLEN;i++){
		printk ("NFSV4:print_fattr4 bits_val[%d] %x \n",i,bm4->BVAL[i]);
		for(j=0;j<32;j++){
			/* if bit is set*/
			if(bm4->BVAL[i] & bit4_mask[j]){
				bit = (i*32)+j;
				printk ("NFSV4:print_fattr4 bit %d is set\n",bit);
				switch(bit){
				case FATTR4_SUPPORTED_ATTRS:
				{
					bitmap4 *b4p = (bitmap4 *) cptr;
					int size = sizeof(u_int) + (b4p->BLEN * sizeof(uint32_t));
					printk("        SUPPORTED\n");
					cptr += size;
				}
				break;
				case FATTR4_TYPE:
				{
					printk ("        TYPE: %d\n", *(fattr4_type *)cptr);
					cptr += sizeof(fattr4_type);
				}
				break;
				case FATTR4_FH_EXPIRE_TYPE:
					printk("        FH_EXPIRE_TYPE\n");
					cptr += sizeof(fattr4_fh_expire_type);
					break;
				case FATTR4_CHANGE:
					printk("        CHANGE\n");
					cptr += sizeof(fattr4_change);
					break;
				case FATTR4_SIZE:
				{
					printk ("        SIZE: %d\n", *(u32 *)cptr);
					cptr += sizeof(fattr4_size);
				}
				break;
				case FATTR4_LINK_SUPPORT:
					printk("        LINK_SUPPORT\n");
					cptr += sizeof(fattr4_link_support);
					break;
				case FATTR4_SYMLINK_SUPPORT:
					printk("        SYMLINK_SUPPORT\n");
					cptr += sizeof(fattr4_symlink_support);
					break;
				case FATTR4_NAMED_ATTR:
					printk("        NAMED_ATTR\n");
					cptr += sizeof(fattr4_named_attr);
					break;
				case FATTR4_FSID:
				{
					/*fattr4_fsid *ptr = (fattr4_fsid *)cptr; */
					u32 *iptr = (u32 *)cptr;

					/****  behavior of next line is compiler-dependent; change it! ****/

					printk (
						"        FSID Major: 0x%x:0x%x  FSID Minor: 0x%x:0x%x\n",
						*iptr++,*iptr++,*iptr++,*iptr);
					cptr += sizeof(fattr4_fsid);
				}
				break;
				case FATTR4_FILEID:
				{
					printk ("        FILEID: %ld\n",*(long int *)cptr);
					cptr += sizeof(fattr4_fileid);
				}
				break;
				case  FATTR4_FILES_FREE:
					printk("        FILES_FREE: %d\n", (int) *((fattr4_files_free *) cptr));
					cptr += sizeof(fattr4_files_free);
					break;
				case  FATTR4_FILES_TOTAL:
					printk("        FILES_TOTAL: %d\n", (int) *((fattr4_files_total *) cptr));
					cptr += sizeof(fattr4_files_total);
					break;
				case FATTR4_MAXNAME:
				{
					printk ("        MAXNAME: %d\n",*(fattr4_maxname *)cptr);
					cptr += sizeof(fattr4_maxname);
				}
				break;
				case FATTR4_MAXREAD:
				{
					printk ("        MAXREAD: %ld\n",*(long int *)cptr);
					cptr += sizeof(fattr4_maxread);
				}
				break;
				case FATTR4_MAXWRITE:
				{
					printk ("        MAXWRITRE: %ld\n",*(long int *)cptr);
					cptr += sizeof(fattr4_maxwrite);
				}
				break;
				case FATTR4_MODE:
				{
					printk ("        MODE: %d\n",*(fattr4_mode *)cptr);
					cptr += sizeof(fattr4_mode);
				}
				break;
				case FATTR4_NUMLINKS:
				{
					printk ("        NUMLINKS: %d\n", *(fattr4_numlinks *)cptr);
					cptr += sizeof(fattr4_numlinks);
				}
				break;
				case FATTR4_OWNER:
				{
					int k;
					int l = *((u_int *) cptr);

					printk ("        OWNER [len %d] ", l);
					cptr += sizeof(fattr4_owner);
					for (k = 0; k < l; k++) {
						printk ("%c", cptr[k]);
					}
					cptr += l;
					printk ("\n");
				}
				break;
				case FATTR4_OWNER_GROUP:
				{
					int k;
					int l = *((u_int *) cptr);

					printk ("        GROUP [len %d] ", l);
					cptr += sizeof(fattr4_owner);
					for (k = 0; k < l; k++) {
						printk ("%c", cptr[k]);
					}
					cptr += l;
					printk ("\n");
				}
				break;
				case FATTR4_RAWDEV:
				{
					u32 *iptr = (u32 *)cptr;

					printk ("        RAWDEV specdata1: 0x%x  specdata2: 0x%x\n",*iptr++,*iptr);
					cptr += sizeof(fattr4_rawdev);
				}
				break;
				case  FATTR4_SPACE_AVAIL:
					printk("        SPACE_AVAIL: %d\n", (int) *((fattr4_space_avail *) cptr));
					cptr += sizeof(fattr4_space_avail);
					break;
				case  FATTR4_SPACE_FREE:
					printk("        SPACE_FREE: %d\n", (int) *((fattr4_space_free *) cptr));
					cptr += sizeof(fattr4_space_free);
					break;
				case  FATTR4_SPACE_TOTAL:
					printk("        SPACE_TOTAL: %d\n", (int) *((fattr4_space_total *) cptr));
					cptr += sizeof(fattr4_space_total);
					break;
				case FATTR4_SPACE_USED:
				{
					printk ("        SPACE_USED: %ld\n", *(long int *)cptr);
					cptr += sizeof(fattr4_space_used);
				}
				break;
				case FATTR4_TIME_ACCESS:
				{
					fattr4_time_access *ptr = (fattr4_time_access *)cptr;
					printk ("        TIME_ACCESS: %ld\n",(long int)ptr->seconds);
					cptr += sizeof(fattr4_time_access);
				}
				break;
				case FATTR4_TIME_CREATE:
				{
					fattr4_time_create *ptr = (fattr4_time_create *)cptr;
					printk ("        TIME_CREATE: %ld\n",(long int)ptr->seconds);
					cptr += sizeof(fattr4_time_create);
				}
				break;
				case FATTR4_TIME_MODIFY:
				{
					fattr4_time_modify *ptr = (fattr4_time_modify *)cptr;
					printk ("        TIME_MODIFY: %ld\n",(long int)ptr->seconds);
					cptr += sizeof(fattr4_time_modify);
				}
				break;
				default:
					printk ("        DEFAULT!!! bit = %d\n",bit);
					break;
				}
			}
		}
	}
}
