/*
 *  linux/fs/nfs/symlink.c
 *
 *  Copyright (C) 1992  Rick Sladkey
 *
 *  Optimization changes Copyright (C) 1994 Florian La Roche
 *
 *  Jun 7 1999, cache symlink lookups in the page cache.  -DaveM
 *
 *  nfs symlink handling code
 */

#define NFS_NEED_XDR_TYPES
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/sunrpc/clnt.h>
#include <linux/nfs_fs.h>
#include <linux/nfs.h>
#include <linux/pagemap.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/string.h>

#include <asm/uaccess.h>

static int nfs_readlink(struct dentry *, char *, int);
static struct dentry *nfs_follow_link(struct dentry *, struct dentry *, unsigned int);

/*
 * symlinks can't do much...
 */
struct inode_operations nfs_symlink_inode_operations = {
	NULL,			/* no file-operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	nfs_readlink,		/* readlink */
	nfs_follow_link,	/* follow_link */
	NULL,			/* get_block */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* flushpage */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL,			/* smap */
	NULL			/* revalidate */
};

/* Symlink caching in the page cache is even more simplistic
 * and straight-forward than readdir caching.
 */
static struct page *try_to_get_symlink_page(struct dentry *dentry, struct inode *inode)
{
#ifdef NFSV4
	char tag[] = "try_to_get_symlink_page() PUTFH GETATTR READLINK";
	nfs_argop4 aop4[3];
	nfs_resop4 rop4[3];
	struct COMPOUND4args cargs;
	struct COMPOUND4res  cres;
	uint32_t *p;
#endif
	struct nfs_fattr fattr;
	struct page *page;
	void * buffer;
	unsigned int error;

	page = nfs_grab_cache_page(inode, 0);
	if (!page)
		goto out;

	if (PageUptodate(page))
		goto unlock_out;

	/* We place the length at the beginning of the page,
	 * in client byte order, followed by the string.
	 */
	buffer = (void *)page_address(page);
	memset(buffer, 0, PAGE_CACHE_SIZE);

#ifdef NFSV4
	setup_compound (3, &cargs, &cres, tag, aop4, rop4);
	/* ----------------- PUTFH ------------------- */
	cargs.OPVAL[0].argop = OP_PUTFH;
	nfs4_setup_putfh (&cargs.OPVAL[0].OPU.opputfh, dentry);
	/* ----------------- GETATTR ----------------- */
	cargs.OPVAL[1].argop = OP_GETATTR;
	nfs4_setupv3_getattr (&cargs.OPVAL[1].OPU.opgetattr,
			      &cres.RESVAL[1].RESU.opgetattr);
	/* ----------------- READLINK ---------------- */
	cargs.OPVAL[2].argop = OP_READLINK;

	error = nfs4_proc_compound (NFS_SERVER(inode), &cargs, &cres);
	if(error > 1000)
		error = -EINVAL;
	else error = -error;

	if (cres.RESCNT > 1) {   // getattr succeeded
					    memset (&fattr, 0, sizeof(fattr));
	nfs4_stuff_nfsfattr (&cres.RESVAL[1].RESU.opgetattr.GFATTR4, &fattr);
	nfs_refresh_inode (inode, &fattr);
	}
	if (cres.RESCNT > 2) {   // readlink succeeded
					    utf8string *s = &cres.RESVAL[2].RESU.opreadlink.READLINK4res_u.resok4.link;
	if (s->utf8string_len > NFS4_MAXNAMLEN)    // truncate string if too long
										     s->utf8string_len = NFS4_MAXNAMLEN;
	p = buffer;
	*p++ = s->utf8string_len;
	memcpy (p, s->utf8string_val, s->utf8string_len);
	}
	if (error < 0)
		goto error;
#else
	error = NFS_CALL(readlink, inode, (dentry, &fattr, buffer,
					   PAGE_CACHE_SIZE-sizeof(u32)-4));
	if (error < 0)
		goto error;
	nfs_refresh_inode(inode, &fattr);
#endif

	if (PageError(page))
		clear_bit(PG_error, &page->flags);
	set_bit(PG_uptodate, &page->flags);

 unlock_out:
	nfs_unlock_page(page);
 out:
	nfs4_kfree_compound_args (&cargs);
	nfs4_kfree_compound_res (&cres);
	return page;

 error:
	set_bit(PG_error, &page->flags);
	goto unlock_out;
}

static int nfs_readlink(struct dentry *dentry, char *buffer, int buflen)
{
	struct inode *inode = dentry->d_inode;
	struct page *page;
	u32 *p, len;

	/* Caller revalidated the directory inode already. */
	page = find_page(inode, 0);
	if (!page)
		goto no_readlink_page;
	if (!PageUptodate(page)) {
		page_cache_release(page);
		goto no_readlink_page;
	}
 success:
	p = (u32 *) page_address(page);
	len = *p++;
	if (len > buflen)
		len = buflen;
	copy_to_user(buffer, p, len);
	page_cache_release(page);
	return len;

 no_readlink_page:
	page = try_to_get_symlink_page(dentry, inode);
	if (!page)
		goto no_page;
	if (PageUptodate(page))
		goto success;
	page_cache_release(page);
	return -EIO;
 no_page:
	return -ENOMEM;
}

static struct dentry *
nfs_follow_link(struct dentry *dentry, struct dentry *base, unsigned int follow)
{
	struct dentry *result;
	struct inode *inode = dentry->d_inode;
	struct page *page;
	u32 *p;

	/* Caller revalidated the directory inode already. */
	page = find_page(inode, 0);
	if (!page)
		goto no_followlink_page;
	if (!PageUptodate(page)) {
		page_cache_release(page);
		goto no_followlink_page;
	}
 success:
	p = (u32 *) page_address(page);
	result = lookup_dentry((char *) (p + 1), base, follow);
	page_cache_release(page);
	return result;

 no_followlink_page:
	page = try_to_get_symlink_page(dentry, inode);
	if (!page)
		goto no_page;
	if (PageUptodate(page))
		goto success;
	page_cache_release(page);
	return ERR_PTR(-EIO);
 no_page:
	return ERR_PTR(-ENOMEM);
}
