/*
 *  linux/fs/nfsd/nfsd4cbxdr.c
 *
 *  server-side xdr for nfsv4 callbacks
 *
 *  by Kendrick Smith <kmsmith@umich.edu>
 *  Copyright (C) 2000 The Regents of the University of Michigan
 */

#include <linux/types.h>
#include <linux/sched.h>

#include <linux/sunrpc/xdr.h>
#include <linux/sunrpc/svc.h>
#include <linux/nfsd/nfsd4_prot.h>
#include <linux/malloc.h>

/*
 *  stuff lifted from fs/nfs/nfs4xdr.c
 */

static inline int
xdr_overflow_check(struct rpc_iov *iov, u32 *p, int arglen)
{
/* for now, just check iov->io_vec[0] */
	if(iov->io_nr == 1)
		return(p + arglen - (u32 *)iov->io_vec[0].iov_base <= iov->io_vec[0].iov_len);
	else return 1;
}

static inline u32 *
xdr_encode_utf8string(struct rpc_rqst *req,u32 *p, utf8string  *utf)
{
	int quadlen = XDR_QUADLEN(utf->utf8string_len);

	if(!(xdr_overflow_check(&req->rq_snd_buf,p,quadlen + sizeof(utf->utf8string_len))))
		return 0;
	p[quadlen] = 0;
	*p++ = htonl(utf->utf8string_len);	
	memcpy(p, utf->utf8string_val, utf->utf8string_len);
	return p + quadlen;
}

static inline u32 *
xdr_encode_64(u32 *p, u64 val)
{
	nfsv4_printk(xdr,"NFSDV4 xdr_encode_64 val: %ld\n",(long int)val);
	*p++ = htonl(val >> 32);
	*p++ = htonl(val & 0xffffffff);
	return p;
}

static inline u32 *
xdr_encode_fh4(struct rpc_rqst *req,u32 *p, nfs_fh4 *fhp)
{
	int len = sizeof(fhp->nfs_fh4_len) + XDR_QUADLEN(fhp->nfs_fh4_len);

	if (fhp->nfs_fh4_len <= NFS4_FHSIZE) {
		if(!(xdr_overflow_check(&req->rq_snd_buf,p,len)))
			return 0;
		*p++ = htonl(fhp->nfs_fh4_len);
		memcpy(p, fhp->nfs_fh4_val, fhp->nfs_fh4_len);
		return p + XDR_QUADLEN(fhp->nfs_fh4_len);
	}
	else 
		return 0;
}

static inline u32 *
xdr_encode_bitmap4(struct rpc_rqst *req,u32 *p, bitmap4  *bm4)
{
	int i,len = sizeof(bm4->BLEN) + (bm4->BLEN * (sizeof(uint32_t)));
	uint32_t *ptr=bm4->BVAL;

	nfsv4_printk(xdr,"NFSV4 xdr_encode_bitmap4 len %d\n",bm4->BLEN);
	if(!(xdr_overflow_check(&req->rq_snd_buf,p,len)))
		return 0;
	*p++ =  htonl(bm4->BLEN);
	for (i=0; i < bm4->BLEN; i++)
		*p++ = htonl(*ptr++);
	return p;
}

static u32 *
xdr_decode_utf8string(u32 *p, utf8string *utf)
{
	utf->utf8string_len = ntohl(*p++);
	utf->utf8string_val = (char *)p;
	return p + XDR_QUADLEN(utf->utf8string_len);
}

static inline u32 *
xdr_decode_32_inplace (u32 *p)
{
	*p = ntohl(*p);
	p++;
	return p;
}

static inline u32 *
xdr_decode_64_inplace (u32 *p)
{
	u32 hi = *p;
	*(p+1) = ntohl(*p);
	*p = ntohl(hi);
	p += 2;
	return p;
}

static u32 *
xdr_decode_bitmap4(u32 *p,bitmap4 *bm4)
{
	int i;

	bm4->BLEN = ntohl(*p++);
	bm4->BVAL = p;
	for (i = 0; i < bm4->BLEN; i++)
		p = xdr_decode_32_inplace(p);
	return p;
}

static u32 *
xdr_decode_fattr4(u32 *p, fattr4 *res)
{
	int i,j;
	int bit;
	bitmap4 *bm4 = &res->attrmask;

	p = xdr_decode_bitmap4(p,bm4);
	res->FATTRSLEN = ntohl(*p++);
	res->FATTRSVAL = (char *)p;

	for (i = 0; i < bm4->BLEN; i++)
		for(j = 0; j < 32; j++)
			if (bm4->BVAL[i] & (1 << j)) {
				bit = (i*32)+j;
				switch (bit) {
					/* not many attributes are allowed under the protocol */
				case FATTR4_CHANGE:
				case FATTR4_SIZE:
					p = xdr_decode_64_inplace(p);
					break;
				case FATTR4_TIME_ACCESS:
				case FATTR4_TIME_MODIFY:
					p = xdr_decode_64_inplace(p);
					p = xdr_decode_32_inplace(p);
					break;
				default:
					printk("!!! default in decode_fattr4()\n");
					return 0;
				}
			}
	
	return p;
}
/*
 *  new stuff
 */

int
xdr_encode_nullcb_args (struct rpc_rqst *rqstp, u32 *p, void *data)
{
	rqstp->rq_slen = xdr_adjust_iovec(rqstp->rq_svec, p);
	return 0;           /* 0 indicates success */
}
int
xdr_decode_nullcb_res (struct rpc_rqst *rqstp, u32 *p, void *data)
{
	return 0;            /* 0 indicates success */
}

static inline u32 *
xdr_encode_cb_getattr4args (struct rpc_rqst *req, u32 *p, CB_GETATTR4args *argp)
{
	if (!(p = xdr_encode_fh4(req, p, &argp->fh)))
		return 0;	    
	p = xdr_encode_bitmap4(req, p, &argp->attr_request);
	return p;
}

static u32 *
xdr_encode_cb_recall4args (struct rpc_rqst *req, u32 *p, CB_RECALL4args *argp)
{
	if (!xdr_overflow_check(&req->rq_snd_buf, p, 12))
		return 0;
	p = xdr_encode_64(p, argp->stateid);
	*p++ = htonl(argp->truncate);
	p = xdr_encode_fh4(req, p, &argp->fh);
	return p;
}

int
xdr_encode_compoundcb_args (struct rpc_rqst *req, u32 *p, CB_COMPOUND4args *argp)
{
	int i;
	int len;
	nfs_cb_argop4 *argop4;
	
	if (!(p = xdr_encode_utf8string(req, p, &argp->tag)))
		return 1;
	if (!xdr_overflow_check(&req->rq_snd_buf, p, 8))
		return 1;
	*p++ = htonl(argp->minorversion);
	len = argp->argarray.argarray_len;
	*p++ = htonl(len);

	for (i = 0; i < len; i++) {
		argop4 = &argp->argarray.argarray_val[i];
		if (!xdr_overflow_check(&req->rq_snd_buf, p, 4))
			return 1;
		*p++ = htonl(argop4->argop);
		switch (argop4->argop) {
		case OP_CB_GETATTR:
			if (!(p = xdr_encode_cb_getattr4args(req, p, &argop4->nfs_cb_argop4_u.opcbgetattr)))
				return 1;
			break;
		case OP_CB_RECALL:
			if (!(p = xdr_encode_cb_recall4args(req, p, &argop4->nfs_cb_argop4_u.opcbrecall)))
				return 1;
			break;
		default:
			printk ("!!! default reached when encoding cb args\n");
		}
	}
	
	req->rq_slen = xdr_adjust_iovec(req->rq_svec, p);
	return 0;      /* 0 indicates success */
}

static inline u32 *
xdr_decode_cb_getattr4res (u32 *p, CB_GETATTR4res *resp)
{
	resp->status = ntohl(*p++);
	if (!resp->status)
		p = xdr_decode_fattr4(p, &resp->CB_GETATTR4res_u.resok4.obj_attributes);
	return p;
}

static inline u32 *
xdr_decode_cb_recall4res (u32 *p, CB_RECALL4res *resp)
{
	resp->status = ntohl(*p++);
	return p;
}

int
xdr_decode_compoundcb_res (struct rpc_rqst *rqstp, u32 *p, CB_COMPOUND4res *resp)
{
	int i;
	int len;
	nfs_cb_resop4 *resop4;
		
	resp->status = ntohl(*p++);
	p = xdr_decode_utf8string(p, &resp->tag);
	len = ntohl(*p++);
	resp->resarray.resarray_len = len;
	resp->resarray.resarray_val = (nfs_cb_resop4 *)
		kmalloc (len * sizeof(nfs_cb_resop4), GFP_KERNEL);

	for (i = 0; i < len; i++) {
		resop4 = &resp->resarray.resarray_val[i];
		resop4->resop = ntohl(*p++);
		switch (resop4->resop) {
		case OP_CB_GETATTR:
			p = xdr_decode_cb_getattr4res (p, &resop4->nfs_cb_resop4_u.opcbgetattr);
			break;
		case OP_CB_RECALL:
			p = xdr_decode_cb_recall4res (p, &resop4->nfs_cb_resop4_u.opcbrecall);
			break;
		default:
			printk ("!!! default reached when decoding cb resp\n");
		}
	}

	return 0;         /* 0 indicates success */
}

void
xdr_free_compoundcb_res (CB_COMPOUND4res *resp)
{
	kfree(resp->resarray.resarray_val);
}
