/*
 *  linux/fs/nfsd/nfsd4deleg.c
 *
 *  server-side delegation support for NFSv4
 *
 *  by Kendrick Smith <kmsmith@umich.edu>
 *  Copyright (C) 2000 The Regents of the University of Michigan
 */

/*  list of things to fix/think about:
 *    - determining correct value for NFSD4_CB_BUFSIZE
 *    - minor version = 1 ?
 *    - change nfs4_proc_nullcb() so that testing of the cb path is
 *      done asynchronously
 *    - check file permission bits when issuing delegations to avoid
 *      security violations!
 *    - the loop in issue_read_delegation(), which walks the lock chain
 *      determining existing share access, duplicates what has already
 *      been done in nfs4_test_lock().  maybe these two functions should
 *      be combined?
 *    - think about interactions between share deny bits on an open and
 *      delegation.. is it always consistent, for all values of the deny
 *      bits, to issue a delegation?
 *    - is it possible for nfs4_issue_delegation() to get a neg. dentry?
 *    - think about OPEN_CONFIRM.. if a delegation is granted and no
 *      OPEN_CONFIRM is received, delegation should be revoked
 *    - permission checking in nfsd4_delegreturn(), to make sure one
 *      client can't return another's delegation
 *    - improve recall-issuing algorithm to be completely asynchronous,
 *      as well as achieving better asymptotic running time.
 *    - how should we deal with a client which misbehaves on callbacks, e.g.
 *      by giving an error or by not responding at all?
 *    - do we really need two semaphores in struct nfs4_file, one for
 *      delegation state, and one for additional state??
 */

#include <linux/fs.h>
#include <linux/inet.h>
#include <linux/sunrpc/svc.h>
#include <linux/sunrpc/clnt.h>
#include <linux/malloc.h>
#include <linux/nfsd/nfsd4_prot.h>
#include <linux/nfsd/nfsd4lock.h>
#include <linux/sunrpc/clnt.h>

#define TEST_CALLBACKS		0
#define ISSUE_READ_DELEGATIONS	0
#define ISSUE_WRITE_DELEGATIONS	0
#define DELEG_PORT		23137

#define DELEG_HASHBITS		8
#define DELEG_HASHSIZE		(1 << DELEG_HASHBITS)
#define DELEG_HASHMASK		(DELEG_HASHSIZE - 1)
#define deleg_hashval(x)	((x) & DELEG_HASHMASK)

#define NFSD4_CB_BUFSIZE	1024

static struct nfs4_delegation *deleg_hashtable[DELEG_HASHSIZE];
static stateid4 global_delegation_stateid = 1;

extern struct rpc_program nfsd4_cb_program;

static char *
delegation_type_str(open_delegation_type4 d)
	/* for debugging only */
{
	switch (d) {
	case OPEN_DELEGATE_NONE:
		return "OPEN_DELEGATE_NONE";
	case OPEN_DELEGATE_READ:
		return "OPEN_DELEGATE_READ";
	case OPEN_DELEGATE_WRITE:
		return "OPEN_DELEGATE_WRITE";
	default:
		return "<<bad open_delegation_type4>>";
	}
}

/*
 *  Callback rpc
 */

static struct rpc_stat nfsd4_cb_stats = { &nfsd4_cb_program,
					  0, 0, 0, 0, 0, 0, 0, 0, 0 };

static struct rpc_procinfo nfsd4_cb_procedures[] = {
	{			/* CB_NULL */
		"cb_null",
		(kxdrproc_t) xdr_encode_nullcb_args,  /* encode */
		(kxdrproc_t) xdr_decode_nullcb_res,   /* decode */
		0,                                    /* bufsize */
		1                                     /* call count */
	},
	{			/* CB_COMPOUND */
		"cb_compound",
		(kxdrproc_t) xdr_encode_compoundcb_args,   /* encode */
		(kxdrproc_t) xdr_decode_compoundcb_res,    /* decode */
		NFSD4_CB_BUFSIZE,                          /* bufsize */
		1                                          /* call count */
	}
};

static struct rpc_version nfsd4_cb_version1 = {
	1,                             /* version number */
	2,                             /* number of procedures */
	nfsd4_cb_procedures            /* table of virtual ops */
};

static struct rpc_version *nfsd4_cb_version[] = {
	NULL,                          /* version 0 */
	&nfsd4_cb_version1             /* version 1 */
};

struct rpc_program nfsd4_cb_program = {
	"nfs4_cb",                     /* protocol name */
	NFS4_CALLBACK,                 /* program number */
	2,                             /* number of versions */
	nfsd4_cb_version,              /* version table (contains virtual ops) */
	&nfsd4_cb_stats
};

/*
 *  Interface for setting up and sending callbacks
 */

void
nfsd4_proc_nullcb(struct svc_client *clp, cb_client4 *argp)
{
	int port;
	struct rpc_xprt *xprt;
	struct rpc_clnt *clnt = NULL;
	struct rpc_timeout timeparms;
	int status;
	struct sockaddr_in callback_sin;
	
	if (!TEST_CALLBACKS) {
		nfsv4_printk(callback, "proc_nullcb(): callbacks disabled; no callback made\n");
		goto out;
	}
	if (argp->cb_program != NFS4_CALLBACK) {
		nfsv4_printk(callback, "proc_nullcb(): program # mismatch; no callback made\n");
		goto out;
	}
	memset(&callback_sin, 0, sizeof(callback_sin));
	callback_sin.sin_family = AF_INET;
	callback_sin.sin_addr = clp->cl_addr;
#if 0
	nfsv4_printk(callback, "proc_nullcb(): calling portmapper\n");
	if ((port = rpc_getport_external(sin, NFS4_CALLBACK, 1, IPPROTO_UDP)) < 0) {
		nfsv4_printk(callback, "proc_nullcb(): portmapper returned error\n");
		goto out;
	}
#else
	nfsv4_printk(callback, "proc_nullcb(): using hardcoded port %d\n",
		     DELEG_PORT);
	port = DELEG_PORT;
#endif
	callback_sin.sin_port = htons(port);

	nfsv4_printk(callback, "proc_nullcb(): using callback addr %s:%d\n",
		     in_ntoa(callback_sin.sin_addr.s_addr),
		     ntohs(callback_sin.sin_port));

	timeparms.to_initval = 70;
	timeparms.to_retries = 3;
	timeparms.to_maxval = 6000;
	timeparms.to_exponential = 1;

	if ( !(xprt = xprt_create_proto (IPPROTO_UDP, &callback_sin, &timeparms))) {
		nfsv4_printk(callback, "proc_nullcb(): couldn't create proto\n");
		goto out;
	}
	if ( !(clnt = rpc_create_client (xprt, in_ntoa(callback_sin.sin_addr.s_addr),
					 &nfsd4_cb_program, 1, 1))) {
		nfsv4_printk(callback, "proc_nullcb(): couldn't create clnt\n");
		xprt_destroy(xprt);
		goto out;
	}
	clnt->cl_softrtry = 1;
	clnt->cl_chatty = 1;

	nfsv4_printk(callback, "proc_nullcb(): starting null callback...\n");
	if ((status = rpc_call(clnt, CB_NULL, 0, 0, 0))) {
		nfsv4_printk(callback, "proc_nullcb(): ...callback returned err %d\n",
			     status);
		rpc_destroy_client(clnt);        /* should xprt be destroyed too? */
		clnt = NULL;
		goto out;
	}
	nfsv4_printk(callback, "proc_nullcb(): ...null callback successful!\n");
	clp->cl_callback_clnt = clnt;
	  
 out:
	return;
}

void
nfsd4_setup_cb_compound(CB_COMPOUND4args *argp, nfs_cb_argop4 *argarray,
			int nops, char *tag)
{
	argp->tag.utf8string_len = strlen(tag);
	argp->tag.utf8string_val = tag;
	argp->minorversion = 1;             /* ??? */
	argp->argarray.argarray_len = nops;
	argp->argarray.argarray_val = argarray;
}

void
nfsd4_setup_cb_getattr(nfs_cb_argop4 *argp, nfs_fh4 *fh, bitmap4 *bm)
{
	CB_GETATTR4args *gp = &argp->nfs_cb_argop4_u.opcbgetattr;

	argp->argop = OP_CB_GETATTR;
	gp->fh = *fh;
	gp->attr_request = *bm;
}

void
nfsd4_setup_cb_recall(nfs_cb_argop4 *argp, stateid4 id, bool_t truncate,
		      nfs_fh4 *fh)
{
	CB_RECALL4args *rp = &argp->nfs_cb_argop4_u.opcbrecall;
	
	argp->argop = OP_CB_RECALL;
	rp->stateid = id;
	rp->truncate = truncate;
	rp->fh = *fh;
}


int
nfsd4_proc_compoundcb(struct rpc_clnt *clnt, CB_COMPOUND4args *argp, CB_COMPOUND4res *resp)
{
	int status;

	nfsv4_printk(callback, "proc_compoundcb(): sending CB_COMPOUND, tag=\"%.*s\"\n",
		     (int) argp->tag.utf8string_len, argp->tag.utf8string_val);
	status = rpc_call(clnt, CB_COMPOUND, argp, resp, 0);
	nfsv4_printk(callback, "proc_compoundcb(): CB_COMPOUND returned status %d\n",
		     status);
	return status;
}

/*
 *  Functions to add/remove delegations
 */

static inline void
issue_recall(struct nfs4_delegation *deleg)
	/* call with file->f_delegsema held */
{
	struct CB_COMPOUND4args cargs;
	struct CB_COMPOUND4res cres;
	char tag[] = "issue_recall(): CB_RECALL";
	nfs_cb_argop4 aop4[1];
	nfs_fh4 fh4 = { sizeof(struct knfs_fh), (char *) &deleg->d_fh };
	
	nfsv4_printk(deleg, "issue_recall(): recalling delegation, stateid=%d\n",
		     (int) deleg->d_stateid);
	
	nfsd4_setup_cb_compound(&cargs, aop4, 1, tag);
	nfsd4_setup_cb_recall(&cargs.argarray.argarray_val[0], deleg->d_stateid, 0, &fh4);
	nfsd4_proc_compoundcb(deleg->d_clnt->cl_callback_clnt, &cargs, &cres);
	
	deleg->d_recalled = 1;

	nfsv4_printk(deleg, "issue_recall(): done\n");
}

void
nfs4_recall_delegations (struct nfs4_file *file, struct svc_client *clp,
			 uint32_t share_access)
{
	struct nfs4_delegation *deleg, *next;
	int pending_recalls;
	struct wait_queue *q = NULL;
	
	nfsv4_printk(deleg, "recall_delegations(): starting on file \"%s\"\n",
		     file->f_file.f_dentry->d_name.name);
	down(&file->f_delegsema);
	
	switch (file->f_delegtype) {
	case OPEN_DELEGATE_NONE:
		goto out;
	case OPEN_DELEGATE_READ:
		if (!(share_access & OPEN4_SHARE_ACCESS_WRITE))
			goto out;
		break;
	case OPEN_DELEGATE_WRITE:
		break;
	default:
		printk("!!! corrupt file->f_delegtype\n");
		goto out;
	}

	/*  this algorithm could be improved; see comment at the beginning
	 *  of the file!
	 */
	for (;;) {
	        pending_recalls = 0;
		for (deleg = file->f_deleglist; deleg; deleg = next) {
			next = deleg->d_next;
			if (deleg->d_clnt != clp) {
				pending_recalls = 1;
				if (!deleg->d_recalled)
					issue_recall(deleg);
			}
		}
		if (!pending_recalls)
			break;

		file->f_wait_count++;
		nfsv4_printk(deleg, "going to sleep waiting for DELEGRETURN\n");
		up(&file->f_delegsema);
		sleep_on(&file->f_wait_queue);   /* will be woken up on DELEGRETURN */
		down(&file->f_delegsema);
		file->f_wait_count--;
		nfsv4_printk(deleg, "received DELEGRETURN; waking up\n");
	}

 out:
	up(&file->f_delegsema);
	nfsv4_printk(deleg, "recall_delegations(): done\n");
}

void
nfs4_issue_delegation (open_delegation4 *resp, struct nfs4_file *file,
		       struct svc_client *clp, uint32_t share_access)
{
	struct inode *inode = file->f_file.f_dentry->d_inode;
	struct nfs4_lock *lock;
	struct nfs4_delegation *deleg;
	uint32_t existing_access;
	unsigned int hashval;
	
	nfsv4_printk(deleg, "issue_delegation(): starting on file \"%s\"\n",
		     file->f_file.f_dentry->d_name.name);

	if (!clp || !clp->cl_callback_clnt) {
		nfsv4_printk(deleg, "issue_delegation(): no callback path\n");
		goto out_no_deleg;
	}
	if (!ISSUE_READ_DELEGATIONS) {
		nfsv4_printk(deleg, "issue_delegation(): delegations disabled\n");
		goto out_no_deleg;
	}
	if (!inode) {
		nfsv4_printk(deleg, "issue_delegation(): dentry is negative?!\n");
		goto out_no_deleg;
	}

	/*
	 *  determine existing share access, taking delegations into account
	 */
	down(&file->f_delegsema);
	
	for (deleg = file->f_deleglist; deleg; deleg = deleg->d_next)
		if (deleg->d_clnt == clp) {       /* XXX: is this ok, or compare clientid's instead? */
			nfsv4_printk(deleg, "issue_delegation(): clnt already has delegation!\n");
			up(&file->f_delegsema);
			goto out_no_deleg;
		}
	switch (file->f_delegtype) {
	case OPEN_DELEGATE_NONE:
		existing_access = 0;
		break;
	case OPEN_DELEGATE_READ:
		existing_access = OPEN4_SHARE_ACCESS_READ;
		break;
	case OPEN_DELEGATE_WRITE:
		existing_access = OPEN4_SHARE_ACCESS_WRITE;
		break;
	default:
		printk("!!! corrupt file->f_delegtype\n");
		up(&file->f_delegsema);
		goto out_no_deleg;
	}
	for (lock = file->f_lock4; lock; lock = lock->l_next)
		if (lock->l_owner->lo_owner->clientid != clp->nfsdclientid)
			existing_access |= lock->l_share_access;
	nfsv4_printk(deleg, "issue_delegation(): access=%d, existing access=%d\n",
		     share_access, existing_access);

	/*
	 *  issue delegation
	 */
	if (existing_access & OPEN4_SHARE_ACCESS_WRITE) {
		nfsv4_printk(deleg, "issue_delegation(): file currently opened for write\n");
		up(&file->f_delegsema);
		goto out_no_deleg;
	}
	if ((share_access & OPEN4_SHARE_ACCESS_WRITE) &&
	    !existing_access && ISSUE_WRITE_DELEGATIONS) {
		/* issue write delegation (not implemented yet) */
		printk("!!! trying to issue a write delegation -- not implemented\n");
		up(&file->f_delegsema);
		goto out_no_deleg;
	}
	else {
		/* issue read delegation */
		open_read_delegation4 *dp = &resp->open_delegation4_u.read;

		nfsv4_printk(deleg, "issue_delegation(): issuing read delegation, stateid=%d\n",
			     (int) global_delegation_stateid);
		
		if (!(deleg = kmalloc(sizeof(*deleg), GFP_KERNEL)))
			goto out_no_deleg;
		hashval = deleg_hashval(global_delegation_stateid);
		deleg->d_stateid = global_delegation_stateid;
		deleg->d_clnt = clp;
		deleg->d_file = file;
		memcpy(&deleg->d_fh, &CURRENT_FH.fh_handle, sizeof(struct knfs_fh));
		deleg->d_next = file->f_deleglist;
		if (deleg->d_next)
			deleg->d_next->d_prev = deleg;
		deleg->d_prev = NULL;
		deleg->d_hashnext = deleg_hashtable[hashval];
		if (deleg->d_hashnext)
			deleg->d_hashnext->d_hashprev = deleg;
		deleg->d_hashprev = NULL;
		deleg_hashtable[hashval] = deleg;
		deleg->d_recalled = 0;
		
		file->f_delegtype = OPEN_DELEGATE_READ;
		file->f_deleglist = deleg;

		resp->delegation_type = OPEN_DELEGATE_READ;
		dp->stateid = global_delegation_stateid;
		dp->recall = 0;
		dp->permissions.type = ACE4_ACCESS_ALLOWED_ACE_TYPE;
		dp->permissions.flag = 0;
		dp->permissions.access_mask = ACE4_GENERIC_READ;
		
		/* the ace is wrong and will be fixed later! */
		dp->permissions.who.utf8string_len = 8;
		dp->permissions.who.utf8string_val = "EVERYONE";

		global_delegation_stateid++;
	}
	
	up(&file->f_delegsema);

 out:
	nfsv4_printk(deleg, "issue_delegation(): returning %s\n",
		     delegation_type_str(resp->delegation_type));
	return;
 out_no_deleg:
	resp->delegation_type = OPEN_DELEGATE_NONE;
	goto out;
}

int
nfsd4_delegreturn(stateid4 stateid)
{
	struct nfs4_delegation *deleg;
	struct nfs4_file *file;
	int hashval = deleg_hashval(stateid);
	int status = 0;
	
	nfsv4_printk(deleg, "nfs4_delegreturn(): called with stateid %d\n", (int) stateid);

	/*  look up delegation by stateid
	 */

	for (deleg = deleg_hashtable[hashval]; deleg; deleg = deleg->d_hashnext)
		if (deleg->d_stateid == stateid)
			goto found;
	nfsv4_printk(deleg, "nfs4_delegreturn(): no such stateid found!\n");
	status = nfserr_bad_stateid;
	goto out;

 found:
	file = deleg->d_file;
	down(&file->f_delegsema);
	if (deleg->d_next)
		deleg->d_next->d_prev = deleg->d_prev;
	if (deleg->d_prev)
		deleg->d_prev->d_next = deleg->d_next;
	else
		file->f_deleglist = deleg->d_next;

	if (deleg->d_hashnext)
		deleg->d_hashnext->d_hashprev = deleg->d_hashprev;
	if (deleg->d_hashprev)
		deleg->d_hashprev->d_hashnext = deleg->d_hashnext;
	else
		deleg_hashtable[hashval]->d_hashnext = deleg->d_hashnext;
	
	kfree(deleg);
	if (!file->f_deleglist)
		file->f_delegtype = OPEN_DELEGATE_NONE;
	wake_up(&file->f_wait_queue);
	up(&file->f_delegsema);
	
 out:
	nfsv4_printk(deleg, "nfs4_delegreturn(): returning status %d\n", status);
	return status;
}
