/*
 * linux/fs/nfsd/nfsdexport.c
 *
 * NFSV4 Implementation
 *  by Andy Adamson <andros@citi.umich.edu>
 *  Copyright (C) 1999 The Regents of the University of Michigan
 *
 * based on work by
 * Copyright (C) 1995, 1996 Olaf Kirch, <okir@monad.swb.de>
 *
 * NFS exporting and validation.
 */

#include <linux/unistd.h>
#include <linux/malloc.h>
#include <linux/stat.h>
#include <linux/in.h>

#include <linux/sunrpc/svc.h>
#include <linux/nfsd/nfsd4_prot.h>
#include <linux/nfsd/nfsfh.h>
#include <linux/nfsd/syscall.h>

#define NFSDDBG_FACILITY	NFSDDBG_EXPORT
#define NFSD_PARANOIA 1

typedef struct svc_client	svc_client;
typedef struct svc_export	svc_export;

/* function prototypes */
static svc_export *	exp_find(svc_client *clp, kdev_t dev);
static svc_export *	exp_parent(svc_client *clp, kdev_t dev,
				   struct dentry *dentry);
static svc_export *	exp_child(svc_client *clp, kdev_t dev,
				  struct dentry *dentry);
static void		exp_unexport_all(svc_client *clp);
static void		exp_do_unexport(svc_export *unexp);
static svc_client *	exp_getclientbyname(char *name,int len);
static void		exp_freeclient(svc_client *clp);
static void		exp_unhashclient(svc_client *clp);
static int		exp_verify_string(char *cp, int max);
static inline int nfs4_cmp_client(nfs_client_id4 *c1,nfs_client_id4 *c2);


/* per client hash table */
#define CLIENT_HASHBITS		6
#define CLIENT_HASHMAX		(1 << CLIENT_HASHBITS)
#define CLIENT_HASHMASK		(CLIENT_HASHMAX - 1)
#define CLIENT_HASH(a) \
		((((a)>>24) ^ ((a)>>16) ^ ((a)>>8) ^(a)) & CLIENT_HASHMASK)
/* XXX: is this adequate for 32bit kdev_t ? */
#define EXPORT_HASH(dev)	((dev) & (NFSCLNT_EXPMAX - 1))

static struct svc_clnthash 	clnt_hash[CLIENT_HASHMAX];

/* Global verifier */
extern char nfsd_verf[]; 

/* new_pseudo_root_exp is stuffed by exportfs ioctl, and
* holds the export struct for the pseudo fs 
*/
struct svc_export        new_pseudo_root_exp;

/* ext2_exported is stuffed by exportfs ioctl, and  holds
* exported ext2 fs's as specified in /etc/exports. these
* are not associated with a client until a user 
* crosses from the pseudo fs into an exported ext2 fs.
*/
struct svc_export        ext2_exported[NFSD4_MAX_EXPORTS];

static int			initialized = 0;

static int			hash_lock = 0;
static int			want_lock = 0;
static int			hash_count = 0;
static struct wait_queue *	hash_wait = NULL;

#define READLOCK		0
#define WRITELOCK		1

/*
 * Find a client's export for a device.
 */
static inline svc_export *
exp_find(svc_client *clp, kdev_t dev)
{
	svc_export *	exp;

	exp = clp->cl_export[EXPORT_HASH(dev)];
	while (exp && exp->ex_dev != dev)
		exp = exp->ex_next;
	return exp;
}

/*
 * Find the export entry matching xdev/xino.
 */
svc_export *
exp_get(svc_client *clp, kdev_t dev, ino_t ino)
{
	svc_export *	exp;
	int i;

	for(i=0;i<NFSD4_MAX_EXPORTS;i++){
		exp = &ext2_exported[i];
		if(exp->ex_dentry == NULL)
			break;
		if (exp->ex_ino == ino && exp->ex_dev == dev)
			goto out;
	}
	exp = NULL;
 out:
	return exp;
}
#if 0
/*
 * Find the client's export entry matching xdev/xino.
 */
svc_export *
exp_get(svc_client *clp, kdev_t dev, ino_t ino)
{
	svc_export *	exp;

	if (!clp)
		return NULL;

	exp = clp->cl_export[EXPORT_HASH(dev)];
	if (exp)
		do {
			if (exp->ex_ino == ino && exp->ex_dev == dev)
				goto out;
		} while (NULL != (exp = exp->ex_next));
	exp = NULL;
 out:
	return exp;
}
#endif /*0*/

/*
 * Look up the device of the parent fs.
 */
static inline int
nfsd_parentdev(kdev_t *devp)
{
	struct super_block	*sb;

	if (!(sb = get_super(*devp)) || !sb->s_root->d_covers)
		return 0;
	if (*devp == sb->s_root->d_covers->d_inode->i_dev)
		return 0;
	*devp = sb->s_root->d_covers->d_inode->i_dev;
	return 1;
}

/*
 * Find the export entry for a given dentry.  <gam3@acm.org>
 */
static svc_export *
exp_parent(svc_client *clp, kdev_t dev, struct dentry *dentry)
{
	svc_export      *exp;
	kdev_t		xdev = dev;
	struct dentry	*xdentry = dentry;
	struct dentry	*ndentry = NULL;

	if (clp == NULL || dentry == NULL)
		return NULL;

	do {
		xdev = dev;
		do {
			exp = clp->cl_export[EXPORT_HASH(xdev)];
			if (exp)
				do {
					ndentry = exp->ex_dentry;
				        if (ndentry == xdentry) {
#ifdef NFSD_PARANOIA
						if (dev == xdev)
							dprintk("nfsd: exp_parent submount over mount.\n");
						else
							dprintk("nfsd: exp_parent found.\n");
#endif
						goto out;
					}
				} while (NULL != (exp = exp->ex_next));
		} while (nfsd_parentdev(&xdev));
		if (IS_ROOT(xdentry))
			break;
	} while ((xdentry = xdentry->d_parent));
	exp = NULL;
 out:
	return exp;
}

/*
 * Find the child export entry for a given fs. This function is used
 * only by the export syscall to keep the export tree consistent.
 * <gam3@acm.org>
 */
static svc_export *
exp_child(svc_client *clp, kdev_t dev, struct dentry *dentry)
{
	svc_export      *exp;
	struct dentry	*xdentry = dentry;
	struct dentry	*ndentry = NULL;

	if (clp == NULL || dentry == NULL)
		return NULL;

	exp = clp->cl_export[EXPORT_HASH(dev)];
	if (exp)
		do {
			ndentry = exp->ex_dentry;
			if (ndentry)
				while ((ndentry = ndentry->d_parent)) {
					if (ndentry == xdentry) {
#ifdef NFSD_PARANOIA
						dprintk("nfsd: exp_child mount under submount.\n");
#endif
						goto out;
					}
					if (IS_ROOT(ndentry))
						break;
				}
		} while (NULL != (exp = exp->ex_next));
	exp = NULL;
 out:
	return exp;
}

static void
create_nfsdclientid(int hash,clientid4 *clid) 
{
	unsigned short *ptr;
	time_t boot = 0;

	boot = *(time_t *)nfsd_verf;  /* server boot time */
	*clid = (long long)boot;
	ptr = (unsigned short *)clid;
	ptr +=2;
	*ptr = hash; /* index into clnt_hash[]  */
	ptr++;
	*ptr = ++clnt_hash[hash].clid_seq_num; /* hash bucket sequence number*/
	nfsv4_printk(setclntid, "NFSV4 create_nfsdclientid [0x%lx , %08lx]\n",
		     (unsigned long)(*clid >> 32),
		     (unsigned long)(*clid & 0xFFFFFFFFUL));    
}

struct svc_client *
exp_getclientbyclientid(clientid4 clid)
{
	struct svc_client *clp = NULL;
	unsigned int hash = GET_NFS4_CLIENTID_HINDEX(clid);

	for (clp = clnt_hash[hash].h_hash_head; clp; clp = clp->cl_next) {
		if (clp->nfsdclientid == clid)
			return clp;
	}
	return NULL;
}
/*
 * Create pseudo fs and ext2_exported[]
 * pay no attention to nxp->ex_client
 */
int
exp_export(struct nfsctl_export *nxp)
{
	svc_export	*exp;
	struct dentry	*dentry = NULL;
	struct inode	*inode = NULL;
	int i,err;
	kdev_t		dev;
	ino_t		ino;

	/* Consistency check */
	err = -EINVAL;
	if (!exp_verify_string(nxp->ex_path, NFS_MAXPATHLEN))
		goto out;

	nfsv4_printk(vfs, "exp_export called for %s:%s (%x/%ld fl %x).\n",
		     nxp->ex_client, nxp->ex_path,
		     nxp->ex_dev, nxp->ex_ino, nxp->ex_flags);
	dev = to_kdev_t(nxp->ex_dev);
	ino = nxp->ex_ino;

	/* Look up the dentry */
	err = -EINVAL;
	dentry = lookup_dentry(nxp->ex_path, NULL, 0);
	if (IS_ERR(dentry))
		goto out;

	err = -ENOENT;
	inode = dentry->d_inode;
	if (!inode)
		goto finish;
	err = -EINVAL;
	if (inode->i_dev != dev || inode->i_ino != nxp->ex_ino) {
		printk(KERN_DEBUG "exp_export: i_dev = %x, dev = %x\n",
		       inode->i_dev, dev); 
		/* I'm just being paranoid... */
		goto finish;
	}

	/* We currently export only dirs and regular files.
	 * This is what umountd does.
	 */
	err = -ENOTDIR;
	if (!S_ISDIR(inode->i_mode) && !S_ISREG(inode->i_mode))
		goto finish;

/* pseudo_fs creates directories in the pseudo fs for each
 * path component, and mounts dentry on it's pseudo equivalent.
 */
	nfsv4_printk(pseudo, "NFSV4 exp_export calling pseudo_exp_it\n");
	err = pseudo_exp_it(dentry,nxp->ex_path);
	if(err != 0){
		printk("NFSV4 exp_export pseudo_exp_it on [%s] ino %ld failed\n",
		       nxp->ex_path,inode->i_ino); 
	}
	exp = get_new_pseudo_root_exp();
	nfsv4_printk(pseudo, "NFSV4 exp_export calling print_pseudo_dentry\n");
	print_pseudo_dentry(exp->ex_dentry);

	err = -EINVAL;
	if (!(inode->i_sb->s_type->fs_flags & FS_REQUIRES_DEV) ||
	    inode->i_sb->s_op->read_inode == NULL) {
		dprintk("exp_export: export of invalid fs type.\n");
		goto finish;
	}
	err = 0;

	/* ex_path has been added to the pseudo fs.  
	 * now add an ext2_exported[] entry
	 */
	for(i=0;i<NFSD4_MAX_EXPORTS;i++){
		if(ext2_exported[i].ex_dentry == dentry)
			/* already there, jobs done */
			goto finish;	
	}
	exp = NULL;
	for(i=0;i<NFSD4_MAX_EXPORTS;i++){
		if(ext2_exported[i].ex_dentry == NULL){
			exp = &ext2_exported[i];
			break;
		}
	}
	if(!exp){
		err = nfserr_resource;
		printk("NFSV4: no more ext2_exported entries! can't export %s\n",
		       nxp->ex_path);
		goto finish;
	}

	err = -EINVAL;
	if (!(inode->i_sb->s_type->fs_flags & FS_REQUIRES_DEV) ||
	    inode->i_sb->s_op->read_inode == NULL) {
		printk("exp_export: export of invalid fs type.\n");
		goto finish;
	}

	printk("NFSV4: exp_export nxp->ex_path %s len %d\n",nxp->ex_path,strlen(nxp->ex_path));
	memset(exp->ex_path,0,NFS_MAXPATHLEN+1);
	memcpy(exp->ex_path, nxp->ex_path,strlen(nxp->ex_path));
	exp->ex_client = NULL;
	exp->ex_sec = 0;
	exp->ex_parent = NULL;
	exp->ex_dentry = dentry;
	exp->ex_flags = nxp->ex_flags;
	exp->ex_dev = dev;
	exp->ex_ino = ino;
	exp->ex_anon_uid = nxp->ex_anon_uid;
	exp->ex_anon_gid = nxp->ex_anon_gid;
	/* initially, we set ex_sec via exportfs nfsctl_export by using
	 * the "10.10.10.10" client string. (ugh). this will change soon
	 * to a new field in /etc/exports. 
	 * we also have one gss_svc mode hardwired - RPC_GSS_SVC_NONE.
	 * futhermore, we have yet to decide where the gss creds will be stored
	 * we think in a slab cache managed by the gssd interface. so, we
	 * just check the cred struct flavor. if it's gss -> RPC_GSS_SVC_NONE.
	 */
	if(!(memcmp(nxp->ex_client,"10.10.10.10",11))){
		printk("NFSV$ exp_export: ADDSEC rpc_gss_svc_none\n");
		exp->ex_sec = RPC_GSS_SVC_NONE;
		printk("NFSV4: exp_export RPC_GSS_SVC_NONE: exp->ex_sec: %d\n",exp->ex_sec);
	}
 	printk("NFSV4: created export entry %s flags: 0x%x EX_SECURE(exp) %d ext2_exported index %d\n", exp->ex_path, exp->ex_flags,EX_SECURE(exp),i);

 out:
	return err;

	/* Release the dentry */
 finish:
	dput(dentry);
	goto out;
}

/*
* Export an ext2 file system for a client. 
* Note this is called under the exp_readlock 
* setup in nfssvc.c:nfsd() surrounding the svc_process call
*/

/* we may want to stuff an unattached (no clp) svc_export per 
* exported fs at exportfs time, then simply attach it 
* to a client at mount point crossing.
* 
* Note: called by nfs4_lookup_pseudo(), and does not dput the dentry.
*/

int
exp_client_export(struct svc_client *clp,struct dentry *dentry)
{
	svc_export	*exp, *parent;
	svc_export	**head;
	struct inode	*inode = NULL;
	int	i,index = -1, err = 0;

	err = nfserr_noent;
	inode = dentry->d_inode;
	if (!inode)
		goto out;

	/* give up exp_readlock (nfssvc.c:nfsd()) */
	exp_unlock();

	/* Try to lock the export table for update */
	if ((err = exp_writelock()) < 0)
		goto out_readlock;

	err = nfserr_inval;
	if (!(inode->i_sb->s_type->fs_flags & FS_REQUIRES_DEV) ||
	    inode->i_sb->s_op->read_inode == NULL) {
		printk("exp_client_export: export of invalid fs type.\n");
		goto out_unlock;
	}

	if ((parent = exp_child(clp, inode->i_dev, dentry)) != NULL) {
		printk("exp_client_export: export not valid (Rule 3).\n");
		goto out_unlock;
	}
	/* Is this is a sub-export, must be a proper subset of FS */
	if ((parent = exp_parent(clp, inode->i_dev, dentry)) != NULL) {
		if (inode->i_dev == parent->ex_dev) {
			printk("exp_client_export: sub-export not valid (Rule 2).\n");
			goto out_unlock;
		}
	}
	err = nfserr_inval;
	/* find file system in ext2_exported[] */
	for(i=0;i<NFSD4_MAX_EXPORTS;i++){
		if(ext2_exported[i].ex_dentry == dentry)
			index = i;	
	}
	if(index == -1)
		goto out_unlock;

	err = nfserr_resource;
	if (!(exp = kmalloc(sizeof(*exp), GFP_USER)))
		goto out_unlock;
	printk("nfsd: created export entry %p for client [%p]\n", exp, clp);

	memcpy(exp->ex_path, ext2_exported[index].ex_path,strlen(ext2_exported[index].ex_path));
	exp->ex_client = clp;
	exp->ex_parent = parent;
	exp->ex_dentry = ext2_exported[index].ex_dentry;
	exp->ex_flags = ext2_exported[index].ex_flags;
	exp->ex_dev = ext2_exported[index].ex_dev;
	exp->ex_ino = ext2_exported[index].ex_ino;
	exp->ex_anon_uid = ext2_exported[index].ex_anon_uid;
	exp->ex_anon_gid = ext2_exported[index].ex_anon_gid;

	nfsv4_printk(export,"NFSV4: created export entry %s flags: 0x%x EX_SECURE(exp) %d\n", exp->ex_path, exp->ex_flags,EX_SECURE(exp));


	/* Update parent pointers of all exports */
	if (parent) {
		for (i = 0; i < NFSCLNT_EXPMAX; i++) {
			svc_export *temp = clp->cl_export[i];

			while (temp) {
				if (temp->ex_parent == parent)
					temp->ex_parent = exp;
				temp = temp->ex_next;
			}
		}
	}

	head = clp->cl_export + EXPORT_HASH(inode->i_dev);
	exp->ex_next = *head;
	*head = exp;

	err = 0;

	/* Unlock write lock on hashtable */
 out_unlock:
	exp_unlock();

/* Restore nfssvc.c:nfsd() readlock */
 out_readlock:
	exp_readlock();

 out:
	return err;
}

/*
 * Unexport a file system. The export entry has already
 * been removed from the client's list of exported fs's.
 */
static void
exp_do_unexport(svc_export *unexp)
{
	svc_export	*exp;
	svc_client	*clp;
	struct dentry	*dentry;
	struct inode	*inode;
	int		i;

	/* Update parent pointers. */
	clp = unexp->ex_client;
	for (i = 0; i < NFSCLNT_EXPMAX; i++) {
		for (exp = clp->cl_export[i]; exp; exp = exp->ex_next){
			if (exp->ex_parent == unexp)
				exp->ex_parent = unexp->ex_parent;
		}
	}

	dentry = unexp->ex_dentry;
	inode = dentry->d_inode;
	if (unexp->ex_dev != inode->i_dev || unexp->ex_ino != inode->i_ino)
		printk(KERN_WARNING "nfsd: bad dentry in unexport!\n");
	dput(dentry);

	kfree(unexp);
}

/*
 * Revoke all exports for a given client.
 * This may look very awkward, but we have to do it this way in order
 * to avoid race conditions (aka mind the parent pointer).
 */
static void
exp_unexport_all(svc_client *clp)
{
	svc_export	*exp;
	int		i;

	printk("unexporting all fs's for clnt %p\n", clp);
	for (i = 0; i < NFSCLNT_EXPMAX; i++) {
		exp = clp->cl_export[i];
		clp->cl_export[i] = NULL;
		while (exp) {
			svc_export *next = exp->ex_next;
			exp_do_unexport(exp);
			exp = next;
		}
	}
}

/*
 * unexport syscall.
 */
int
exp_unexport(struct nfsctl_export *nxp)
{
	svc_client	*clp;
	svc_export	**expp, *exp = NULL;
	int		err;

	/* Consistency check */
	if (!exp_verify_string(nxp->ex_client, NFSCLNT_IDMAX))
		return -EINVAL;

	if ((err = exp_writelock()) < 0)
		goto out;

	err = -EINVAL;
	clp = exp_getclientbyname(nxp->ex_client,strlen(nxp->ex_client));
	if (clp) {
		expp = clp->cl_export + EXPORT_HASH(nxp->ex_dev);
		while ((exp = *expp) != NULL) {
			if (exp->ex_dev == nxp->ex_dev) {
				if (exp->ex_ino == nxp->ex_ino) {
					*expp = exp->ex_next;
					exp_do_unexport(exp);
					err = 0;
					break;
				}
			}
			expp = &(exp->ex_next);
		}
	}

	exp_unlock();
 out:
	return err;
}

/*
 * Obtain the root fh on behalf of a client.
 * This could be done in user space, but I feel that it adds some safety
 * since its harder to fool a kernel module than a user space program.
 */
int
exp_rootfh(struct svc_client *clp, kdev_t dev, ino_t ino,
	   char *path, struct knfs_fh *f)
{
	struct svc_export	*exp;
	struct dentry		*dentry = NULL;
	struct inode		*inode;
	struct svc_fh		fh;
	int			err;

	err = -EPERM;
	if (path) {
		if (!(dentry = lookup_dentry(path, NULL, 0))) {
			printk("nfsd: exp_rootfh path not found %s", path);
			return -EPERM;
		}
		dev = dentry->d_inode->i_dev;
		ino = dentry->d_inode->i_ino;
	
		dprintk("nfsd: exp_rootfh(%s [%p] %s:%x/%ld)\n",
			path, dentry, clp->cl_ident, dev, ino);
		exp = exp_parent(clp, dev, dentry);
	} else {
		dprintk("nfsd: exp_rootfh(%s:%x/%ld)\n",
			clp->cl_ident, dev, ino);
		if ((exp = exp_get(clp, dev, ino)))
			if (!(dentry = dget(exp->ex_dentry))) {
				printk("exp_rootfh: Aieee, NULL dentry\n");
				return -EPERM;
			}
	}
	if (!exp) {
		dprintk("nfsd: exp_rootfh export not found.\n");
		goto out;
	}

	inode = dentry->d_inode;
	if (!inode) {
		printk("exp_rootfh: Aieee, NULL d_inode\n");
		goto out;
	}
	if (inode->i_dev != dev || inode->i_ino != ino) {
		printk("exp_rootfh: Aieee, ino/dev mismatch\n");
		printk("exp_rootfh: arg[dev(%x):ino(%ld)]"
		       " inode[dev(%x):ino(%ld)]\n",
		       dev, ino, inode->i_dev, inode->i_ino);
	}

	/*
	 * fh must be initialized before calling fh_compose
	 */
	fh_init(&fh);
	fh_compose(&fh, exp, dentry);
	memcpy(f, &fh.fh_handle, sizeof(struct knfs_fh));
	fh_put(&fh);
	return 0;

 out:
	dput(dentry);
	return err;
}

/*
 * Hashtable locking. Write locks are placed only by user processes
 * wanting to modify export information.
 */
void
exp_readlock(void)
{
	while (hash_lock || want_lock)
		sleep_on(&hash_wait);
	hash_count++;
}

int
exp_writelock(void)
{
	/* fast track */
	if (!hash_count && !hash_lock) {
	lock_it:
		hash_lock = 1;
		return 0;
	}

	current->sigpending = 0;
	want_lock++;
	while (hash_count || hash_lock) {
		interruptible_sleep_on(&hash_wait);
		if (signal_pending(current))
			break;
	}
	want_lock--;

	/* restore the task's signals */
	spin_lock_irq(&current->sigmask_lock);
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);

	if (!hash_count && !hash_lock)
		goto lock_it;
	return nfserr_serverfault;
}

void
exp_unlock(void)
{
	if (!hash_count && !hash_lock)
		printk(KERN_WARNING "exp_unlock: not locked!\n");
	if (hash_count)
		hash_count--;
	else
		hash_lock = 0;
	wake_up(&hash_wait);
}

/*
 * Find a valid client given an inet address and a clientid. 
 * We always move the most
 * recently used client to the front of the hash chain to speed up
 * future lookups.
 * Locking against other processes is the responsibility of the caller.
 */
struct svc_client *
exp_getclient(struct sockaddr_in *sin,clientid4 clid)
{
	struct svc_client	**hp,*client;
	unsigned long		addr = sin->sin_addr.s_addr;
	int hash = CLIENT_HASH(addr);

	if (!initialized)
		return NULL;

	for(hp = &clnt_hash[hash].h_hash_head;(client = *hp);hp= &client->cl_next){
		if((client->cl_addr.s_addr == addr) && (client->nfsdclientid == clid)){
			if(hp != &clnt_hash[hash].h_hash_head){
				*hp = client->cl_next;
				client->cl_next = clnt_hash[hash].h_hash_head;
				clnt_hash[hash].h_hash_head = client;   
			}
			return client;
		}
	}
	return NULL;
}
struct svc_client *
exp_getclientbyident(struct sockaddr_in *sin,nfs_client_id4 *c1)
{
	struct svc_client	**hp,*client;
	unsigned long		addr = sin->sin_addr.s_addr;
	int hash = CLIENT_HASH(addr);

	if (!initialized)
		return NULL;

	for(hp = &clnt_hash[hash].h_hash_head;(client = *hp);hp= &client->cl_next){
		if((client->cl_addr.s_addr == addr) && 
		   (nfs4_cmp_client(&client->clnt4,c1))){
			if(hp != &clnt_hash[hash].h_hash_head){
				*hp = client->cl_next;
				client->cl_next = clnt_hash[hash].h_hash_head;
				clnt_hash[hash].h_hash_head = client;   
			}
			return client;
		}
	}
	return NULL;
}

/*
 * Find a client given its identifier.
 */
static svc_client *
exp_getclientbyname(char *ident,int len)
{
	svc_client *	clp;
	int i;

	if(len == 0) 
		return NULL;
	for(i=0;i<CLIENT_HASHMAX;i++){
		for (clp = clnt_hash[i].h_hash_head; clp; clp = clp->cl_next) {
			if (!memcmp(clp->cl_ident,ident,len))
				return clp;
		}
	}
	return NULL;
}

struct flags {
	int flag;
	char *name[2];
} expflags[] = {
	{ NFSEXP_READONLY, {"ro", "rw"}},
	{ NFSEXP_INSECURE_PORT, {"insecure", ""}},
	{ NFSEXP_ROOTSQUASH, {"root_squash", "no_root_squash"}},
	{ NFSEXP_ALLSQUASH, {"all_squash", ""}},
	{ NFSEXP_ASYNC, {"async", ""}},
	{ NFSEXP_GATHERED_WRITES, {"wdelay", ""}},
	{ NFSEXP_UIDMAP, {"uidmap", ""}},
	{ NFSEXP_KERBEROS, { "kerberos", ""}},
	{ NFSEXP_SUNSECURE, { "sunsecure", ""}},
	{ NFSEXP_CROSSMNT, {"nohide", ""}},
	{ NFSEXP_NOSUBTREECHECK, {"no_subtree_check", ""}},
	{ NFSEXP_NOAUTHNLM, {"no_auth_nlm", ""}},
	{ 0, {"", ""}}
};

#if 0
static int
exp_flags(char *buffer, int flag)
{
	int len = 0, first = 0;
	struct flags *flg = expflags;

	for (;flg->flag;flg++) {
		int state = (flg->flag & flag)?0:1;
		if (!flg->flag)
			break;
		if (*flg->name[state]) {
			len += sprintf(buffer + len, "%s%s",
				       first++?",":"", flg->name[state]);
		}
	}
	return len;
}

int
exp_procfs_exports(char *buffer, char **start, off_t offset,
		   int length, int *eof, void *data)
{
	struct svc_client	**hp, **head, *tmp;
	struct svc_client	*clp;
	svc_export *exp;
	off_t	pos = 0;
        off_t	begin = 0;
        int	len = 0;
	int	i;

        len += sprintf(buffer, "# Version 1.0\n");
        len += sprintf(buffer+len, "# Path Client(Flags) # IPs\n");

	for (clp = clients; clp; clp = clp->cl_next) {
		for (i = 0; i < NFSCLNT_EXPMAX; i++) {
			exp = clp->cl_export[i];
			while (exp) {
				int first = 0;
				len += sprintf(buffer+len, "%s\t", exp->ex_path);
				len += sprintf(buffer+len, "%s", clp->cl_ident);
				len += sprintf(buffer+len, "(");

				len += exp_flags(buffer+len, exp->ex_flags);
				len += sprintf(buffer+len, ") # ");

				head = &clnt_hash[CLIENT_HASH(clp->cl_addr.s_addr)].h_hash_head;
				for (hp = head; (tmp = *hp) != NULL; hp = &(tmp->cl_next)) {
					if (tmp->cl_addr.s_addr == clp->cl_addr.s_addr) {
						if (first++) len += sprintf(buffer+len, "%s", " ");
						if (tmp->h_client != clp)
							len += sprintf(buffer+len, "(");
						len += sprintf(buffer+len, "%d.%d.%d.%d",
							       htonl(clp->cl_addr.s_addr) >> 24 & 0xff,
							       htonl(clp->cl_addr.s_addr) >> 16 & 0xff,
							       htonl(clp->cl_addr.s_addr) >>  8 & 0xff,
							       htonl(clp->cl_addr.s_addr) >>  0 & 0xff);
						if (tmp->h_client != clp)
							len += sprintf(buffer+len, ")");
						break;
					}
				}
				exp = exp->ex_next;

				buffer[len++]='\n';

				pos=begin+len;
				if(pos<offset) {
					len=0;
					begin=pos;
				}
				if (pos > offset + length)
					goto done;
			}
		}
	}

	*eof = 1;

 done:
	*start = buffer + (offset - begin);
	len -= (offset - begin);
	if ( len > length )
		len = length;
	return len;
}
#endif /*0*/

/*
* stub out until change exportfs
*/
int
exp_addclient(struct nfsctl_client *ncp)
{
	int i;
	printk("NFSV4 ADDCLIENT. cl_ident: %s\n",ncp->cl_ident);
	printk("                 cl_naddr: %d\n",ncp->cl_naddr);
	for (i = 0; i < ncp->cl_naddr; i++) 
		printk("                 cl_addrlist[%d]: %d\n",i,ncp->cl_addrlist[i].s_addr);
	printk("                 cl_fhkeytype: %d\n",ncp->cl_fhkeytype);
	printk("                 cl_fhkeylen: %d\n",ncp->cl_fhkeylen);

	return 0;
}

/* 
* add a new client 
* called from vfs.c
*/
int
exp_newaddclient(struct svc_rqst *rqstp)
{
	svc_client *		clp;
	int err,hash;

	printk("NFSV4 new ADDCLIENT \n");

	/* give up exp_readlock (nfssvc.c:nfsd()) */
	exp_unlock();

	/* Try to lock the export table for update */
	if ((err = exp_writelock()) < 0)
		goto out_readlock;

	err = nfserr_resource;
	if (!(clp = kmalloc(sizeof(*clp), GFP_KERNEL)))
		goto out_unlock;
	memset(clp, 0, sizeof(*clp));

	printk("created client (%p)\n",clp);

	/* Copy address. */
	clp->cl_addr = rqstp->rq_addr.sin_addr;

	/* initialize semaphore */
	clp->cl_sema = MUTEX;

	/* Insert client into hashtable. */
	hash = CLIENT_HASH(rqstp->rq_addr.sin_addr.s_addr);

	clp->cl_next = clnt_hash[hash].h_hash_head;
	clnt_hash[hash].h_hash_head = clp;

	if(!rqstp->rq_client)
		rqstp->rq_client = clp;
	err = 0;

	/* Unlock write lock on hashtable */
 out_unlock:
	exp_unlock();

/* Restore nfssvc.c:nfsd() readlock */
 out_readlock:
	exp_readlock();
	return err;
}

/*
 * Delete a client given an identifier.
 */
int
exp_delclient(struct nfsctl_client *ncp)
{
	int		err = 0;
#if 0
	svc_client	**clpp, *clp;

	err = -EINVAL;
	if (!exp_verify_string(ncp->cl_ident, NFSCLNT_IDMAX))
		goto out;

	/* Lock the hashtable */
	if ((err = exp_writelock()) < 0)
		goto out;

	err = -EINVAL;
	for (clpp = &clients; (clp = *clpp); clpp = &(clp->cl_next))
		if (!strcmp(ncp->cl_ident, clp->cl_ident))
			break;

	if (clp) {
		*clpp = clp->cl_next;
		exp_freeclient(clp);
		err = 0;
	}

	exp_unlock();
 out:
#endif /*0*/
	return err;
}

/*
 * Free a client. 
 */
static void
exp_freeclient(svc_client *clp)
{
	exp_unhashclient(clp);
	exp_unexport_all(clp);
	kfree (clp);
}

/*
 * Remove client from hashtable.
 */
static void
exp_unhashclient(svc_client *clp)
{
	struct svc_client **hp, *client;
	int hash;
	
	hash = CLIENT_HASH(clp->cl_addr.s_addr);
	for(hp = &clnt_hash[hash].h_hash_head;(client = *hp);hp= &client->cl_next){
		if(client == clp){
			*hp = client->cl_next;
		}	
	}
}

/*
 * Verify that string is non-empty and does not exceed max length.
 */
static int
exp_verify_string(char *cp, int max)
{
	int	i;

	for (i = 0; i < max; i++)
		if (!cp[i])
			return i;
	cp[i] = 0;
	printk(KERN_NOTICE "nfsd: couldn't validate string %s\n", cp);
	return 0;
}

/*
 * Initialize the exports module.
 */
void
nfsd_export_init(void)
{
	int		i;

	printk("NFSV4: initializing export module.\n");
	if (initialized)
		return;
	for (i = 0; i < CLIENT_HASHMAX; i++){
		clnt_hash[i].h_hash_head = NULL;
		clnt_hash[i].clid_seq_num = 0;	
	}
	initialized = 1;
}

/*
 * Shutdown the exports module.
 */
void
nfsd_export_shutdown(void)
{
	int	i;

	printk("nfsd: EXPORT: shutting down export module.\n");
	if (!initialized)
		return;
	if (exp_writelock() < 0) {
		printk(KERN_WARNING "Weird: hashtable locked in exp_shutdown");
		return;
	}
	for (i = 0; i < CLIENT_HASHMAX; i++) {
		while (clnt_hash[i].h_hash_head)
			exp_freeclient(clnt_hash[i].h_hash_head);
	}
	exp_unlock();
	printk("nfsd: export shutdown complete.\n");
}

int
nfsd_pseudo_exp_init(struct dentry *dentry)
{
	struct inode    *inode;
	int i,err = 0;

	printk("NFSV4 nfsd_pseudo_exp_init\n");
	inode = dentry->d_inode;
	if (!inode) {
		err = 1;
		printk("nfsd_pseudo_exp_init: Aieee, NULL d_inode\n");
		goto out;
	}
	strncpy(new_pseudo_root_exp.ex_path, "/",1);
	/* pseudo fs root not associated with any client */
	new_pseudo_root_exp.ex_client = NULL;
	/* not a mount point so no parent */
	new_pseudo_root_exp.ex_parent = NULL;
	/* no rpc_gss_svc security type */
	new_pseudo_root_exp.ex_sec = 0;
	
	new_pseudo_root_exp.ex_dentry = dentry;
	/* more than readonly. actually readdir only */
	new_pseudo_root_exp.ex_flags = NFSEXP_READONLY | NFSEXP_PSEUDO_EXP;
	new_pseudo_root_exp.ex_dev = dentry->d_inode->i_dev;
	new_pseudo_root_exp.ex_ino = dentry->d_inode->i_ino;
	/* uid,gid. hmmm. needs to be readable by all lookups..
	 * lets use zero, and do no access checks for now.
	 */
	new_pseudo_root_exp.ex_anon_uid = 0;
	new_pseudo_root_exp.ex_anon_gid = 0;

	nfsv4_printk(export,"NFSV4 nfsd_pseudo_exp_init: dev/ino (%d/%ld)\n",new_pseudo_root_exp.ex_dev,new_pseudo_root_exp.ex_ino);

	/* initialize ext2_exports array */
	for(i=0;i<NFSD4_MAX_EXPORTS;i++)
		ext2_exported[i].ex_dentry = NULL;
 out:
	return err;
}

/* SETCLIENTID_CONFIRM
*
*/
int
nfsd4_setclientid_confirm(struct svc_rqst *rqstp,SETCLIENTID_CONFIRM4args *argp)
{
	struct svc_client *clp;
	int err = nfserr_inval;
	clientid4 clid,clpcid;

	nfsv4_printk(setclntid,"NFSV4 nfsd4_setclientid_confirm\n");
	
	/* unpack the clientid from setclientid_confirm*/
	clid = *(clientid4 *)argp->setclientid_confirm;
	nfsv4_printk(setclntid,
		     "NFSD4: nfsd4_setclientid_confirm: args->setclientid_confirm:(clientid) [0x%lx , %08lx]\n",
		     (unsigned long)(clid >> 32),
		     (unsigned long)(clid & 0xFFFFFFFFUL));

	/* release the exp_readlock() called in nfssvc.c: nfsd() 
	 * that surrounds the svc_process() call.
	 */
	exp_unlock();

	/* lock the hash table - writing */
	if ((err = exp_writelock()) < 0)
		goto out;

	/* find client in clnt_hash */
	err = nfserr_inval;
	clp = exp_getclient(&rqstp->rq_addr,clid);  
	if(!clp){
		nfsv4_printk(setclntid,"NFSV4: nfsd4_setclintid_confirm exp_getclient failed addr 0x%x\n",rqstp->rq_addr.sin_addr.s_addr); 
		goto out_unlock;
	}

	err = nfserr_inval; /*???*/
	if(clp->clid_state != CLID_NEGOTIATING){
		printk("NFSV4: setclientid_confirm: wrong clid_state: %d\n",
		       clp->clid_state);
		goto out_unlock;
	}

	clpcid = *(clientid4 *)clp->setclientid_confirm;
	nfsv4_printk(setclntid,
		     "NFSD4: nfsd4_setclientid_confirm: clp->setclientid_confirm:(clientid) [0x%lx , %08lx]\n",
		     (unsigned long)(clpcid >> 32),
		     (unsigned long)(clpcid & 0xFFFFFFFFUL));

	/* compare verifier with verifier stored in clp */
	err = cmp_verifier(clp->setclientid_confirm,argp->setclientid_confirm);
	if(!err){
		printk("NFSV4: setclientid_confirm: verifier mismatch\n");
		err = nfserr_inval;
		goto out_unlock;
	}
	else
		clp->clid_state = CLID_CONFIRMED;
	err = 0;

 out_unlock:
	/* unlock hash table - writing */
	exp_unlock();

 out:
	/* restore the nfssvc:nfsd() exp_readlock */
	exp_readlock();
	nfsv4_printk(setclntid,"NFSV4 setclientid_confirm returns %d\n",
		     ntohl(err));
	return err;
}

static inline int
nfs4_cmp_client(nfs_client_id4 *c1,nfs_client_id4 *c2)
{
	nfsv4_printk(lock,
		     "NFSV4: nfs4_cmp_client: c1 len %d c2 len %d memcmp %d\n",
		     c1->id.id_len,c2->id.id_len,memcmp(c1->id.id_val,c2->id.id_val,c2->id.id_len));
	return ((c1->id.id_len == c2->id.id_len) &&
		!memcmp(c1->id.id_val,c2->id.id_val,c2->id.id_len));  
}



/* 
* creation of the svc_client's has been removed from the 
* ADDCLIENT ioctl - NFSD will no longer parse the /etc/exports at 
* boot to look for clients (NFSCTL_ADDCLIENT).
* instead, svc_client creation is now split between LOOKUP and SETCLIENTID.
* this is because LOOKUP doesn't require a clientid and therefore
* a client should be able to browse an exported fs (given correct permission) 
* without setting a clientid.
*
* so, at startup, ext2_exports[] stuffed with the exported fs's.
* and when the pseudo fs/ext2 boundry is crossed in LOOKUP, newaddclient()
* and then exp_client_export() are called to both create the svc_client and 
* associate the correct export with the svc_client. note that the clientid
* portions of the svc_client are NOT set. in this case, all that needs to be
* done is the setting of the clientid. 
* 
* a client could attempt an open of an ext2 exported file without having called
* lookup. in this case the vfs layer calls the lookup. 
*
* we reference svc_client structs both by address and clientid.
* when we create the svc_client via the LOOKUP path, there is no clientid.
* therefore, we must search both by address and by name to make sure we 
* we locate all possible previous svc_clients and client id useage.
*/

int
nfsd4_setclientid(struct svc_rqst *rqstp, SETCLIENTID4args *argp,SETCLIENTID4res *rptr)
{
	struct svc_client *clp=NULL; 
	int err = 0;
	char dummy[] = "";
 
	nfsv4_printk(setclntid,"NFSV4 nfsd4_setclientid\n");

	if(argp->client.id.id_len > NFSCLNT_IDMAX)
		argp->client.id.id_len = NFSCLNT_IDMAX;

	/*
	 * the export hash table is already locked for reading - nfssvc.c:nfsd()
	 */

	/* LOOKUP creates svc_clients with no setclientid state.
	 *
	 * OK. need to create multiple svc_clients per address,
	 * because you can have multiple nfsclients on one ipaddr.
	 * so, in nfssvc.c we call exp_getclient with a  
	 * clientid of 0. if found, this svc_client was created by nfsd4_lookup
	 * when a SETCLIENTID has yet to be performed. note that this
	 * svc_client struct will service lookups for any client associated 
	 * with the ipaddress. here, we nail the svc_client struct to 
	 * the nfsclient running behind the ipaddr by stuffing the clientid4.
	 * the next lookup from the ipaddr will create a new svc_client with
	 * a zero clientid, and will again service all nfsclients behind the
	 * ipaddr - even those that have called setclientid .
	 */

	/* search the clientid space for a match in nfs_client_id4.id */
	clp = exp_getclientbyname(argp->client.id.id_val,argp->client.id.id_len);
	if(clp){ 
		nfsv4_printk(setclntid,
			     "NFSV4 nfsd4_setclientid found client struct with same client.id\n");

		/* check client addr */
		if(!(rqstp->rq_addr.sin_addr.s_addr == clp->cl_addr.s_addr)) {
			nfsv4_printk(setclntid,
				     "NFSV4 nfsd4_setclientid same client.id: different addr\n");
			/* not same client */
			err = nfserr_clid_inuse;	
			goto out_inuse;
		}
		/* same client*/
		nfsv4_printk(setclntid,
			     "NFSV4 nfsd4_setclientid same client.id,addr\n");
		/* check client verifier */
		if(cmp_verifier(clp->clnt4.verifier,argp->client.verifier)){
			nfsv4_printk(setclntid,
				     "NFSV4 nfsd4_setclientid same client.id, addr, verifier\n");
			/* check princ:uid */
			if(clp->cl_uid == rqstp->rq_cred.cr_uid){ /* RETRANS*/
				nfsv4_printk(setclntid,
					     "NFSV4 nfsd4_setclientid same client.id, addr, verifier,princuid\n");
				/* Don't know what the heck to do here....*/
				err = nfserr_inval;
				goto out_err;
			}else{ 
				nfsv4_printk(setclntid,
					     "NFSV4 nfsd4_setclientid same client.id, addr, verifier:different princuid\n");
				err = nfserr_clid_inuse;	
				goto out_inuse;
			}	
		}else{ /* cmp verifiers */
			nfsv4_printk(setclntid,
				     "NFSV4 nfsd4_setclientid same client.id, addr:different verifier\n");
			/* check princ:uid */
			if(clp->cl_uid == rqstp->rq_cred.cr_uid){
				nfsv4_printk(setclntid,
					     "NFSV4 nfsd4_setclientid same client.id,addr,princuid:different verifier\n");
				release_locks(clp->nfsdclientid);
				goto haveclp;
			}else{ /* ROGUE */
				nfsv4_printk(setclntid,
					     "NFSV4 nfsd4_setclientid same client.id,addr:different verifier,princuid\n");
				/* which error? not NFS4ERR_CLID_INUSE - we believe this is a rogue!
				 * so we don't want to encourage more SETCLIENTID negotiation. */
				err = nfserr_inval;
				goto out_err;
			}
		} /* cmp verifier */
	}else {/* clp?*/

		/* this will occur if the client has yet to call LOOKUP (i.e. mount)
		 * which is the only other place exp_newaddclient is called.
		 */
		if(!rqstp->rq_client){
			err = exp_newaddclient(rqstp);
			printk("NFSV4 exp_newaddclient returns %d\n",err);
			if(err){
				goto out;
			}
		}
		clp = rqstp->rq_client;
	} /* clp */

	/* all checks passed. go ahead and add this client addr/clid  */

	/* release the exp_readlock() called in nfssvc.c: nfsd() 
	 * that surrounds the svc_process() call.
	 */
	exp_unlock();

	/* lock the hash table - writing */
	if ((err = exp_writelock()) < 0)
		goto out;

	if(!clp){ /* even OPEN4 goes through lookup which creates the svc_client
		   * - at least until i implement atomic create.
		   */
		printk("NFSV4 nfsd4_setclientid - NEED TO CREATE SVC_CLIENT\n");
		err =  nfserr_inval;
		goto out_unlock;
	}
 haveclp:

	/* test callback path */
	nfsd4_proc_nullcb(clp, &argp->callback);
	
	/* set nfs_client_id4 client id */
	set_verifier(clp->clnt4.verifier,argp->client.verifier);	
	memcpy(clp->cl_ident,argp->client.id.id_val,argp->client.id.id_len);
	clp->clnt4.id.id_len = argp->client.id.id_len;
	clp->clnt4.id.id_val = clp->cl_ident;        

	/* create a clientid4 to pass back to the client */
	create_nfsdclientid(CLIENT_HASH(rqstp->rq_addr.sin_addr.s_addr),&clp->nfsdclientid);	

	nfsv4_printk(setclntid,"NFSV4 nfsd4_setclientid. clp->nfsdclientid: %Ld verifier:\n",clp->nfsdclientid);
	print_verifier4(clp->clnt4.verifier);

	nfsv4_printk(setclntid,
		     "NFSV4 nfsd4_setclientid. cl_ident %s\n",clp->cl_ident);

	/* set the NFSV4 client id */
	clp->cl_uid = rqstp->rq_cred.cr_uid; 
	clp->cl_lease = CURRENT_TIME;
	clp->cl_lease_len = 600; /* 10 minute lease, grace period */
	clp->clid_state = CLID_NEGOTIATING;

	/* set the clp in the request here - since the clp didn't exist
	 * yet in nfssvc : nfsd() 
	 */
	rqstp->rq_client = clp;
	
	/* stuff return results */

	/* normal resok4 case */
	rptr->CLIDOK.clientid = clp->nfsdclientid;

	/* use the clientid a verifier to allow nfsd4_setclientid_confirm 
	 * to locate the correct svc_client. send to the client and keep in clp */
	memcpy(rptr->CLIDOK.setclientid_confirm,&clp->nfsdclientid,NFS4_VERIFIER_SIZE);
	/* used to use current time...
	 * memset(rptr->CLIDOK.setclientid_confirm,0,NFS4_VERIFIER_SIZE);
	 *  *(unsigned int *)rptr->CLIDOK.setclientid_confirm = CURRENT_TIME;
	 */

	*(stateid4 *)clp->setclientid_confirm = clp->nfsdclientid;

	nfsv4_printk(setclntid,"NFSV4 nfsd4_setclientid. save and send setclientid_confirms:\n");
	print_verifier4(clp->setclientid_confirm);
	print_verifier4(rptr->CLIDOK.setclientid_confirm);
	err =  0;
	goto out_unlock;

 out_inuse:
	/* NOTE XXX set the following fields correctly */ 
	rptr->SETCLIENTID4res_u.client_using.r_netid = dummy;
	rptr->SETCLIENTID4res_u.client_using.r_addr = dummy;
	
 out_unlock:
	/* unlock hash table - writing */
	exp_unlock();

 out:
	/* restore the nfssvc:nfsd() exp_readlock */
	exp_readlock();
 out_err:
	return err;
}

/*
 * Pseudo fs
 * create a file handle for the pseudo fs root 
 */
u32
exp_pseudorootfh(struct svc_fh *f)
{

	struct dentry           *dentry;
	struct inode            *inode;
	struct svc_fh           fh;
	u32                     err = 0;

	err = nfserr_noent;
	if (!(dentry = new_pseudo_root_exp.ex_dentry)){
		printk("nfsd: exp_pseudorootfh path [/] not found\n");
		return err;
	}
	inode = dentry->d_inode;
	if (!inode) {
		printk("exp_pseudorootfh: Aieee, NULL d_inode\n");
		goto out;
	}
	print_pseudo_inode(inode);

	fh_init(&fh);
	fh_compose(&fh, &new_pseudo_root_exp, dentry);
	memcpy(f, &fh, sizeof(struct svc_fh));

	return 0;

 out:
	return err;
}

/* create an fh - used by pseudo fs */
/*
 * check to make sure the file handle (i.e. dev,ino)
 * has the pseudo root export. if so, return pointer to it.
 */
svc_export *
get_new_pseudo_root_exp(void)
{
	return &new_pseudo_root_exp;
}

/* 
* caller places current time into <now>
* check if now is within the valid lease interval
*
* find client using addr, the hash of which is stored
* in the input clientid.
* check clientid.
* return of 0 indicates now is with in lease interval
*/
u32 
check_and_set_lease(struct sockaddr_in *sin, clientid4 clid, time_t now)
{
	struct svc_client *clp;
	u32 nfserr;

	/* the exp_readlock is held: nfssvc.c:nfsd() */

	nfsv4_printk(lease,"NFSV4: check_and_set_lease.\n");

	/* this should never happen */
	clp = exp_getclient(sin,clid);
	if(!clp){
		nfsv4_printk(lease,"NFSV4: check_and_set_lease. no client struct\n");
		nfserr = nfserr_inval;
		goto out;
	}

	nfserr = nfserr_stale_clientid;
	if((clp->clid_state != CLID_CONFIRMED) || 
	   (clp->nfsdclientid != clid)){
		nfsv4_printk(lease,"NFSV4: check_and_set_lease. invalid clientid 1 \n");
		/* setclientid_confirm not done yet,  */
		goto out;
	}

	if(GET_NFS4_CLIENTID_VERF(clid) != *(time_t *)nfsd_verf){
		nfsv4_printk(lease,"NFSV4: check_and_set_lease. invalid clientid verifier\n");
		nfserr = nfserr_stale_clientid;
		goto out;
	}

	nfserr = 0;
	nfsv4_printk(lease,"NFSV4: check_and_set_lease. now 0x%x, lease 0x%x, lease len %d, lease+len 0x%x\n ",
		     (unsigned int)now,(unsigned int)clp->cl_lease, clp->cl_lease_len,
		     (unsigned int)clp->cl_lease + clp->cl_lease_len);

	/* lock the clp NOTE! Need to make sure that this
	 * guarentees that the clp won't be mucked with - 
	 * should probably put this lock on the hash bucket, or
	 * use the clnt_hash chicken lock. */
	nfserr = nfs_ok;
	down(&clp->cl_sema);

	if(!(now <= (clp->cl_lease + clp->cl_lease_len))){
		nfsv4_printk(lease,"NFSV4: check_and_set_lease. lease expired\n");
		nfserr = nfserr_expired;
		up(&clp->cl_sema);
		goto out;
	}
	nfsv4_printk(lease,"NFSV4: check_and_set_lease. now 0x%x, lease 0x%x, lease len %d, lease+len 0x%x\n ",
		     (unsigned int)now,(unsigned int)clp->cl_lease, clp->cl_lease_len, 
		     (unsigned int)clp->cl_lease + clp->cl_lease_len);

	/* reset the lease */
	clp->cl_lease = now;
	up(&clp->cl_sema);
 out:
	return nfserr;
}
#if 0
/* renew a clients lease
* 
* note: check_lease needs to be called first.
* lease reset each time an operation with a valid stateid
* is performed:
* CLOSE, DELEGRETURN, LOCK, LOCKU, OPEN, OPEN_CONFIRM, 
* READ, RENEW, SETATTR, WRITE 
*/
int
renew_lease(struct sockaddr_in *sin,clientid4 clid,time_t now)
{
	int err;
	struct svc_client *clp;

	nfsv4_printk(lock,"NFSV4: renew lease: addr 0x%x\n",sin->sin_addr.s_addr);
	/* find clp */
	err = nfserr_inval;
	clp = exp_getclient(sin,clid);
	if(!clp)
		goto out;

	nfserr = nfserr_stale_clientid;
	if((clp->clid_state != CLID_CONFIRMED) || 
	   (clp->nfsdclientid != clid)){
		/* setclientid_confirm not done yet,  */
		goto out;
	}
	err = nfs_ok;
	nfsv4_printk(lease,"NFSV4: renew lease. now 0x%x, lease 0x%x, lease len %d, lease+len 0x%x\n ",
		     (unsigned int)now,(unsigned int)clp->cl_lease, clp->cl_lease_len, 
		     (unsigned int)clp->cl_lease + clp->cl_lease_len);

	/* lock the clp NOTE! Need to make sure that this
	 * guarentees that the clp won't be mucked with - 
	 * should probably put this lock on the hash bucket, or
	 * use the clnt_hash chicken lock. */
	down(&clp->cl_sema);
	/* reset the lease */
	clp->cl_lease = now;
	up(&clp->cl_sema);

 out:
	return err;
}

/* given stateid, renew the lease associated with the svc_client.
 * 
 * stateid -> lock -> lockowner -> clientid -> svc_client.
 */
u32
nfsd4_renew(struct svc_rqst *rqst,RENEW4args *args,time_t now)
{
	u32 nfserr = 0;
	clientid4 clid;
	svc_client *clnt;

	/* nfsd3_process_lock checks stateid, and sets the current_lockonwer */
	nfserr = nfs4_process_lock(rqstp,NULL,stid,0,OP_OPEN_CONFIRM,now);
	if(nfserr)
		goto out;

	clid = ((find_global_state(current->pid))->current_lockowner)->lo_owner.clientid; 	
	/* stateid is good. nfserr_old_stateid is used if exp_getclientbyclientid()
	 * fails because this means the lock is found, but the client is gone,
	 * which should never happen....
	 */
	nfserr = nfserr_old_stateid;
	clnt = exp_getclientbyclientid(clid);	
	if(!clnt)
		goto out;
	clnt->cl_lease = now;	
	nfserr = 0;
 out:
	return nfserr;
}
#endif /*0*/
