/*
 * linux/fs/nfsd/nfsdlock.c
 * 
 * based on:
 * linux/fs/lockd/svcsubs.c
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 *
 */                            
#include <linux/fs.h>
#include <linux/sunrpc/svc.h>
#include <linux/nfsd/nfsd4_prot.h>
#include <linux/malloc.h> 

/* forward declarations */
static void create_lock_stateid(unsigned int hash,unsigned short f_seq,unsigned short generation,stateid4 *stateid);
int verify_lock_stateid(stateid4 stateid);
u32 resendit(struct svc_rqst *rqstp,struct nfs4_lock_owner *lowner);
void nfs4_delete_lockowner(struct nfs4_lock_owner *lo);

/*
 * Global hash tables
 */

/* NFSD4_NRHASH = number of hash buckets for open files on nfsd 
* as well as number of hash buckets for lockowners.
* i may seperate these. 
*/

/* nfs4_file[] hash stuff */
#define NFSD4_F_NRHASH  1024 
#define NFSD4_BITS_IN_F_NRHASH 11 /* 11 bits to represent 0-1023 */
#define NFSD4_F_HASH_BITS    3  /* used in nfs4_hash */

/* nfs4_lock_owners[] hash stuff */
#define NFSD4_L_NRHASH  256 
#define NFSD4_BITS_L_IN_NRHASH 9 /* 9 bits to represent 0-1023 */
#define NFSD4_L_HASH_BITS    5  /* used in nfs4_hash */

/* the 64 bit lock stateid is as follows:
* | 32 bits nfsd boot (seconds) |
* | 12 bits nfs4_files[] hash (on ino, dev) |
* | 10 bits file f_seq, unique per nfs4_files[] hash |
* | 10 bits lock l_generation unique per nfs4_file |
*/
#define GET_NFS4_STATEID_VERF(sid)  (unsigned int)(sid & 0xFFFFFFFFUL)
#define GET_NFS4_STATEID_HINDEX(sid)  (unsigned int)(((sid >> 32) & 0xFFF00000UL)>>20)
#define GET_NFS4_STATEID_FSEQ(sid)  (unsigned int)(((sid >> 32) & 0x000FFF00UL) >>10)
#define GET_NFS4_STATEID_LGEN(sid) (unsigned int)((sid >> 32) & 0x000003FFUL)  

static struct nfs4_file_element   nfs4_files[NFSD4_F_NRHASH];
static struct nfs4_lock_owner * nfs4_lock_owners[NFSD4_L_NRHASH];
static struct semaphore   nfs4_file_sema = MUTEX;         
static struct semaphore   nfs4_lowner_sema = MUTEX;         
long	nfs4_filecnt, nfs4_lock_ownercnt, nfs4_lockcnt;

/* Global State */
extern char nfsd_verf[]; /* server boot time */

/* used for both nfs4_file and nfs4_lock_owner */
static unsigned int nfs4_hash(dev_t dev, ino_t ino,int bits,int nrhash)
{
	unsigned long tmp = (unsigned long) ino | (unsigned long) dev;
	tmp = tmp + (tmp >> bits) + (tmp >> bits*2);
	return tmp & (nrhash - 1);
} 
/*
* find an nfs4_file lock using the CURRENT filehandle and stateid.
* check the stateid and the seqid.
* check and perhaps renew the lease.
* return the new stateid with error = nfs_ok, or nfserr.
* if nfserr = BAD_SEQID, then caller should still 
* call renew_lease(). i don't here, because of deadlock
* possibilities between file table lock and export lock.
*
* the operation parameter indicates whether or not process_lock was
* called from a lock mutating operation requiring the stateid to be 
* updated, and the current_lockowner set for lockowner replay cache
* update during the xdr encode phase. OP_CLOSE is a special case
* because lock state is removed.
*/
int
nfs4_process_lock(struct svc_rqst *rqstp,stateid4 *result, stateid4 stateid,seqid4 seqid,int operation,time_t now)
{
	struct knfs_fh  *fh = &CURRENT_FH.fh_handle;
	struct nfs4_file  *file = NULL,**fpp = NULL;
	struct nfs4_lock  *lock = NULL,**lpp = NULL;
	struct nfs4_lock_owner *lowner = NULL;
	unsigned int  hash,shash;
	u32   nfserr;

	nfsv4_printk(lock,"NFSV4: nfs4_process_lock for file(%s/%u), stateid [0x%lx , %08lx]\n",
		     kdevname(u32_to_kdev_t(fh->fh_dev)), fh->fh_ino,
		     (unsigned long)(stateid >> 32),
		     (unsigned long)(stateid & 0xFFFFFFFFUL));

	/* is verifier in stateid valid? */
	if ((operation != OP_READ) && (operation != OP_WRITE)) {
		nfserr = verify_lock_stateid(stateid);
		if (nfserr == 0) {
			nfserr = nfserr_stale_stateid;
			goto out;
		}                                
	}
		
	/* perhaps check the calculated hash against the input stateid hindex
	 * component*/
	hash = nfs4_hash(u32_to_kdev_t(fh->fh_dev), u32_to_ino_t(fh->fh_ino),
			 NFSD4_F_HASH_BITS, NFSD4_F_NRHASH);
	shash = GET_NFS4_STATEID_HINDEX(stateid);

	nfsv4_printk(lock,"NFSV4: nfs4_process_lock: calc hash %d stateid hash %d\n",
		     hash, shash);
	/* filehandle - stateid mismatch */
	if (hash != shash) {
		nfserr = nfserr_bad_stateid;
		goto out;
	}
	/* Lock file table */
	down(&nfs4_file_sema);

	for (fpp = &nfs4_files[shash].f_hash_head; (file = *fpp); fpp = &file->f_next) {
		if (file->f_handle.fh_dcookie == fh->fh_dcookie &&
		    !memcmp(&file->f_handle, fh, sizeof(*fh)))
			goto found_file;
	}
	/* file not found much less the lock */
	nfsv4_printk(lock,"NFSV4: nfs4_process_lock: file not found\n");
	nfserr = nfserr_bad_stateid;
	up(&nfs4_file_sema);
	goto out;

                                                                           
 found_file:
	/* NOTE!! need to examine race conditions involved with holding the
	 * filetable lock vrs holding the file lock. try this:
	 * lock file, and unlock file table 
	 */
	down(&file->f_sema);
	up(&nfs4_file_sema);
	nfsv4_printk(lock,"NFSV4:nfs4_process_lock found file %p (count %d) file->f_lock4 %p\n", 
		     file, file->f_count,file->f_lock4);

	for (lpp = &file->f_lock4;(lock = *lpp); lpp = &lock->l_next) {
		if(lock->l_stateid == stateid){
			if (lpp != &file->f_lock4) {
				/* move found lock to front of linked list */
				*lpp = lock->l_next;
				lock->l_next = file->f_lock4;
				file->f_lock4 = lock;
			}
			goto found_lock;
		}
	}

	/* the verifier in the stateid is good, and the state id
	 * doesn't match any lock for the file. see if the  
	 * file->f_seq stored in the stateid matches the found file's f_seq.
	 * if it does match, then this is the lockowner file pair in
	 * question, but the lock->l_generation, the last component of the
	 * stateid doesn't match, so return nfserr_old_stateid. else, return
	 * nfserr_bad_stateid.
	 */
	if(GET_NFS4_STATEID_FSEQ(stateid) == file->f_seq){
		nfsv4_printk(lock,"NFSV4: nfs4_process_lock: old stateid GET_NFS4_STATEID_LGEN(stateid) %d\n",GET_NFS4_STATEID_LGEN(stateid));
		nfserr = nfserr_old_stateid;
		goto out_unlock;
	}
	nfserr = nfserr_bad_stateid;
	goto out_unlock;

 found_lock:
	nfsv4_printk(lock,"NFSV4: nfs4_process_lock: found_lock, lowner %p lock->l_owner %p\n",lowner,lock->l_owner);

	lowner = lock->l_owner;
	/* check and perhaps set the lease*/
	nfserr = check_and_set_lease(&rqstp->rq_addr,lowner->lo_owner->clientid,now);	
	if (nfserr) {
		if ((lowner->lo_state == OPEN_CONFIRM_REQ_NEGOTIATING) && 
		    (nfserr == nfserr_expired)) {
			/* release lockowner state */
			file = lowner->confirm_lock->l_file;
			down(&file->f_sema);
			delete_lock(file,lowner->confirm_lock);
			up(&file->f_sema);
			/* we hold the nfs4_lowner_sema */
			nfs4_delete_lockowner(lowner);
		}
		goto out_unlock;
	}

	/* good stateid. check lockowner state */
	if (lowner->lo_state == OPEN_CONFIRM_REQ_NEGOTIATING) {
		if (operation != OP_OPEN_CONFIRM) {
			nfserr = nfserr_bad_stateid;
			goto out_unlock;
		}
		/* operation is OPEN_CONFIRM - is this correct open? */
		if (lowner->confirm_lock != lock) {
			nfserr = nfserr_bad_stateid;
			goto out_unlock;
		}
	}

	/* divide the operations into two catagories: those which
	 * bump the seqid and which return a new stateid and those that don't
	 */
	switch(operation){
	case OP_CLOSE:
	case OP_LOCK:
	case OP_OPEN_CONFIRM:
	{
		if (SEQID_LOCKOWNER) {
			if ((lowner != SEQID_LOCKOWNER) ||
			    (lowner->lo_seq_num != seqid)) {
				nfserr = nfserr_bad_seqid;
				goto out_unlock;
			}
		}
		else {
			down(&lowner->lo_seqid_sema);
			SEQID_LOCKOWNER = lowner;
		}

		/* set current_lockowner */
		CURRENT_LOCKOWNER = lowner;

		/* common case, correct seqid */
		if (operation == OP_OPEN_CONFIRM) {
			if (seqid == lowner->lo_confirm_seq_num + 1) {
				nfsv4_printk(lock,"NFSV4: nfs4_process_lock: found_lock, good seqid %d lo_confirm_seq_num %d lo_seq_num %d\n",seqid,lowner->lo_confirm_seq_num, lowner->lo_seq_num);
				lowner->lo_state = OPEN_CONFIRM_REQ_CONFIRMED;
				lowner->lo_seq_num++; 
				goto new_stateid;
			}
		} else if (seqid == lowner->lo_seq_num + 1) {
			nfsv4_printk(lock,"NFSV4: nfs4_process_lock: found_lock, good seqid %d\n",seqid);
			lowner->lo_seq_num++; 
			goto new_stateid;
		}

		if (seqid == lowner->lo_seq_num) {
			nfsv4_printk(lock,
				     "NFSV4: nfs4_process_lock: REPLAY equal seqid. sequid's %d \n",seqid);
			nfserr = nfserr_replay_me;
			goto out_unlock;
		}
		else {
			nfsv4_printk(lock,"NFSV4 nfs4_process_lock:bad seqid.\n");
			nfserr = nfserr_bad_seqid;
			goto out_unlock;
		}
	new_stateid:
		/* success ! generate a new stateid */
		nfsv4_printk(lock,"NFSV4 nfs4_process_lock : generate new state id. lock->l_generation: %d\n",lock->l_generation);
		lock->l_generation = ++file->f_lock4_seq;
		create_lock_stateid(hash,file->f_seq,lock->l_generation, result);
		lock->l_stateid = *result;
		nfsv4_printk(lock,
			     "NFSV4 nfs4_process_lock :new lock->l_stateid [0x%lx , %08lx]\n",
			     (unsigned long)(lock->l_stateid >> 32),
			     (unsigned long)(lock->l_stateid & 0xFFFFFFFFUL)); 
		nfserr = 0;
	} 
	break; 
	case OP_READ:
	case OP_WRITE:
		nfserr = 0;
		nfsv4_printk(lock,"NFSV4 nfs4_process_lock : case OP_READ,OP_WRITE \n");
		break;
	case OP_RENEW:
		/* set current_lockowner */
		CURRENT_LOCKOWNER = lowner;
		break;
	default:
		nfsv4_printk(lock,"NFSV4: nfs4_process_lock: invalid operation %d\n",operation);
		nfserr = nfserr_inval;
		
		break;
	} // switch operation

		/* if nfs4_process_lock is called from OP_CLOSE, then release 
		 *  all the lock state
		 *	 if    - the file reference count is one
		 *              - no delegations exist
		 *              - no threads are waiting on the file's wait queue
		 *        remove the nfs4_file struct.
		 */
		if (operation == OP_CLOSE) {
			nfsv4_printk(lock,"NFSV4: nfs4_process_lock called from OP_CLOSE\n");	
			nfs4_file_close(&file->f_file);
			delete_lock(file,lock);
			/* delete_lock decrements f_count */
			if ((file->f_count == 0) && (file->f_delegtype == OPEN_DELEGATE_NONE) &&
			    (file->f_wait_count == 0)) {
				nfsv4_printk(lock, "NFSV4: nfs4_process_lock: delete file. f_count %d\n",
					     file->f_count);
				*fpp = file->f_next;	
				kfree(file);
			}
		}
 out_unlock:
	up(&file->f_sema);
 out:
	return nfserr;
}

/*
* boolean - check for a stale stateid: 
* does stateid boot time = global nfsd boot time?
*/
int
verify_lock_stateid(stateid4 stateid)
{
	time_t nfsd_boot;

	nfsv4_printk(lock," verify_lock_stateid [0x%lx , %08lx]\n",
		     (unsigned long)(stateid >> 32),
		     (unsigned long)(stateid & 0xFFFFFFFFUL)); 
	nfsd_boot = *(time_t *)nfsd_verf;
	nfsv4_printk(lock,"NFSV4: verify_lock_stateid. stateid verifier 0x%x\n",(unsigned int)GET_NFS4_STATEID_VERF(stateid));
	nfsv4_printk(lock,"NFSV4: verify_lock_stateid. server verifier 0x%x\n", (unsigned int)nfsd_boot);
	return(GET_NFS4_STATEID_VERF(stateid) ==  nfsd_boot);
}

/* 
* Create a new stateid4
* 
* stateid4: 64 bits.
*		server verifier            		32 bits
*   nfs4_file.f_seq		            16 bits
*		nfs4_lock.l_generation      	16 bits
*/
static void
create_lock_stateid(unsigned int hash,unsigned short f_seq,unsigned short l_generation,stateid4 *stateid)
{
	nfsv4_printk(lock,"NFSV4: create_lock_stateid. hash index %d f_seq %d l_generation %d stateid [0x%lx , %08lx]\n",hash,f_seq,l_generation,(unsigned long)(*stateid >> 32),(unsigned long)(*stateid & 0xFFFFFFFFUL));
	*stateid = 0;
	*stateid = *stateid | hash;
	*stateid = *stateid << 10;
	*stateid = *stateid | f_seq;
	*stateid =*stateid << 10;
	*stateid = *stateid | l_generation;
	*stateid =*stateid << 32;
	*stateid = *stateid | *(time_t *)nfsd_verf;

#if 0
	boot = *(time_t *)nfsd_verf;
	*stateid = (long long)boot;
	ptr = (unsigned short *)stateid;
	ptr += 2;
	*ptr = f_seq; /* which nfs4_file in hash bucket */
	ptr++;
	*ptr = generation;  /* which lock structure */
#endif 

	nfsv4_printk(lock," create_lock_stateid [0x%lx , %08lx]\n",
		     (unsigned long)(*stateid >> 32),
		     (unsigned long)(*stateid & 0xFFFFFFFFUL));

	nfsv4_printk(lock," create_lock_stateid verf 0x%x hash %d fseq %d lgen %d\n",
		     GET_NFS4_STATEID_VERF(*stateid),GET_NFS4_STATEID_HINDEX(*stateid),
		     GET_NFS4_STATEID_FSEQ(*stateid),GET_NFS4_STATEID_LGEN(*stateid));
}

/*
* Compare lockowner4's
*/
static inline int
nfs4_cmp_owner(struct nfs_lockowner4 *owner1, struct nfs_lockowner4 *owner2)
{
	return ((owner1->clientid == owner2->clientid) &&
		(owner1->owner.owner_len == owner2->owner.owner_len) && 
		(((owner1->owner.owner_val) && (owner2->owner.owner_val))?
		 !memcmp(owner1->owner.owner_val, 
			 owner2->owner.owner_val, 
			 owner2->owner.owner_len):0));
}

/* 
* Initalize lock state
*/
void init_lock_state(void)
{
	int i;
	nfsv4_printk(lock,"NFSV4 Initalize Lock State\n");
	nfs4_filecnt = nfs4_lock_ownercnt = 0;
	for(i=0;i<NFSD4_F_NRHASH;i++){
		nfs4_files[i].f_hash_head = NULL;
		nfs4_files[i].f_uniq_seq = 0;
		nfs4_files[i].f_pad = 0;
	}
	for(i=0;i<NFSD4_L_NRHASH;i++){
		nfs4_lock_owners[i] = NULL;	
	}
	nfs4_filecnt = nfs4_lock_ownercnt = nfs4_lockcnt = 0;
}


/*
* Clean up. This is the last nfsd thread, so no
* locking needed.
*/
void shutdown_lock_state(void)
{
	struct nfs4_file *file,*f;
	struct nfs4_lock *lock;
	struct nfs4_lock_owner *owner,*o;
	int i;
	int freelockcnt = 0;

	nfsv4_printk(lock,"NFSV4 Shutdown Lock State\n");
	for(i=0;i<NFSD4_F_NRHASH;i++){
		for(file = nfs4_files[i].f_hash_head;file;){
			for(lock = file->f_lock4;lock;lock = lock->l_next){
				/* if not NULL, must have been kmalloc'ed */
				freelockcnt++;
				kfree(lock);
			}
			f = file;
			kfree(file);
			file = f->f_next;
		} 
	}
	for(i=0;i<NFSD4_L_NRHASH;i++){
		for(owner = nfs4_lock_owners[i];owner;){
			if(owner->lo_owner->owner.owner_val){
				kfree(owner->lo_owner->owner.owner_val);
				owner->lo_owner->owner.owner_val = NULL;
			}
			o =owner->lo_next;
			kfree(owner);
			owner = o;
		}
	}
	nfsv4_printk(lock,"NFSV4 shutdown_lock_state: freelockcnt %d\n",
		     freelockcnt);
}

/* 
* Test for share conflicts and perhaps create lock
* the nfs4_lowner_sema is held if lockowner state is 
* OPEN_CONFIRM_REQ_NEGOTIATING.
*/
u32
nfs4_test_lock(struct nfs4_lock_owner *lowner, struct nfs4_file *file,
	       u_int share_access, u_int share_deny,stateid4 *stateid)
{
	struct nfs4_lock	*lock;
	int err = nfs_ok;

	nfsv4_printk(lock,"NFSV4 nfs4_test_lock. lowner %p\n",lowner);

	/* lock the nfs4_file */
	down(&file->f_sema);

	for (lock = file->f_lock4; lock; lock = lock->l_next) {
		nfsv4_printk(lock,"NFSV4 nfs4_test_lock for loop. lock %p\n",lock);
		if (nfs4_cmp_owner(lock->l_owner->lo_owner, lowner->lo_owner))
			goto update;
		if ((share_access & lock->l_share_deny)
		    || (share_deny   & lock->l_share_access ))
			err = nfserr_share_denied;
		goto unlock;
	}

	nfsv4_printk(lock,"NFSV4 nfs4_test_lock creating lock\n");
	err = nfserr_resource;
	lock = (struct nfs4_lock *) kmalloc(sizeof(*lock), GFP_KERNEL);
	if (lock == NULL)
		goto unlock;

	err = nfs_ok;
	file->f_count++;   /* # references to a file = # lock structs */
	lock->l_generation  = ++file->f_lock4_seq;
	lock->l_pad         = 0;
	lock->l_file        = file;
	lock->l_owner       = lowner;
	lock->l_next        = file->f_lock4;
	file->f_lock4       = lock;
	create_lock_stateid(file->hashindex,file->f_seq,lock->l_generation,&lock->l_stateid);
	*stateid = lock->l_stateid;
	nfsv4_printk(lock, "NFSV4 nfs4_test_lock lock created. file->f_lock4 %p lock->l_owner%p\n",
		     file->f_lock4,lock->l_owner);

	if (lowner->lo_state == OPEN_CONFIRM_REQ_NEGOTIATING) {
		nfsv4_printk(lock, "NFSV4 nfs4_test_lock setting confirm_lock\n");
		lowner->confirm_lock = lock;
	}
	
 update:
	lock->l_share_access = share_access;
	lock->l_share_deny   = share_deny;
 unlock:
	up(&file->f_sema);
	return err;
}

/*
 * for a file, delete a lockowners lock.
 */
u32
nfs4_delete_lock(struct nfs4_lock_owner *lowner, struct nfs4_file *file)
{
	struct nfs4_lock	*lock, **lpp;

	for (lpp = &file->f_lock4; (lock = *lpp); lpp = &lock->l_next) {
		if (nfs4_cmp_owner(lock->l_owner->lo_owner, lowner->lo_owner)) {
			*lpp = lock->l_next;
			kfree(lock);
			return nfs_ok; 
		}
	}
	return nfs_ok; 
}

/* 
* remove all locks associated with a clientid4
*
* note that both the lockowner hash table and the file hash
* table need to be locked. i don't think there's a race condition
* here - so far there is no case where the file lock is held 
* waiting for the lockowner lock.
*
* FUTURE:
* may want to link all locks with the same clientid so that 
* the full file table won't have to be linearly searched.
*/
void
release_locks(clientid4 clid)
{
	struct nfs4_lock_owner **hp,*lowner;
	struct nfs4_file *file;
	int i,hash;

	hash = nfs4_hash((unsigned long)(clid >> 32),
			 (unsigned long)(clid & 0xFFFFFFFFUL),NFSD4_L_HASH_BITS,NFSD4_L_NRHASH);
	
	down(&nfs4_lowner_sema);
	for (hp=&nfs4_lock_owners[hash];(lowner = *hp);hp =&lowner->lo_next) {
		if(lowner->lo_owner->clientid == clid){
			for(i=0;i<NFSD4_F_NRHASH;i++){
				down(&nfs4_file_sema);
				for(file = nfs4_files[i].f_hash_head;file;file = file->f_next){
					nfs4_delete_lock(lowner,file);	
				}
				up(&nfs4_file_sema);
			}
		}
	}
	up(&nfs4_lowner_sema);
}

static inline int
fhcmp(struct knfs_fh *fh1, struct knfs_fh *fh2)
{
#if 0
	return memcmp(fh1, fh2, sizeof(struct knfs_fh));
#else
	return (fh1->fh_base.fb_dev != fh2->fh_base.fb_dev) ||
		(fh1->fh_base.fb_ino != fh2->fh_base.fb_ino);
#endif
}

/*
 * Lookup NFSD4 file info. If it doesn't exist, create a file info struct
 * and open a (VFS) file for the given inode.
 */
u32
nfs4_get_file(struct svc_rqst *rqstp, struct nfs4_file **result,
	      struct svc_fh *f)
{
	struct knfs_fh	*fh = &f->fh_handle;
	struct nfs4_file	*file=NULL;
	unsigned int	hash;
	u32		nfserr;

	nfsv4_printk(lock,"NFSV4: nfs4_get_file(%s/%u)\n",
		     kdevname(u32_to_kdev_t(fh->fh_dev)), fh->fh_ino);

	hash = nfs4_hash(u32_to_kdev_t(fh->fh_dev), u32_to_ino_t(fh->fh_ino),
			 NFSD4_F_HASH_BITS,NFSD4_F_NRHASH);

	/* Lock file table */
	down(&nfs4_file_sema);

	for (file = nfs4_files[hash].f_hash_head; file; file = file->f_next)
		if (!fhcmp(&file->f_handle, fh))
			goto found;
	
	nfsv4_printk(lock,"NFSV4: nfs4_get_file: creating file for %s/%u\n",
		     kdevname(u32_to_kdev_t(fh->fh_dev)), fh->fh_ino);
	nfserr = nfserr_resource;
	file = (struct nfs4_file *) kmalloc(sizeof(*file), GFP_KERNEL);
	if (!file)
		goto out_unlock;

	nfs4_filecnt++;
	memset(file, 0, sizeof(*file));
	file->f_handle = *fh;
	file->f_sema   = MUTEX;

	/* Open the file. 
	 * We have to make sure we have the right credential to open
	 * the file.
	 */
	if ((nfserr = nfs4_file_open(rqstp, f, &file->f_file)) != 0) {
		nfsv4_printk(lock,"NFSV4: open failed (nfserr %d)\n", ntohl(nfserr));
		goto out_free;
	}

	file->hashindex = hash;
	file->f_seq = ++nfs4_files[hash].f_uniq_seq;
	file->f_lock4_seq = 0;
	file->f_next = nfs4_files[hash].f_hash_head;
	nfs4_files[hash].f_hash_head = file;
	
	file->f_delegsema = MUTEX;
	file->f_delegtype = OPEN_DELEGATE_NONE;
	file->f_deleglist = NULL;
	file->f_wait_count = 0;
	file->f_wait_queue = NULL;
 found:
	nfsv4_printk(lock,"NFSV4: found file %p (count %d)\n", file, file->f_count);
	*result = file;
	nfserr = 0;

 out_unlock:
	up(&nfs4_file_sema);
	return nfserr;

 out_free:
	kfree(file);
	nfserr = nfserr_denied;
	goto out_unlock;
}

/*
* Create and stuff nfs_lockowner4
*/
int
create_lockowner4(struct nfs4_lock_owner *target, nfs_lockowner4 *src)
{
	int err = nfserr_inval;

	nfsv4_printk(lock,"NFSV4 create_lockowner4: src %p target %p\n",src,target);
	if(!src)
		goto out;
	/* sanity check */
	if(src->owner.owner_len > 1023){
		/* 2^31 -  1 is max opaque xdr size */
		nfsv4_printk(lock,"NFSDV4 create_lockowner4 owner_len %d too big \n",src->owner.owner_len);
		goto out;	
	}
	err = nfserr_resource;
	if (!(target->lo_owner=(struct nfs_lockowner4 *)kmalloc(sizeof(*target->lo_owner),GFP_KERNEL))){
		goto out;
	}
	err = 0;
	memset(target->lo_owner,0,sizeof(*target->lo_owner));
	nfsv4_printk(lock,"NFSV4 create_lockowner4: setting clientid  target->lo_owner %p\n",target->lo_owner);
	target->lo_owner->clientid = src->clientid;
	target->lo_owner->owner.owner_len = src->owner.owner_len;
	target->lo_owner->owner.owner_val = kmalloc(target->lo_owner->owner.owner_len, GFP_KERNEL);
	memcpy(target->lo_owner->owner.owner_val,src->owner.owner_val,src->owner.owner_len);
	target->lo_seqid_sema = MUTEX;
	
 out:
	nfsv4_printk(lock,"NFSDV4 create_lockowner4: returning %p\n",target->lo_owner);
	return err;
}

/*
 * Lock owner lookup or creation
 *
 * Called only by OPEN. All other lock mutating operations call 
 * nfs4_process_lock(). 
 * 
 */
u32
nfs4_get_lockowner(struct svc_rqst *rqstp, nfs_lockowner4 *owner, seqid4 seqid,
		   struct nfs4_lock_owner **result, time_t now)
{
	struct nfs4_lock_owner *lowner, **hp;
	struct nfs4_file *file = NULL;
	int hash;
	int err = 0;

	nfsv4_printk(lock, "NFSV4: nfs4_get_lockowner\n");

	if (!owner) {
		nfsv4_printk(lock,"NFSV4: no owner in nfs4_get_lockowner!\n");
		return nfserr_inval;
	}
	nfsv4_printk(lock, "NFSV4: nfs4_get_lockowner clientid: [0x%lx , %08lx]\n",
		     (unsigned long)(owner->clientid >> 32),
		     (unsigned long)(owner->clientid & 0xFFFFFFFFUL));

	/* a compound call with two lock sequence number mutating
	 * operations -> both operations have the same sequence 
	 * number.
	 */ 
	if (SEQID_LOCKOWNER) {
		if ((!nfs4_cmp_owner(owner, SEQID_LOCKOWNER->lo_owner) ||
		     (seqid != SEQID_LOCKOWNER->lo_seq_num))) {
			*result = NULL;
			return nfserr_bad_seqid;
		}
		else {
			*result = SEQID_LOCKOWNER;
			return 0;
		}
	}

	hash = nfs4_hash((unsigned long)(owner->clientid >> 32),
			 (unsigned long)(owner->clientid & 0xFFFFFFFFUL),
			 NFSD4_L_HASH_BITS,NFSD4_L_NRHASH);

	/* Lock hash table */
	down(&nfs4_lowner_sema);

	for (hp = &nfs4_lock_owners[hash]; (lowner = *hp); hp = &lowner->lo_next) {
		if (nfs4_cmp_owner(owner, lowner->lo_owner)) {
			down(&lowner->lo_seqid_sema);
			SEQID_LOCKOWNER = lowner;
			
			if (hp != &nfs4_lock_owners[hash]) {
				/* move found lock_owner to front of hash linked list (MRU) */
				*hp = lowner->lo_next;
				lowner->lo_next = nfs4_lock_owners[hash];
				nfs4_lock_owners[hash] = lowner;
			}
			nfsv4_printk(lock,
				     "NFSV4: nfs4_get_lockowner: request sequid %d stored seqid %d\n",
				     seqid,lowner->lo_seq_num);

			/* have lockowner. set current_lockowner and check seqid */
			CURRENT_LOCKOWNER = lowner;

			/* replay case. return a 'fake' error (not in spec) to 
			 * indicate to the nfsd dispatcher (nfssvc.c) to use the
			 * SEQID_LOCKOWNER replay cache 
			 */
			if (seqid == lowner->lo_seq_num) {
				nfsv4_printk(lock,"NFSV4: nfs4_get_lockonwer: equal seqid. sequid's %d \n",seqid);
				err = nfserr_replay_me;
				goto out;
			}

			if (lowner->lo_state == OPEN_CONFIRM_REQ_NEGOTIATING) {
				/* spec (draft07) states that given an OPEN on an unconfirmed 
				 * lockowner where the seqnum is out of sequence, drop the
				 * old unconfirmed OPEN state, use this new OPEN and indicate
				 * OPEN_CONFIRM needed.
				 * NOTE: equal seqnum -> replay cache, all other OPENS
				 * -> drop the old unconfirmed OPEN state and use this new
				 * OPEN and indicate OPEN_CONFIRM needed.
				 */

				/* delete old lowner->confirm_lock */
				file = lowner->confirm_lock->l_file;
				down(&file->f_sema);
				delete_lock(file,lowner->confirm_lock);
				up(&file->f_sema);
				/* set lowner->lo_confirm_seq_num and lo_seq_num */
				lowner->lo_confirm_seq_num = seqid;
				lowner->lo_seq_num = seqid;
				goto out;
			}


			/* the common correct case.*/
			if (seqid == lowner->lo_seq_num + 1) {
				lowner->lo_seq_num = seqid;
				goto out;
			}

			/* the "bad seqid" case */
			nfsv4_printk(lock,"NFSV4 nfs4_get_lockowner: bad seqid.\n");
			err = nfserr_bad_seqid;
			goto out;
		}
	}

	/* No lockowner found, create it */
	nfsv4_printk(lock, "NFSV4: creating lockowner entry\n");

	err = nfserr_resource;
	if (!(lowner = (struct nfs4_lock_owner *) kmalloc(sizeof(*lowner), GFP_KERNEL)))
		goto out;
	memset(lowner, 0, sizeof(*lowner));

	err = create_lockowner4(lowner, owner);
	nfsv4_printk(lock, "NFSV4: nfs4_get_lockowner create_lockowner returns %d lowner->lo_owner %p\n",
		     err, lowner->lo_owner);
	if (err != 0) {
		kfree(lowner);
		lowner = NULL;
		goto out;
	}
	lowner->lo_seq_num = seqid;

	/* allocate lockowner replay cache. 
	 * see nfssvc.c: svc_create - NFSD_BUFSIZE is used as the 
	 * svc_buf recv size.
	 */
	nfsv4_printk(lock, "NFSV4: nfs4_get_lockowner calling svc_init_buffer\n");
	if (!svc_init_buffer(&lowner->lo_seq_response, NFSD_BUFSIZE)) {
		kfree(lowner);
		lowner = NULL;
		err = nfserr_resource;
		goto out;
	}
	down(&lowner->lo_seqid_sema);
	SEQID_LOCKOWNER = lowner;

	lowner->lo_next = nfs4_lock_owners[hash];
	nfs4_lock_owners[hash]  = lowner;
	nfs4_lock_ownercnt++;
	/* lo_state set to zero in create_lockowner*/
	lowner->lo_state = OPEN_CONFIRM_REQ_NEGOTIATING;
	nfsv4_printk(andros, "ANDROS: CREATED LOWNER: lowner->lo_state %d\n", lowner->lo_state);
	lowner->lo_confirm_seq_num = seqid;

	CURRENT_LOCKOWNER = lowner;
 out:
	up(&nfs4_lowner_sema);
	*result = lowner;
	return err;
}

#if 0
/* find a file given stateid. 
* 	verifystateid() has been called.
*/
u32
nfs4_find_file(stateid4 stateid, nfs4_file **result)
{
	nfs4_file **fpp,*file = NULL;
	uint fhash = GET_NFS4_STATEID_HINDEX(stateid);
	unsigned short fseq = GET_NFSD4_STATEID_FSEQ(stateid);
	u32 nfserr = nfserr_bad_stateid; /* nfs4_file not found */

	down(&nfs4_file_sema);
	for(fpp = &nfs4_files[fhash].f_hash_head;(file = *fpp);fpp = &file->f_next){
		if(file->f_seq == fseq){
			nfserr = 0;
			if(fpp != &nfs4_files[fhash].f_hash_head){
				/* move found file to front of hash bucket linked list */
				*fpp = file->f_next;
				file->f_next = nfs4_files[fhash].f_hash_head;
				nfs4_files[fhash].f_hash_head = file;
			}
		}
	}
	up(&nfs4_file_sema);
	*result = file;
	return nfserr;
}

/* find a lock given nfs4_file.  this routine is part of
* the following sequence.
* 	start with a stateid.
*		passes verifystateid() 
* 	the file's filehandle is good
* therefore, if the lock is not found, it has been
* reaped by the server -> lease expired
*/
u32
nfs4_find_lock(nfs4_file *file, nfs4_lock **result)
{
	nfs4_lock **lpp,*lock = NULL;
	u32 nfserr = nfserr_expired;

	down(&file->f_sema);
	for (lpp = &file->f_lock4;(lock = *lpp); lpp = &lock->l_next) {
		if(lock->l_stateid == stateid){
			nfserr = 0;
			if (lpp != &file->f_lock4) {
				/* move found lock to front of linked list */
				*lpp = lock->l_next;
				lock->l_next = file->f_lock4;
				file->f_lock4 = lock;
			}
		}
	}
	*result = lock;
	up(&file->f_sema);
	return nfserr;
}

/*
 * find lockowner given stateid
 */
u32
nfsd4_find_lockowner(stateid4 stateid, struct nfs4_lock_owner **lowner)
{
	nfs4_file *file = NULL;
	nfs4_lock *lock = NULL;
	u32 nfserr = 0;
	
	file = nfs4_find_file(stateid,&file);
	if(!file){
		nfserr = nfserr_inval;	
		oto err_out;
	}
	lock = nfs4_find_lock(file,&lock);	
	f(!lock){
		nfserr = 
			goto err_out;
	}
	return lock->l_owner;
 err_out:
	return(struct nfs4_lock_owner *)NULL;
}

/*
 * find a lockowner given a clientid
 */
struct nfs4_lock_owner *
nfsd4_find_lockowner_clientid(clientid4 clientid)
{
	struct nfs4_lock_owner *lowner = NULL;
	unsigned int hash = GET_NFS4_CLIENTID_HINDEX(clid);

	down(&nfs4_lowner_sema);
	for (hp=&nfs4_lock_owners[hash];(lowner = *hp);) {
		if(lowner->lo_owner->clientid == clid){
			goto out;
		}
	}
 out:
	up(&nfs4_lowner_sema);
	return lowner;
}

#endif /*0*/


/*
 * Delete a single lockowner
 * must be called with nfs4_lowner_sema held
 */
void
nfs4_delete_lockowner(struct nfs4_lock_owner *lo)
{
	struct nfs4_lock_owner **hp, *lowner;
	unsigned int hash = GET_NFS4_CLIENTID_HINDEX(lo->lo_owner->clientid);
 
	nfsv4_printk(lock, "NFSV4 nfs4_delete_lockowners clientid: [0x%lx , %08lx]\n",
		     (unsigned long)(lo->lo_owner->clientid >> 32),
		     (unsigned long)(lo->lo_owner->clientid & 0xFFFFFFFFUL));
	
	for (hp=&nfs4_lock_owners[hash]; (lowner = *hp); ) {
		if (lowner == lo) {
			*hp = lowner->lo_next;
			kfree(&lowner->lo_seq_response.buf);
			kfree(lowner);	
		} else
			hp = &lowner->lo_next;
	}
}

/*
 * Delete all lockowners of a clientid
 */
void
nfs4_delete_lockowners(clientid4 clid)
{
	struct nfs4_lock_owner	**hp, *lowner;
	unsigned int hash = GET_NFS4_CLIENTID_HINDEX(clid);

	nfsv4_printk(lock, "NFSV4 nfs4_delete_lockowners clientid: [0x%lx , %08lx]\n",
		     (unsigned long)(clid >> 32),
		     (unsigned long)(clid & 0xFFFFFFFFUL));

	down(&nfs4_lowner_sema);
	for (hp=&nfs4_lock_owners[hash];(lowner = *hp);) {
		if(lowner->lo_owner->clientid == clid){
			*hp = lowner->lo_next;
			kfree(&lowner->lo_seq_response.buf);
			kfree(lowner);	
		}else
			hp = &lowner->lo_next;
	}
	up(&nfs4_lowner_sema);
}

/*
* delete lock from a file
* must be called with the file f_sema held.
*/
void
delete_lock(struct nfs4_file *file,struct nfs4_lock *lock)
{
	nfsv4_printk(lock,"NFSV4: delete_lock - stateid [0x%lx , %08lx]\n",
		     (unsigned long)(lock->l_stateid >> 32),
		     (unsigned long)(lock->l_stateid & 0xFFFFFFFFUL));
	/* lock is guarenteed to be first on the linked list */
	file->f_lock4 = lock->l_next;
	file->f_count--; 
	kfree(lock);
}

/*
* Open a file.
* must be called with the either file->f_sema OR filetable lock held
* Note: we hold the dentry use count while the file is open.
*/

u32
nfs4_file_open(struct svc_rqst *rqstp, struct svc_fh *fh, struct file *filp)
{
	u32   nfserr = 0;

	nfsv4_printk(lock,"NFSV4: nfs4_file_open calling nfsd_open\n");
	nfserr = nfsd_open(rqstp, fh, S_IFREG, MAY_LOCK, filp);
	if (!nfserr)
		dget(filp->f_dentry);
	fh_put(fh);
	return nfserr;
}

/*
* Close a file
* must be called with the file f_sema held
*/
void
nfs4_file_close(struct file *filp)
{
	nfsv4_printk(lock,"NFSV4: nfs4_file_close calling nfsd_close\n");
	nfsd_close(filp);
	dput(filp->f_dentry);

}

/*
* Resend reply when locking sequence numbers match.
*
* Sequence number matches -> 
*			1) responding to correct compound operation i.e. not just the 
*        lock sequence num mutating operation, but all other 
*        operations in the compound rpc.
*			2) client receive buffer size is exactly the same.
*
*   perhaps i should check these assumptions, and relax the
*   second from exact size to 'big enough'.
*
*   reply buffer sent on the wire: rqstp->rq_resbuf, an svc_buf.
*   svc_buf.base - start of buffer.
*   svc_buf.buflen - total kmalloc'ed length of buffer.
*   svc_buf.len - actual used portion of buffer, set in xdr_ressize_check.
*
*   the reason buflen could be different than len is that the client
*   uses the same buffer to send and receive - so the buflen is set to 
*   the larger of the possible send/receive size for the compound call.
*   
*/

u32
resendit(struct svc_rqst *rqstp,struct nfs4_lock_owner *lowner)
{
	struct svc_buf *resp, *lobufp;
	int size;

	nfsv4_printk(lock,"NFSV4 resendit\n");
	if(!rqstp || !lowner)
		return nfserr_inval;

	resp = &rqstp->rq_resbuf;
	lobufp = &lowner->lo_seq_response;
	size = rqstp->rq_procinfo->pc_ressize;
	nfsv4_printk(lock,"NFSV4 resendit. resp->len %d : cache->len %d\n",
		     resp->len,lobufp->len);
	
	resp->len = lobufp->len;
	memcpy(resp->buf, lobufp->buf, lobufp->len << 2);
	return 0;
}


void 
set_current_lockowner_reply(struct svc_rqst *rqstp, struct nfs4_lock_owner *lo)
{
	
	nfsv4_printk(lock, "NFSV4: set_current_lockowner_reply. lo %p rq_resp.len %d\n",
		     lo, rqstp->rq_resbuf.len);
	lo->lo_seq_response.len = rqstp->rq_resbuf.len;
	memcpy(lo->lo_seq_response.buf, rqstp->rq_resbuf.buf, rqstp->rq_resbuf.len << 2);

	nfsv4_printk(lock, "NFSV4: set_current_lockowner_reply. lo->lo_seq_response.area %p lo->lo_seq_response.base %p lo->lo_seq_response.buf %p lo->lo_seq_response.buflen %d lo->lo_seq_response.len %d \n", 
		     lo->lo_seq_response.area,lo->lo_seq_response.base, 
		     lo->lo_seq_response.buf,lo->lo_seq_response.buflen,
		     lo->lo_seq_response.len);

	nfsv4_printk(lock, "NFSV4: set_current_lockowner_reply. rqstp->rq_resbuf.area %p rqstp->rq_resbuf.base %p rqstp->rq_resbuf.buf %p rqstp->rq_resbuf.buflen %d  rqstp->rq_resbuf.len %d \n",
		     rqstp->rq_resbuf.area, rqstp->rq_resbuf.base, rqstp->rq_resbuf.buf,
		     rqstp->rq_resbuf.buflen,rqstp->rq_resbuf.len);
	nfsv4_printk(lock, "NFSV4 set_current_lockowner_reply(): done\n");
}

u32
nfsd4_open_confirm(struct svc_rqst *rqstp,OPEN_CONFIRM4args * args,OPEN_CONFIRM4res *res, time_t now)
{
	stateid4 stid;
	u32 nfserr = 0;

	nfsv4_printk(open,"NFSD4: nfsd4_open_confirm\n");
	/* verify current filehandle */
	nfserr = fh_verify(rqstp,&CURRENT_FH, 0, MAY_NOP);
	if(nfserr)
		goto out;			

	/* unpack the stateid from the verifier */
	stid = *(stateid4 *)args->open_confirm;
	nfsv4_printk(open,
		     "NFSD4: nfsd4_open_confirm: open_confirm (stateid) [0x%lx , %08lx]\n",
		     (unsigned long)(stid >> 32),
		     (unsigned long)(stid & 0xFFFFFFFFUL));

	nfserr = nfs4_process_lock(rqstp,
	                           &res->OPEN_CONFIRM4res_u.resok4.stateid,
	                           stid,args->seqid,OP_OPEN_CONFIRM,now);
	if(nfserr)
		goto out;			

	nfsv4_printk(open,
		     "NFSD4: nfsd4_open_confirm: open_confirm (result) [0x%lx , %08lx]\n",
		     (unsigned long)(res->OPEN_CONFIRM4res_u.resok4.stateid >> 32),
		     (unsigned long)(res->OPEN_CONFIRM4res_u.resok4.stateid & 0xFFFFFFFFUL));
 out:
	nfsv4_printk(open,"NFSD4: nfsd4_open_confirm returns %d\n",htonl(nfserr));
	return nfserr;
}
