/*
 * linux/fs/nfsd/nfsd4proc.c
 *
 * Process version 4 NFS requests.
 *
 * by Andy Adamson <andros@citi.umich.edu>
 * Copyright (C) 1999 The Regents of the University of Michigan
 */

#include <linux/linkage.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/locks.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/version.h>
#include <linux/unistd.h>
#include <linux/malloc.h>
#include <linux/major.h>

#include <linux/sunrpc/svc.h>
#include <linux/nfsd/cache.h>
#include <linux/nfsd/nfsd4_prot.h>
#include <linux/nfsd/n2s_cache.h>

#define NFSDDBG_FACILITY		NFSDDBG_PROC


/* NFSD global state */
/* Verifier: time nfsd boots
 */
char nfsd_verf[NFS4_VERIFIER_SIZE] = {0,0,0,0,0,0,0,0};

/* per nfsd thread global state table */
struct nfsd_state_tbl_head nfsd_state;

/*
 * ATTR_BITMAP: indicates which attributes are supported by
 * this server. 
 * bits_to_set[]: this is a list of the bits to set in ATTR_BITMAP,
 * terminated by the -1. as attribute functionality changes, change
 * the bits_to_set array.
 */

bitmap4 ATTR_BITMAP = {(u_int)0,(uint32_t *)NULL};
static uint32_t attr_bitmap_val[2] = {(uint32_t)0,(uint32_t)0};
static int bits_to_set[] = {0,1,2,3,4,5,6,7,8,9,10,11,15,16,19,25,26,28,29,30,32,34,35,36,40,44,46,48,51,-1};

static struct {
	unsigned int    mode;
	nfs_ftype4      nfstype;
} nfs_type4fmt[] = {
	{ 0,              NF4NON   },
	{ S_IFREG,        NF4REG   },
	{ S_IFDIR,        NF4DIR   },
	{ S_IFBLK,        NF4BLK   },
	{ S_IFCHR,        NF4CHR   },
	{ S_IFLNK,        NF4LNK   },
	{ S_IFSOCK,       NF4SOCK  },
	{ S_IFIFO,        NF4FIFO  },
	{ 0,              NF4BAD   }
};



int
init_global_state(pid_t pid)
{
	int i;
	for(i=0;i<nfsd_state.nrservs;i++){
		if(nfsd_state.tbl[i] == NULL){
			nfsd_state.tbl[i] = kmalloc(sizeof(struct nfsd_global_state),GFP_KERNEL);
			if(nfsd_state.tbl[i] == NULL)
				return -ENOMEM;
			memset(nfsd_state.tbl[i],0,sizeof(struct nfsd_global_state));
			nfsd_state.tbl[i]->pid = pid;
			return 0;
		}
	}
	printk("NFSDV4 nfsd_state ran out of slots - EINVAL\n");
	return -EINVAL;
}
			
struct nfsd_global_state *
find_global_state(pid_t pid)
{
	int i;
	for(i=0;i<nfsd_state.nrservs;i++){
		if(nfsd_state.tbl[i] && nfsd_state.tbl[i]->pid == pid)
			return nfsd_state.tbl[i];
	}
	printk("AHHH NFSV4 can't find global state for pid %d\n",pid);
	return NULL;
}


void
free_global_state(pid_t pid)
{
	struct nfsd_global_state *s=NULL;

	printk("free global state for pid %d\n",pid);
	s = find_global_state(pid);
	if(s==NULL)
		return; 
	kfree(s);
}

void
verf_init(void)
{
	time_t boot_time = CURRENT_TIME;

/* stuff the 4 bytes of time into the first 4 bytes
 * of the nfs4 verifier
 */
	*(time_t *)nfsd_verf = boot_time;
}

void
set_verifier(verifier4 target,verifier4 src)
{
	int i;

	if(!src)
		for(i=0;i<NFS4_VERIFIER_SIZE;i++)
			target[i] = 0;
	else
		for(i=0;i<NFS4_VERIFIER_SIZE;i++)
			target[i] = src[i];
}

/* boolean compare */
int
cmp_verifier(verifier4 v1, verifier4 v2)
{
	int i;

	for(i=0;i<NFS4_VERIFIER_SIZE;i++)
		if(v1[i] != v2[i])
			return 0;
	return 1;
}
	
/* NOTE: need to really do utf8...*/

int
utf8cmp(utf8string *s1,utf8string *s2)
{
	nfsv4_printk(utf8,"NFSDV4 utf8cmp\n");
	print_utf8string(s1);
	print_utf8string(s2);
	return strncmp(s1->utf8string_val,s2->utf8string_val,s1->utf8string_len);
}
void
utf8copy(utf8string *dest, utf8string *src)
{
	dest->utf8string_len = src->utf8string_len;
	if(dest->utf8string_len > 0){
		dest->utf8string_val = kmalloc(dest->utf8string_len,GFP_KERNEL);
		strncpy(dest->utf8string_val,src->utf8string_val,dest->utf8string_len);
	}
}
/* NOTE: caller must free */
void
ascii_2_utf8(utf8string *utf8, char *str)
{
	utf8->utf8string_len = strlen(str);
	if(utf8->utf8string_len > 0){
		utf8->utf8string_val = kmalloc(utf8->utf8string_len,GFP_KERNEL);
		memset(utf8->utf8string_val,0,utf8->utf8string_len);
		memcpy(utf8->utf8string_val,str,utf8->utf8string_len);
	}
}
void
utf8_2_ascii(utf8string *utf8, char **str)
{
	*str = kmalloc(utf8->utf8string_len+1,GFP_KERNEL);
	memset(*str,0,utf8->utf8string_len+1);
	memcpy(*str,utf8->utf8string_val,utf8->utf8string_len);
	nfsv4_printk(utf8,"NFSV4 utf8_2_ascii: %s\n",*str);
}

void
qstr_to_ascii(struct qstr *qs, char **str)
{
	*str = kmalloc(qs->len+1, GFP_KERNEL);
	memcpy(*str,qs->name, qs->len);
	(*str)[qs->len] = 0;
}

unsigned int
get_bit(unsigned int x,int p)
{
        return((x >> p) & 0x00000001UL);
}
int
get_bitmap4_bit(bitmap4 *bm4,int pos)
{
	int index,which_bit,end = (bm4->BLEN * 32)-1;

        if((pos >= MAX_GETATTR_BITS) || (pos > end))
                return -1;
        index = pos/32;
        which_bit = pos % 32;
        return(get_bit(bm4->BVAL[index],which_bit));
}
unsigned int
set_bit4(unsigned int *x, int p)
{
	unsigned int mask = 1;

        mask = (mask << p);
        *x |= mask;
        return 0;
}

int
set_bitmap4_bit(int b,bitmap4 *bm)
{
	int index,which_bit;

	index = (b/32);
	which_bit = b % 32;
	if(index >= bm->BLEN){
		return -1;
	}
        set_bit4(&bm->BVAL[index],b);
	return 0;
}

void
setup_bitmap4(bitmap4 *bm,int len)
{
	bm->BLEN = len;
	bm->BVAL = kmalloc(len * sizeof(__u32),GFP_KERNEL);
	memset(bm->BVAL,0,len * sizeof(__u32));
}

void
nfsd4_attr_bitmap_init(void)
{
	int i;
	ATTR_BITMAP.BLEN = ATTR_BITMAP_LEN;
	ATTR_BITMAP.BVAL = attr_bitmap_val;

	for(i=0;i<MAX_GETATTR_BITS;i++){
		if(bits_to_set[i] == -1)
			break;
		set_bitmap4_bit(bits_to_set[i],&ATTR_BITMAP);
	}
}


/* given a request bitmap4, calculate the maximum 
 * fattr4.attr_vals.attrlist4_len in bytes. 
 * So, char lengths such as utf8string_len uses NFS4_MAXNAMLEN.
 * This function is called by GETATTR and by 
 * READDIR - in vfs.c: fill_entry().
 */
int
get_fattr4_res_buf_size(bitmap4 *bm4)
{
	int i=0;
	int size=0;
	int bit;

	if(!bm4)
		return 0;
	nfsv4_printk(attributes,"NFSDV4 get_fattr4_res_buf_size: bm4 LENGTH = %d\n",bm4->BLEN);
	do{
		bit = get_bitmap4_bit(bm4,i);
		i++;
		if(!bit || bit == -1)
			continue;
		/* calculate size for correct return argument*/
		switch(i-1){
		case FATTR4_SUPPORTED_ATTRS:
			size += sizeof(u_int); /* bitmap4_len*/
			size += sizeof(u_int) * ATTR_BITMAP.BLEN;
			break;
		case FATTR4_TYPE:
			size += sizeof(fattr4_type);
			break;
		case FATTR4_FH_EXPIRE_TYPE:
			size += sizeof(fattr4_fh_expire_type);
			break;
		case FATTR4_CHANGE:
			size += sizeof(fattr4_change);
			break;
		case FATTR4_SIZE:
			size += sizeof(fattr4_size);
			break;
		case FATTR4_LINK_SUPPORT:
			size += sizeof(fattr4_link_support);
			break;
		case FATTR4_SYMLINK_SUPPORT:
			size += sizeof(fattr4_symlink_support);
			break;
		case FATTR4_NAMED_ATTR:
			size += sizeof(fattr4_named_attr);
			break;
		case FATTR4_UNIQUE_HANDLES:
			size += sizeof(fattr4_unique_handles);
			break;
		case FATTR4_LEASE_TIME:
			size += sizeof(fattr4_lease_time);
			break;
		case FATTR4_RDATTR_ERROR:
			size += sizeof(fattr4_rdattr_error);
			break;
		case FATTR4_FSID:
			size += sizeof(fattr4_fsid);
			break;
		case FATTR4_ACL:
/*NOTE NEED TO FIX  not supported*/
			size += sizeof(fattr4_acl);
			break;
		case FATTR4_ACLSUPPORT:
			size += sizeof(fattr4_aclsupport);
			break;
		case FATTR4_ARCHIVE:
			size += sizeof(fattr4_archive);
			break;
		case FATTR4_CANSETTIME:
			size += sizeof(fattr4_cansettime);
			break;
		case FATTR4_CASE_INSENSITIVE:
			size += sizeof(fattr4_case_insensitive);
			break;
		case FATTR4_CASE_PRESERVING:
			size += sizeof(fattr4_case_preserving);
			break;
		case FATTR4_CHOWN_RESTRICTED:
			size += sizeof(fattr4_chown_restricted);
			break;
		case FATTR4_FILEHANDLE:
			size += sizeof(u_int); /* nfs_fh4_len */
			size += NFS4_FHSIZE;
			break;
		case FATTR4_FILEID:
			size += sizeof(fattr4_fileid);
			break;
		case FATTR4_FILES_AVAIL:
			size += sizeof(fattr4_files_avail);
			break;
		case FATTR4_FILES_FREE:
			size += sizeof(fattr4_files_free);
			break;
		case FATTR4_FILES_TOTAL:
			size += sizeof(fattr4_files_total);
			break;
		case FATTR4_FS_LOCATIONS:
/* NOTE NEED TO CODE */
			break;
		case FATTR4_HIDDEN:
			size += sizeof(fattr4_hidden);
			break;
		case FATTR4_HOMOGENEOUS:
			size += sizeof(fattr4_homogeneous);
			break;
		case FATTR4_MAXFILESIZE:
			size += sizeof(fattr4_maxfilesize);
			break;
		case FATTR4_MAXLINK:
			size += sizeof(fattr4_maxlink);
			break;
		case FATTR4_MAXNAME:
			size += sizeof(fattr4_maxname);
			break;
		case FATTR4_MAXREAD:
			size += sizeof(fattr4_maxread);
			break;
		case FATTR4_MAXWRITE:
			size += sizeof(fattr4_maxwrite);
			break;
		case FATTR4_MIMETYPE:
			size += sizeof(fattr4_mimetype);
			size += NFS4_MAXNAMLEN;
			break;
		case FATTR4_MODE:
			size += sizeof(fattr4_mode);
			break;
		case FATTR4_NO_TRUNC:
			size += sizeof(fattr4_no_trunc);
			break;
		case FATTR4_NUMLINKS:
			size += sizeof(fattr4_numlinks);
			break;
		case FATTR4_OWNER:
			size += sizeof(fattr4_owner);
			size += NFS4_MAXNAMLEN;
			break;
		case FATTR4_OWNER_GROUP:
			size += sizeof(fattr4_owner_group);
			size += NFS4_MAXNAMLEN;
			break;
		case FATTR4_QUOTA_HARD:
			size += sizeof(fattr4_quota_hard);
			break;
		case FATTR4_QUOTA_SOFT:
			size += sizeof(fattr4_quota_soft);
			break;
		case FATTR4_QUOTA_USED:
			size += sizeof(fattr4_quota_used);
			break;
		case FATTR4_RAWDEV:
			size += sizeof(fattr4_rawdev);
			break;
		case FATTR4_SPACE_AVAIL:
			size += sizeof(fattr4_space_avail);
			break;
		case FATTR4_SPACE_FREE:
			size += sizeof(fattr4_space_free);
			break;
		case FATTR4_SPACE_TOTAL:
			size += sizeof(fattr4_space_total);
			break;
		case FATTR4_SPACE_USED:
			size += sizeof(fattr4_space_used);
			break;
		case FATTR4_SYSTEM:
			size += sizeof(fattr4_system);
			break;
		case FATTR4_TIME_ACCESS:
			size += sizeof(fattr4_time_access);
			break;
		case FATTR4_TIME_ACCESS_SET:
			size += sizeof(fattr4_time_access_set);
			break;
		case FATTR4_TIME_BACKUP:
			size += sizeof(fattr4_time_backup);
			break;
		case FATTR4_TIME_CREATE:
			size += sizeof(fattr4_time_create);
			break;
		case FATTR4_TIME_DELTA:
			size += sizeof(fattr4_time_delta);
			break;
		case FATTR4_TIME_METADATA:
			size += sizeof(fattr4_time_metadata);
			break;
		case FATTR4_TIME_MODIFY:
			size += sizeof(fattr4_time_modify);
			break;
		case FATTR4_TIME_MODIFY_SET:
			size += sizeof(fattr4_time_modify_set);
			break;
		default:
			break;
		}
	}while (bit != -1);
	nfsv4_printk(attributes,"NFSV4 get_fattr4_res_buf_size: #bytes: %d\n",size);
	return size;
}

/*
 * Map errnos to NFSV4 errnos.
 */
int
nfsd4errno (int errno)
{
	static struct {
		int     nfserr;
		int     syserr;
	} nfsd4_errtbl[] = {
		{ NFS4_OK, 0 },
		{ NFS4ERR_PERM, EPERM },
		{ NFS4ERR_NOENT, ENOENT },
		{ NFS4ERR_IO, EIO },
		{ NFS4ERR_NXIO, ENXIO },
		{ NFS4ERR_ACCES, EACCES },
		{ NFS4ERR_EXIST, EEXIST },
		{ NFS4ERR_XDEV, EXDEV },
		{ NFS4ERR_MLINK, EMLINK },
		{ NFS4ERR_NODEV, ENODEV },
		{ NFS4ERR_NOTDIR, ENOTDIR },
		{ NFS4ERR_ISDIR, EISDIR },
		{ NFS4ERR_INVAL, EINVAL },
		{ NFS4ERR_FBIG, EFBIG },
		{ NFS4ERR_NOSPC, ENOSPC },
		{ NFS4ERR_ROFS, EROFS },
		{ NFS4ERR_NAMETOOLONG, ENAMETOOLONG },
		{ NFS4ERR_NOTEMPTY, ENOTEMPTY },
#ifdef EDQUOT
		{ NFS4ERR_DQUOT, EDQUOT },
#endif
		{ NFS4ERR_STALE, ESTALE },
		{ -1, EIO }
	};
	int     i;

	for (i = 0; nfsd4_errtbl[i].nfserr != -1; i++) {
		if (nfsd4_errtbl[i].syserr == errno)
			return htonl(nfsd4_errtbl[i].nfserr);
	}
	printk (KERN_INFO "NFSDV4 nfsd: non-standard errno: %d\n", errno);
	return nfserr_io; 
}

/* given fattr4 information, stuff the vfs iattr structure, and also
 * pull out type from fattr4 if available
 */
u32
nfs4_set_iattr(struct fattr4 *f4,struct iattr *iap,fattr4_type *f4type)
{
	u32 nfserr = 0;
	int i = 0;
	int bit;
	bitmap4 *bm4 = &f4->attrmask;
	void *cptr = f4->FATTRSVAL;
	void *end = (char *)cptr + f4->FATTRSLEN;
	fattr4_type type = 0;
	fattr4_size size = 0;
	fattr4_mode mode = 0;

	nfsv4_printk(vfs,"NFSV4 nfs4_set_iattr. len: %d\n",f4->FATTRSLEN);

	memset (iap, 0, sizeof(*iap));

	if(!bm4) {
		nfserr = nfserr_inval;
		goto out;
	}

	do{
		bit = get_bitmap4_bit(bm4,i);
		i++;
		if(!bit || bit == -1)
			continue;
		
		switch(i-1){
		case FATTR4_TYPE:
		{
			type = *(fattr4_type *)cptr;

			nfsv4_printk(vfs,"NFSV4 nfs4_set_iattr FATTR4_TYPE \n");
			if(cptr + sizeof(fattr4_type) > end)
				goto overrun;
			if(type > NF4BAD)
				type=NF4BAD;
			*f4type=nfs_type4fmt[type].mode;
			cptr += sizeof(fattr4_type);         
		}
		break;
		case FATTR4_SIZE:
		{
			size = *(fattr4_size *)cptr;

			if(cptr + sizeof(fattr4_size) > end)
				goto overrun;
			iap->ia_valid |= ATTR_SIZE;
			iap->ia_size = size;
			nfsv4_printk(vfs,
				     "NFSV4 nfs4_set_iattr FATTR4_SIZE: %d \n",(int)iap->ia_size);
			cptr += sizeof(fattr4_size);
		}
		break;
		case FATTR4_MODE:
		{
			mode = *(fattr4_mode *)cptr;

			nfsv4_printk(vfs,"NFSV4 nfs4_set_iattr FATTR4_MODE. mode 0x%x\n",mode);
			if(cptr + sizeof(fattr4_mode) > end)
				goto overrun;
			iap->ia_valid |= ATTR_MODE;
			iap->ia_mode = mode & (S_IFMT | S_IALLUGO);
			cptr += sizeof(fattr4_mode);
		}
		break;
		case FATTR4_OWNER:
		{
			int k;
			utf8string *sp = (utf8string *) cptr;
			int err;
			u_int uid;

			nfsv4_printk (vfs, "set_iattr() calling get_num() with user name \"");
			for (k = 0; k < sp->utf8string_len; k++)
				nfsv4_printk (vfs, "%c", sp->utf8string_val[k]);
			nfsv4_printk (vfs, "\"...");
			err = get_num (NUM2STR_UID, sp, &uid);
			if (err) {
				printk ("get_num() returned error\n");
				nfserr = nfserr_inval;        // what is the right error to return?
									 goto out;
			}
			iap->ia_valid |= ATTR_UID;
			iap->ia_uid = uid;
			nfsv4_printk (vfs, "returned uid %d\n", iap->ia_uid);
			cptr += sizeof(utf8string) + sp->utf8string_len;
		}
		break;
		case FATTR4_OWNER_GROUP:
		{
			int k;
			utf8string *sp = (utf8string *) cptr;
			int err;
			u_int gid;

			nfsv4_printk (vfs, "set_iattr() calling get_num() with group name \"");
			for (k = 0; k < sp->utf8string_len; k++)
				nfsv4_printk (vfs, "%c", sp->utf8string_val[k]);
			nfsv4_printk (vfs, "\"...");
			err = get_num (NUM2STR_GID, sp, &gid);
			if (err) {
				printk ("get_num() returned error\n");
				nfserr = nfserr_inval;        // what is the right error to return?
									 goto out;
			}
			iap->ia_valid |= ATTR_GID;
			iap->ia_gid = gid;
			nfsv4_printk (vfs, "returned gid %d\n", iap->ia_gid);
			cptr += sizeof(utf8string) + sp->utf8string_len;
		}
		break;
		case FATTR4_TIME_ACCESS:
		{
			fattr4_time_access *time_access = (fattr4_time_access *)cptr;

			nfsv4_printk(vfs,"NFSV4 nfs4_set_iattr FATTR4_TIME_ACCESS \n");
			if(cptr + sizeof(fattr4_time_access) > end)
				goto overrun;
			iap->ia_valid |= ATTR_ATIME;
			iap->ia_atime = time_access->seconds;
			cptr += sizeof(fattr4_time_access);
		}
		break;
		case FATTR4_TIME_MODIFY:
		{
			fattr4_time_modify *time_modify = (fattr4_time_modify *)cptr;

			nfsv4_printk(vfs,"NFSV4 nfs4_set_iattr FATTR4_TIME_MODIFY \n");
			if(cptr + sizeof(fattr4_time_modify) > end)
				goto overrun;
			iap->ia_valid |= ATTR_MTIME;
			iap->ia_mtime = time_modify->seconds;
			cptr += sizeof(fattr4_time_modify);
		}
		break;
		case FATTR4_TIME_CREATE:
		{
			fattr4_time_create *time_create = (fattr4_time_create *)cptr;

			nfsv4_printk(vfs,"NFSV4 nfs4_set_iattr FATTR4_TIME_CREATE \n");
			if(cptr + sizeof(fattr4_time_create) > end)
				goto overrun;
			iap->ia_valid |= ATTR_MTIME;
			iap->ia_mtime = time_create->seconds;	
			cptr += sizeof(fattr4_time_create);
		}
		break;
		default:
			printk("!!! NFSV4: nfs4_set_iattr: unsupported case %d\n",bit);
			break;
		}
	}while (bit != -1);

	//	iap->ia_valid = ATTR_MODE | ATTR_UID | ATTR_GID;
	//	nfsv4_printk(vfs,"NFSV4 nfs4_set_iattr no S_IFMT, ia_mode: 0x%x\n",iap->ia_mode);

	//	if(mode){
		//		iap->ia_mode &= ~S_IFMT;
		//		if (!(iap->ia_valid & ATTR_MODE)) {
			//			iap->ia_valid |= ATTR_MODE;
			//			iap->ia_mode = S_IFREG;
			//		} else {
				//			iap->ia_mode = (iap->ia_mode & ~S_IFMT) | S_IFREG;
				//		}                                                      
		//	}else
			//		nfserr = nfserr_inval;

	//	nfsv4_printk(vfs,"NFSV4 nfs4_set_iattr ia_mode: 0x%x\n",iap->ia_mode);

 out:
	return nfserr;

 overrun:
	nfsv4_printk(vfs,"NFSV4 nfs4_set_iattr OVERRUN error\n");
	nfserr = nfserr_resource;
	goto out;
}

void
hard_code_secinfo4res(SECINFO4res *resp)
{
	secinfo4 *secinfo;
	rpcsec_gss_info *gss_info;
	char krb5[] = "1.2.840.113554.1.2.2";
	int len = 0;

	nfsv4_printk(secinfo,"NFSV4 hard_code_secinfo4\n");	

	resp->status = 0;
	resp->SECRES.SECINFO4resok_len = 1;
	resp->SECRES.SECINFO4resok_val = kmalloc(1*sizeof(secinfo4),GFP_KERNEL);

	secinfo = &resp->SECRES.SECINFO4resok_val[0];
	secinfo->flavor = RPCSEC_GSS;

	secinfo->flavor_info.flavor_info_val = kmalloc(1*sizeof(rpcsec_gss_info),GFP_KERNEL);

	gss_info = (rpcsec_gss_info *)secinfo->flavor_info.flavor_info_val;
	gss_info->oid.sec_oid4_len = strlen(krb5);
	gss_info->oid.sec_oid4_val = krb5;
	
	gss_info->qop = 42; /* fake qop */	
	gss_info->service = RPC_GSS_SVC_NONE;
	len = gss_info->oid.sec_oid4_len + sizeof(qop4) + sizeof(rpc_gss_svc_t);
	secinfo->flavor_info.flavor_info_len = XDR_QUADLEN(len) << 2;

	nfsv4_printk(secinfo,
		     "NFSDV4 hard_code_secinfo4 &gss_info 0x%x resp->SECRES.SECINFO4resok_val.flavor_info.flavor_info_len %d\n",(unsigned int)&gss_info,
		     (unsigned int)resp->SECRES.SECINFO4resok_val->flavor_info.flavor_info_len);
	print_rpcsec_gss_info(gss_info);
}

/*
 * NULL call.
 */
static int
nfsd4_proc_null(struct svc_rqst *rqstp, void *argp, void *resp)
{
	/*nfsv4_printk(null,"NFSV4 nfsd: NULL proc\n");*/
	return 0;
}

/*
 * Compound call.
 */

static int
nfsd4_proc_compound(struct svc_rqst *rqstp, struct COMPOUND4args  *argp,
		    struct COMPOUND4res *resp)
{
	int i;
	int status = NFS4_OK;
	int numops_done = 0;
	time_t now = CURRENT_TIME;

	nfsv4_printk(compound,"NFSV4 nfsd: COMPOUND proc. opcnt: %d\n",argp->OPCNT);
	nfsv4_printk(kendrick, "nfsd4_proc_compound() called with tag \"%.*s\"\n",
		     (int) argp->tag.utf8string_len, argp->tag.utf8string_val);

	/* copy the args tag into res.NOTE: utf8copy calls kmalloc... */
	utf8copy(&resp->tag,&argp->tag);

	resp->RESCNT = argp->OPCNT;
	resp->RESVAL = kmalloc(resp->RESCNT * sizeof(struct nfs_resop4),GFP_KERNEL);
	memset(resp->RESVAL,0,resp->RESCNT * sizeof(struct nfs_resop4));
	for(i=0;i<argp->OPCNT && status == NFS4_OK;i++){
		switch (argp->OPVAL[i].argop){
		case OP_ACCESS:
			nfsv4_printk(compound,"NFSDV4 proc compound case ACCESS\n"); 
			status = nfsd4_access(rqstp,&argp->OPVAL[i].OPU.opaccess,&resp->RESVAL[i].RESU.opaccess);
			resp->RESVAL[i].resop = OP_ACCESS;
			resp->RESVAL[i].RESU.opaccess.status = status;
			break;
		case OP_CLOSE:
		{
			CLOSE4args *aptr = &argp->OPVAL[i].OPU.opclose;
			CLOSE4res *rptr = &resp->RESVAL[i].RESU.opclose;
				
			nfsv4_printk(compound,"NFSDV4 proc compound case CLOSE. seqid: %d\n",
				     aptr->seqid); 
				/* find the lock given CURRENT_FH and args->stateid. 
				 * check the stateid against open file locking state.
				 * check and possibily renew the lease
				 * check the seqence number. 
				 */
			status = nfs4_process_lock(rqstp,&rptr->CLOSE4res_u.stateid,
						   aptr->stateid,aptr->seqid,OP_CLOSE,now);	
			if(status)
				goto out_close;

		out_close:
			resp->RESVAL[i].resop = OP_CLOSE;
			resp->RESVAL[i].RESU.opclose.status = status;
		}
		break;
		case OP_COMMIT:
			status = nfserr_notsupp;
			break;
		case OP_CREATE:
		{
			CREATE4args *aptr = &argp->OPVAL[i].OPU.opcreate;
			CREATE4res *rptr = &resp->RESVAL[i].RESU.opcreate;

			nfsv4_printk(compound,"NFSDV4 proc compound case CREATE\n");
			status = nfsd4_Create (rqstp, aptr, rptr, &TEMP_FH);
			if (status)
				goto create_out;
			fh_copy (&CURRENT_FH, &TEMP_FH);

		create_out:
			resp->status = status;
			resp->RESVAL[i].resop = OP_CREATE;
			resp->RESVAL[i].RESU.opcreate.status = status;
		}
		break;
		case OP_DELEGPURGE:
			status = nfserr_notsupp;
			break;
		case OP_DELEGRETURN:
			status = nfsd4_delegreturn(argp->OPVAL[i].OPU.opdelegreturn.stateid);
			resp->RESVAL[i].resop = OP_DELEGRETURN;
			resp->RESVAL[i].RESU.opdelegreturn.status = status;
			break;
		case OP_GETATTR:
		{
			GETATTR4res *rptr = &resp->RESVAL[i].RESU.opgetattr;
			GETATTR4args *aptr = &argp->OPVAL[i].OPU.opgetattr;
			int attrslen = 0;

				/* bitmap4 allocation: */
			nfsv4_printk(compound,"NFSDV4 proc COMPOUND case OP_GETATTR\n");
			rptr->GFATTR4.attrmask.BLEN = aptr->attr_request.BLEN;
			nfsv4_printk(compound,
				     "NFSDV4 nfsd4_proc_compound GETATTR bitmap4 allocation. len: %d\n",
				     rptr->GFATTR4.attrmask.BLEN); 
			rptr->GFATTR4.attrmask.BVAL = kmalloc(rptr->GFATTR4.attrmask.BLEN * sizeof(uint32_t),GFP_KERNEL);
			memset(rptr->GFATTR4.attrmask.BVAL,0,rptr->GFATTR4.attrmask.BLEN * sizeof(uint32_t));

			/* calculate size of a successful call in words, 
			 * and set rptr->GFATTR.FATTRSLEN for size in bytes.
			 * rptr->GFATTR4.FATTRSLEN is later zeroed then filled 
			 * in by
			 * nfsd4_getattr() as each indicated attribute is
			 * successfully aquired
			 */
			attrslen = get_fattr4_res_buf_size(&aptr->attr_request);
			/* switch from words to char */
			rptr->GFATTR4.FATTRSLEN = (XDR_QUADLEN(attrslen) << 2);

			rptr->GFATTR4.FATTRSVAL = kmalloc(rptr->GFATTR4.FATTRSLEN,GFP_KERNEL);
			memset(rptr->GFATTR4.FATTRSVAL,0,XDR_QUADLEN(attrslen) * sizeof(char));
				/* Finally, call getattr... */
			status = nfsd4_getattr(rqstp, aptr,rptr);

				/* stuff the reply status */
			resp->RESVAL[i].resop = OP_GETATTR;
			resp->RESVAL[i].RESU.opgetattr.status = status;
		}
		break;
		case OP_GETFH:
			/* no OPU.opgetfh - void argp  */
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_GETFH\n");
			status = nfsd4_getfh(rqstp,
					     &resp->RESVAL[i].RESU.opgetfh);
			resp->RESVAL[i].resop = OP_GETFH;
			resp->RESVAL[i].RESU.opgetfh.status = status;
			break;
		case OP_LINK:
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_LINK\n");
			status = nfsd4_link(rqstp, &argp->OPVAL[i].OPU.oplink, &resp->RESVAL[i].RESU.oplink);
			resp->RESVAL[i].resop = OP_LINK;
			resp->RESVAL[i].RESU.oplink.status = status;
			break;
		case OP_LOCK:
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_LOCK\n");
			status = nfserr_notsupp;
			break;
		case OP_LOCKT:
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_LOCKTTTTT\n");
			status = nfserr_notsupp;
			break;
		case OP_LOCKU:
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_LOCKU\n");
			status = nfserr_notsupp;
			break;
		case OP_LOOKUP:
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_LOOKUP\n");
			status = nfsd4_lookup(rqstp, argp->OPVAL[i].OPU.oplookup.path);
			resp->RESVAL[i].RESU.oplookup.status = status;
			resp->RESVAL[i].resop = OP_LOOKUP;
			break;
		case OP_LOOKUPP:
			// create phony lookup arg/res structs and call nfsd4_lookup()
				{
					component4 dot_dot = { 2, ".." };
					LOOKUP4args args = {{ 1, &dot_dot }};
					LOOKUP4res res;
					nfsv4_printk(compound,
						     "NFSDV4 proc COMPOUND case OP_LOOKUPP\n");
					status = nfsd4_lookup (rqstp, args.path);
					res.status = status;
					resp->RESVAL[i].resop = OP_LOOKUPP;
					resp->RESVAL[i].RESU.oplookupp.status = status;
				}
		break;
		case OP_NVERIFY:
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_VERIFY\n");
			status = nfserr_notsupp;
			break;
		case OP_OPEN:
			resp->RESVAL[i].resop = OP_OPEN;
			status = nfs4_open(rqstp, &argp->OPVAL[i].OPU.opopen, &resp->RESVAL[i].RESU.opopen);
			break;
		case OP_OPENATTR:
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_OPENATTR\n");
			status = nfserr_notsupp;
			break;
		case OP_OPEN_CONFIRM:
		{
			OPEN_CONFIRM4res *rptr = &resp->RESVAL[i].RESU.opopen_confirm;
			OPEN_CONFIRM4args *aptr = &argp->OPVAL[i].OPU.opopen_confirm;
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_OPEN_CONFIRM\n");
			status = nfsd4_open_confirm(rqstp,aptr,rptr,now);
			resp->RESVAL[i].resop = OP_OPEN_CONFIRM;
			resp->RESVAL[i].RESU.opopen_confirm.status = status;
		}
		break;
		case OP_PUTFH:
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_PUTFH\n");
			status = nfsd4_putfh(rqstp,
					     &argp->OPVAL[i].OPU.opputfh);
			resp->RESVAL[i].resop = OP_PUTFH;
			resp->RESVAL[i].RESU.opputfh.status = status;
			break;
		case OP_PUTPUBFH:
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_PUTPUBFH\n");
			status = nfserr_notsupp;
			break;
		case OP_PUTROOTFH:
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_PUTROOTFH\n");
			status = nfsd4_putrootfh(rqstp);
			resp->RESVAL[i].resop = OP_PUTROOTFH;
			resp->RESVAL[i].RESU.opputrootfh.status = status;
			break;
		case OP_READ:
		{
			READ4res *rptr = &resp->RESVAL[i].RESU.opread;
			READ4args *aptr = &argp->OPVAL[i].OPU.opread;
			char prntbuf[32];

			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_READ\n");

				/* find the lock given CURRENT_FH and args->stateid. 
				 * check the stateid against open file locking state.
				 * check and possibly renew lease.
				 */
/* XXX ANDROS what about 1 */
			if(aptr->stateid != 0){
				status = nfs4_process_lock(rqstp,NULL,aptr->stateid,0,OP_READ,now);	
				if(status)
					goto out_read;
			}

			/* NOTE: for now, kmalloc off a PAGESIZE (4096)
			 * of space for the READ4res.READVAL buffer.
			 * once read works, i'll use 
			 * the svcbuf_reserve() and return iovec...
			 * to eliminate the copy.
			 */
			
			rptr->READVAL = kmalloc(aptr->count,GFP_KERNEL);
			if(rptr->READVAL){
				/* yeah, i know, this is slowwwww */
				memset(rptr->READVAL,0,aptr->count);

				rptr->READLEN = aptr->count;
				status = nfsd4_read(rqstp,aptr,rptr);
				nfsv4_printk(compound,"NFSDV4 proc COMPOUND after nfsd4_read");
				memset(prntbuf,0,32);
				memcpy(prntbuf,rptr->READVAL,31);
				nfsv4_printk(compound,"first 31 bytes: %s\n", prntbuf);
					
				nfsv4_printk(compound," READLEN %d status %d  EOF %d\n",
					     rptr->READLEN, status,
					     rptr->READEOF);
			} else
				status = nfserr_resource;
		out_read:
			nfsv4_printk(compound,"NFSV4 compound OP_READ returns status %d \n",status);
			resp->RESVAL[i].resop = OP_READ;
			rptr->status = status;
		}
		break;
		case OP_READDIR:
		{
			READDIR4res *rptr = &resp->RESVAL[i].RESU.opreaddir;
			READDIR4args *aptr = &argp->OPVAL[i].OPU.opreaddir;
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_READDIR\n");
			/* NOTE: for now, kmalloc off a PAGESIZE (4096)
			 * of space for the READDIR4res.ENTRIES buffer.
			 * once readdir works, i'll use 
			 * the svcbuf_reserve() and return iovec...
			 * to eliminate the copy.
			 */
			
			rptr->ENTRIES = kmalloc(aptr->maxcount,GFP_KERNEL);
			if(rptr->ENTRIES){
				/* yeah, i know, this is slowwwww */
				memset(rptr->ENTRIES,0,aptr->maxcount);
				status = nfsd4_readdir(rqstp, &argp->OPVAL[i].OPU.opreaddir,rptr);
				nfsv4_printk(compound,
					     "NFSDV4 proc COMPOUND after nfsd4_readdir\n");
			}else
				status = nfserr_resource;

			resp->RESVAL[i].resop = OP_READDIR;
			resp->RESVAL[i].RESU.opreaddir.status = status;
		}
		break;
		case OP_READLINK:
		{
			READLINK4res *rptr = &resp->RESVAL[i].RESU.opreadlink;

			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_READLINK\n");
			rptr->READLINK4res_u.resok4.link.utf8string_val = kmalloc (NFS4_MAXNAMLEN, GFP_KERNEL);
			status = nfsd4_readlink (rqstp, rptr);

			resp->RESVAL[i].resop = OP_READLINK;
			resp->RESVAL[i].RESU.opreadlink.status = status;
		}
		break;
		case OP_REMOVE:
        		nfsv4_printk(compound, 
				     "NFSDv4: proc COMPOUND case OP_REMOVE \n");
        		status = nfsd4_remove(rqstp,&argp->OPVAL[i].OPU.opremove,&resp->RESVAL[i].RESU.opremove);
        		resp->RESVAL[i].resop = OP_REMOVE;
        		resp->RESVAL[i].RESU.opremove.status = status;
      			break;
		case OP_RENAME:
			nfsv4_printk(compound, 
				     "NFSDV4: proc COMPOUND case OP_RENAME\n");
			status = nfsd4_rename(rqstp,&argp->OPVAL[i].OPU.oprename,&resp->RESVAL[i].RESU.oprename);
			resp->RESVAL[i].resop = OP_RENAME;
			resp->RESVAL[i].RESU.oprename.status = status;
			break;
		case OP_RENEW:
		{
			RENEW4args *args = &argp->OPVAL[i].OPU.oprenew;

			nfsv4_printk(compound, 
				     "NFSDV4: proc COMPOUND case OP_RENEW\n");
			status = nfs4_process_lock(rqstp,NULL,args->stateid,0,OP_RENEW,now);
			resp->RESVAL[i].resop = OP_RENEW;
			resp->RESVAL[i].RESU.oprenew.status = status;
		}
		break;
		case OP_RESTOREFH:
			nfsv4_printk(compound, 
				     "NFSDv4: proc COMPOUND case RESTOREFH \n");
			status = nfsd4_restorefh(rqstp);
			resp->RESVAL[i].resop = OP_RESTOREFH;
			nfsv4_printk(rename, "NFSDv4: RESTOREFH status %d\n", status);
			resp->RESVAL[i].RESU.oprestorefh.status = status;
			break;
		case OP_SECINFO:
			nfsv4_printk(compound,
				     "NFSV4: proc COMPOUND case OP_SECINFO\n");
			print_utf8string(&argp->OPVAL[i].OPU.opsecinfo.name);
			hard_code_secinfo4res(&resp->RESVAL[i].RESU.opsecinfo);
			nfsv4_printk(secinfo,"secinfo FLAVOR: %d\n", 
				     resp->RESVAL[i].RESU.opsecinfo.SECRES.SECINFO4resok_val[0].flavor);
			nfsv4_printk(secinfo,"secinfo FLAVOR_INFO_LEN: %d\n", 
				     resp->RESVAL[i].RESU.opsecinfo.SECRES.SECINFO4resok_val[0].flavor_info.flavor_info_len);
			nfsv4_printk(secinfo,"secinfo FLAVOR_INFO_VAL: 0x%x\n", 
				     (unsigned int)&resp->RESVAL[i].RESU.opsecinfo.SECRES.SECINFO4resok_val[0].flavor_info.flavor_info_val[0]);

/*
  print_rpcsec_gss_info((rpcsec_gss_info *)resp->RESVAL[i].RESU.opsecinfo.SECRES.SECINFO4resok_val[0].flavor_info.flavor_info_val);  
*/
			resp->RESVAL[i].resop = OP_SECINFO;
			resp->RESVAL[i].RESU.oprename.status = 0;
			break;
		case OP_SAVEFH:
			nfsv4_printk(compound, 
				     "NFSDv4: proc COMPOUND case OP_SAVEFH\n");
			status = nfsd4_savefh(rqstp);
			resp->RESVAL[i].resop = OP_SAVEFH;
			nfsv4_printk(rename, "NFSDv4: SAVEFH status %d\n", status);
			resp->RESVAL[i].RESU.opsavefh.status = status;
			break;
		case OP_SETATTR:
		{
			int blen = argp->OPVAL[i].OPU.opsetattr.obj_attributes.attrmask.bitmap4_len;
			nfsv4_printk(compound, 
				     "NFSDv4: proc COMPOUND case OP_SETATTR\n");

			// allocate resp bitmap
				   resp->RESVAL[i].RESU.opsetattr.attrsset.bitmap4_len = blen;
			resp->RESVAL[i].RESU.opsetattr.attrsset.bitmap4_val =
				(uint32_t *) kmalloc (blen * sizeof(uint32_t), GFP_KERNEL);
			memset (resp->RESVAL[i].RESU.opsetattr.attrsset.bitmap4_val, 0, blen * sizeof(uint32_t));
			  
			status = nfsd4_setattr(rqstp, &argp->OPVAL[i].OPU.opsetattr, &resp->RESVAL[i].RESU.opsetattr);
			resp->RESVAL[i].resop = OP_SETATTR;
			resp->RESVAL[i].RESU.opsetattr.status = status;
		}
		break;
		case OP_SETCLIENTID:
		{
			SETCLIENTID4args *args = &argp->OPVAL[i].OPU.opsetclientid;
			SETCLIENTID4res *rptr = &resp->RESVAL[i].RESU.opsetclientid;

			nfsv4_printk(compound, 
				     "NFSDv4: proc COMPOUND case OP_SETCLIENTID\n");
			nfsv4_printk(setclntid, "OP_SETCLIENTID: verifier:");
			print_verifier4(args->client.verifier);
			nfsv4_printk(setclntid, "OP_SETCLIENTID: id %s\n",
				     args->client.id.id_val);
			status = nfsd4_setclientid(rqstp,args,rptr);
			resp->RESVAL[i].resop = OP_SETCLIENTID;
			nfsv4_printk(setclntid, "NFSDv4: SETCLIENTID status %d\n", status);
			resp->RESVAL[i].RESU.opsetclientid.status = status;
		}
		break;
		case OP_SETCLIENTID_CONFIRM:
			nfsv4_printk(compound, 
				     "NFSDv4: proc COMPOUND case OP_SETCLIENTID_CONFIRM\n");
			status = nfsd4_setclientid_confirm(rqstp,&argp->OPVAL[i].OPU.opsetclientid_confirm);
			resp->RESVAL[i].resop = OP_SETCLIENTID_CONFIRM;
			nfsv4_printk(setclntid, 
				     "NFSDv4: SETCLIENTID_CONFIRM status %d\n", status);
			resp->RESVAL[i].RESU.opsetclientid_confirm.status = status;
			break;
		case OP_VERIFY:
			nfsv4_printk(compound, 
				     "NFSDv4: proc COMPOUND case OP_VERIFY\n");
			status = nfserr_notsupp;
			break;
		case OP_WRITE:
		{
			WRITE4res *rptr = &resp->RESVAL[i].RESU.opwrite;
			WRITE4args *aptr = &argp->OPVAL[i].OPU.opwrite;
			char prntbuf[32];

			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case OP_WRITE\n");
			memset(prntbuf,0,32);
			memcpy(prntbuf,aptr->WRITEVAL,31);
			nfsv4_printk(compound,"OP_WRITE: first 31 bytes: %s\n",
				     prntbuf);

				/* find the lock given CURRENT_FH and args->stateid. 
				 * check the stateid against open file locking state.
				 * check and possibly renew lease.
				 */
			if(aptr->stateid != 0){
				status = nfs4_process_lock(rqstp,NULL,aptr->stateid,0,OP_WRITE,now);	
				if(status)
					goto out_write;
			}

			status = nfsd4_write(rqstp,aptr,rptr);

		out_write:
			nfsv4_printk(compound,"NFSDV4 proc COMPOUND after nfsd4_write. status: %d bytes written: %d\n",status,rptr->WRITERES.count);
					
			resp->RESVAL[i].resop = OP_WRITE;
			rptr->status = status;
		}
		break;
		default:
			nfsv4_printk(compound,
				     "NFSDV4 proc COMPOUND case default, BAD OPCODE\n");
			status = nfserr_inval;
			break;
		}
		if(status == NFS4_OK) numops_done++;
	}
	resp->RESCNT = numops_done;
	resp->status = status;
	nfsv4_printk(compound,
		     "NFSV4: proc_compound returns status: %d numops_done %d\n", 
		     resp->status,numops_done);
	nfsd4_kfree_compound_args(argp);
	nfsv4_printk (kendrick, "nfsd4_proc_compound() returning status %d\n", status);
	return(status);
}


/*
 * NFSv4 Server procedures.
 */
/*
#define nfs4svc_decode_voidargs         NULL
#define nfs4svc_encode_voidres          NULL
*/
#define nfs4svc_release_void            NULL
#define nfsd4_voidres                   nfsd4_voidargs

#define nfs4svc_decode_COMPOUND4argsargs nfs4svc_decode_COMPOUND4args
#define nfs4svc_encode_COMPOUND4resres nfs4svc_encode_COMPOUND4res
#define nfsd4_COMPOUND4argsargs		COMPOUND4args
#define nfsd4_COMPOUND4resres		COMPOUND4res

struct nfsd4_voidargs { int dummy; };

#define PROC(name, argt, rest, relt, cache)	\
 { (svc_procfunc) nfsd4_proc_##name,		\
   (kxdrproc_t) nfs4svc_decode_##argt##args,	\
   (kxdrproc_t) nfs4svc_encode_##rest##res,	\
   (kxdrproc_t) nfs4svc_release_##relt,		\
   sizeof(struct nfsd4_##argt##args),		\
   sizeof(struct nfsd4_##rest##res),		\
   0,						\
   cache					\
 }
struct svc_procedure		nfsd_procedures4[2] = {
	PROC(null,	 void,		void,		void,	 RC_NOCACHE),
	PROC(compound, COMPOUND4args, COMPOUND4res,   compound, RC_NOCACHE)
};
