/*
 * linux/fs/nfsd/nfsd4xdr.c
 *
 * XDR support for nfsd/protocol version 4.
 *
 * by Andy Adamson <andros@citi.umich.edu>
 * Copyright (C) 1999 The Regents of the University of Michigan
 */

#include <linux/types.h>
#include <linux/sched.h>

#include <linux/sunrpc/xdr.h>
#include <linux/sunrpc/svc.h>
#include <linux/nfsd/nfsd4_prot.h>
#include <linux/malloc.h>
extern bitmap4 ATTR_BITMAP;
extern struct nfs4_lock_owner *current_lockowner;


#define NFSDDBG_FACILITY		NFSDDBG_XDR

/*
 * Initialization of NFSV4 status variables
 */

void
nfsd4_xdr_init(void)
{
	static int      inited = 0;

	if (inited)
		return;

	nfs_ok          = htonl(NFS4_OK);
	nfserr_perm     = htonl(NFS4ERR_PERM);
	nfserr_noent    = htonl(NFS4ERR_NOENT);
	nfserr_io       = htonl(NFS4ERR_IO);
	nfserr_inval    = htonl(NFS4ERR_INVAL);
	nfserr_nxio     = htonl(NFS4ERR_NXIO);
	nfserr_acces    = htonl(NFS4ERR_ACCES);
	nfserr_exist    = htonl(NFS4ERR_EXIST);
	nfserr_xdev     = htonl(NFS4ERR_XDEV);
	nfserr_nodev    = htonl(NFS4ERR_NODEV);
	nfserr_notdir   = htonl(NFS4ERR_NOTDIR);
	nfserr_isdir    = htonl(NFS4ERR_ISDIR);
	nfserr_fbig     = htonl(NFS4ERR_FBIG);
	nfserr_nospc    = htonl(NFS4ERR_NOSPC);
	nfserr_rofs     = htonl(NFS4ERR_ROFS);
	nfserr_mlink    = htonl(NFS4ERR_MLINK);
	nfserr_nametoolong = htonl(NFS4ERR_NAMETOOLONG);
	nfserr_notempty = htonl(NFS4ERR_NOTEMPTY);
	nfserr_dquot    = htonl(NFS4ERR_DQUOT);
	nfserr_stale    = htonl(NFS4ERR_STALE);
	nfserr_badhandle = htonl(NFS4ERR_BADHANDLE);
	nfserr_badcookie = htonl(NFS4ERR_BAD_COOKIE);
	nfserr_notsupp  = htonl(NFS4ERR_NOTSUPP);
	nfserr_toosmall = htonl(NFS4ERR_TOOSMALL);
	nfserr_serverfault = htonl(NFS4ERR_SERVERFAULT);
	nfserr_badtype  = htonl(NFS4ERR_BADTYPE);
	nfserr_delay  = htonl(NFS4ERR_DELAY);
	nfserr_same	= htonl(NFS4ERR_SAME);
	nfserr_denied	= htonl(NFS4ERR_DENIED);
	nfserr_expired	= htonl(NFS4ERR_EXPIRED);
	nfserr_locked	= htonl(NFS4ERR_LOCKED);
	nfserr_grace	= htonl(NFS4ERR_GRACE);
	nfserr_fhexpired	= htonl(NFS4ERR_FHEXPIRED);
	nfserr_share_denied	= htonl(NFS4ERR_SHARE_DENIED);
	nfserr_wrongsec	= htonl(NFS4ERR_WRONGSEC);
	nfserr_clid_inuse	= htonl(NFS4ERR_CLID_INUSE);
	nfserr_resource	= htonl(NFS4ERR_RESOURCE);
	nfserr_moved	= htonl(NFS4ERR_MOVED);
	nfserr_nofilehandle	= htonl(NFS4ERR_NOFILEHANDLE);
	nfserr_minor_vers_mismatch = htonl(NFS4ERR_MINOR_VERS_MISMATCH);
	nfserr_stale_clientid = htonl(NFS4ERR_STALE_CLIENTID);
	nfserr_stale_stateid = htonl(NFS4ERR_STALE_STATEID);
	nfserr_old_stateid = htonl(NFS4ERR_OLD_STATEID);
	nfserr_bad_stateid = htonl(NFS4ERR_BAD_STATEID);
	nfserr_bad_seqid = htonl(NFS4ERR_BAD_SEQID);
	nfserr_not_same = htonl(NFS4ERR_NOT_SAME);
	nfserr_lock_range = htonl(NFS4ERR_LOCK_RANGE);
	nfserr_symlink = htonl(NFS4ERR_SYMLINK);
	nfserr_replay_me = htonl(NFS4ERR_REPLAY_ME);
	  
	inited = 1;
}


#ifdef NFSD_OPTIMIZE_SPACE
# define inline
#endif

static inline int
xdr_overflow_check(struct svc_buf *buf, u32 *p, int arglen)
{
	return p + (arglen >>2) - buf->base <= buf->buflen;
}


/* just like xdr_encode/decode_hyper*/
static inline u32 *
xdr_encode_64(u32 *p, u64 val)
{
	nfsv4_printk(xdr,"NFSDV4 xdr_encode_64 val: %ld\n",(long int)val);
	*p++ = htonl(val >> 32);
	*p++ = htonl(val & 0xffffffff);
	return p;
}

static inline u32 *
xdr_decode_64(u32 *p, u64 *valp)
{
	*valp  = ((u64) ntohl(*p++)) << 32;
	*valp |= ntohl(*p++);
	return p;
}
static inline u32 *
xdr_decode_fh4(struct svc_rqst *rqstp,u32 *p, nfs_fh4 *fhp)
{
	int len = sizeof(fhp->nfs_fh4_len) + XDR_QUADLEN(sizeof(struct knfs_fh));

	nfsv4_printk(fh4,"NFSDV4: xdr_decode_fh4");
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	fhp->nfs_fh4_len = ntohl(*p++);
	if (fhp->nfs_fh4_len == sizeof(struct knfs_fh)) {
		fhp->nfs_fh4_val= kmalloc(fhp->nfs_fh4_len,GFP_KERNEL);
		if(!fhp->nfs_fh4_val){
			fhp->nfs_fh4_val = NULL;
			return 0;
		}
		memset(fhp->nfs_fh4_val,0,fhp->nfs_fh4_len);
		memcpy(fhp->nfs_fh4_val, p, fhp->nfs_fh4_len);
		print_nfs_fh4(fhp);
		return p + XDR_QUADLEN(fhp->nfs_fh4_len);
	} 
	nfsv4_printk(fh4,"NFSDV4 xdr_decode_fh: incorrect fh size: %d\n",fhp->nfs_fh4_len);
	return 0;
}

static inline u32 *
xdr_encode_fh4(struct svc_rqst *rqstp,u32 *p, nfs_fh4 *fhp)
{
	int len = sizeof(fhp->nfs_fh4_len) + XDR_QUADLEN(fhp->nfs_fh4_len);

	nfsv4_printk(fh4,"NFSDV4: xdr_encode_fh4: FILE HANDLE:\n");
	print_nfs_fh4(fhp);

	if(fhp->nfs_fh4_len > NFS4_FHSIZE){
		nfsv4_printk(fh4,
			     "NFSDV4 xdr_encode_fh: incorrect fh size: %d\n",fhp->nfs_fh4_len);
		return 0;
	}
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
		return 0;
	*p++ = htonl(fhp->nfs_fh4_len);
	memcpy(p, fhp->nfs_fh4_val, fhp->nfs_fh4_len);
	return p + XDR_QUADLEN(fhp->nfs_fh4_len);
}

static u32 *
xdr_encode_verifier4(u32 *p,verifier4 ch4)
{
	nfsv4_printk(setclntid, "xdr_encode_verifier: \n");
	print_verifier4(ch4);
	memcpy(p,ch4,NFS4_VERIFIER_SIZE);
	return p + XDR_QUADLEN(NFS4_VERIFIER_SIZE);
}

static u32 *
xdr_decode_verifier4(u32 *p,verifier4 ch4)
{
	nfsv4_printk(setclntid,"NFS4 xdr_decode_verifier  verifier: %p\n",ch4);
	memcpy(ch4,(char *)p,NFS4_VERIFIER_SIZE);
	print_verifier4(ch4);
	p =  p + XDR_QUADLEN(NFS4_VERIFIER_SIZE);
	nfsv4_printk(setclntid,"NFS4 xdr_decode_verifier2  verifier: %p\n",ch4);
	print_verifier4(ch4);
	return p;
}


static inline u32 *
xdr_encode_nfstime4(u32 *p, nfstime4 time)
{
	p = xdr_encode_64(p,time.seconds);
	*p++ = 0;
	return p;
}


static inline u32 *
xdr_decode_nfstime4(u32 *p, nfstime4 *timep)
{
	p = xdr_decode_64(p,&timep->seconds);
	p++;
	return p;
}

static inline u32 *
xdr_encode_utf8string(struct svc_rqst *rqstp,u32 *p, utf8string  *utf)
{
	int len = sizeof(utf->utf8string_len) + XDR_QUADLEN(utf->utf8string_len); 

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len )))
		return 0;
	*p++=htonl(utf->utf8string_len);
	memcpy(p,utf->utf8string_val,utf->utf8string_len);
	return p + XDR_QUADLEN(utf->utf8string_len);
}

static u32 *
xdr_decode_utf8string(struct svc_rqst *rqstp, u32 *p, utf8string *utf)
{
	int len = sizeof(utf->utf8string_len) + ntohl(*p);
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	utf->utf8string_len = ntohl(*p++);
	utf->utf8string_val = (char *)p;
	return p + XDR_QUADLEN(utf->utf8string_len);
}

static u32*
xdr_decode_bitmap4(struct svc_rqst *rqstp,u32 *p, bitmap4 *bm4)
{
/* ANDROS note: instead of kmalloc, decode BVAL's in place. */

	int i,len = sizeof(bm4->BLEN) + ntohl(*p);
	uint32_t *ptr;

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	bm4->BLEN = ntohl(*p++);
	bm4->BVAL = kmalloc(bm4->BLEN *sizeof(uint32_t),
			    GFP_KERNEL);
	if(!bm4->BVAL)
		return 0;
	memset(bm4->BVAL,0,bm4->BLEN *sizeof(uint32_t));
	ptr = bm4->BVAL;
	for(i=0;i<bm4->BLEN;i++){
		*ptr++ = ntohl(*p++); 
	}
	return p;
}

static u32*
xdr_encode_bitmap4(struct svc_rqst *rqstp,u32 *p, bitmap4 *bm4)
{
	int i,len = sizeof(bm4->BLEN) + (bm4->BLEN * sizeof(uint32_t));
	uint32_t *ptr;

	nfsv4_printk(xdr,"NFSDV4 xdr_encode_bitmap4: BLEN = %d\n",bm4->BLEN);
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len )))
		return 0;
	*p++ =  htonl(bm4->BLEN);
	for(i=0;i<bm4->BLEN;i++){
		ptr=&bm4->BVAL[i];
		nfsv4_printk(xdr,
			     "NFSDV4 xdr_encode_bitmap4: val[%d] = 0x%x\n",i,bm4->BVAL[i]);
		*p++ = htonl(*ptr++);	
	}
	return p;
}

static u32*
xdr_encode_specdata4(u32 *p,fattr4_rawdev *rd)
{
	nfsv4_printk(xdr,"NFSV4 xdr_encode_specdata4  ");
	*p = htonl(rd->specdata1);
	nfsv4_printk(xdr,"specdata1 0x%x ",*p);
	p++;
	*p = htonl(rd->specdata2);
	nfsv4_printk(xdr,"specdata2 0x%x\n",*p);
	p++;
	return p;
}
static u32*
xdr_decode_specdata(u32 *p, fattr4_rawdev *rd)
{
	rd->specdata1 = ntohl(*p++);
	rd->specdata2 = ntohl(*p++);
	return p;
} 

static u32*
xdr_encode_fsid4(u32 *p,fattr4_fsid *f4)
{
	p = xdr_encode_64(p,f4->major);
	p = xdr_encode_64(p,f4->minor);
	return p;
}

static u32*
xdr_decode_fsid4(u32 *p, fattr4_fsid *f4)
{
	p = xdr_decode_64(p,&f4->major);
	p = xdr_decode_64(p,&f4->minor);
	return p;
}

static u32 *
xdr_decode_fattr4(struct svc_rqst *rqstp,u32 *p,fattr4 *res)
{
	int i = 0;
	int bit;
	bitmap4 *bm4 = &res->attrmask;
	void *cptr;
	void *end;
	
	nfsv4_printk(attributes,"NFSV4 xdr_decode_fattr: calling decode bitmap4\n");
	if(!(p = xdr_decode_bitmap4(rqstp,p,bm4)))
		return 0;

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,sizeof(res->FATTRSLEN))))
		return 0;
	res->FATTRSLEN = ntohl(*p++);
	
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,res->FATTRSLEN)))
		return 0;
	res->FATTRSVAL = kmalloc(res->FATTRSLEN, GFP_KERNEL);

	if(!res->FATTRSVAL){
		nfsv4_printk(attributes,"NFSV4 xdr_decode_fattr4: kmalloc failed for res->FATTRSVAL\n");
		res->FATTRSVAL = NULL;
		return 0;
	}
	memset(res->FATTRSVAL,0,res->FATTRSLEN);
	cptr = res->FATTRSVAL;
	if(!cptr){
		nfsv4_printk(attributes,"NFSV4 xdr_decode_fattr4: cptr is 0\n");
		return 0;
	}
	/* p has res->FATTRSLEN room, so end checks for both cptr and p overflow */
	end = (char *)cptr + res->FATTRSLEN;
	nfsv4_printk(attributes,"NFSV4 xdr_decode_fattr4: cptr: 0x%x end 0x%x attrs_len = %d p = 0x%x\n",(unsigned int)cptr, (unsigned int)end,res->FATTRSLEN,(unsigned int)p);
	do {
		bit = get_bitmap4_bit(bm4,i);
		i++;
		if(!bit || bit == -1)
			continue;
		switch(i-1){
		case FATTR4_SUPPORTED_ATTRS:
		{
			int i,size;
			nfsv4_printk(attributes,
				     "NFSV4 xdr_decode_fattr4 FATTR4_SUPPORTED_ATTRS.\n");
			size = ntohl(*p);
			if((cptr + sizeof(u_int) + (size * sizeof(uint32_t))) > end)
				goto out_overrun;
			*(u_int *)cptr = ntohl(*p++); /* bitmap_len*/	
			for(i=0;i<size;i++){
				*(uint32_t *)cptr = ntohl(*p++);	 /* bitmap_val(s)*/
			}
			cptr += sizeof(u_int) + (size * sizeof(uint32_t));
		}
		break;
		case FATTR4_TYPE:
			nfsv4_printk(attributes,"NFSV4 xdr_decode_fattr4 FATTR4_TYPE. p: 0x%x \n",(unsigned int)p);
			if((cptr + sizeof(fattr4_type)) > end)
				goto out_overrun;
			*(fattr4_type *)cptr  = ntohl(*p++);
			cptr += sizeof(fattr4_type);
			break;
		case FATTR4_FH_EXPIRE_TYPE:
			if((cptr + sizeof(fattr4_fh_expire_type)) > end)
				goto out_overrun;
			*(fattr4_fh_expire_type *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_fh_expire_type);
			break;
		case FATTR4_CHANGE:
			if((cptr + sizeof(fattr4_change)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_change *)cptr);
			cptr += sizeof(fattr4_change);
			break;
		case FATTR4_SIZE:
			nfsv4_printk(attributes,"NFSV4 xdr_decode_fattr4 FATTR4_SIZE\n");
			if((cptr + sizeof(fattr4_size)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_size *)cptr);
			cptr += sizeof(fattr4_size);
			break;
		case FATTR4_LINK_SUPPORT:
			if((cptr + sizeof(fattr4_link_support)) > end)
				goto out_overrun;
			*(fattr4_link_support *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_link_support);
			break;
		case FATTR4_SYMLINK_SUPPORT:
			if((cptr + sizeof(fattr4_symlink_support)) > end)
				goto out_overrun;
			*(fattr4_symlink_support *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_symlink_support);
			break;
		case FATTR4_NAMED_ATTR:
			if((cptr + sizeof(fattr4_named_attr)) > end)
				goto out_overrun;
			*(fattr4_named_attr *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_named_attr);
			break;
		case FATTR4_FSID:
			if((cptr + sizeof(fattr4_fsid)) > end)
				goto out_overrun;
			p = xdr_decode_fsid4(p,(fattr4_fsid *)cptr);
			cptr += sizeof(fattr4_fsid);
			break;
		case FATTR4_UNIQUE_HANDLES:
			if((cptr + sizeof(fattr4_unique_handles)) > end)
				goto out_overrun;
			*(fattr4_unique_handles *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_unique_handles);
			break;
		case FATTR4_LEASE_TIME:
			if((cptr + sizeof(fattr4_lease_time)) > end)
				goto out_overrun;
			*(fattr4_lease_time *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_lease_time);
			break;
		case FATTR4_RDATTR_ERROR:
			if((cptr + sizeof(fattr4_rdattr_error)) > end)
				goto out_overrun;
			*(fattr4_rdattr_error *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_rdattr_error);
			break;
		case FATTR4_ACL:
			/* NOT SUPPORTED */
			break;
		case FATTR4_ACLSUPPORT:
			/* NOT SUPPORTED */
			break;
		case FATTR4_ARCHIVE:
			if((cptr + sizeof(fattr4_archive)) > end)
				goto out_overrun;
			*(fattr4_archive *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_archive);
			break;
		case FATTR4_CANSETTIME:
			if((cptr + sizeof(fattr4_cansettime)) > end)
				goto out_overrun;
			*(fattr4_cansettime *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_cansettime);
			break;
		case FATTR4_CASE_INSENSITIVE:
			if((cptr + sizeof(fattr4_case_insensitive)) > end)
				goto out_overrun;
			*(fattr4_case_insensitive *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_case_insensitive);
			break;
		case FATTR4_CASE_PRESERVING:
			if((cptr + sizeof(fattr4_case_preserving)) > end)
				goto out_overrun;
			*(fattr4_case_preserving *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_case_preserving);
			break;
		case FATTR4_CHOWN_RESTRICTED:
			if((cptr + sizeof(fattr4_chown_restricted)) > end)
				goto out_overrun;
			*(fattr4_chown_restricted *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_chown_restricted);
			break;
		case FATTR4_FILEHANDLE:
				/* xdr_decode_fh4 kmalloc's space, so cptr 
				 * needs to advance size of pointer, not length of fh4_val
				 */
			if((cptr + sizeof(nfs_fh4)) > end)
				goto out_overrun;
			if(!(p = xdr_decode_fh4(rqstp,p,(nfs_fh4 *)cptr)))
				return 0;
			cptr += sizeof(nfs_fh4);
			break;
		case FATTR4_FILEID:
			if((cptr + sizeof(fattr4_fileid)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_fileid *)cptr);
			nfsv4_printk(attributes,"NFSV4 xdr_decode_fattr4 : FILEID: %ld\n", *(long int*)cptr);
			cptr += sizeof(fattr4_fileid);
			break;
		case FATTR4_FILES_AVAIL:
			if((cptr + sizeof(fattr4_files_avail)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_files_avail *)cptr);
			cptr += sizeof(fattr4_files_avail);
			break;
		case FATTR4_FILES_FREE:
			if((cptr + sizeof(fattr4_files_free)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_files_free *)cptr);
			cptr += sizeof(fattr4_files_free);
			break;
		case FATTR4_FILES_TOTAL:
			if((cptr + sizeof(fattr4_files_total)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_files_total *)cptr);
			cptr += sizeof(fattr4_files_total);
			break;
		case FATTR4_FS_LOCATIONS:
			/* NOT SUPPORTED*/
			break;
		case FATTR4_HIDDEN:
			if((cptr + sizeof(fattr4_hidden)) > end)
				goto out_overrun;
			*(fattr4_hidden *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_hidden);
			break;
		case FATTR4_HOMOGENEOUS:
			if((cptr + sizeof(fattr4_homogeneous)) > end)
				goto out_overrun;
			*(fattr4_homogeneous *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_homogeneous);
			break;
		case FATTR4_MAXFILESIZE:
			if((cptr + sizeof(fattr4_maxfilesize)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_maxfilesize *)cptr);
			cptr += sizeof(fattr4_maxfilesize);
			break;
		case FATTR4_MAXLINK:
			if((cptr + sizeof(fattr4_maxlink)) > end)
				goto out_overrun;
			*(fattr4_maxlink *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_maxlink);
			break;
		case FATTR4_MAXNAME:
			if((cptr + sizeof(fattr4_maxname)) > end)
				goto out_overrun;
			*(fattr4_maxname *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_maxname);
			break;
		case FATTR4_MAXREAD:
			if((cptr + sizeof(fattr4_maxread)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_maxread *)cptr);
			cptr += sizeof(fattr4_maxread);
			break;
		case FATTR4_MAXWRITE:
			if((cptr + sizeof(fattr4_maxwrite)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_maxwrite *)cptr);
			cptr += sizeof(fattr4_maxwrite);
			break;
		case FATTR4_MIMETYPE:
		{
			fattr4_mimetype *ptr = (fattr4_mimetype *)cptr;
			nfsv4_printk(attributes,"NFSV4: xdr_decode_fattr4 - FATTR4_MIMETYPE\n");
			if((cptr + sizeof(u_int)) > end)
				goto out_overrun;
			cptr += sizeof(u_int);
			ptr->utf8string_val = cptr;
			if((cptr + ptr->utf8string_len) > end)
				goto out_overrun;
			p = xdr_decode_utf8string(rqstp,p,ptr);
			print_utf8string(ptr);
			cptr += ptr->utf8string_len;
		}
		break;
		case FATTR4_MODE:
			if((cptr + sizeof(fattr4_mode)) > end)
				goto out_overrun;
			nfsv4_printk(attributes,"NFSV4 xdr_decode_fattr4 FATTR4_MODE\n");
			*(fattr4_mode *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_mode);
			break;
		case FATTR4_NO_TRUNC:
			if((cptr + sizeof(fattr4_no_trunc)) > end)
				goto out_overrun;
			*(fattr4_no_trunc *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_no_trunc);
			break;
		case FATTR4_NUMLINKS:
			if((cptr + sizeof(fattr4_numlinks)) > end)
				goto out_overrun;
			*(fattr4_numlinks *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_numlinks);
			break;
		case FATTR4_OWNER:
		{
			int l;

			nfsv4_printk(attributes,"NFSV4: xdr_decode_fattr4 - FATTR4_OWNER\n");
			if((cptr + sizeof(u_int)) > end)
				goto out_overrun;
			*((u_int *) cptr) = l = ntohl(*p++);
			cptr += sizeof(u_int);
			*((char **) cptr) = cptr + sizeof(char *);
			cptr += sizeof(char *);
			memcpy (cptr, p, l);
			cptr += l;
			p += XDR_QUADLEN(l);
		}
		break;
		case FATTR4_OWNER_GROUP:
		{
			int l;

			nfsv4_printk(attributes,"NFSV4: xdr_decode_fattr4 - FATTR4_OWNER_GROUP\n");
			if((cptr + sizeof(u_int)) > end)
				goto out_overrun;
			*((u_int *) cptr) = l = ntohl(*p++);
			cptr += sizeof(u_int);
			*((char **) cptr) = cptr + sizeof(char *);
			cptr += sizeof(char *);
			memcpy (cptr, p, l);
			cptr += l;
			p += XDR_QUADLEN(l);
		}
		break;
		case FATTR4_QUOTA_HARD:
			if((cptr + sizeof(fattr4_quota_hard)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_quota_hard *)cptr);
			cptr += sizeof(fattr4_quota_hard);
			break;
		case FATTR4_QUOTA_SOFT:
			if((cptr + sizeof(fattr4_quota_soft)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_quota_soft *)cptr);
			cptr += sizeof(fattr4_quota_soft);
			break;
		case FATTR4_QUOTA_USED:
			if((cptr + sizeof(fattr4_quota_used)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_quota_used *)cptr);
			cptr += sizeof(fattr4_quota_used);
			break;
		case FATTR4_RAWDEV:
			if((cptr + sizeof(fattr4_rawdev)) > end)
				goto out_overrun;
			p = xdr_decode_specdata(p,(fattr4_rawdev *)cptr);
			cptr += sizeof(fattr4_rawdev);
			break;
		case FATTR4_SPACE_AVAIL:
			if((cptr + sizeof(fattr4_space_avail)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_space_avail *)cptr);
			cptr += sizeof(fattr4_space_avail);
			break;
		case FATTR4_SPACE_FREE:
			if((cptr + sizeof(fattr4_space_free)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_space_free *)cptr);
			cptr += sizeof(fattr4_space_free);
			break;
		case FATTR4_SPACE_TOTAL:
			if((cptr + sizeof(fattr4_space_total)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_space_total *)cptr);
			cptr += sizeof(fattr4_space_total);
			break;
		case FATTR4_SPACE_USED:
			if((cptr + sizeof(fattr4_space_used)) > end)
				goto out_overrun;
			p = xdr_decode_64(p,(fattr4_space_used *)cptr);
			cptr += sizeof(fattr4_space_used);
			break;
		case FATTR4_SYSTEM:
			if((cptr + sizeof(fattr4_system)) > end)
				goto out_overrun;
			*(fattr4_system *)cptr = ntohl(*p++);
			cptr += sizeof(fattr4_system);
			break;
		case FATTR4_TIME_ACCESS:
			if((cptr + sizeof(fattr4_time_access)) > end)
				goto out_overrun;
			p = xdr_decode_nfstime4(p,(fattr4_time_access *)cptr);
			cptr += sizeof(fattr4_time_access);
			break;
		case FATTR4_TIME_ACCESS_SET:
			break;
		case FATTR4_TIME_BACKUP:
			if((cptr + sizeof(fattr4_time_backup)) > end)
				goto out_overrun;
			p = xdr_decode_nfstime4(p,(fattr4_time_backup *)cptr);
			cptr += sizeof(fattr4_time_backup);
			break;
		case FATTR4_TIME_CREATE:
			if((cptr + sizeof(fattr4_time_create)) > end)
				goto out_overrun;
			p = xdr_decode_nfstime4(p,(fattr4_time_create *)cptr);
			cptr += sizeof(fattr4_time_create);
			break;
		case FATTR4_TIME_DELTA:
			if((cptr + sizeof(fattr4_time_delta)) > end)
				goto out_overrun;
			p = xdr_decode_nfstime4(p,(fattr4_time_delta *)cptr);
			cptr += sizeof(fattr4_time_delta);
			break;
		case FATTR4_TIME_METADATA:
			if((cptr + sizeof(fattr4_time_metadata)) > end)
				goto out_overrun;
			p = xdr_decode_nfstime4(p,(fattr4_time_metadata *)cptr);
			cptr += sizeof(fattr4_time_metadata);
			break;
		case FATTR4_TIME_MODIFY:
			if((cptr + sizeof(fattr4_time_modify)) > end)
				goto out_overrun;
			p = xdr_decode_nfstime4(p,(fattr4_time_modify *)cptr);
			cptr += sizeof(fattr4_time_modify);
			break;
		case FATTR4_TIME_MODIFY_SET:
			break;
		default:
			break;
		}
	}while (bit != -1);
	nfsv4_printk(attributes,"NFSV4: xdr_decode_fattr4: FINISH. cptr = 0x%x\n",(unsigned int)cptr);
	return p;
 out_overrun:
	nfsv4_printk(attributes,"NFSDV4: xdr_decode_fattr4: OVERRUN, return NULL\n");
	return 0;  
}


static u32*
xdr_encode_fattr4(struct svc_rqst *rqstp,u32 *p,struct fattr4 *res)
{
	bitmap4 *bm4 = &res->attrmask;
	int i = 0,bit;
	void *cptr = res->FATTRSVAL;
	void *end = (char *)cptr + res->FATTRSLEN;
	u32 *start;
	u32 *save;
				
	nfsv4_printk(attributes,"encode_fattr4 cptr 0x%x, end 0x%x, 
	res->FATTRSLEN %d\n", (unsigned int)cptr, 
		     (unsigned int)end,res->FATTRSLEN);

	p = xdr_encode_bitmap4(rqstp,p,bm4);
	
	save = p;
	*p++ = htonl(res->FATTRSLEN);
	start = p;
	nfsv4_printk(attributes,
		     "NFSDV4: xdr_encode_fattr4: p, beginning. 0x%x\n",(unsigned int)p);
	nfsv4_printk(attributes,
		     "NFSDV4: xdr_encode_fattr4: res->FATTRSLEN %d\n", res->FATTRSLEN);

	do{
		bit = get_bitmap4_bit(bm4,i);
		i++;
		if(!bit || bit == -1)
			continue;
		nfsv4_printk(attributes,
			     "NFSDV4: xdr_encode_fattr4:bit %d is set get_bitmap4 return %d\n",
			     i-1,bit);

		switch(i-1){
		case FATTR4_SUPPORTED_ATTRS:
		{		
			int i,size = sizeof(u_int) + (ATTR_BITMAP_LEN * sizeof(uint32_t));
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_SUPPORTED_ATTRS\n");
			if((cptr + size) > end) 
				goto out_overrun;
			*p++ =  htonl(*(u_int *)cptr); /* bitmap4_len */
			for(i=0;i<ATTR_BITMAP_LEN;i++){
				*p++ =  htonl(*(uint32_t *)cptr); /* bitmap4_val  */
			}
			cptr += size; 
		}
		break;
		case FATTR4_TYPE:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_TYPE\n");
			if((cptr + sizeof(fattr4_type)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_type *)cptr);
			cptr += sizeof(fattr4_type); 
			break;
		case FATTR4_FH_EXPIRE_TYPE:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_FH_EXPIRE_TYPE\n");
			if((cptr + sizeof(fattr4_fh_expire_type)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_fh_expire_type *)cptr);
			cptr += sizeof(fattr4_fh_expire_type); 
			break;
		case FATTR4_CHANGE:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_CHANGE\n");
			if((cptr + sizeof(fattr4_change)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_change *)cptr);
			cptr += sizeof(fattr4_change);
			break;
		case FATTR4_SIZE:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_SIZE\n");
			if((cptr + sizeof(fattr4_size)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_size *)cptr);
			cptr += sizeof(fattr4_size); 
			break;
		case FATTR4_LINK_SUPPORT:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_LINK_SUPPORT\n");
			if((cptr + sizeof(fattr4_link_support)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_link_support *)cptr);
			cptr += sizeof(fattr4_link_support); 
			break;
		case FATTR4_SYMLINK_SUPPORT:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_SYMLINK_SUPPORT\n");
			if((cptr + sizeof(fattr4_symlink_support)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_symlink_support *)cptr);
			cptr += sizeof(fattr4_symlink_support); 
			break;
		case FATTR4_NAMED_ATTR:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_NAMED_ATTR\n");
			if((cptr + sizeof(fattr4_named_attr)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_named_attr *)cptr);
			cptr += sizeof(fattr4_named_attr); 
			break;
		case FATTR4_FSID:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_FSID\n");
			if((cptr + sizeof(fattr4_fsid)) > end) 
				goto out_overrun;
			p = xdr_encode_fsid4(p,(fattr4_fsid *)cptr);
			cptr += sizeof(fattr4_fsid); 
			break;
		case FATTR4_UNIQUE_HANDLES:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_UNIQUE_HANDLES\n");
			if((cptr + sizeof(fattr4_unique_handles)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_unique_handles *)cptr);
			cptr += sizeof(fattr4_unique_handles); 
			break;
		case FATTR4_LEASE_TIME:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_LEASE_TIME\n");
			if((cptr + sizeof(fattr4_lease_time)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_lease_time *)cptr);
			cptr += sizeof(fattr4_lease_time); 
			break;
		case FATTR4_RDATTR_ERROR:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_RDATTR_ERROR\n");
			if((cptr + sizeof(fattr4_rdattr_error)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_rdattr_error *)cptr);
			cptr += sizeof(fattr4_rdattr_error); 
			break;
		case FATTR4_ACL:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_ACL\n");
			/* NOTE: fattr4_acl is a placeholdeer until futher notice */
			break;
		case FATTR4_ACLSUPPORT:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_ACLSUPPORT\n");
			if((cptr + sizeof(fattr4_rdattr_error)) > end)
				goto out_overrun;
			*p++ = htonl(*(fattr4_rdattr_error *)cptr);
			cptr += sizeof(fattr4_rdattr_error);
			break;
		case FATTR4_ARCHIVE:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_ARCHIVE\n");
			if((cptr + sizeof(fattr4_archive)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_archive *)cptr);
			cptr += sizeof(fattr4_archive); 
			break;
		case FATTR4_CANSETTIME:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_CANSETTIME\n");
			if((cptr + sizeof(fattr4_cansettime)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_cansettime *)cptr);
			cptr += sizeof(fattr4_cansettime); 
			break;
		case FATTR4_CASE_INSENSITIVE:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_CASE_INSENSITIVE\n");
			if((cptr + sizeof(fattr4_case_insensitive)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_case_insensitive *)cptr);
			cptr += sizeof(fattr4_case_insensitive); 
			break;
		case FATTR4_CASE_PRESERVING:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_CASE_PRESERVING\n");
			if((cptr + sizeof(fattr4_case_preserving)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_case_preserving *)cptr);
			cptr += sizeof(fattr4_case_preserving); 
			break;
		case FATTR4_CHOWN_RESTRICTED:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_CHOWN_RESTRICTED\n");
			if((cptr + sizeof(fattr4_chown_restricted)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_chown_restricted *)cptr);
			cptr += sizeof(fattr4_chown_restricted);
			break;
		case FATTR4_FILEHANDLE:
		{
			int len = (int)*(int *)cptr;
			int size = sizeof(u_int) + len;
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_FILEHANDLE\n");
			if((cptr + size) > end) 
				goto out_overrun;
			if(!(p = xdr_encode_fh4(rqstp,p,(nfs_fh4 *)cptr)))
				return NULL;
			else
				cptr += size; 
		}
		break;
		case FATTR4_FILEID:
			nfsv4_printk(attributes,
				     "NFSDV4 xdr_encode_fattr4: FATTR_FILEID = %ld\n",*(long int *)cptr);
			if((cptr + sizeof(fattr4_fileid)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p, *(fattr4_fileid *)cptr);
			cptr = cptr + sizeof(fattr4_fileid);
			break;
		case FATTR4_FILES_AVAIL:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_FILES_AVAIL\n");
			if((cptr + sizeof(fattr4_files_avail)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_files_avail *)cptr);
			cptr += sizeof(fattr4_files_avail); 
			break;
		case FATTR4_FILES_FREE:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_FILES_FREE\n");
			if((cptr + sizeof(fattr4_files_free)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_files_free *)cptr);
			cptr += sizeof(fattr4_files_free); 
			break;
		case FATTR4_FILES_TOTAL:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_FILES_TOTAL\n");
			if((cptr + sizeof(fattr4_files_total)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_files_total *)cptr);
			cptr += sizeof(fattr4_files_total); 
			break;
		case FATTR4_FS_LOCATIONS:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_FS_LOCATIONS: NOTE NOT YET CODED !!!!!!!!!!\n");
			break;
		case FATTR4_HIDDEN:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_HIDDEN\n");
			if((cptr + sizeof(fattr4_hidden)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_hidden *)cptr);
			cptr += sizeof(fattr4_hidden);; 
			break;
		case FATTR4_HOMOGENEOUS:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_HOMOGENEOUS\n");
			if((cptr + sizeof(fattr4_homogeneous)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_homogeneous *)cptr);
			cptr += sizeof(fattr4_homogeneous); 
			break;
		case FATTR4_MAXFILESIZE:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_MAXFILESIZE\n");
			if((cptr + sizeof(fattr4_maxfilesize)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_maxfilesize *)cptr);
			cptr += sizeof(fattr4_maxfilesize); 
			break;
		case FATTR4_MAXLINK:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_MAXLINK\n");
			if((cptr + sizeof(fattr4_maxlink)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_maxlink *)cptr);
			cptr += sizeof(fattr4_maxlink);
			break;
		case FATTR4_MAXNAME:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_MAXNAME\n");
			if((cptr + sizeof(fattr4_maxname)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_maxname *)cptr);
			cptr += sizeof(fattr4_maxname);
			break;
		case FATTR4_MAXREAD:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_MAXREAD\n");
			if((cptr + sizeof(fattr4_maxread)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_maxread *)cptr);
			cptr += sizeof(fattr4_maxread);
			break;
		case FATTR4_MAXWRITE:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_MAXWRITE\n");
			if((cptr + sizeof(fattr4_maxwrite)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_maxwrite *)cptr);
			cptr += sizeof(fattr4_maxwrite);
			break;
		case FATTR4_MIMETYPE:
		{
			u_int len;
			fattr4_owner *ptr = (fattr4_mimetype *)cptr;

			len = ptr->utf8string_len;
			nfsv4_printk(attributes,"NFSV4 xdr_encode_fattr4 FATTR4_MIMETYPE: len %d \n",len);
			print_utf8string((fattr4_mimetype *)cptr);
			if((cptr + sizeof(fattr4_mimetype) + len ) > end)
				goto out_overrun;
			p = xdr_encode_utf8string(rqstp,p,ptr);
			cptr += sizeof(fattr4_mimetype) +len;
		}
		break;                                       
		case FATTR4_MODE:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_MODE. value %d\n",*(fattr4_mode *)cptr);
			if((cptr + sizeof(fattr4_mode)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_mode *)cptr);
			cptr += sizeof(fattr4_mode);
			break;
		case FATTR4_NO_TRUNC:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_NO_TRUNC\n");
			if((cptr + sizeof(fattr4_no_trunc)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_no_trunc *)cptr);
			cptr += sizeof(fattr4_no_trunc);
			break;
		case FATTR4_NUMLINKS:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_NUMLINKS value %d\n",*(fattr4_numlinks *)cptr);
			if((cptr + sizeof(fattr4_numlinks)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_numlinks *)cptr);
			cptr += sizeof(fattr4_numlinks);
			break;
		case FATTR4_OWNER:
		{
			u_int len;
			fattr4_owner *ptr = (fattr4_owner *)cptr;

			len = ptr->utf8string_len;
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_OWNER: len %d \n",len);
			p = xdr_encode_utf8string(rqstp,p,ptr);
			if((cptr + sizeof(fattr4_owner) + len ) > end) 
				goto out_overrun;
			cptr += sizeof(fattr4_owner) + len;
		}
		break;
		case FATTR4_OWNER_GROUP:
		{
			u_int len;
			fattr4_owner_group *ptr = (fattr4_owner_group *)cptr;

			len = ptr->utf8string_len;
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_OWNER: len %d \n",len);
			p = xdr_encode_utf8string(rqstp,p,ptr);
			if((cptr + sizeof(fattr4_owner_group) + len ) > end) 
				goto out_overrun;
			cptr += sizeof(fattr4_owner_group) + len;
		}
		break;
		case FATTR4_QUOTA_HARD:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_QUOTA_HARD\n");
			if((cptr + sizeof(fattr4_quota_hard)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_quota_hard *)cptr);
			cptr += sizeof(fattr4_quota_hard);
			break;
		case FATTR4_QUOTA_SOFT:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_QUOTA_SOFT\n");
			if((cptr + sizeof(fattr4_quota_soft)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_quota_soft *)cptr);
			cptr += sizeof(fattr4_quota_soft);
			break;
		case FATTR4_QUOTA_USED:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_QUOTA_USED\n");
			if((cptr + sizeof(fattr4_quota_used)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_quota_used *)cptr);
			cptr += sizeof(fattr4_quota_used);
			break;
		case FATTR4_RAWDEV:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_RAWDEV\n");
			if((cptr + sizeof(fattr4_rawdev)) > end) 
				goto out_overrun;
			p = xdr_encode_specdata4(p,(fattr4_rawdev *)cptr);
			cptr += sizeof(fattr4_rawdev);
			break;
		case FATTR4_SPACE_AVAIL:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_SPACE_AVAIL\n");
			if((cptr + sizeof(fattr4_space_avail)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_space_avail *)cptr);
			cptr += sizeof(fattr4_space_avail);
			break;
		case FATTR4_SPACE_FREE:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_SPACE_FREE\n");
			if((cptr + sizeof(fattr4_space_free)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_space_free *)cptr);
			cptr +=sizeof(fattr4_space_free);
			break;
		case FATTR4_SPACE_TOTAL:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_SPACE_TOTAL\n");
			if((cptr + sizeof(fattr4_space_total)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_space_total *)cptr);
			cptr +=sizeof(fattr4_space_total);
			break;
		case FATTR4_SPACE_USED:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_SPACE_USED\n");
			if((cptr + sizeof(fattr4_space_used)) > end) 
				goto out_overrun;
			p = xdr_encode_64(p,*(fattr4_space_used *)cptr);
			cptr +=sizeof(fattr4_space_used);
			break;
		case FATTR4_SYSTEM:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_SYSTEM\n");
			if((cptr + sizeof(fattr4_system)) > end) 
				goto out_overrun;
			*p++ = htonl(*(fattr4_system *)cptr);
			cptr += sizeof(fattr4_system);
			break;
		case FATTR4_TIME_ACCESS:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_TIME_ACCESS\n");
			if((cptr + sizeof(fattr4_time_access)) > end) 
				goto out_overrun;
			p = xdr_encode_nfstime4(p,*(fattr4_time_access *)cptr);
			cptr += sizeof(fattr4_time_access);
			break;
		case FATTR4_TIME_ACCESS_SET:
			break;
		case FATTR4_TIME_BACKUP:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_TIME_BACKUP\n");
			if((cptr + sizeof(fattr4_time_backup)) > end) 
				goto out_overrun;
			p = xdr_encode_nfstime4(p,*(fattr4_time_backup *)cptr);
			cptr += sizeof(fattr4_time_backup);
			break;
		case FATTR4_TIME_CREATE:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_TIME_CREATE\n");
			if((cptr + sizeof(fattr4_time_create)) > end) 
				goto out_overrun;
			p = xdr_encode_nfstime4(p,*(fattr4_time_create *)cptr);
			cptr += sizeof(fattr4_time_create);
			break;
		case FATTR4_TIME_DELTA:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_TIME_DELTA\n");
			if((cptr + sizeof(fattr4_time_delta)) > end) 
				goto out_overrun;
			p = xdr_encode_nfstime4(p,*(fattr4_time_delta *)cptr);
			cptr += sizeof(fattr4_time_delta);
			break;
		case FATTR4_TIME_METADATA:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_TIME_METADATA\n");
			if((cptr + sizeof(fattr4_time_metadata)) > end) 
				goto out_overrun;
			p = xdr_encode_nfstime4(p,*(fattr4_time_metadata *)cptr);
			cptr += sizeof(fattr4_time_metadata);
			break;
		case FATTR4_TIME_MODIFY:
			nfsv4_printk(attributes,
				     "NFSV4 xdr_encode_fattr4 FATTR4_TIME_MODIFY:\n");
			if((cptr + sizeof(fattr4_time_modify)) > end) 
				goto out_overrun;
			p = xdr_encode_nfstime4(p,*(fattr4_time_modify *)cptr);
			cptr += sizeof(fattr4_time_modify);
			break;
		case FATTR4_TIME_MODIFY_SET:
			break;
		default:
			break;
		}
	}while (bit != -1);
	nfsv4_printk(attributes,
		     "NFSDV4:xdr_encode_fattr4: p 0x%x res->FATTRSLEN %d p-start %d\n",
		     (unsigned int)p,res->FATTRSLEN,p-start);
	*save = htonl((p-start)*4);
	return p;

 out_overrun:
	nfsv4_printk(attributes,
		     "NFSDV4: xdr_encode_fattr4: OVERRUN, return NULL\n");
	return NULL;
}
/* NOTE: presently, no overflow checking done in xdr_encode_dirlist4 */
static u32*
xdr_encode_dirlist4(struct svc_rqst *rqstp,u32 *p,dirlist4 *res)
{
	u32 	*ptr = (u32 *)res->entries;
	int i,nentry;
	struct fattr4 attrs;
	u32    bm4val[2] = {0,0};
	utf8string utf8;

	nfsv4_printk(xdr,"NFSDV4 xdr_encode_dirlist4");
	nentry = *(int *)ptr;
	nfsv4_printk(xdr," first nextentry 0x%x\n",*(unsigned int *)ptr);
	while(nentry == 1){
		nfsv4_printk(xdr,"      nextentry 0x%x\n",*(unsigned int *)ptr);
		*p++ = htonl(*(u32 *)ptr);
		nfsv4_printk(xdr,"        \n");
		ptr += 1;

		/* cookie */
		nfsv4_printk(xdr,"        cookie %ld\n",*(long int *)ptr);
		p = xdr_encode_64(p,*(u64 *)ptr);
		ptr += 2;

		utf8.utf8string_len = *(u_int *)ptr;
		nfsv4_printk(xdr,"        utf8.utf8string_len %d\n",*(u_int *)ptr);
		ptr += 1;

		utf8.utf8string_val = kmalloc(utf8.utf8string_len,GFP_KERNEL);
		nfsv4_printk(xdr,"        name: [");
		for(i=0;i<utf8.utf8string_len;i++){
			nfsv4_printk(xdr,"%c",*(char *)ptr);
			utf8.utf8string_val[i] = *(char *)ptr;
			(char *)ptr += sizeof(char);
		}
		nfsv4_printk(xdr,"]\n");
		p = xdr_encode_utf8string(rqstp,p,&utf8);
		kfree(utf8.utf8string_val);

		nfsv4_printk(xdr,"        attrs_val\n");
	
		attrs.attrmask.bitmap4_len = *(u_int *)ptr;
		ptr += 1;

		for(i=0;i<attrs.attrmask.bitmap4_len;i++){
			bm4val[i] = *(u32 *)ptr;
			ptr += 1;
		}
		attrs.attrmask.bitmap4_val = &bm4val[0];

		attrs.FATTRSLEN = *(u_int *)ptr;
		ptr += 1;
		attrs.FATTRSVAL = (char *)ptr;
		print_fattr4(&attrs);

		if(!(p = xdr_encode_fattr4(rqstp,p,&attrs)))
			return NULL;

		(char *)ptr += attrs.FATTRSLEN;

		nentry =  *(int *)ptr;
	}
	/* last entry */
	nfsv4_printk(xdr,"      nextentry 0x%x\n",*(unsigned int *)ptr);
	*p++ = htonl(*(u32 *)ptr);
	nfsv4_printk(xdr,"        \n");
	ptr += 1;
	
	/* dirlist eof */
	*p++ = htonl(res->eof);
	return p;
}
static u32*
xdr_decode_write4data(struct svc_rqst *rqstp,u32 *p,WRITE4args *args)
{
	int len = sizeof(args->WRITELEN);

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	args->WRITELEN = ntohl(*p++);

	nfsv4_printk(vfs,
		     "NFSDV4 xdr_decode_write4data WRITELEN: %d\n",args->WRITELEN);

	if(nfsv4_debug_write) {
		/* debug: look at data in p */
		char prntbuf[32];
		memset(prntbuf,0,32);
		memcpy(prntbuf,(u8 *)p,31);
		nfsv4_printk(vfs,
			     "NFSDV4 xdr_decode_write4data first 31 bytes: %s\n",prntbuf);
	}
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,XDR_QUADLEN(args->WRITELEN))))
		return 0;
	args->WRITEVAL = (char *)p;

#if 0
/*NOTE: get rid of kmalloc, memset, memcpy. use rpc return buffer */
	args->WRITEVAL = kmalloc(args->WRITELEN,GFP_KERNEL);
	memset(args->WRITEVAL,0,args->WRITELEN);
	memcpy((u8 *)args->WRITEVAL,(u8 *)p,args->WRITELEN);
#endif /*0*/

	if(nfsv4_debug_write){
		/* deubg: look at data in args */
		char prntbuf[32];
		memset(prntbuf,0,32);
		memcpy(prntbuf,args->WRITEVAL,(args->WRITELEN <= 31?args->WRITELEN:31));
		nfsv4_printk(vfs,
			     "NFSDV4 xdr_decode_writedata first 31 bytes: %s\n",prntbuf);
	}
	p += XDR_QUADLEN(args->WRITELEN);
	return p;
}

static u32*
xdr_encode_read4resok(struct svc_rqst *rqstp, u32 *p,READ4res *res)
{
	int len = sizeof(res->READEOF) + sizeof(res->READLEN) + XDR_QUADLEN(res->READLEN);

	nfsv4_printk(read,
		     "NFSDV4 xdr_encode_read4resok READEOF: %d READLEN: %d\n",
		     res->READEOF,res->READLEN);

	if(nfsv4_debug_read){
		char prntbuf[32];
		/* deubg: look at data in res */
		memset(prntbuf,0,32);
		memcpy(prntbuf,res->READVAL,31);
		nfsv4_printk(read,
			     "NFSDV4 xdr_encode_read4resok first 31 bytes in READ4res: %s\n",prntbuf);
	}
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
		return 0;
	*p++ = htonl(res->READEOF);
	*p++ = htonl(res->READLEN);
	memcpy((u8 *)p,(u8 *)res->READVAL,res->READLEN);
	p += XDR_QUADLEN(res->READLEN);

	if(nfsv4_debug_read){
		char prntbuf[32];
		/* deubg: look at data in p */
		memset(prntbuf,0,32);
		memcpy(prntbuf,(u8 *)p,31);
		nfsv4_printk(read,
			     "NFSDV4 xdr_encode_read4resok first 31 bytes in p: %s\n",prntbuf);
	}
	return p;
}

static u32*
xdr_decode_nfs_client_id4(struct svc_rqst *rqstp,u32 *p,nfs_client_id4 *id)
{
	int len = NFS4_VERIFIER_SIZE + sizeof(id->id.id_len);

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	p = xdr_decode_verifier4(p,id->verifier);
	nfsv4_printk(setclntid,
		     "NFSV4 xdr_decode_nfs_client_id4 verifier: %p\n",id->verifier);
	print_verifier4(id->verifier);

	id->id.id_len = ntohl(*p++);
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,XDR_QUADLEN(id->id.id_len))))
		return 0;
	id->id.id_val = (char *)p;
	return p + XDR_QUADLEN(id->id.id_len);
}

static u32*
xdr_encode_clientaddr4(struct svc_rqst *rqstp,u32 *p,clientaddr4 *c4)
{
	int len = 0;
	len = strlen(c4->r_netid);
	if(len > 1023)
		return 0;
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(len) + XDR_QUADLEN(len))))
		return 0;
	*p++ = htonl(len);
	memcpy((char *)p,c4->r_netid,len);
	p += XDR_QUADLEN(len);

	len = strlen(c4->r_addr);
	if(len > 1023)
		return 0;
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(len) + XDR_QUADLEN(len))))
		return 0;
	*p++ = htonl(len);
	memcpy((char *)p,c4->r_addr,len);
	p += XDR_QUADLEN(len);

	return p;
}
static u32*
xdr_decode_clientaddr4(struct svc_rqst *rqstp,u32 *p,clientaddr4 *c4)
{
/*length comes first. when  not specified it's max is 1023 (2^32)-1 */
	int cblen,len = sizeof(int);
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	cblen = ntohl(*p++);
	if(cblen > 1023)
		return 0;
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,XDR_QUADLEN(cblen) + len)))
		return 0;
	c4->r_netid = (char *)p;
	p += XDR_QUADLEN(cblen);
	cblen = ntohl(*p++);
	if(cblen > 1023)
		return 0;
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,XDR_QUADLEN(cblen))))
		return 0;
	c4->r_addr = (char *)p;
	p += XDR_QUADLEN(cblen);
	return p;
}

static u32*
xdr_decode_cb_client4(struct svc_rqst *rqstp,u32 *p,cb_client4 *cb)
{
	int len = sizeof(cb->cb_program);

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	cb->cb_program = ntohl(*p++);	
	if(!(p = xdr_decode_clientaddr4(rqstp,p,&cb->cb_location)))
		return 0;
	return p;
}

static u32*
xdr_decode_pathname4(struct svc_rqst *rqstp,u32 *p,pathname4 *p4)
{
	int j,len = sizeof(p4->pathname4_len);

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	p4->pathname4_len = ntohl(*p++);
	p4->pathname4_val = kmalloc(p4->pathname4_len,GFP_KERNEL);
	for(j=0;j<p4->pathname4_len;j++)
		if(!(p = xdr_decode_utf8string(rqstp,p,&p4->pathname4_val[j])))
			return 0;
	return p;
}

static u32 *
xdr_decode_open_claim_delegate_cur4(struct svc_rqst *rqstp,u32 *p,open_claim_delegate_cur4  *cur4)
{
	int len = sizeof(cur4->delegate_stateid);

	if(!(p = xdr_decode_pathname4(rqstp,p,&cur4->file)))
		return 0;	
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	p = xdr_decode_64(p,&cur4->delegate_stateid);
	return p;
}

static u32 *
xdr_encode_nfs_lockowner4(u32 *p,nfs_lockowner4 *lown)
{
	p = xdr_encode_64(p,lown->clientid);
	*p++ = htonl(lown->owner.owner_len);
	memcpy(p,lown->owner.owner_val,lown->owner.owner_len);
	return p + XDR_QUADLEN(lown->owner.owner_len);
}
static u32 *
xdr_decode_nfs_lockowner4(struct svc_rqst *rqstp,u32 *p,nfs_lockowner4 *lown)
{
	int len = sizeof(lown->clientid) + sizeof(lown->owner.owner_len);

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	p = xdr_decode_64(p,&lown->clientid);
	
	lown->owner.owner_len = ntohl(*p++);
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,XDR_QUADLEN(lown->owner.owner_len))))
		return 0;
	lown->owner.owner_val = (char *)p;
	return p + XDR_QUADLEN(lown->owner.owner_len);
}


static u32 *
xdr_decode_createhow4(struct svc_rqst *rqstp,u32 *p,createhow4 *ch4)
{
	int len = sizeof(ch4->mode);

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	ch4->mode = ntohl(*p++);

	switch(ch4->mode){
	case UNCHECKED4:
	case GUARDED4:
		if(!(p = xdr_decode_fattr4(rqstp,p,&ch4->createhow4_u.createattrs)))
			return 0;
		break;
	case EXCLUSIVE4:
		if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,NFS4_VERIFIER_SIZE)))
			return 0;
		p = xdr_decode_verifier4(p,ch4->createhow4_u.createverf);
		break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_decode_createhow4: invlaid mode %d\n",
			     ch4->mode);
		return 0;
	}
	return p;
}

static u32 *
xdr_decode_openflag4(struct svc_rqst *rqstp,u32 *p,openflag4 *of4)
{
	int len = sizeof(of4->opentype);

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	of4->opentype = ntohl(*p++);

	switch(of4->opentype){
	case OPEN4_NOCREATE:
		break;
	case OPEN4_CREATE:
		if(!(p = xdr_decode_createhow4(rqstp,p,&of4->openflag4_u.how)))
			return 0;
		break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_decode_openflag4: invalid type %d\n",of4->opentype);
		return 0;
		break;
	}
	return p;
}                                                              

static inline u32 *
xdr_decode_open_claim4(struct svc_rqst *rqstp,u32 *p, open_claim4 *c4)
{
	int len = sizeof(c4->claim);

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	c4->claim = ntohl(*p++);
	switch(c4->claim){
	case CLAIM_NULL:
		if(!(p = xdr_decode_pathname4(rqstp,p,&c4->open_claim4_u.file)))
			return 0;
		break;
	case CLAIM_PREVIOUS:
		len = sizeof(c4->claim);
		if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
			return 0;
		c4->open_claim4_u.delegate_type = ntohl(*p++);
		break;
	case CLAIM_DELEGATE_CUR:
		if(!(p = xdr_decode_open_claim_delegate_cur4(rqstp,p,&c4->open_claim4_u.delegate_cur_info)))
			return 0;
		break;
	case CLAIM_DELEGATE_PREV:
		if(!(p = xdr_decode_pathname4(rqstp,p,&c4->open_claim4_u.file_delegate_prev)))
			return 0;
		break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_decode_open_claim4: invalid claim %d\n",c4->claim);
		return 0;
		break;
	}
	return p;
}

static u32*
xdr_decode_createtype4(struct svc_rqst *rqstp,u32 *p, createtype4 *objt)
{
	objt->type = ntohl(*p++);
	switch (objt->type){
	case NF4LNK:
		p = xdr_decode_utf8string(rqstp,p, &objt->createtype4_u.linkdata);
		break;
	case NF4BLK:
	case NF4CHR:
		objt->createtype4_u.devdata.specdata1 = ntohl(*p++);
		objt->createtype4_u.devdata.specdata2 = ntohl(*p++);
		break;
	default:
		break;
	}
	return p;
}

static u32*
xdr_encode_change_info4(struct svc_rqst *rqstp,u32 *p, change_info4 *c4)
{
	int len = sizeof(c4->atomic) + sizeof(c4->before) + sizeof(c4->after);

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
		return 0;
	*p++= htonl(c4->atomic);
	p = xdr_encode_64(p, c4->before);
	p = xdr_encode_64(p, c4->after);
	return p;
}

static u32*
xdr_decode_change_info4(u32 *p, change_info4 *c4)
{
	c4->atomic = ntohl(*p++);
	p = xdr_decode_64(p, &c4->before);
	p = xdr_decode_64(p, &c4->after);
	return p;
}

static u32*
xdr_decode_nfs_modified_limit4(u32 *p,nfs_modified_limit4 *m4)
{
	m4->num_blocks = ntohl(*p++);
	m4->bytes_per_block = ntohl(*p++);
	return p;
}

static u32*
xdr_encode_nfs_modified_limit4(u32 *p,nfs_modified_limit4 *m4)
{
	*p++ = htonl(m4->num_blocks);
	*p++ = htonl(m4->bytes_per_block);
	return p;
}

static u32*
xdr_decode_nfs_space_limit4(u32 *p, nfs_space_limit4 *s4)
{
	s4->limitby = ntohl(*p++);
	switch(s4->limitby){
	case NFS_LIMIT_SIZE:
		p = xdr_decode_64(p,&s4->nfs_space_limit4_u.filesize);
		break;
	case NFS_LIMIT_BLOCKS:
		p = xdr_decode_nfs_modified_limit4(p,&s4->nfs_space_limit4_u.mod_blocks);
		break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_decode_nfs_space_limit4: invalid switch %d\n",s4->limitby);
		return 0;
		break;
	}
	return p;
}

static u32*
xdr_encode_nfs_space_limit4(u32 *p, nfs_space_limit4 *s4)
{
	*p++ = htonl(s4->limitby);
	switch(s4->limitby){
	case NFS_LIMIT_SIZE:
		p = xdr_encode_64(p,s4->nfs_space_limit4_u.filesize);
		break;
	case NFS_LIMIT_BLOCKS:
		p = xdr_encode_nfs_modified_limit4(p,&s4->nfs_space_limit4_u.mod_blocks);
		break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_encode_nfs_space_limit4: invalid switch %d\n",s4->limitby);
		return 0;
		break;
	}
	return p;
}

static u32*
xdr_decode_nfsace4(struct svc_rqst *rqstp,u32 *p, nfsace4* acl)
{
	acl->type = ntohl(*p++);
	acl->flag = ntohl(*p++);
	acl->access_mask = ntohl(*p++);
	p = xdr_decode_utf8string(rqstp,p,&acl->who);
	return p;
}

static u32*
xdr_encode_nfsace4(struct svc_rqst *rqstp,u32 *p, nfsace4* acl)
{
	*p++ = htonl(acl->type);
	*p++ = htonl(acl->flag);
	*p++ = htonl(acl->access_mask);
	p = xdr_encode_utf8string(rqstp,p, &acl->who);
	return p;
}

static u32*
xdr_decode_open_write_delegation4(struct svc_rqst *rqstp,u32 *p, open_write_delegation4 *w4)
{
	p = xdr_decode_64(p, &w4->stateid);
	w4->recall = ntohl(*p++);
	if(!(p  = xdr_decode_nfs_space_limit4(p, &w4->space_limit)))
		return 0;
	p = xdr_decode_nfsace4(rqstp,p, &w4->permissions);
	return p;
}

static u32*
xdr_encode_open_write_delegation4(struct svc_rqst *rqstp,u32 *p, open_write_delegation4 *w4)
{
	p = xdr_encode_64(p, w4->stateid);
	*p++ = htonl(w4->recall);
	if(!(p  = xdr_encode_nfs_space_limit4(p, &w4->space_limit)))
		return 0;
	p = xdr_encode_nfsace4(rqstp,p, &w4->permissions);
	return p;
}

static u32*
xdr_decode_open_read_delegation4(struct svc_rqst *rqstp, u32 *p, open_read_delegation4 *r4)
{
	p = xdr_decode_64(p, &r4->stateid);
	r4->recall = ntohl(*p++);
	p = xdr_decode_nfsace4(rqstp,p, &r4->permissions);
	return p;
}

static u32*
xdr_encode_open_read_delegation4(struct svc_rqst *rqstp,u32 *p, open_read_delegation4 *r4)
{
	p = xdr_encode_64(p, r4->stateid);
	*p++ = htonl(r4->recall);
	p = xdr_encode_nfsace4(rqstp,p, &r4->permissions);
	return p;
}

static u32*
xdr_decode_open_delegation4(struct svc_rqst *rqstp,u32 *p, open_delegation4 *d4)
{
	d4->delegation_type = ntohl(*p++); 
	switch(d4->delegation_type){
	case OPEN_DELEGATE_NONE:
		break;
	case OPEN_DELEGATE_READ:
		p = xdr_decode_open_read_delegation4(rqstp,p,&d4->open_delegation4_u.read);
		break;
	case OPEN_DELEGATE_WRITE:
		if(!(p = xdr_decode_open_write_delegation4(rqstp,p,&d4->open_delegation4_u.write)))
			return 0;
		break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_decode_open_delegation4: invalid type %d\n",d4->delegation_type);
		return 0;
		break;
	} 
	return p;
}

static u32*
xdr_encode_open_delegation4(struct svc_rqst *rqstp,u32 *p, open_delegation4 *d4)
{
	*p++ = htonl(d4->delegation_type); 
	switch(d4->delegation_type){
	case OPEN_DELEGATE_NONE:
		break;
	case OPEN_DELEGATE_READ:
		p = xdr_encode_open_read_delegation4(rqstp,p,&d4->open_delegation4_u.read);
		break;
	case OPEN_DELEGATE_WRITE:
		if(!(p = xdr_encode_open_write_delegation4(rqstp,p,&d4->open_delegation4_u.write)))
			return 0;
		break;
	default:
		nfsv4_printk(xdr,"NFSV4 xdr_decode_open_delegation4: invalid type %d\n",d4->delegation_type);
		return 0;
		break;
	} 
	return p;
}

static u32*
xdr_encode_sec_oid4(struct svc_rqst *rqstp,u32 *p, sec_oid4 *s4)
{
	int len = sizeof(s4->sec_oid4_len) + XDR_QUADLEN(s4->sec_oid4_len);

	nfsv4_printk(secinfo,"NFSDV4 xdr_encode_sec_oid4. sec_oid4_len %d\n",
		     s4->sec_oid4_len);
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
		return 0;
	*p++ = htonl(s4->sec_oid4_len);	
	memcpy(p,s4->sec_oid4_val,s4->sec_oid4_len);
	return p + XDR_QUADLEN(s4->sec_oid4_len);
}

static u32*
xdr_encode_secinfo4resok(struct svc_rqst *rqstp,u32 *p, secinfo4 *info)
{
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(info->flavor))))
		return 0;
	*p++ = htonl(info->flavor);
	nfsv4_printk(secinfo,"NFSDv4 xdr_encode_secinfo4resok flavor: %d\n",
		     info->flavor);
	switch(info->flavor){
	case RPC_AUTH_NULL:
	case RPC_AUTH_UNIX:
		break;
	case RPCSEC_GSS:
	{
		/* note: flavor_info_len = xdr size of the opaque flavor_info_val data */
		rpcsec_gss_info *rpcinfo = (rpcsec_gss_info *)info->flavor_info.flavor_info_val;
		int len = sizeof(info->flavor_info.flavor_info_len) + info->flavor_info.flavor_info_len;

		nfsv4_printk(secinfo,
			     "NFSDv4 xdr_encode_secinfo4resok flavor_info_len: %d\n",
			     info->flavor_info.flavor_info_len);

		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
			return 0;
		*p++ = htonl(info->flavor_info.flavor_info_len);
		if(!(p = xdr_encode_sec_oid4(rqstp,p,&rpcinfo->oid)))
			return 0;
		*p++ = htonl(rpcinfo->qop);
		*p++ = htonl(rpcinfo->service);
	}
	break;
	}                                                                
	return p;
}

/*
 * Check buffer bounds after decoding arguments
 */
static inline int
xdr_argsize_check(struct svc_rqst *rqstp, u32 *p)
{
	struct svc_buf	*buf = &rqstp->rq_argbuf;

	return p - buf->base <= buf->buflen;
}

static inline int
xdr_ressize_check(struct svc_rqst *rqstp, u32 *p)
{
	struct svc_buf	*buf = &rqstp->rq_resbuf;

	buf->len = p - buf->base;
	nfsv4_printk(xdr,"nfsd: ressize_check p %p base %p len %d\n",
		     p, buf->base, buf->buflen);
	return (buf->len <= buf->buflen);
}

/*
* XDR decode functions
*/

u32 *
xdr_decode_access4(struct svc_rqst *rqstp, u32 * p, ACCESS4args *argp) 
{
	int len = sizeof(argp->access);
	nfsv4_printk(compound, "NFSDV4 decode compound ACCESS\n");

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	argp->access = ntohl(*p++);
	nfsv4_printk(compound,"      access: 0x%x \n", argp->access);
	return p;
}

u32 *
xdr_decode_close4(struct svc_rqst *rqstp, u32 * p, CLOSE4args *argp) 
{
	int len =  sizeof(argp->seqid) + sizeof(argp->stateid);
	nfsv4_printk(compound, "NFSDV4 decode compound CLOSE\n");
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	argp->seqid = ntohl(*p++);
	p = xdr_decode_64(p,&argp->stateid);
	nfsv4_printk(compound,"NFSV4: decode CLOSE stateid: [0x%lx , %08lx]\n",
		     (unsigned long)(argp->stateid >> 32),
		     (unsigned long)(argp->stateid & 0xFFFFFFFFUL));
	return p;
}

static u32 *
xdr_decode_delegreturn4(struct svc_rqst *rqstp, u32 *p, DELEGRETURN4args *argp)
{
	nfsv4_printk(compound, "NFSDV4 decode compound DELEGRETURN\n");
	if (!xdr_overflow_check(&rqstp->rq_argbuf, p, 8))
		return 0;
	p = xdr_decode_64(p, &argp->stateid);
	return p;
}

u32 *
xdr_decode_create4(struct svc_rqst *rqstp, u32 * p, CREATE4args *argp) 
{
	nfsv4_printk(compound, "NFSDV4 decode compound CREATE\n");
	if(!(p = xdr_decode_utf8string(rqstp,p, &argp->objname)))
		return 0;
	print_utf8string(&argp->objname);
	if(!(p = xdr_decode_createtype4(rqstp,p, &argp->objtype)))
		return 0;
	return p;
}
u32 *
xdr_decode_getattr4(struct svc_rqst *rqstp, u32 * p, GETATTR4args *argp) 
{
	nfsv4_printk(compound, "NFSDV4 decode compound GETATTR4\n");
	if(!(p = xdr_decode_bitmap4(rqstp,p,&argp->attr_request)))
		return 0;
	return p;
}
u32 *
xdr_decode_link4(struct svc_rqst *rqstp, u32 * p, LINK4args *argp) 
{
	nfsv4_printk(compound, "NFSDV4 decode compound LINK\n");
	if(!(p = xdr_decode_utf8string(rqstp,p,&argp->newname)))
		return 0;
	print_utf8string(&argp->newname);
	return p;
}
u32 *
xdr_decode_lookup4(struct svc_rqst *rqstp, u32 * p, LOOKUP4args *argp) 
{
	int j, len = sizeof(argp->PLEN);
	nfsv4_printk(compound, "NFSDV4 decode compound LOOKUP\n");

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;

	argp->PLEN = ntohl(*p++);
	argp->PVAL = kmalloc(argp->PLEN * sizeof(component4), GFP_KERNEL);
	if(!argp->PVAL)
		return 0;
	memset(argp->PVAL,0,argp->PLEN * sizeof(component4));
	for(j=0;j<argp->PLEN;j++)
	{
		if(!(p = xdr_decode_utf8string(rqstp,p,&argp->PVAL[j])))
			return 0;
	}
	return p;
}
u32 *
xdr_decode_open4(struct svc_rqst *rqstp, u32 * p, OPEN4args *argp) 
{
	int len = sizeof(argp->seqid) + sizeof(argp->share_access) + sizeof(argp->share_deny);

	if(!(p = xdr_decode_open_claim4(rqstp,p, &argp->claim)))
		return 0;
	if(!(p = xdr_decode_openflag4(rqstp,p,&argp->openhow)))
		return 0;
	if(!(p = xdr_decode_nfs_lockowner4(rqstp,p,&argp->owner)))  
		return 0;

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	argp->seqid = ntohl(*p++);
	argp->share_access = ntohl(*p++);
	argp->share_deny = ntohl(*p++);

	nfsv4_printk(open, 
		     "got FOR OPEN seqid %d access %d deny %d\n", 
		     argp->seqid, argp->share_access, argp->share_deny);
	return p;
}
u32 *
xdr_decode_open_confirm4(struct svc_rqst *rqstp,u32 * p,OPEN_CONFIRM4args *argp)
{
	int len = sizeof(argp->seqid) + NFS4_VERIFIER_SIZE;

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	argp->seqid = ntohl(*p++);
	p = xdr_decode_verifier4(p,argp->open_confirm);
	nfsv4_printk(open, "OPEN_CONFIRM seqid %d verifier\n", argp->seqid);
	print_verifier4(argp->open_confirm);
	return p;
}
u32 *
xdr_decode_putfh4(struct svc_rqst *rqstp,u32 * p,PUTFH4args *argp)
{
	nfsv4_printk(compound, "NFSDV4 decode compound PUTFH\n");
	if(!(p = xdr_decode_fh4(rqstp,p,&argp->object)))
		return 0;	
	print_nfs_fh4(&argp->object);
	return p;
}
u32 *
xdr_decode_read4(struct svc_rqst *rqstp,u32 * p,READ4args *argp)
{
	int len = sizeof(argp->stateid) + sizeof(argp->offset) + sizeof(argp->count);

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	nfsv4_printk(compound, "NFSDV4 decode compound READ\n");
	p = xdr_decode_64(p,&argp->stateid);
	p = xdr_decode_64(p,&argp->offset);
	argp->count = ntohl(*p++);
	return p;
}
u32 *
xdr_decode_readdir4(struct svc_rqst *rqstp,u32 * p,READDIR4args *argp)
{
	int len = sizeof(argp->cookie) + NFS4_VERIFIER_SIZE + sizeof(argp->dircount) + sizeof(argp->maxcount);

	nfsv4_printk(compound, "NFSDV4 decode compound READDIR\n");
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	p = xdr_decode_64(p,&argp->cookie);	
	p = xdr_decode_verifier4(p,argp->cookieverf);	
	argp->dircount = ntohl(*p++);
	argp->maxcount = ntohl(*p++);
			
	if(!(p = xdr_decode_bitmap4(rqstp,p,&argp->attr_request)))	
		return 0;
	nfsv4_printk(compound, "NFSDV4 decode, READDIR cookie %ld maxcount %d\n",(long int)argp->cookie,argp->maxcount);
	return p;
}
u32 *
xdr_decode_remove4(struct svc_rqst *rqstp,u32 * p,REMOVE4args *argp)
{
	nfsv4_printk(compound, "NFSDv4 decode compound REMOVE\n");
	if(!(p = xdr_decode_utf8string(rqstp,p,&argp->target)))
		return 0;
	print_utf8string(&argp->target);
	return p;
}
u32 *
xdr_decode_rename4(struct svc_rqst *rqstp,u32 * p,RENAME4args *argp)
{
	nfsv4_printk(rename, 
		     "NFSDv4 decode compound RENAME: old length %d\n", ntohl(*p));

/*ANDROS: hey, shouldn't have to kmalloc: xdr_decode_utf8string 
 * uses the rpc buffer.. check this.

 argp->OLDNAME_VAL = kmalloc(ntohl(*p)+1, GFP_KERNEL);
 memset(argp->OLDNAME_VAL, 0, ntohl(*p)+1);
*/

	if(!(p = xdr_decode_utf8string(rqstp,p,&argp->oldname)))
		return 0;
	if (nfsv4_debug_rename)
		print_utf8string(&argp->oldname);
	nfsv4_printk(rename, 
		     "NFSDv4 decode compound RENAME: new length %d\n", ntohl(*p));

/*ANDROS: hey, shouldn't have to kmalloc: xdr_decode_utf8string 
 * uses the rpc buffer.. check this.

 argp->NEWNAME_VAL = kmalloc(ntohl(*p)+1, GFP_KERNEL);
 memset(argp->NEWNAME_VAL, 0, ntohl(*p)+1);
*/
	if(!(p = xdr_decode_utf8string(rqstp,p,&argp->newname)))
		return 0;
	if (nfsv4_debug_rename)
		print_utf8string(&argp->newname);
	nfsv4_printk(rename, "NFSDv4 decode RENAME done\n");
	return p;
}
u32 *
xdr_decode_renew4(struct svc_rqst *rqstp,u32 * p,RENEW4args *argp)
{
	int len = sizeof(argp->stateid);

	nfsv4_printk(compound, "NFSDV4 decode compound RENEW\n");
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	p = xdr_decode_64(p,&argp->stateid);
	return p;
}
u32 *
xdr_decode_secinfo4(struct svc_rqst *rqstp,u32 * p,SECINFO4args *argp)
{
	nfsv4_printk(compound, "NFSDV4 decode compound SECINFO\n");
	if(!(p = xdr_decode_utf8string(rqstp,p,&argp->name)))
		return 0;
	return p;
}
u32 *
xdr_decode_setclientid4(struct svc_rqst *rqstp,u32 * p,SETCLIENTID4args *argp)
{
	nfsv4_printk(setclntid, "NFSDv4 decode SETCLIENTID args\n");
	if(!(p = xdr_decode_nfs_client_id4(rqstp,p,&argp->client)))
		return 0;
	if(!(p = xdr_decode_cb_client4(rqstp,p,&argp->callback)))
		return 0;
	return p;
}
u32 *
xdr_decode_setclientid_confirm4(struct svc_rqst *rqstp,u32 * p,SETCLIENTID_CONFIRM4args *argp)
{
	nfsv4_printk(setclntid, "NFSDv4 decode SETCLIENTID_CONFIRM args\n");
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,NFS4_VERIFIER_SIZE)))
		return 0;
	p = xdr_decode_verifier4(p,argp->setclientid_confirm);
	return p;
}
u32 *
xdr_decode_setattr4(struct svc_rqst *rqstp,u32 * p,SETATTR4args *argp)
{
	int len = sizeof(argp->stateid);
	nfsv4_printk(compound, "NFSDV4 decode compound SETATTR\n");
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	p = xdr_decode_64(p, &argp->stateid);
	if(!(p = xdr_decode_fattr4(rqstp,p, &argp->obj_attributes)))
		return 0;
	return p;
}
u32 *
xdr_decode_write4(struct svc_rqst *rqstp,u32 * p,WRITE4args *argp)
{
	int len = sizeof(argp->stateid) + sizeof(argp->offset) + sizeof(argp->stable);

	nfsv4_printk(compound,"NFSDV4 decode compound WRITE\n");
	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	p = xdr_decode_64(p,&argp->stateid);	
	p = xdr_decode_64(p,&argp->offset);	
	argp->stable = ntohl(*p++);		
	if(!(p = xdr_decode_write4data(rqstp,p,argp)))
		return 0;
	return p;
}

/* this is a boolean function */
int
nfs4svc_decode_voidargs(struct svc_rqst *rqstp, u32 *p, void *dummy)
{
	nfsv4_printk(xdr,"NFSDV4 decode voidargs\n");
	return xdr_argsize_check(rqstp, p);
}

/* this is a boolean function */
int
nfs4svc_decode_COMPOUND4args(struct svc_rqst *rqstp, u32 *p, struct COMPOUND4args *cargs)
{
	int i,len = sizeof(cargs->minorversion) + sizeof(cargs->OPCNT);
	int oplen = sizeof(nfs_opnum4);

	nfsv4_printk(compound,"NFSDV4 decode compoundargs ");
	nfsv4_printk(compound,
		     "rqstp->rq_resbuf.buflen: %d\n",rqstp->rq_resbuf.buflen);

	if(!(p = xdr_decode_utf8string(rqstp,p, &cargs->tag)))
		return 0;
	print_utf8string(&cargs->tag);

	if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,len)))
		return 0;
	cargs->minorversion = ntohl(*p++);
	cargs->OPCNT = ntohl(*p++);
	nfsv4_printk(compound,
		     "NFSDV4: decode compoundarsg: minorversion %d opcnt: %d\n",
		     cargs->minorversion,cargs->OPCNT);

	cargs->OPVAL = kmalloc(cargs->OPCNT * sizeof(struct nfs_argop4),GFP_KERNEL);
	if(!cargs->OPVAL){
		nfsv4_printk(compound,"NFSDV4: decode compound kmalloc failed\n");
		return 0;
	}
	memset(cargs->OPVAL,0,cargs->OPCNT * sizeof(struct nfs_argop4));

	for(i=0;i<cargs->OPCNT;i++){
		if(!(xdr_overflow_check(&rqstp->rq_argbuf,p,oplen)))
			return 0;
		cargs->OPVAL[i].argop = ntohl(*p++);
		nfsv4_printk(compound,"cargs->OPVAL[i].argop %d\n",cargs->OPVAL[i].argop);

		switch (cargs->OPVAL[i].argop){
		case OP_ACCESS:
			if(!(p = xdr_decode_access4(rqstp,p,&cargs->OPVAL[i].OPU.opaccess)))
				return 0;
			break;
		case OP_CLOSE:
			if(!(p = xdr_decode_close4(rqstp,p,&cargs->OPVAL[i].OPU.opclose)))
				return 0;
			break;
		case OP_CREATE:
			if(!(p = xdr_decode_create4(rqstp,p,&cargs->OPVAL[i].OPU.opcreate)))
				return 0;
			break;
		case OP_DELEGRETURN:
			if (!(p = xdr_decode_delegreturn4(rqstp, p, &cargs->OPVAL[i].OPU.opdelegreturn)))
				return 0;
			break;
		case OP_GETATTR:
			if(!(p = xdr_decode_getattr4(rqstp,p,&cargs->OPVAL[i].OPU.opgetattr)))
				return 0;
			break;
		case OP_GETFH: 
			/* nothing to do */
			break;
		case OP_LINK:
			if(!(p = xdr_decode_link4(rqstp,p,&cargs->OPVAL[i].OPU.oplink)))
				return 0;
			break;
		case OP_LOOKUP:
			if(!(p = xdr_decode_lookup4(rqstp,p,&cargs->OPVAL[i].OPU.oplookup)))
				return 0;
			break;
		case OP_LOOKUPP:
			/* nothing to do */
			break;
		case OP_OPEN:
			if(!(p = xdr_decode_open4(rqstp,p,&cargs->OPVAL[i].OPU.opopen)))
				return 0;
			break;
		case OP_OPEN_CONFIRM:
			if(!(p = xdr_decode_open_confirm4(rqstp,p,&cargs->OPVAL[i].OPU.opopen_confirm)))
				return 0;
			break;
		case OP_PUTFH:
			if(!(p = xdr_decode_putfh4(rqstp,p,&cargs->OPVAL[i].OPU.opputfh)))
				return 0;
			break;
		case OP_PUTROOTFH: 
			break;
		case OP_READ:
			if(!(p = xdr_decode_read4(rqstp,p,&cargs->OPVAL[i].OPU.opread)))
				return 0;
			break;
		case OP_READDIR:
			if(!(p = xdr_decode_readdir4(rqstp,p,&cargs->OPVAL[i].OPU.opreaddir)))
				return 0;
			break;
		case OP_READLINK:
			// nothing to do
				break;
		case OP_REMOVE:
			if(!(p = xdr_decode_remove4(rqstp,p,&cargs->OPVAL[i].OPU.opremove)))
				return 0;
			break;
		case OP_RENAME:
			if(!(p = xdr_decode_rename4(rqstp,p,&cargs->OPVAL[i].OPU.oprename)))
				return 0;
			break;
		case OP_RENEW:
			if(!(p = xdr_decode_renew4(rqstp,p,&cargs->OPVAL[i].OPU.oprenew)))
				return 0;
			break;
			break;
		case OP_RESTOREFH:
			nfsv4_printk (rename, "NFSDv4 decode RESTOREFH args, nothing to do\n");
			break;
		case OP_SECINFO:
			if(!(p = xdr_decode_secinfo4(rqstp,p,&cargs->OPVAL[i].OPU.opsecinfo)))
				return 0;
			break;
		case OP_SAVEFH:
			nfsv4_printk(rename, "NFSDv4 decode SAVEFH args, nothing to do\n");
			break;
		case OP_SETCLIENTID:
			if(!(p = xdr_decode_setclientid4(rqstp,p,&cargs->OPVAL[i].OPU.opsetclientid)))
				return 0;
			break;
		case OP_SETCLIENTID_CONFIRM:
			if(!(p = xdr_decode_setclientid_confirm4(rqstp,p,&cargs->OPVAL[i].OPU.opsetclientid_confirm)))
				return 0;
			break;
		case OP_SETATTR:
			if(!(p = xdr_decode_setattr4(rqstp,p,&cargs->OPVAL[i].OPU.opsetattr)))
				return 0;
			break;
		case OP_WRITE:
		{
			WRITE4args *args = &cargs->OPVAL[i].OPU.opwrite;

			nfsv4_printk(compound,"NFSDV4 decode compound WRITE\n");
			p = xdr_decode_64(p,&args->stateid);	
			p = xdr_decode_64(p,&args->offset);	
			args->stable = ntohl(*p++);		
			p = xdr_decode_write4data(rqstp,p,args);
		}
		break;
		default:
			if((cargs->OPVAL[i].argop >= 0) && (cargs->OPVAL[i].argop < 35)){
				nfsv4_printk(compound,"NFSV4: unimplemented opcode %d\n",
					     cargs->OPVAL[i].argop);
				return xdr_argsize_check(rqstp, p);
			}else {
				nfsv4_printk(compound,"NFSV4: bad opcode %d\n",
					     cargs->OPVAL[i].argop);
				return 0;
			}
			break;
		}
	}
	nfsv4_printk(compound,"NFSDV4 decode returning\n");
	return xdr_argsize_check(rqstp, p);
}

/*
 * XDR encode functions
 */

u32 *
xdr_encode_access4(struct svc_rqst *rqstp, u32 * p, ACCESS4res *resp) 
{
	int len = sizeof(resp->SUPPORTED) + sizeof(resp->ACCESS);

	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres ACCESS status: %d\n", ntohl(resp->status));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed */
	if(resp->status == NFS4_OK){
		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
			return 0;
		*p++ = htonl(resp->SUPPORTED);
		*p++ = htonl(resp->ACCESS);
	}
	return p;
}

u32 *
xdr_encode_close4(struct svc_rqst *rqstp, u32 * p, CLOSE4res *resp) 
{
	int len = sizeof(resp->CLOSE4res_u.stateid);

	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres CLOSE status: %d\n",ntohl(resp->status));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed */
	if(resp->status == NFS4_OK){
		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
			return 0;
		p = xdr_encode_64(p,resp->CLOSE4res_u.stateid);
	}
	return p;
}
u32 *
xdr_encode_create4(struct svc_rqst *rqstp, u32 * p, CREATE4res *resp) 
{
	nfsv4_printk(compound, 
		     "NFSDv4 encode_compoundres CREATE status: %d\n", ntohl(resp->status));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status;	/* already htonl'ed */
	if (resp->status == NFS4_OK){
		if(!(p = xdr_encode_change_info4(rqstp,p, &resp->CREATEinfo)))
			return 0;
	}
	return p;
}

static u32 *
xdr_encode_delegreturn4(struct svc_rqst *rqstp, u32 *p, DELEGRETURN4res *resp)
{
	nfsv4_printk(compound, "NFSDv4 encode_compoundres DELEGRETURN status %d\n",
		     ntohl(resp->status));

	if (!xdr_overflow_check(&rqstp->rq_resbuf, p, 4))
		return 0;
	*p++ = resp->status;    /* already htonl'ed */
	return p;
}

u32 *
xdr_encode_getattr4(struct svc_rqst *rqstp, u32 * p, GETATTR4res *resp) 
{
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres GETATTR status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed */
	if(resp->status == NFS4_OK){
		if(!(p = xdr_encode_fattr4(rqstp,p,&resp->GFATTR4)))
			return 0; 
	}
	return p;
}
u32 *
xdr_encode_getfh4(struct svc_rqst *rqstp, u32 * p, GETFH4res *resp) 
{
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres GETFH status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed */
	if(resp->status == NFS4_OK){
		if(!(p = xdr_encode_fh4(rqstp,p,&resp->GETFH4FH)))
			return 0; 
	}
	return p;
}
u32 *
xdr_encode_link4(struct svc_rqst *rqstp, u32 * p, LINK4res *resp) 
{
	int len = sizeof(resp->LINK_RES_INFO.atomic) + sizeof(resp->LINK_RES_INFO.before) + sizeof(resp->LINK_RES_INFO.after);

	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres LINK status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = htonl(resp->status);
	if(resp->status == NFS4_OK){
		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
			return 0;
		*p++ = htonl(resp->LINK_RES_INFO.atomic);
		p = xdr_encode_64(p, resp->LINK_RES_INFO.before);
		p = xdr_encode_64(p, resp->LINK_RES_INFO.after);
	}
	return p;
}
u32 *
xdr_encode_lookup4(struct svc_rqst *rqstp, u32 * p, LOOKUP4res *resp) 
{
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres LOOKUP status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status;
	return p;
}
u32 *
xdr_encode_lookupp4(struct svc_rqst *rqstp, u32 * p, LOOKUPP4res *resp) 
{
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres LOOKUPP status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status;
	return p;
}
u32 *
xdr_encode_open4(struct svc_rqst *rqstp, u32 * p, OPEN4res *resp) 
{
	int len = sizeof(resp->OPEN_RES.rflags) + NFS4_VERIFIER_SIZE;

	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres OPEN status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; 
	if(resp->status == NFS4_OK){
		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->OPEN_RES.stateid))))
			return 0;
		p = xdr_encode_64(p,resp->OPEN_RES.stateid);

		if(!(p = xdr_encode_change_info4(rqstp,p, &resp->OPEN_RES.cinfo)))
			return 0;

		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
			return 0;
		*p++ = htonl(resp->OPEN_RES.rflags);
		p = xdr_encode_verifier4(p, resp->OPEN_RES.open_confirm);

		if(!(p = xdr_encode_open_delegation4(rqstp,p,&resp->OPEN_RES.delegation)))
			return 0;
	}
	return p;
}
u32 *
xdr_encode_open_confirm4(struct svc_rqst *rqstp,u32 * p,OPEN_CONFIRM4res *resp) 
{
	int len = sizeof(resp->OPEN_CONFIRM4res_u.resok4.stateid);

	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres OPEN_CONFIRM status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed */
	if(resp->status == NFS4_OK){
		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
			return 0;
		p = xdr_encode_64(p,resp->OPEN_CONFIRM4res_u.resok4.stateid);
	}
	return p;
}
u32 *
xdr_encode_putfh4(struct svc_rqst *rqstp,u32 * p, PUTFH4res *resp) 
{
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres PUTFH status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed */
	return p;
}
u32 *
xdr_encode_putrootfh4(struct svc_rqst *rqstp,u32 * p, PUTROOTFH4res *resp) 
{
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres PUTROOTFH status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed */
	return p;
}
u32 *
xdr_encode_read4(struct svc_rqst *rqstp,u32 * p, READ4res *resp) 
{
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres READ status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; 
	if(resp->status == NFS4_OK){
		if(!(p = xdr_encode_read4resok(rqstp,p,resp)))
			return 0;
	}
	return p;
}
u32 *
xdr_encode_readdir4(struct svc_rqst *rqstp,u32 * p, READDIR4res *resp) 
{
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres READDIR status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed */
	if(resp->status == NFS4_OK){
		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,NFS4_VERIFIER_SIZE)))
			return 0;
		p = xdr_encode_verifier4(p, resp->READDIR4res_u.resok4.cookieverf);
		/* NOTE: have not done overflow checking in xdr_encode_dirlist4 */
		if(!(p = xdr_encode_dirlist4(rqstp,p,&resp->DIR4)))
			return 0; 
	}
	return p;
}
u32 *
xdr_encode_readlink4(struct svc_rqst *rqstp,u32 * p, READLINK4res *resp) 
{
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres READLINK status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status;   
	if (resp->status == NFS4_OK){
		if(!(p = xdr_encode_utf8string(rqstp,p,&resp->READLINK4res_u.resok4.link)))
			return 0;
	}
	return p;
}
u32 *
xdr_encode_remove4(struct svc_rqst *rqstp,u32 * p, REMOVE4res *resp) 
{
	int len = sizeof(resp->REMOVE4res_u.resok4.cinfo.atomic) + sizeof(resp->REMOVE4res_u.resok4.cinfo.before) + sizeof(resp->REMOVE4res_u.resok4.cinfo.after);

	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres REMOVE status: %d\n",ntohl(resp->status));
		
	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; 
	if (resp->status == NFS4_OK){
		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
			return 0;
		*p++ = htonl(resp->REMOVE4res_u.resok4.cinfo.atomic);
		p = xdr_encode_64(p, resp->REMOVE4res_u.resok4.cinfo.before);
		p = xdr_encode_64(p, resp->REMOVE4res_u.resok4.cinfo.after);
	}
	return p;
}

u32 *
xdr_encode_rename4(struct svc_rqst *rqstp,u32 * p, RENAME4res *resp) 
{
	int len = sizeof(resp->RENAME_RES_SRC.atomic) + sizeof(resp->RENAME_RES_SRC.before) + sizeof(resp->RENAME_RES_SRC.after) + sizeof(resp->RENAME_RES_TAR.atomic) + sizeof(resp->RENAME_RES_TAR.before) + sizeof(resp->RENAME_RES_TAR.after);

	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres RENAME status: %d\n",ntohl(resp->status));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed*/
	if (resp->status == NFS4_OK){
		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
			return 0;
		*p++ = htonl(resp->RENAME_RES_SRC.atomic);
		p = xdr_encode_64(p, resp->RENAME_RES_SRC.before);
		p = xdr_encode_64(p, resp->RENAME_RES_SRC.after);
		*p++ = htonl(resp->RENAME_RES_TAR.atomic);
		p = xdr_encode_64(p, resp->RENAME_RES_TAR.before);
		p = xdr_encode_64(p, resp->RENAME_RES_TAR.after);
	}
	return p;
}
u32 *
xdr_encode_renew4(struct svc_rqst *rqstp,u32 * p, RENEW4res *resp) 
{

	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres RENEW status: %d\n",ntohl(resp->status));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed*/
	return p;
}

u32 *
xdr_encode_restorefh4(struct svc_rqst *rqstp,u32 * p, RESTOREFH4res *resp) 
{
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres RESTOREFH status: %d\n",ntohl(resp->status));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed*/
	return p;
}
u32 *
xdr_encode_secinfo4(struct svc_rqst *rqstp,u32 * p, SECINFO4res *resp) 
{
	int i;

	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres SECINFO status: %d\n",ntohl(resp->status));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed */
	if (resp->status == NFS4_OK){
		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->SECRES.SECINFO4resok_len))))
			return 0;
		*p++ = htonl(resp->SECRES.SECINFO4resok_len);
		for(i=0;i<resp->SECRES.SECINFO4resok_len;i++){
			if(!(p=xdr_encode_secinfo4resok(rqstp,p,&resp->SECRES.SECINFO4resok_val[i])))
				return 0;
		}
	}
	return p;
}
u32 *
xdr_encode_savefh4(struct svc_rqst *rqstp,u32 * p, SAVEFH4res *resp) 
{

	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres SAVEFH status: %d\n",ntohl(resp->status));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status;
	return p;
}

u32 *
xdr_encode_setclientid4(struct svc_rqst *rqstp,u32 * p, SETCLIENTID4res *resp) 
{
	int len = sizeof(resp->CLIDOK.clientid) + NFS4_VERIFIER_SIZE;

	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres SETCLIENTID status: %d\n",ntohl(resp->status));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status;
	switch(resp->status){
	case NFS4_OK:
		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
			return 0;
		p = xdr_encode_64(p,resp->CLIDOK.clientid);
		p = xdr_encode_verifier4(p,resp->CLIDOK.setclientid_confirm);
		break;
	case NFS4ERR_CLID_INUSE:
		if(!(p = xdr_encode_clientaddr4(rqstp,p,&resp->SETCLIENTID4res_u.client_using)))
			return 0;
		break;
	default:
		break;
	}

	return p;
}
u32 *
xdr_encode_setclientid_confirm4(struct svc_rqst *rqstp,u32 * p, SETCLIENTID_CONFIRM4res *resp) 
{
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres SETCLIENTID_CONFIRM status: %d\n",ntohl(resp->status));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status;
	return p;
}
u32 *
xdr_encode_setattr4(struct svc_rqst *rqstp,u32 * p, SETATTR4res *resp) 
{
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres SETATTR status: %d\n",ntohl(resp->status));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status; /* already htonl'ed */
	if (resp->status == NFS4_OK){
		if(!(p = xdr_encode_bitmap4(rqstp,p, &resp->attrsset)))
			return 0;
	}
	return p;
}
u32 *
xdr_encode_write4(struct svc_rqst *rqstp,u32 * p, WRITE4res *resp) 
{
	int len = sizeof(resp->WRITERES.count) + sizeof(resp->WRITERES.committed) + NFS4_VERIFIER_SIZE;
	nfsv4_printk(compound,
		     "NFSDV4 encode_compoundres WRITE status: %d\n",ntohl(resp->status));

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(resp->status))))
		return 0;
	*p++ = resp->status;
	if(resp->status == NFS4_OK){
		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
			return 0;
		*p++ = htonl(resp->WRITERES.count);
		*p++ = htonl(resp->WRITERES.committed);
		p = xdr_encode_verifier4(p,resp->WRITERES.writeverf);
	}
	return p;
}

int
nfs4svc_encode_voidres(struct svc_rqst *rqstp, u32 *p, void *dummy)
{
	nfsv4_printk(xdr,"NFSDV4 encode voidargs\n");
	return xdr_argsize_check(rqstp, p);
}


int
nfs4svc_encode_COMPOUND4res(struct svc_rqst *rqstp, u32 *p, struct COMPOUND4res *cres)
{
	int i, status = NFS4_OK;
	struct svc_buf *buf = &rqstp->rq_resbuf;
	int len = sizeof(cres->RESCNT);

	nfsv4_printk(compound,
		     "NFSDV4 encode compoundres. rescnt: %d\n",cres->RESCNT);

	nfsv4_printk(compound,"NFSDV4 encode compoundres. buf->base 0x%x p 0x%x, len %d\n",(unsigned int)buf->base,(unsigned int)p,(p - buf->base));

	/* note, the COMPOUND4res status is already htonl'ed
	 * see nfsd4errno(). 
	 */

	*p++ = cres->status;
	print_utf8string(&cres->tag);
	if(!(p = xdr_encode_utf8string(rqstp,p,&cres->tag)))
		return 0;

	if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,len)))
		return 0;
	*p++ = htonl(cres->RESCNT);
	for(i=0;i<cres->RESCNT && status == NFS4_OK;i++){
		if(!(xdr_overflow_check(&rqstp->rq_resbuf,p,sizeof(nfsstat4))))
			return 0;
		*p++ = htonl(cres->RESVAL[i].resop);
		switch (cres->RESVAL[i].resop){
		case OP_ACCESS:
			if(!(p = xdr_encode_access4(rqstp,p,&cres->RESVAL[i].RESU.opaccess)))
				return 0;
			status = cres->RESVAL[i].RESU.opaccess.status;
			break;
		case OP_CLOSE:
			if(!(p = xdr_encode_close4(rqstp,p,&cres->RESVAL[i].RESU.opclose)))
				return 0;
			status = cres->RESVAL[i].RESU.opclose.status;
			break;
		case OP_CREATE:
			if(!(p = xdr_encode_create4(rqstp,p,&cres->RESVAL[i].RESU.opcreate)))
				return 0;
			status = cres->RESVAL[i].RESU.opcreate.status;
			break;
		case OP_DELEGRETURN:
			if (!(p = xdr_encode_delegreturn4(rqstp, p, &cres->RESVAL[i].RESU.opdelegreturn)))
				return 0;
			status = cres->RESVAL[i].RESU.opdelegreturn.status;
			break;
		case OP_GETATTR:
			if(!(p = xdr_encode_getattr4(rqstp,p,&cres->RESVAL[i].RESU.opgetattr)))
				return 0;
			status = cres->RESVAL[i].RESU.opgetattr.status;
			break;
		case OP_GETFH:
			if(!(p = xdr_encode_getfh4(rqstp,p,&cres->RESVAL[i].RESU.opgetfh)))
				return 0;
			status = cres->RESVAL[i].RESU.opgetfh.status;
			break;
		case OP_LINK:
			if(!(p = xdr_encode_link4(rqstp,p,&cres->RESVAL[i].RESU.oplink)))
				return 0;
			status = cres->RESVAL[i].RESU.oplink.status;
			break;
		case OP_LOOKUP:
			if(!(p = xdr_encode_lookup4(rqstp,p,&cres->RESVAL[i].RESU.oplookup)))
				return 0;
			status = cres->RESVAL[i].RESU.oplookup.status;
			break;
		case OP_LOOKUPP:
			if(!(p = xdr_encode_lookupp4(rqstp,p,&cres->RESVAL[i].RESU.oplookupp)))
				return 0;
			status = cres->RESVAL[i].RESU.oplookupp.status;
			break;
		case OP_OPEN:
			if(!(p = xdr_encode_open4(rqstp,p,&cres->RESVAL[i].RESU.opopen)))
				return 0;
			status = cres->RESVAL[i].RESU.opopen.status;
			break;
		case OP_OPEN_CONFIRM:
			if(!(p = xdr_encode_open_confirm4(rqstp,p,&cres->RESVAL[i].RESU.opopen_confirm)))
				return 0;
			status = cres->RESVAL[i].RESU.opopen_confirm.status;
			break;
		case OP_PUTFH:
			if(!(p = xdr_encode_putfh4(rqstp,p,&cres->RESVAL[i].RESU.opputfh)))
				return 0;
			status = cres->RESVAL[i].RESU.opputfh.status;
			break;
		case OP_PUTROOTFH:
			if(!(p = xdr_encode_putfh4(rqstp,p,&cres->RESVAL[i].RESU.opputfh)))
				return 0;
			status = cres->RESVAL[i].RESU.opputfh.status;
			break;
		case OP_READ:
			if(!(p = xdr_encode_read4(rqstp,p,&cres->RESVAL[i].RESU.opread)))
				return 0;
			status = cres->RESVAL[i].RESU.opread.status;
			break;
		case OP_READDIR:
			if(!(p = xdr_encode_readdir4(rqstp,p,&cres->RESVAL[i].RESU.opreaddir)))
				return 0;
			status = cres->RESVAL[i].RESU.opreaddir.status;
			break;
		case OP_READLINK:
			if(!(p = xdr_encode_readlink4(rqstp,p,&cres->RESVAL[i].RESU.opreadlink)))
				return 0;
			status = cres->RESVAL[i].RESU.opreadlink.status;
			break;
		case OP_REMOVE:
			if(!(p = xdr_encode_remove4(rqstp,p,&cres->RESVAL[i].RESU.opremove)))
				return 0;
			status = cres->RESVAL[i].RESU.opremove.status;
			break;
		case OP_RENAME:
			if(!(p = xdr_encode_rename4(rqstp,p,&cres->RESVAL[i].RESU.oprename)))
				return 0;
			status = cres->RESVAL[i].RESU.oprename.status;
			break;
		case OP_RENEW:
			if(!(p = xdr_encode_renew4(rqstp,p,&cres->RESVAL[i].RESU.oprenew)))
				return 0;
			status = cres->RESVAL[i].RESU.oprenew.status;
			break;
		case OP_RESTOREFH:
			if(!(p =xdr_encode_restorefh4(rqstp,p,&cres->RESVAL[i].RESU.oprestorefh)))
				return 0;
			status = cres->RESVAL[i].RESU.oprestorefh.status;
			break;
		case OP_SECINFO:
			if(!(p =xdr_encode_secinfo4(rqstp,p,&cres->RESVAL[i].RESU.opsecinfo)))
				return 0;
			status = cres->RESVAL[i].RESU.opsecinfo.status;
			break;
		case OP_SAVEFH:
			if(!(p =xdr_encode_savefh4(rqstp,p,&cres->RESVAL[i].RESU.opsavefh)))
				return 0;
			status = cres->RESVAL[i].RESU.opsavefh.status;
			break;
		case OP_SETCLIENTID:
			if(!(p =xdr_encode_setclientid4(rqstp,p,&cres->RESVAL[i].RESU.opsetclientid)))
				return 0;
			status = cres->RESVAL[i].RESU.opsetclientid.status;
			break;
		case OP_SETCLIENTID_CONFIRM:
			if(!(p =xdr_encode_setclientid_confirm4(rqstp,p,&cres->RESVAL[i].RESU.opsetclientid_confirm)))
				return 0;
			status = cres->RESVAL[i].RESU.opsetclientid_confirm.status;
			break;
		case OP_SETATTR:
			if(!(p =xdr_encode_setattr4(rqstp,p,&cres->RESVAL[i].RESU.opsetattr)))
				return 0;
			status = cres->RESVAL[i].RESU.opsetattr.status;
			break;	
		case OP_WRITE:
			if(!(p =xdr_encode_write4(rqstp,p,&cres->RESVAL[i].RESU.opwrite)))
				return 0;
			status = cres->RESVAL[i].RESU.opwrite.status;
			break;
		default:
			if((cres->RESVAL[i].resop >= 0) &&(cres->RESVAL[i].resop < 39)){
				nfsv4_printk(compound,
					     "NFSV4: unimplemented opcode %d\n", cres->RESVAL[i].resop);
				return 0;
			}else {
				nfsv4_printk(compound,
					     "NFSV4: bad opcode: %d\n", cres->RESVAL[i].resop);
				return 0;
			}
			break;
		}
	}
	
	return xdr_ressize_check(rqstp, p);
}


/*
 * XDR free args function
 */
void
nfsd4_kfree_compound_args(COMPOUND4args *args)
{
	int i;

	nfsv4_printk(free,"NFSV4 nfsd4_kfree_compound_args\n");
	for(i=0;i<args->OPCNT;i++){
		switch(args->OPVAL[i].argop){
		case OP_ACCESS: /*nothing to do */
			break;
		case OP_CREATE: /*nothing to do */
			break;
		case OP_DELEGRETURN: /* nothing to do */
			break;
		case OP_GETATTR:
		{
			GETATTR4args *argp= &args->OPVAL[i].OPU.opgetattr;

			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OP_GETATTR4\n");

			if(argp->attr_request.BVAL)
				kfree(argp->attr_request.BVAL);
		}
		break;
		case OP_GETFH:
			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OP_GETFH\n");
			break;
		case OP_LINK:
			nfsv4_printk(free, "NFSV4: nfsd4_kfree_compound_args OP_LINK\n");
			/* nothing to do */
			break;
		case OP_LOOKUP:
		{
			LOOKUP4args *argp= &args->OPVAL[i].OPU.oplookup;
	
			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OP_LOOKUP\n");
			if(argp->PVAL)
				kfree(argp->PVAL);
		}
		break;
		case OP_LOOKUPP:
			/* nothing to do */
			break;
		case OP_OPEN:
		{
			OPEN4args *argp= &args->OPVAL[i].OPU.opopen;

			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OP_OPEN\n");
			switch(argp->openhow.opentype){
			case OPEN4_CREATE:
				switch(argp->openhow.FLAG_HOW_MODE){
				case GUARDED4:
					if(argp->openhow.FLAG_HOW.CREATE_FATTR.attrmask.BVAL)					
						kfree(argp->openhow.FLAG_HOW.CREATE_FATTR.attrmask.BVAL);
					if(argp->openhow.FLAG_HOW.CREATE_FATTR.FATTRSVAL)
						kfree(argp->openhow.FLAG_HOW.CREATE_FATTR.FATTRSVAL);	
					break;
				case UNCHECKED4:
				case EXCLUSIVE4:
					break;
				}
				break;
			case OPEN4_NOCREATE:
				break;
			}
		}
		break;
		case OP_PUTFH:
		{
			PUTFH4args *argp= &args->OPVAL[i].OPU.opputfh;

			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OP_PUTFH\n");
			if(argp->object.nfs_fh4_val)
				kfree(argp->object.nfs_fh4_val);
		}
		break;
		case OP_PUTROOTFH: /*nothing to do */
			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OP_PUTROOTFH\n");
			break;
		case  OP_READ: /* nothing to do */
			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OP_READ\n");
			break;
		case OP_READDIR:
		{
			READDIR4args *argp= &args->OPVAL[i].OPU.opreaddir;
			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OP_READDIR\n");
	
			if(argp->attr_request.BVAL)
				kfree(argp->attr_request.BVAL);
		}
		break;
		case OP_REMOVE:
			nfsv4_printk(free, "NFSV4: nfsd4_kfree_compound_args OP_REMOVE\n");
			break;
		case OP_RENAME:
			nfsv4_printk(free, "NFSV4: nfsd4_kfree_compound_args OP_RENAME\n");
			/* nothing to do */
			break;
		case OP_RESTOREFH:
			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OP_RESTOREFH\n");
			/* nothing to do */
			break;
		case OP_SAVEFH:
			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OP_SAVEFH\n");
			/* nothing to do */
			break;
		case OP_SECINFO:
			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OP_SECINFO\n");
			/* nothing to do*/
			break;
		case OP_SETATTR:
		{
			fattr4 *fattrp = &args->OPVAL[i].OPU.opsetattr.obj_attributes;
			kfree (fattrp->attrmask.bitmap4_val);
			kfree (fattrp->attr_vals.attrlist4_val);
		}
		break;
		case OP_WRITE: /*nothing to do */
			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OP_WRITE\n");
			break;
		default:
			nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args:  bad opcode - DEFAULT\n");
			break;
		} /*switch*/
	}
/* NOTE: args->tag.utf8string_val is assigned to the rpc buffer
 * string location in xdr_decode_utf8string, and therefore
 * is non-null and not kmalloc'ed - so don't kfree it!
 */
	nfsv4_printk(free,"NFSV4: nfsd4_kfree_compound_args OPVAL\n");
	if(args->OPVAL) {
		kfree(args->OPVAL);
	}
}


/*
 * XDR release function - free response
 */

int
nfs4svc_release_compound(struct svc_rqst *rqstp, u32 *p, struct COMPOUND4res *res)
{
	int i;
	nfsv4_printk(compound,
		     "NFSDV4 release compoundrel. rescnt: %d\n",res->RESCNT);
	for(i=0;i<res->RESCNT;i++){
		switch (res->RESVAL[i].resop){
		case OP_ACCESS: /* nothing to do */
			break;
		case OP_CREATE: /* nothing to do */
			break;
		case OP_DELEGRETURN:
			break;
		case OP_GETATTR:
		{
			GETATTR4res *resp= &res->RESVAL[i].RESU.opgetattr;

			nfsv4_printk(compound,
				     "NFSDV4 release compoundrel GETATTR\n");
			kfree(resp->GFATTR4.FATTRSVAL);
			kfree(resp->GFATTR4.attrmask.BVAL);
		}
		break;
		case OP_GETFH:
		{
			GETFH4res *resp= &res->RESVAL[i].RESU.opgetfh;
	
			nfsv4_printk(compound,
				     "NFSDV4 release compoundrel GETFH: nfs_fh4_len: %d\n",resp->GETFH4FH.nfs_fh4_len);
			
			/* why does this cause nfsd to blow up with 'kfree: bad object?
			   if(resp->GETFH4FH.nfs_fh4_val)
			   kfree(resp->GETFH4FH.nfs_fh4_val);
			*/
		}
		break;
		case OP_LINK:
			/* nothing to do */
			break;
		case OP_LOOKUP: /* nothing to do */
			break;
		case OP_LOOKUPP: /* nothing to do */
			break;
		case OP_OPEN:
		{
			OPEN4res *resp= &res->RESVAL[i].RESU.opopen;
			open_delegation4 *d4 = &resp->OPEN_RES.delegation;

 			switch (d4->delegation_type) {
			case OPEN_DELEGATE_NONE:
				break;
			case OPEN_DELEGATE_READ:
			case OPEN_DELEGATE_WRITE:
				if(d4->open_delegation4_u.write.permissions.who.utf8string_val)
					kfree(d4->open_delegation4_u.write.permissions.who.utf8string_val);
				break;
			}
		}
		break;
		case OP_PUTFH: /* nothing to do */
			break;
		case OP_PUTROOTFH: /* nothing to do */
			break;
		case OP_READ: /* nothing to do */
			break;
		case OP_READDIR:
		{
			READDIR4res *resp= &res->RESVAL[i].RESU.opreaddir;
		
			if(resp->ENTRIES)
				kfree(resp->ENTRIES);
		}
		break;
		case OP_READLINK:
		{
			READLINK4res *resp = &res->RESVAL[i].RESU.opreadlink;
			kfree (resp->READLINK4res_u.resok4.link.utf8string_val);
		}
		break;
		case OP_REMOVE:
			break;
		case OP_RENAME:
			/* nothing to do */
			break;
		case OP_RESTOREFH:
			/* nothing to do */
			break;
		case OP_SAVEFH:
			/* nothing to do */
			break;
		case OP_SETATTR:
			kfree (res->RESVAL[i].RESU.opsetattr.attrsset.bitmap4_val);
			break;
		case OP_WRITE: /* nothing to do */
			break;
		default:
			break;
		}
	}
	if(res->tag.utf8string_val)
		kfree(res->tag.utf8string_val);
	kfree(res->RESVAL);
	return 0;
}
