/*
 * linux/fs/nfsd/nfssvc.c
 *
 * Central processing for nfsd.
 *
 * Authors:	Olaf Kirch (okir@monad.swb.de)
 *
 * Copyright (C) 1995-1999 Olaf Kirch <okir@monad.swb.de>
 *
 * NFSV4 Implementation
 *      by Andy Adamson <abdros@citi.umich.edu>
 *      Copyright (C) 1999 The Regents of the University of Michigan
 */

#define __NO_VERSION__
#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/nfs.h>
#include <linux/in.h>
#include <linux/uio.h>
#include <linux/version.h>
#include <linux/unistd.h>
#include <linux/malloc.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>

#include <linux/sunrpc/types.h>
#include <linux/sunrpc/stats.h>
#include <linux/sunrpc/svc.h>
#include <linux/sunrpc/svcsock.h>
#ifdef NFSV4
#include <linux/sunrpc/auth_gss.h>
#include <linux/nfsd/nfsd4_prot.h>
#else /*NFSV4*/
#include <linux/nfsd/nfsd.h>
#include <linux/nfsd/xdr.h>
#endif /*NFSV4*/
#include <linux/nfsd/stats.h>
#include <linux/nfsd/cache.h>
#ifndef NFSv4
#include <linux/lockd/bind.h>
#endif /*NFSV4*/

#define NFSDDBG_FACILITY	NFSDDBG_SVC
#ifndef NFSV4 /* placed in nfsd4_prot.h */
#define NFSD_BUFSIZE		(1024 + NFSSVC_MAXBLKSIZE)
#endif /*NFSV4*/

#define ALLOWED_SIGS	(sigmask(SIGKILL))
#define SHUTDOWN_SIGS	(sigmask(SIGKILL) | sigmask(SIGINT) | sigmask(SIGQUIT))

extern struct svc_program	nfsd_program;
static void			nfsd(struct svc_rqst *rqstp);
struct timeval			nfssvc_boot = { 0, 0 };
static atomic_t			nfsd_active = ATOMIC_INIT(0);

#ifndef NFSV4
/* moved into nfsd4_prot.h
 *
 * Maximum number of nfsd processes
 */
#define	NFSD_MAXSERVS		128
#endif /*NFSV4*/

int
nfsd_svc(unsigned short port, int nrservs)
{
	struct svc_serv *	serv;
	int			error;
#ifdef NFSV4
	int i;
#endif

	error = -EINVAL;
	if (atomic_read(&nfsd_active))
		goto out;	/* already running */
	if (nrservs <= 0)
		goto out;
	if (nrservs > NFSD_MAXSERVS)
		nrservs = NFSD_MAXSERVS;

	dprintk("nfsd: creating service (%d)\n", nrservs);
	nfsd_nservers = nrservs;
	nfssvc_boot = xtime;

#ifdef NFSV4
	/* set up per thread state array */
	nfsd_state.nrservs = nrservs;
	for(i=0;i<NFSD_MAXSERVS;i++)
		nfsd_state.tbl[i] = NULL;
	nfsv4_printk(svc,"NFSD_SVC global state array init. #thds %d\n",nrservs);
#endif /*NFSV4*/

	error = -ENOMEM;
	serv = svc_create(&nfsd_program, NFSD_BUFSIZE, NFSSVC_XDRSIZE);
	if (serv == NULL)
		goto out;

	error = svc_makesock(serv, IPPROTO_UDP, port);
	if (error < 0)
		goto failure;

#if 0	/* Don't even pretend that TCP works. It doesn't. */
	error = svc_makesock(serv, IPPROTO_TCP, port);
	if (error < 0)
		goto failure;
#endif

	nfsd_racache_init();	/* Readahead param cache */

	while (nrservs--) {
		error = svc_create_thread(nfsd, serv);
		if (error < 0)
			break;
	}

 failure:
	svc_destroy(serv);		/* Release server */
 out:
	return error;
}

/*
 * This is the NFS server kernel thread
 */
static void
nfsd(struct svc_rqst *rqstp)
{
	struct svc_serv	*serv = rqstp->rq_server;
	int		err;

	/* Lock module */
	MOD_INC_USE_COUNT;

	/* Set up kernel thread */
	lock_kernel();
	exit_mm(current);
	sprintf(current->comm, "nfsd");
	current->session = 1;
	current->pgrp = 1;
	current->fs->umask = 0;

#ifdef NFSV4
	init_global_state(current->pid);
#endif /*NFSV4*/

	/* Count active threads */
	atomic_inc(&nfsd_active);

#ifndef NFSV4
	/* Start lockd */
	lockd_up();
#endif /*NFSV4*/

	/* Let svc_process check client's authentication. */
	rqstp->rq_auth = 1;

	/*
	 * The main request loop
	 */
	for (;;) {
		/* Block all but the shutdown signals */
		spin_lock_irq(&current->sigmask_lock);
		siginitsetinv(&current->blocked, SHUTDOWN_SIGS);
		recalc_sigpending(current);
		spin_unlock_irq(&current->sigmask_lock);

		/*
		 * Find a socket with data available and call its
		 * recvfrom routine.
		 */
		while ((err = svc_recv(serv, rqstp,
				       MAX_SCHEDULE_TIMEOUT)) == -EAGAIN)
			;
		if (err < 0)
			break;

		/* Lock the export hash tables for reading. */
		exp_readlock();

		/* Validate the client's address. This will also defeat
		 * port probes on port 2049 by unauthorized clients.
		 */
/*NFSV4 - we don't have a client struct until the client performs a 
 * SETCLIENTID, so the exp_getclient() will fail. see svc_process()
 */
		rqstp->rq_client = exp_getclient(&rqstp->rq_addr,0);
		/* Process request with signals blocked.  */
		spin_lock_irq(&current->sigmask_lock);
		siginitsetinv(&current->blocked, ALLOWED_SIGS);
		recalc_sigpending(current);
		spin_unlock_irq(&current->sigmask_lock);

		svc_process(serv, rqstp);

		/* Unlock export hash tables */
		exp_unlock();
	}

	if (err != -EINTR) {
		printk(KERN_WARNING "nfsd: terminating on error %d\n", -err);
	} else {
		unsigned int	signo;

		for (signo = 1; signo <= _NSIG; signo++)
			if (sigismember(&current->signal, signo) &&
			    !sigismember(&current->blocked, signo))
				break;
		printk(KERN_WARNING "nfsd: terminating on signal %d\n", signo);
	}

	/* Count active threads */
	printk("Calling atomic_dec_and_test\n");
	if (atomic_dec_and_test(&nfsd_active)) {
		printk("Called atomic_dec_and_test SHUTDOWN NFSD\n");
		nfsd_export_shutdown();		/* revoke all exports */
		nfsd_racache_shutdown();	/* release read-ahead cache */
#ifdef NFSV4
		n2s_cache_shutdown();
		shutdown_pseudo_fs();
		shutdown_lock_state();
#endif /*NFSV4*/
	}

#ifdef NFSV4
/* free threads global state */
	free_global_state(current->pid);
#else /*NFSV4*/
	/* Release lockd */
	lockd_down();
#endif /*NFSV4*/

	/* Destroy the thread's resources */
	svc_exit_thread(rqstp);

	/* Release module */
	MOD_DEC_USE_COUNT;
}

static int
nfsd_dispatch(struct svc_rqst *rqstp, u32 *statp)
{
	struct svc_procedure	*proc;
	kxdrproc_t		xdr;
	u32			nfserr;
	struct nfs4_lock_owner	*lowner;
	
	dprintk("nfsd_dispatch: vers %d proc %d\n",
		rqstp->rq_vers, rqstp->rq_proc);
	proc = rqstp->rq_procinfo;

#ifdef NFSV4
	/* XXX - NULLPROC decode stage handled by svcauth_gss(). */
	if (rqstp->rq_proc == 0) {
		if (!xdr_encode_rpc_gss_init_res(rqstp, rqstp->rq_resbuf.buf,
						 (struct rpc_gss_init_res *)
						 rqstp->rq_resp))
			*statp = rpc_system_err;
		return 1;
	}
	/* we don't cache reply's to compound calls that contain 
	 * an OPEN, CLOSE, LOCK, etc here - handled in nfs4lock.c: nfs4_process_lock.
	 * since we don't know at this time whether or not the compound call 
	 * contains a lock seq num mutating operation, the v3 nfsd cache can
	 * not be used as is. futhermore, the potential size of the compound call return
	 * is problematic. what i'll do is stuff the cache from nfsd4xdr.c when the
	 * size and operations are appropriate, so the cache will only contain these
	 * approved of elements.
	 */

	/* HEY! where is the release function call! sheeze. add a call to 
	 * the release functions a la the svc.c generic dispatcher....
	 */

	/* Decode arguments */
	nfsv4_printk(dispatcher, "nfsd_dispatch() xdr decoding packet\n");
	xdr = proc->pc_decode;
	if (xdr && !xdr(rqstp, rqstp->rq_argbuf.buf, rqstp->rq_argp)) {
		dprintk("nfsd: failed to decode arguments!\n");
		/* NFSV4 - need to call nfsv4_free_args function here to free any potential
		 * kmallocs */
		nfsd_cache_update(rqstp, RC_NOCACHE, NULL);
		*statp = rpc_garbage_args;
		return 1;
	}

	
	/* Now call the procedure handler, and encode NFS status. */
	nfsv4_printk(dispatcher, "nfsd_dispatch() calling nfsd4_proc_compound()\n");
	SEQID_LOCKOWNER = NULL;
     	nfserr = proc->pc_func(rqstp, rqstp->rq_argp, rqstp->rq_resp);
	lowner = SEQID_LOCKOWNER;
	
	if (nfserr == nfserr_replay_me) {
		nfsv4_printk(dispatcher, "nfsd_dispatch() replaying packet\n");
		rqstp->rq_resbuf.len = lowner->lo_seq_response.len;
		memcpy(rqstp->rq_resbuf.buf, lowner->lo_seq_response.buf, lowner->lo_seq_response.len << 2);
		nfsv4_printk(dispatcher, "nfsd_dispatch() raising semaphore\n");
		up(&lowner->lo_seqid_sema);
	}
	else {
		nfsv4_printk(dispatcher, "nfsd_dispatch() xdr encoding packet\n");

		/* Encode results even when nfserr != 0 because
		 * Compound requests can partially succeed
		 */
		xdr = proc->pc_encode;
		if (xdr && !xdr(rqstp, rqstp->rq_resbuf.buf, rqstp->rq_resp)) {
			/* Failed to encode result. Release cache entry */
			/* NFSV4 - need to call nfsv4_free_args and nfsv4_free_res  function here 
			 * to free any potential kmallocs */
			nfsv4_printk(svc,"NFSV4 nfsd_dispatch: failed to encode result!\n");
			nfsd_cache_update(rqstp, RC_NOCACHE, NULL);
			*statp = rpc_system_err;
		}
		if (lowner) {
			nfsv4_printk(dispatcher, "nfsd_dispatch() SEQID_LOCKOWNER is set\n");
			if (nfserr != nfserr_bad_seqid) {
				/* cache response */
				nfsv4_printk(dispatcher, "nfsd_dispatch() caching response\n");
				lowner->lo_seq_response.len = rqstp->rq_resbuf.len;
				memcpy(lowner->lo_seq_response.buf, rqstp->rq_resbuf.buf, rqstp->rq_resbuf.len << 2);
			}
			nfsv4_printk(dispatcher, "nfsd_dispatch() raising semaphore\n");
			up(&lowner->lo_seqid_sema);
		}
		else
			nfsv4_printk(dispatcher, "nfsd_dispatch() SEQID_LOCKOWNER is not set\n");
	}
	nfsv4_printk(dispatcher, "nfsd_dispatch() done\n");
	
	return 1;
	
#else /*NFSV4*/
	/* Check whether we have this call in the cache. */
	switch (nfsd_cache_lookup(rqstp, proc->pc_cachetype)) {
	case RC_INTR:
	case RC_DROPIT:
		return 0;
	case RC_REPLY:
		return 1;
	case RC_DOIT:
		/* do it */
	}

	/* Decode arguments */
	xdr = proc->pc_decode;
	if (xdr && !xdr(rqstp, rqstp->rq_argbuf.buf, rqstp->rq_argp)) {
		dprintk("nfsd: failed to decode arguments!\n");
		/* NFSV4 - need to call nfsv4_free_args function here to free any potential
		 * kmallocs */
		nfsd_cache_update(rqstp, RC_NOCACHE, NULL);
		*statp = rpc_garbage_args;
		return 1;
	}

	/* Now call the procedure handler, and encode NFS status. */
	nfserr = proc->pc_func(rqstp, rqstp->rq_argp, rqstp->rq_resp);
	if (rqstp->rq_proc != 0)
		svc_putlong(&rqstp->rq_resbuf, nfserr);

#if defined(CONFIG_NFSD_V3)
	/* Encode result.
	 * For NFSv2, additional info is never returned in case of an error.
	 */
	if (!(nfserr && rqstp->rq_vers == 2)) {
		xdr = proc->pc_encode;
		if (xdr && !xdr(rqstp, rqstp->rq_resbuf.buf, rqstp->rq_resp)) {
			/* Failed to encode result. Release cache entry */
			dprintk("nfsd: failed to encode result!\n");
			nfsd_cache_update(rqstp, RC_NOCACHE, NULL);
			*statp = rpc_system_err;
			return 1;
		}
	}
#else
	xdr = proc->pc_encode;
	if (!nfserr && xdr
	    && !xdr(rqstp, rqstp->rq_resbuf.buf, rqstp->rq_resp)) {
		/* Failed to encode result. Release cache entry */
		dprintk("nfsd: failed to encode result!\n");
		nfsd_cache_update(rqstp, RC_NOCACHE, NULL);
		*statp = rpc_system_err;
		return 1;
	}
#endif /* CONFIG_NFSD_V3 */

	/* Store reply in cache. */
	nfsd_cache_update(rqstp, proc->pc_cachetype, statp + 1);
	return 1;

#endif /*NFSV4*/

}

#ifdef CONFIG_NFS_V4
static struct svc_version nfsd_version4 = {
	4, 2, nfsd_procedures4, nfsd_dispatch
};
#else /* NFSV4*/
static struct svc_version nfsd_version2 = {
	2, 18, nfsd_procedures2, nfsd_dispatch
};
#endif /*NFSV4*/ 

#ifdef CONFIG_NFSD_V3
static struct svc_version	nfsd_version3 = {
	3, 22, nfsd_procedures3, nfsd_dispatch
};
#endif
static struct svc_version *	nfsd_version[] = {
	NULL,
	NULL,
#ifndef CONFIG_NFS_V4
	&nfsd_version2,
#endif /*CONFIG_NFS_V4*/
#ifdef CONFIG_NFSD_V3
	&nfsd_version3,
#endif
#ifdef CONFIG_NFS_V4
	NULL,
	NULL,
	&nfsd_version4,
#endif /*CONFIG_NFS_V4*/
};

#define NFSD_NRVERS		(sizeof(nfsd_version)/sizeof(nfsd_version[0]))
struct svc_program		nfsd_program = {
	NFS_PROGRAM,		/* program number */
	2, NFSD_NRVERS-1,	/* version range */
	NFSD_NRVERS,		/* nr of entries in nfsd_version */
	nfsd_version,		/* version table */
	"nfsd",			/* program name */
	&nfsd_svcstats,		/* version table */
};
