#include <linux/string.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/sunrpc/svc.h>
#include <linux/fs.h>
#include <linux/nfsd/nfsd4_prot.h>
#include <linux/nfsd/pseudo.h>

/* pseudo_fs consists of a root directory which contains
* one directory per nfsd exported directory from 
* /etc/exports which will be the mount point 
* of said exported directory.
*
* the pseudo fs is a virtual fs, no disk.
* the pseudo fs is not hooked into the VFS layer,
* only nfsd knows about it. only nfsd can
* create pseudo_fs inodes and write pseudo_fs data.
* currently, this only happens at startup, when
* the pseudo fs is created - therefore don't needc
* locks. eventually, it will be possible to change
* the export tree without cycling nfsd and locks
* will then be required - perhaps just a chicken lock..
*
* pseudo inodes and their data (pseudo_fs_entries) are
* managed by nfsd - get_empty_inode() is used, so inodes
* are not hashed, iget() won't find them. instead, 
* they are placed an inode ptr * array which is 
* used both to find them (pseudo_iget) and 
* reap them (shutdown_pseudo_fs).
* 
* need a kdev_t for the pseudo fs -  
* use an unnamed block device.
*/

extern struct super_block *get_empty_super(void);
extern kdev_t get_unnamed_dev(void);
extern void put_unnamed_dev(kdev_t dev);
extern int nfsd_pseudo_exp_init(struct dentry *dentry);

/*Globals*/
static struct super_block *sbp;
static int next_pseudo_i_num = PSEUDO_FS_ROOT_INO; /* root inode# = 2 */
static struct pseudo_fs_entry *all_fs_entries[TOTAL_NFSV4_PSEUDO_ENTRIES];
static struct inode *all_pseudo_inodes[TOTAL_NFSV4_PSEUDO_INODES];

/* Function Prototypes */
static void pseudo_read_inode(struct inode *);
static void pseudo_put_inode(struct inode *);
static void pseudo_delete_inode(struct inode *);
static void init_pseudo_inode_dots(struct inode * dir_inode,struct inode *inode);
static struct inode * pseudofs_get_dir(struct inode *dir_inode,char *path_comp);
static struct inode * pseudo_make_inode(struct super_block *sb); 
static int add_path(pseudo_fs_entry *ep, char *path_comp,unsigned long ino);
static int set_next_fs_entry(pseudo_fs_entry *ep);
static int set_next_inode(struct inode *inode);
void print_pseudo_inode(struct inode *inode);
void print_pseudo_dentry(struct dentry *dentry);
static void print_pseudo_entry(pseudo_fs_entry *ep);
struct inode * pseudo_lookup(struct inode *iparent, char *name);
struct inode * pseudo_iget(unsigned int ino);
int pseudo_open(struct inode *inode, struct file *filp);
int pseudo_readdir(struct file *filep, void *cd, filldir_t filldir);

/* these all return -EACCES */
static loff_t pseudo_llseek(struct file *file,loff_t offset,int flag);
static ssize_t pseudo_read(struct file *file,char *name,size_t count,loff_t *offset);
static ssize_t pseudo_write(struct file *file,const char *name,size_t count,loff_t *offset);
static unsigned int pseudo_poll(struct file *file, struct poll_table_struct *pt);
static int pseudo_ioctl(struct inode *inode,struct file *file,unsigned int which,unsigned long why);
static int pseudo_mmap(struct file *file, struct vm_area_struct *vma);
static int pseudo_flush(struct file *file);
static int pseudo_release(struct inode *inode, struct file *file);
static int pseudo_fsync(struct file *file, struct dentry *dentry);
static int pseudo_fasync(int dummy, struct file *file, int flag);
static int pseudo_check_media_change(kdev_t dev);
static int pseudo_revalidate(kdev_t dev);
static int pseudo_lock(struct file *file, int flag, struct file_lock *flock);  

/* all of these return -EACCES */
static int pseudo_inode_create(struct inode *ino,struct dentry *den,int dummy);
static struct dentry * pseudo_inode_lookup(struct inode *ino,struct dentry *den);
static int pseudo_inode_link(struct dentry *den,struct inode *ino,struct dentry *den2);
static int pseudo_inode_unlink(struct inode *ino,struct dentry *den);
static int pseudo_inode_symlink(struct inode *ino,struct dentry *den,const char *str);
static int pseudo_inode_mkdir(struct inode *ino,struct dentry *den,int dummy);
static int pseudo_inode_rmdir(struct inode *ino,struct dentry *den);
static int pseudo_inode_mknod(struct inode *ino,struct dentry *den,int dummy,int dummy2);
static int pseudo_inode_rename(struct inode *ino, struct dentry *den, struct inode *ino2, struct dentry *den2);
static int pseudo_inode_readlink(struct dentry *den, char *str,int dummy);
static struct dentry * pseudo_inode_follow_link(struct dentry *den, struct dentry *den2, unsigned int dummy);
static int pseudo_inode_readpage(struct file *filp, struct page *page);
static int pseudo_inode_writepage(struct file *filp, struct page *page);
static int pseudo_inode_bmap(struct inode *ino,int dummy);
static void pseudo_inode_truncate(struct inode *ino);
static int pseudo_inode_permission(struct inode *ino, int dummy);
static int pseudo_inode_smap(struct inode *ino,int dummy);
static int pseudo_inode_updatepage(struct file *filp, struct page *page, const char *str, unsigned long dummy, unsigned int dummy2, int dummy3);
static int pseudo_inode_revalidate(struct dentry *den);

static int pseudo_statfs (struct super_block *sb, struct statfs *buf, int buflen)
{
	// temporary hack: assume that, whenever the client does a statfs(),
		   //   the result of statfs()'ing the server's root directory can be
		   //   used
		   struct fs_struct *root_fs;
	struct dentry *root_d;
	struct inode *root_i;
	struct super_block *root_sb;
	struct super_operations *root_op;
	int (*root_statfs) (struct super_block *, struct statfs *, int);

	if ( (root_fs = current->fs) &&
	     (root_d = root_fs->root) &&
	     (root_i = root_d->d_inode) &&
	     (root_sb = root_i->i_sb) &&
	     (root_op = root_sb->s_op) &&
	     (root_statfs = root_op->statfs) )
		return root_statfs (root_sb, buf, buflen);
	return -EINVAL;
}

static struct super_operations pseudo_fs_sops = { 
	pseudo_read_inode,		/* read inode */
	NULL,			/* write inode */
	pseudo_put_inode,		/* put inode */
	pseudo_delete_inode,	/* delete inode */
	NULL,	/* notify change */
	NULL,		/* put superblock */
	NULL,			/* write superblock */
	pseudo_statfs,		/* stat filesystem */
	NULL,			/* remount */
	NULL,			/* clear inode */
	NULL,	/* umount attempt begin */
};

static struct file_operations pseudo_file_ops = {
	pseudo_llseek, /* llseek */
	pseudo_read, /* read */
	pseudo_write, /* write */
	pseudo_readdir, /* readdir */
	pseudo_poll, /* poll */
	pseudo_ioctl, /* ioctl */
	pseudo_mmap, /* mmap */
	pseudo_open, /* open */
	pseudo_flush, /* flush */
	pseudo_release, /* release */
	pseudo_fsync, /* fsync */
	pseudo_fasync, /* fasync */
	pseudo_check_media_change, /* check_media_change */
	pseudo_revalidate, /* revalidate */
	pseudo_lock, /* lock */
};

struct inode_operations pseudo_dir_inode_operations = {
	&pseudo_file_ops,  /* default directory dir-ops */
	pseudo_inode_create,   /* create */
	pseudo_inode_lookup,   /* lookup */
	pseudo_inode_link,   /* link */
	pseudo_inode_unlink,   /* unlink */
	pseudo_inode_symlink,    /* symlink */
	pseudo_inode_mkdir,    /* mkdir */
	pseudo_inode_rmdir,    /* rmdir */
	pseudo_inode_mknod,    /* mknod */
	pseudo_inode_rename,   /* rename */
	pseudo_inode_readlink,     /* readlink */
	pseudo_inode_follow_link,     /* follow_link */
	pseudo_inode_readpage,     /* readpage */
	pseudo_inode_writepage,     /* writepage */
	pseudo_inode_bmap,     /* bmap */
	pseudo_inode_truncate,     /* truncate */
	pseudo_inode_permission,   /* permission */
	pseudo_inode_smap,     /* smap */
	pseudo_inode_updatepage,     /* updatepage */
	pseudo_inode_revalidate,   /* revalidate */
};


/* don't think i want to call register_filesystem.
* pseudo_fs has no read_super, because nobody can
* see it except nfsd.
*/
static struct file_system_type pseudo_fs_type = {
	"nfsd_pseudo_fs",
	0,
	NULL,
	NULL
};


/*
* 1) get a super_block. initialize only what you have to. the pseudo fs
* has no brown and round, and is not plugged into the VFS layer.
*
*	2) get an inode. initialize it to be the root inode. 
* 
* 3) get a dentry for the root inode and stuff super_block.root_inode.
* 
*	4) for each exported directory in /etc/exports
*			create pseudo_fs inode as a directory.
*			add to pseudo root
* 		use as a mount point for exported directory (i.e. mount the sucker).
*/
void
create_pseudo_fs(void)
{
	struct inode *inode;
	struct dentry *dentry;
	int i,err;

	printk("ANDROS CREATE PSEUDO\n");
	nfsv4_printk(pseudo,"NFSV4 create_pseudo_fs\n");
	sbp = get_empty_super();
	if(sbp == NULL){
		nfsv4_printk(pseudo,"NFSV4 create_pseudo_fs UNABLE TO GET SUPER BLOCK n");
	}
	sbp->s_dev = get_unnamed_dev();
	nfsv4_printk(pseudo,"NFSV4 create_pseudo_fs dev: major %d minor %d\n",
		     MAJOR(sbp->s_dev),MINOR(sbp->s_dev));

	/* no device. but i believe we need a blksize to get
	 * through code checks. 
	 */ 
	sbp->s_blocksize = 512;
	sbp->s_blocksize_bits = 0;

	sbp->s_dirt = 0;
	sbp->s_rd_only = 0; /* i don't see this used anywhere...*/

	/* see comments in fs.h, struct super_block typedef 
	 * i don't know if i need this yet.
	 */
	sema_init(&sbp->s_nfsd_free_path_sem,1);

	/* since no vfs layer hookup, do i need this?*/
	sbp->s_time =0; /* time of last super_block update*/

	/* i need to more fully understand these flags. this seems
	 * a reasonable first guess.
	 */
	sbp->s_flags = MS_RDONLY || MS_NOEXEC || MS_NOSUID || MS_NODEV || MS_NOATIME || MS_NODIRATIME;

	sbp->s_type = &pseudo_fs_type;
	sbp->s_magic = 0xface;	/* version */
	sbp->s_op = &pseudo_fs_sops;

	/* create root inode */
	inode = pseudo_make_inode(sbp);

	/* place . and .. into root inode */
	init_pseudo_inode_dots(inode,inode);

	/* create pseudoroot dentry */
	dentry = d_alloc_root(inode,NULL);
	if(dentry == NULL){
		nfsv4_printk(pseudo,"NFSV4 create_pseudo_fs UNABLE TO GET ALLOC DENTRY\n");
	}
	dentry->d_sb = sbp;
	dentry->d_inode = inode;
	dentry->d_flags |= DCACHE_NFSDV4_PSEUDO_FS; 
	sbp->s_root = dentry;
	
	/* create pseudoroot export */
	err = nfsd_pseudo_exp_init(dentry);
	nfsv4_printk(pseudo,"NFSV4 create_pseudo_fs returning: err %d\n",err);

	print_pseudo_dentry(dentry);
	/* initialize the fs_entry pointer array used by shutdown */
	for(i=0;i<TOTAL_NFSV4_PSEUDO_ENTRIES;i++)
		all_fs_entries[i] = NULL;
	/* initialize the inode pointer array used by pseudo_iget and shutdown */
	for(i=0;i<TOTAL_NFSV4_PSEUDO_INODES;i++)
		all_pseudo_inodes[i] = NULL;
		
	/* place the root inode in the inode array*/
	all_pseudo_inodes[0] = inode;
}

int
is_pseudo(kdev_t dev)
{
	if(dev == sbp->s_dev)
		return 1;
	return 0;
}

static void
init_pseudo_inode_dots(struct inode * dir_inode, struct inode *inode)
{
	struct nfsd_pseudo_inode_info *ptr;

	nfsv4_printk(pseudo,"NFSV4 init_pseudo_inode_dots\n");

	ptr = &dir_inode->u.nfsd_pseudo_i;
	ptr->dot.inode = inode->i_ino;
	ptr->dot.name_len = 1;
	ptr->dot.name[0] = '.';
	ptr->dot.name[1] = 0;
	ptr->dot.name[2] = 0;
	ptr->dot.name[3] = 0;

	ptr->dot_dot.inode = inode->i_ino;
	ptr->dot_dot.name_len = 1;
	ptr->dot_dot.name[0] = '.';
	ptr->dot_dot.name[1] = '.';
	ptr->dot_dot.name[2] = 0;
	ptr->dot_dot.name[3] = 0;
	
	inode->i_size += 3; /* name lengths */
	nfsv4_printk(pseudo,"NFSV4 init_pseudo_inode_dots [%s] [%s]\n",ptr->dot.name,ptr->dot_dot.name);
}

static void
print_pseudo_entry(pseudo_fs_entry *ep)
{
	nfsv4_printk(pseudo,"NFSV4 print_pseudo_entry\n");
	if(!ep)
		return;
	nfsv4_printk(pseudo,"        inode number: %d\n",ep->inode);
	nfsv4_printk(pseudo,"        name len: %d\n",ep->name_len);
	if(ep->name)
		nfsv4_printk(pseudo,"        name: [%s]\n",ep->name);
	else
		nfsv4_printk(pseudo,"        name is NULL\n");
}

void
print_pseudo_dentry(struct dentry *dentry)
{
	nfsv4_printk(pseudo,"NFSV4 print_pseudo_dentry\n");
	if(!dentry){
		nfsv4_printk(pseudo,"NFSV4 print_pseudo_dentry NULL dentry\n");
		return;
	}

	nfsv4_printk(pseudo,"   d_count: %d\n",dentry->d_count);
	nfsv4_printk(pseudo,"   d_inode: %p\n",dentry->d_inode);
	if(dentry->d_inode)
		nfsv4_printk(pseudo,"   d_inode->i_ino: %ld\n",dentry->d_inode->i_ino);
	nfsv4_printk(pseudo,"   d_parent: %p\n",dentry->d_parent);
	nfsv4_printk(pseudo,"   d_mounts: %p\n",dentry->d_mounts);
	nfsv4_printk(pseudo,"   d_covers: %p\n",dentry->d_covers);
	nfsv4_printk(pseudo,"   d_sb: %p\n",dentry->d_sb);
	nfsv4_printk(pseudo,"   d_name.len: %d\n",dentry->d_name.len);
	if(dentry->d_name.len > 0)
		nfsv4_printk(pseudo,"   d_name.name[0]: [%c]\n",dentry->d_name.name[0]);

}
void
print_pseudo_inode(struct inode *inode)
{
	int i;
	pseudo_fs_entry *ep;

	nfsv4_printk(pseudo,"NFSV4 print_pseudo_inode\n");
	if(!inode){
		nfsv4_printk(pseudo,"NFSV4 print_pseudo_inode NULL inode\n");
		return;
	}

	nfsv4_printk(pseudo,"   i_sb: %p\n",inode->i_sb);
	nfsv4_printk(pseudo,"   i_dev: (%d/%d)\n",MAJOR(inode->i_dev),MINOR(inode->i_dev));
	nfsv4_printk(pseudo,"   i_flags: %d\n", inode->i_flags);
	nfsv4_printk(pseudo,"   i_ino: %ld\n", inode->i_ino);
	nfsv4_printk(pseudo,"   i_count: %d\n", inode->i_count);
	nfsv4_printk(pseudo,"   i_blksize: %ld\n",inode->i_blksize);
	nfsv4_printk(pseudo,"   i_mode: %d\n",inode->i_mode);
	nfsv4_printk(pseudo,"   i_rdev: (%d/%d)\n",MAJOR(inode->i_rdev),MINOR(inode->i_rdev));
	nfsv4_printk(pseudo,"   i_op: %p\n",inode->i_op);
	nfsv4_printk(pseudo,"   i_size: %ld\n",inode->i_size);
	nfsv4_printk(pseudo,"   i_mtime: %ld\n",inode->i_mtime);
	nfsv4_printk(pseudo,"   i_ctime: %ld\n",inode->i_ctime);
	nfsv4_printk(pseudo,"   i_atime: %ld\n",inode->i_atime);

	for(i=0;i<MAX_NFSV4_PSEUDO_ENTRIES;i++){
		ep = inode->u.nfsd_pseudo_i.fs_entry[i];
		if(ep == NULL)
			break;
		print_pseudo_entry(ep);
	}
}

/* i don't think there are any races here.. 
* maybe on the global next_pseudo_i_num. 
* for now, this is only called at nfsd_init.
* later on, as the /etc/exports changes, exportfs
* should cause pseudo_fs inodes to be created or
* distroyed.
*
* NOTE: maybe we should not hash these inodes - 
* we want them to exist as long as nfsd is up. if they get iput,
* we loose the pseudo fs! 
*/
static struct inode *
pseudo_make_inode(struct super_block *sb)
{
	struct inode *inode = get_empty_inode();
	int err = 0;

	nfsv4_printk(pseudo,"NFSV4 pseudo_make_inode\n");
	if(!inode){
		nfsv4_printk(pseudo,"NFSV4 pseudo_make_inode - get_empty_inode() retuns NULL\n");
		return NULL;
	}
	if(!sb){
		nfsv4_printk(pseudo,"NFSV4 pseudo_make_inode - NULL superblock pointer\n");
		return NULL;
	}
	inode->i_sb = sb;
	inode->i_dev = sb->s_dev;
	inode->i_flags |= DCACHE_NFSDV4_PSEUDO_FS;

/* for now, just blindly increment the inode num.
 * eventually, will want to reap unused numbers...
 * race condition here? how do real fs's do this..
 */
	inode->i_ino = next_pseudo_i_num++;

/* clients want to calculate space used...*/
	inode->i_blksize = inode->i_sb->s_blocksize;
	inode->i_blocks = 1;

/* want the mode to be readable by the world, and 
 * be a directory.
 */
	inode->i_mode = 0;
	inode->i_mode |= S_IFDIR;
	inode->i_mode |= S_IRUGO;
	inode->i_mode |= S_IXUGO;

/* virtural fs, rdev = i_dev */
	inode->i_rdev = sb->s_dev;

	inode->i_op = &pseudo_dir_inode_operations;
	inode->i_size = 2 * sizeof(pseudo_fs_entry_dot);
	inode->i_mtime = inode->i_ctime = inode->i_atime = CURRENT_TIME;

/* place inode into inode ptr array */
	err = set_next_inode(inode);
	if(err)
		return NULL;

	nfsv4_printk(pseudo,"NFSV4: pseudo_make_inode done\n");
	print_pseudo_inode(inode);
	return inode;
}

/* the only reason this is here is because
* get_new_inode() wants it to be.
*/
static void pseudo_read_inode(struct inode *inode)
{
	nfsv4_printk(pseudo,"NFSV4 pseudo_read_inode\n");
}

static void pseudo_put_inode(struct inode *inode)
{
	nfsv4_printk(pseudo,"NFSV4 pseudo_put_inode inode->i_count %d inode->i_nlink %d\n",inode->i_count,inode->i_nlink);
/*
  need to keep the inodes around - they have all the data! 
  if(inode->i_count == 1)
  inode->i_nlink = 0;
*/
}

/* delete a pseudo_fs inode */
static void pseudo_delete_inode(struct inode *inode)
{
/*read only fs, so just clean */
	clear_inode(inode);
}

int
pseudo_exp_it(struct dentry *ext2dentry, char *ex_path)
{
	int err = 0;
	char *path_comp;
	struct inode *next_dir_inode = NULL, *dir_inode = NULL;
	struct dentry *dentry = NULL,*dir_dentry = NULL;
	char path[256];
	struct qstr this;

	nfsv4_printk(pseudo,"NFSV4 pseudo_exp_it ex_path %s calling iget\n",ex_path);
	/* get the root inode */
	dir_inode = pseudo_iget(2);

	if(!dir_inode){
		nfsv4_printk(pseudo,"NFSV4 pseudo_exp_it iget failed\n");
		return -ENOENT;
	}
	print_pseudo_inode(dir_inode);

	/* start with the already allocated root inode dentry */
	dir_dentry = sbp->s_root;

	/* parse ex_path, and either find or create
	 * all entries in pseudo fs
	 */
	memset(path,0,NAME_MAX);
	memcpy(path,ex_path,strlen(ex_path));
	nfsv4_printk(pseudo,"NFSV4 pseudo_exp_it path: %s\n",path);
	path_comp = strtok(path,"/");	
	while(path_comp){

#if 0
		nfsv4_printk(pseudo,"NFSV4 pseudo_exp_it dir_inode->u.nfsd_pseudo_i.fs_entry[0]: %p\n",dir_inode->u.nfsd_pseudo_i.fs_entry[0]);
#endif 

		next_dir_inode = pseudofs_get_dir(dir_inode,path_comp);
		if(next_dir_inode == NULL){
			nfsv4_printk(pseudo,"NFSV4 pseudo_exp_it pseudofs_get_dir ERROR %d\n",err);
			break;
		}
		print_pseudo_inode(next_dir_inode);

		/* stuff a qstr - actually, pseudo_fs_entry should have a qstr... */
		this.name = path_comp;
		this.len = strlen(path_comp);
		this.hash = full_name_hash(this.name,this.len);                           

		dentry = d_alloc(dir_dentry,&this);
		if(!dentry)
			return nfserr_resource;
		d_instantiate(dentry,next_dir_inode);
		dentry->d_parent = dir_dentry;
		dentry->d_sb = sbp;
		dentry->d_flags |= DCACHE_NFSDV4_PSEUDO_FS;
		print_pseudo_dentry(dentry);

		/* next please....*/
		path_comp = strtok(NULL,"/");	
		dir_inode = next_dir_inode;
		dir_dentry = dentry;
		next_dir_inode = NULL;
		dentry = NULL;
	}

	/* dir_inode/dir_dentry is now the pseudo inode/dentry of the 
	 * directory to export it gets the real directory mounted on it.
	 */

	/*
	 * mount dentries (in super.c).
	 * static void d_mount(struct dentry *covered, struct dentry *mounted) 
	 * 
	 * covered: pseudo fs dentry (dir_dentry)
	 * mounted: real ext2 dentry of exported ext2 directory.
	 * 
	 */
	

	if(dir_dentry->d_mounts != dir_dentry){
		nfsv4_printk(pseudo,"NFSV4: pseudo_exp_it: pseudo dentry %p already mounted\n",dir_dentry);
		return nfserr_noent;
	}
	dir_dentry->d_mounts =  ext2dentry; 
	ext2dentry->d_covers = dir_dentry;
	print_pseudo_dentry(dir_dentry);
	print_pseudo_dentry(ext2dentry);

	return err;
}


static int
add_path(pseudo_fs_entry *ep, char *path_comp,unsigned long ino)
{
	nfsv4_printk(pseudo,"NFSV4 add_path  fs_entry %p path: [%s], ino %ld\n",
		     ep,path_comp,ino);

	if(!ep)
		return nfserr_noent;
	ep->name = kmalloc(ep->name_len+1,GFP_KERNEL);
	if(!ep->name)
		return nfserr_resource;
	ep->name_len = strlen(path_comp);
	memcpy(ep->name,path_comp,ep->name_len);
	ep->name[ep->name_len] = '\0';
	ep->inode = ino;
	return nfs_ok;
}

/* read the directory ps_inode looking for path_comp.
* if found, return inode. else, create inode
*/

static struct inode * 
pseudofs_get_dir(struct inode *dir_inode,char *path_comp)
{
	int i,err;
	pseudo_fs_entry *ep;
	struct inode *next_ps_inode = NULL;

	nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir path_comp [%s]  dir ino %ld\n",path_comp, dir_inode->i_ino);
	if(!dir_inode)
		return NULL;

	for(i=0;i<MAX_NFSV4_PSEUDO_ENTRIES;i++){
		ep = dir_inode->u.nfsd_pseudo_i.fs_entry[i];

		nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir dir_ino %ld dir_inode->u.nfsd_pseudo_i.fs_entry[%d] %p\n",dir_inode->i_ino,i,dir_inode->u.nfsd_pseudo_i.fs_entry[i]);

		if(ep == NULL){
			/* not found, create it */

			nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir CREATING name [%s] in dir %ld\n",path_comp, dir_inode->i_ino);
			ep = kmalloc(sizeof(pseudo_fs_entry),GFP_KERNEL);

			if(ep == NULL)
				return NULL;
			dir_inode->u.nfsd_pseudo_i.fs_entry[i] = ep;

			nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir dir_inode->u.nfsd_pseudo_i.fs_entry[%d] ONE%p\n",i,dir_inode->u.nfsd_pseudo_i.fs_entry[i]);

			set_next_fs_entry(ep);
			nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir dir_inode->u.nfsd_pseudo_i.fs_entry[%d] ONE&1/2 %p\n",i,dir_inode->u.nfsd_pseudo_i.fs_entry[i]);

			memset(ep,0,sizeof(pseudo_fs_entry));

			/* create inode */
			next_ps_inode = pseudo_make_inode(sbp); 
			if(!next_ps_inode){
				nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir pseudo_make_inode failed\n");
				return NULL;
			}

			/* set . and .. */
			init_pseudo_inode_dots(dir_inode,next_ps_inode);
			nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir dir_inode->u.nfsd_pseudo_i.fs_entry[%d] ONE&3/4 %p\n",i,dir_inode->u.nfsd_pseudo_i.fs_entry[i]);

			nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir inode from pseudo_make_inode:\n");
			print_pseudo_inode(next_ps_inode);

			nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir next_ps_inode->i_count %d\n",next_ps_inode->i_count);
			/* place path_comp into dir_inode */
			err = add_path(ep,path_comp,next_ps_inode->i_ino);
			if(err != 0){
				nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir add_path failed - err = %d return NULL\n",err);
				iput(next_ps_inode);
				return NULL;
			}
			/* update inode size */
			next_ps_inode->i_size += sizeof(struct pseudo_fs_entry) + ep->name_len;

			nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir i %d dir_inode->u.nfsd_pseudo_i.fs_entry[i] TWO%p\n",i,dir_inode->u.nfsd_pseudo_i.fs_entry[i]);

			break;
		}

		nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir ep NOT NULL. name [%s] in dir %ld\n",path_comp, dir_inode->i_ino);

		if(strlen(path_comp) == ep->name_len){
			if(!strncmp(path_comp,ep->name,ep->name_len)){
				/* found it. */
				nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir FOUND name [%s] in dir %ld\n",path_comp, dir_inode->i_ino);
				nfsv4_printk(pseudo,"NFSV4 pseudo_get_dir calling pseudo_iget\n");
				next_ps_inode = pseudo_iget(ep->inode); 
				if(!next_ps_inode){
					nfsv4_printk(pseudo,"NFSV4 pseudofs_get_dir pseudo_iget failed\n");
					return NULL;
				}
				break;
			}
		}
	}
	return next_ps_inode;
}

/* all_fs_entries[] is for kfree to have an easy time of it. */
static int
set_next_inode(struct inode *inode)
{
	int i;

	for(i=0;i<TOTAL_NFSV4_PSEUDO_INODES;i++){
		if(all_pseudo_inodes[i] == NULL){
			nfsv4_printk(pseudo,"NFSV4 set_next_inode setting inode index: %d\n",i);
			all_pseudo_inodes[i] = inode;
			return nfs_ok;
		}
	}
	return nfserr_resource;
}

/* all_fs_entries[] is for kfree to have an easy time of it. */
static int
set_next_fs_entry(pseudo_fs_entry *ep)
{
	int i;

	for(i=0;i<TOTAL_NFSV4_PSEUDO_ENTRIES;i++){
		if(all_fs_entries[i] == NULL){
			nfsv4_printk(pseudo,"NFSV4 set_next_fs_entry setting fs_entry index: %d\n",i);
			all_fs_entries[i] = ep;
			return nfs_ok;
		}
	}
	return nfserr_resource;
}

#if 0
/* walk the i_dentry list and dput pseudo_fs dentries. this
*dies a horrible death - i need to figure out if the pseudo
*dentries are left dangling, or if i need to do something like
*this function intended
 */
void
pseudo_put_dentries(struct inode *inode)
{
	struct list_head *lp;
	struct dentry *dentry;

	for (lp = inode->i_dentry.next; lp != &inode->i_dentry ; lp=lp->next) {
		dentry = list_entry(lp,struct dentry, d_alias);
		nfsv4_printk(pseudo,"NFSV4 find_peudo_dentry found dentry: \n");
		print_pseudo_dentry(dentry);
		while(dentry->d_count !=0)
			dput(dentry);
	}                                                                  
}
#endif /*0*/
/* walk the array of fs_entry pointers to 
*  free all kmalloc'ed fs_entry's and their names.
*/
void
shutdown_pseudo_fs()
{
	int i;

/*  
* do iputs at shutdown - start from next_pseudo_i_num and work down to 2.
*		iput(dir_inode);
*/
	nfsv4_printk(pseudo,"NFSV4 shutdown_pseudo_fs\n");

	/* return virtual device */
	put_unnamed_dev(sbp->s_dev);

	for(i=0;i<TOTAL_NFSV4_PSEUDO_ENTRIES;i++){
		if(all_fs_entries[i] != NULL){
			nfsv4_printk(pseudo,"NFSV4 shutdown_pseudo_fs calling kfree on fs_entry index: %d\n",i);
			kfree(all_fs_entries[i]->name);
			kfree(all_fs_entries[i]);
			all_fs_entries[i] = NULL;
		}
	}
/* walk each inode's i_dentry list and dput all dentries 
* then 'delete' inodes
*/
	for(i=0;i<TOTAL_NFSV4_PSEUDO_INODES;i++){
		if(all_pseudo_inodes[i] != NULL){
#if 0
			(pseudo_put_dentries(all_pseudo_inodes[i]);
#endif /*0*/
			 nfsv4_printk(pseudo,"NFSV4 shutdown_pseudo_fs calling pseudo_delete_inode on ino: %ld\n",all_pseudo_inodes[i]->i_ino);
			 pseudo_delete_inode(all_pseudo_inodes[i]);
			 all_pseudo_inodes[i] = NULL;
			 }
		}
	}

	struct inode *
		pseudo_lookup(struct inode *iparent,char *name) 
		{
			struct inode *inode;
			struct nfsd_pseudo_inode_info *ptr;
			int i;

			nfsv4_printk(pseudo,"NFSV4 pseudo_lookup for [%s]\n",name);

			if(!iparent){
				nfsv4_printk(pseudo,"NFSV4 pseudo_lookup: no parent inode\n");
				return ERR_PTR(-ENOENT);
			}
			nfsv4_printk(pseudo,"      in inode %ld\n",iparent->i_ino);

			if(!strcmp(name,".")){
				nfsv4_printk(pseudo,"NFSV4 pseudo_lookup DOT\n");
				return iparent;
			}
			ptr = &iparent->u.nfsd_pseudo_i;        
			if(!strcmp(name,"..")){
				nfsv4_printk(pseudo,"NFSV4 pseudo_lookup DOT_DOT\n");
				inode = pseudo_iget(ptr->dot_dot.inode);
				if(!inode){
					nfsv4_printk(pseudo,"NFSV4 pseudo_lookup no child inode\n");
					return ERR_PTR(-ENOENT);
				}
				return inode;	
			}
			for(i=0;i<MAX_NFSV4_PSEUDO_ENTRIES;i++){    
				nfsv4_printk(pseudo,"NFSV4 pseudo_lookup 1\n");
				if(ptr->fs_entry[i]){
					nfsv4_printk(pseudo,"NFSV4 pseudo_lookup 2\n");
					if(!strcmp(ptr->fs_entry[i]->name,name)){
						nfsv4_printk(pseudo,"NFSV4 pseudo_lookup 3\n");
						inode =pseudo_iget(ptr->fs_entry[i]->inode);
						if(!inode){
							nfsv4_printk(pseudo,"NFSV4 pseudo_lookup no child inode\n");
							return ERR_PTR(-ENOENT);
						}
						print_pseudo_inode(inode);
						return inode;
					}
				}
			}
			return ERR_PTR(-ENOENT);
		}

	struct inode *
		pseudo_iget(unsigned int ino)
		{
			int i;
			nfsv4_printk(pseudo,"NFSV4 pseudo_iget\n");

			for(i=0;i<TOTAL_NFSV4_PSEUDO_INODES;i++){
				if(!all_pseudo_inodes[i])
					break;
				if(all_pseudo_inodes[i]->i_ino == ino){
					nfsv4_printk(pseudo,"NFSV4 pseudo_iget FOUND ino %u at index: %d\n",ino,i);
					return all_pseudo_inodes[i];
				}
			}
			nfsv4_printk(pseudo,"NFSV4 pseudo_iget FAILED to FIND inode ino: %u\n",ino);
			return NULL;
		}


	int 
		pseudo_open(struct inode *inode, struct file *filp)
		{
			nfsv4_printk(pseudo,"NFSV4: pseudo_open\n");
			if((filp->f_dentry) && (filp->f_dentry->d_iname))
				nfsv4_printk(pseudo,"NFSV4: pseudo_open %s \n",filp->f_dentry->d_iname);
			return 0;
		}


/* pseudo fs doesn't allow any operation beyond lookup
* open and readdir
*/
	loff_t 
		pseudo_llseek(struct file *file, loff_t offset, int flag)
		{
			printk("NFSV4: pseudo_llseek\n");
			return -EACCES;
		}

	ssize_t 
		pseudo_read(struct file *file, char *name, size_t count, loff_t *offset)
		{
			printk("NFSV4: pseudo_read\n");
			return -EACCES;
		}

	ssize_t
		pseudo_write(struct file *file, const char *name, size_t count, loff_t *offset)
		{
			printk("NFSV4: pseudo_write\n");
			return -EACCES;
		}

	unsigned int
		pseudo_poll(struct file *file, struct poll_table_struct *pt)
		{
			printk("NFSV4: pseudo_poll\n");
			return -EACCES;
		}

	int
		pseudo_ioctl(struct inode *inode, struct file *file, unsigned int which, unsigned long why)
		{
			printk("NFSV4: pseudo_ioctl\n");
			return -EACCES;
		}

	int
		pseudo_mmap(struct file *file, struct vm_area_struct *vma)
		{
			printk("NFSV4: pseudo_mmap\n");
			return -EACCES;
		}

	int
		pseudo_flush(struct file *file)
		{
			printk("NFSV4: pseudo_flush\n");
			return -EACCES;
		}

	int
		pseudo_release(struct inode *inode, struct file *file)
		{
			printk("NFSV4: pseudo_release\n");
			return -EACCES;
		}

	int
		pseudo_fsync(struct file *file, struct dentry *dentry)
		{
			printk("NFSV4: pseudo_fsync\n");
			return -EACCES;
		}

	int
		pseudo_fasync(int dummy, struct file *file, int flag)
		{
			printk("NFSV4: pseudo_fasync\n");
			return -EACCES;
		}

	int
		pseudo_check_media_change(kdev_t dev)
		{
			printk("NFSV4: pseudo_check_media\n");
			return -EACCES;
		}

	int
		pseudo_revalidate(kdev_t dev)
		{
			printk("NFSV4: pseudo_revalidate\n");
			return -EACCES;
		}

	int
		pseudo_lock(struct file *file, int flag, struct file_lock *flock)
		{
			printk("NFSV4: pseudo_lock\n");
			return -EACCES;
		}


/* call the filldir function with name, namelen and ino
* of all the inode entries.
*
* filep->f_pos will equal the index into the fs_entries
* 	i.e. f_pos = 1 -> start at dot_dot
*				f_pos = 2 -> start at fs_entries[0]
*				f_pos = n, n > 1 -> start at fs_entries[n-2]
*/
	int 
		pseudo_readdir(struct file *filep, void *cd, filldir_t filldir)
		{
			int i,err = 0, index = 0;
			pseudo_fs_entry *ep = NULL;
			struct inode *inode = filep->f_dentry->d_inode;

			nfsv4_printk(pseudo,"NFSV4: pseudo_readdir\n");
			if(!inode){
				nfsv4_printk(pseudo,"NFSV4: pseudo_readdir NO INODE\n");
				return -ENOENT;	
			}
			print_pseudo_inode(inode);
			if(!is_pseudo(inode->i_dev)){
				nfsv4_printk(pseudo,"NFSV4: pseudo_readdir NOT PSEUDO INODE\n");
				return -ENOENT;	
			}

			index = filep->f_pos;
			for(i=index;i<MAX_NFSV4_PSEUDO_ENTRIES+2;i++){
				if(i==0){
					nfsv4_printk(pseudo,"NFSV4: pseudo_readdir calling filldir on %s, inode %d, with f_pos %d\n",inode->u.nfsd_pseudo_i.dot.name,inode->u.nfsd_pseudo_i.dot.inode,i+1);
					err = filldir(cd,inode->u.nfsd_pseudo_i.dot.name,
						      inode->u.nfsd_pseudo_i.dot.name_len,
						      i+1,inode->u.nfsd_pseudo_i.dot.inode);
					filep->f_pos = i+1;
					if(err != 0)
						break;
					continue;
				}
				if(i==1){
					nfsv4_printk(pseudo,"NFSV4: pseudo_readdir calling filldir on %s, inode %d, with f_pos %d\n",inode->u.nfsd_pseudo_i.dot_dot.name,inode->u.nfsd_pseudo_i.dot_dot.inode,i+1);
					err = filldir(cd,inode->u.nfsd_pseudo_i.dot_dot.name,
						      inode->u.nfsd_pseudo_i.dot_dot.name_len,
						      i+1,inode->u.nfsd_pseudo_i.dot_dot.inode);
					filep->f_pos = i+1;
					if(err != 0)
						break;
					continue;
				}
				if(i > 1){	
					nfsv4_printk(pseudo,"NFSV4 pseudo_readdir i %d\n",i);
					print_pseudo_inode(inode);
					ep = inode->u.nfsd_pseudo_i.fs_entry[i-2];
				}
				if(ep){
					nfsv4_printk(pseudo,"NFSV4: pseudo_readdir calling filldir on %s, inode %d, with f_pos %d\n",ep->name,ep->inode,i+1);
					err = filldir(cd,ep->name,ep->name_len,i+1,ep->inode);
					filep->f_pos = i+1;
					if(err != 0)
						break;
					continue;
				}
				else if(!ep){
					nfsv4_printk(pseudo,"NFSV4: pseudo_readdir no more entries index %d \n",i);
					nfsv4_printk(pseudo,"NFSV4: pseudo_readdir calling filldir with name == 0 and f_pos %d\n",i+1);
					err = filldir(cd,0,0,i+1,0);
					filep->f_pos = i+1;
					((struct readdir_cd *)cd)->eob = 0;
					break;
				}
			}
			return err;
			;}


/* dummy inode ops */
	int 
		pseudo_inode_create(struct inode *ino,struct dentry *den,int dummy)
		{
			printk("NFSV$: pseudo_inode_create\n");
			return -EACCES;
		}
	struct dentry * 
		pseudo_inode_lookup(struct inode *ino,struct dentry *den)
		{
			printk("NFSV$: pseudo_inode_lookup\n");
			return ERR_PTR(-EACCES);
		}
	int 
		pseudo_inode_link(struct dentry *den,struct inode *ino,struct dentry *den2)
		{
			printk("NFSV$: pseudo_inode_link\n");
			return -EACCES;
		}
	int 
		pseudo_inode_unlink(struct inode *ino,struct dentry *den)
		{
			printk("NFSV$: pseudo_inode_unlink\n");
			return -EACCES;
		}
	int 
		pseudo_inode_symlink(struct inode *ino,struct dentry *den,const char *str)
		{
			printk("NFSV$: pseudo_inode_symlink\n");
			return -EACCES;
		}
	int 
		pseudo_inode_mkdir(struct inode *ino,struct dentry *den,int dummy)
		{
			printk("NFSV$: pseudo_inode_mkdir\n");
			return -EACCES;
		}
	int 
		pseudo_inode_rmdir(struct inode *ino,struct dentry *den)
		{
			printk("NFSV$: pseudo_inode_rmdir\n");
			return -EACCES;
		}
	int 
		pseudo_inode_mknod(struct inode *ino,struct dentry *den,int dummy,int dummy2)
		{
			printk("NFSV$: pseudo_inode_mknod\n");
			return -EACCES;
		}
	int 
		pseudo_inode_rename(struct inode *ino, struct dentry *den, struct inode *ino2, struct dentry *den2)
		{
			printk("NFSV$: pseudo_inode_rename\n");
			return -EACCES;
		}
	int 
		pseudo_inode_readlink(struct dentry *den, char *str,int dummy)
		{
			printk("NFSV$: pseudo_inode_readlink\n");
			return -EACCES;
		}
	struct dentry * 
		pseudo_inode_follow_link(struct dentry *den, struct dentry *den2, unsigned int dummy)
		{
			printk("NFSV$: pseudo_inode_follow_link\n");
			return ERR_PTR(-EACCES);
		}
	int 
		pseudo_inode_readpage(struct file *filp, struct page *page)
		{
			printk("NFSV$: pseudo_inode_readpage\n");
			return -EACCES;
		}
	int 
		pseudo_inode_writepage(struct file *filp, struct page *page)
		{
			printk("NFSV$: pseudo_inode_writepage\n");
			return -EACCES;
		}
	int 
		pseudo_inode_bmap(struct inode *ino,int dummy)
		{
			printk("NFSV$: pseudo_inode_bmap\n");
			return -EACCES;
		}
	void 
		pseudo_inode_truncate(struct inode *ino)
		{
			printk("NFSV$: pseudo_inode_truncate\n");
		}
	int 
		pseudo_inode_permission(struct inode *ino, int acc)
		{
			printk("NFSV$: pseudo_inode_permission\n");
			return 0;
		}
	int 
		pseudo_inode_smap(struct inode *ino,int dummy)
		{
			printk("NFSV$: pseudo_inode_smap\n");
			return -EACCES;
		}
	int 
		pseudo_inode_updatepage(struct file *filp, struct page *page, const char *str, unsigned long dummy, unsigned int dummy2, int dummy3)
		{
			printk("NFSV$: pseudo_inode_updatepage\n");
			return -EACCES;
		}
	int 
		pseudo_inode_revalidate(struct dentry *den)
		{
			nfsv4_printk(pseudo,"NFSV$: pseudo_inode_revalidate\n");
			return -EACCES;
		}
