/*
 * linux/fs/nfsd/vfs.c
 *
 * File operations used by nfsd. Some of these have been ripped from
 * other parts of the kernel because they weren't in ksyms.c, others
 * are partial duplicates with added or changed functionality.
 *
 * Note that several functions dget() the dentry upon which they want
 * to act, most notably those that create directory entries. Response
 * dentry's are dput()'d if necessary in the release callback.
 * So if you notice code paths that apparently fail to dput() the
 * dentry, don't worry--they have been taken care of.
 *
 * Copyright (C) 1995-1999 Olaf Kirch <okir@monad.swb.de>
 *
 * NFSV4 Implementation
 *      by Andy Adamson <abdros@citi.umich.edu>
 *      Copyright (C) 1999 The Regents of the University of Michigan
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/locks.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/net.h>
#include <linux/unistd.h>
#include <linux/malloc.h>
#include <linux/in.h>

#include <linux/sunrpc/svc.h>
#ifdef NFSV4
#include <linux/nfsd/nfsd4_prot.h>
#include <linux/nfsd/n2s_cache.h>
#include <linux/nfsd/nfsfh.h>
#else /*NFSV4*/
#include <linux/nfsd/nfsd.h>
#endif /*NFSV4*/
#ifdef CONFIG_NFSD_V3
#include <linux/nfs3.h>
#include <linux/nfsd/xdr3.h>
#endif /* CONFIG_NFSD_V3 */
#include <linux/nfsd/nfsfh.h>
#include <linux/quotaops.h>

#if LINUX_VERSION_CODE >= 0x020100
#include <asm/uaccess.h>
#endif

#define NFSDDBG_FACILITY		NFSDDBG_FILEOP
#define NFSD_PARANOIA


/* We must ignore files (but only files) which might have mandatory
 * locks on them because there is no way to know if the accesser has
 * the lock.
 */
/* MANDATORY_LOCK taken from 2.3 */
#define MANDATORY_LOCK(inode) \
	(IS_MANDLOCK(inode) && ((inode)->i_mode & (S_ISGID | S_IXGRP)) == S_ISGID)
#define IS_ISMNDLK(i)	(S_ISREG((i)->i_mode) && MANDATORY_LOCK(i))
/* Time difference margin in seconds for comparison. It is a
   dynamically-tunable parameter via /proc/fs/nfs/time-diff-margin.
 */
extern long nfsd_time_diff_margin;

/* Check for dir entries '.' and '..' */
#define isdotent(n, l)	(l < 3 && n[0] == '.' && (l == 1 || n[1] == '.'))

/*
 * This is a cache of readahead params that help us choose the proper
 * readahead strategy. Initially, we set all readahead parameters to 0
 * and let the VFS handle things.
 * If you increase the number of cached files very much, you'll need to
 * add a hash table here.
 */
struct raparms {
	struct raparms		*p_next;
	unsigned int		p_count;
	ino_t			p_ino;
	dev_t			p_dev;
	unsigned long		p_reada,
				p_ramax,
				p_raend,
				p_ralen,
				p_rawin;
};

int nfsd_nservers = 0;
#define FILECACHE_MAX		(2 * nfsd_nservers) 
static struct raparms *		raparml = NULL;
static struct raparms *		raparm_cache = NULL;

#ifdef NFSV4
/* supported attributes. see nfsd4proc.c */
extern bitmap4 ATTR_BITMAP;

static u32      nfs4_ftypes[] = {
        NF4NON,  NF4FIFO, NF4CHR, NF4BAD,
        NF4DIR,  NF4BAD,  NF4BLK, NF4BAD,
        NF4REG,  NF4BAD,  NF4LNK, NF4BAD,
        NF4SOCK, NF4BAD,  NF4LNK, NF4BAD,
};
#endif /*NFSV4*/

/*
 * We need to do a check-parent every time
 * after we have locked the parent - to verify
 * that the parent is still our parent and
 * that we are still hashed onto it..
 *
 * This is required in case two processes race
 * on removing (or moving) the same entry: the
 * parent lock will serialize them, but the
 * other process will be too late..
 *
 * Note that this nfsd_check_parent is identical
 * the check_parent in linux/fs/namei.c.
 */
#define nfsd_check_parent(dir, dentry) \
	((dir) == (dentry)->d_parent && !list_empty(&dentry->d_hash))

/*
 * Lock a parent directory following the VFS locking protocol.
 */
int
fh_lock_parent(struct svc_fh *parent_fh, struct dentry *dchild)
{
	fh_lock(parent_fh);
	/*
	 * Make sure the parent->child relationship still holds,
	 * and that the child is still hashed.
	 */
	if (nfsd_check_parent(parent_fh->fh_dentry, dchild))
	    return 0;

	printk(KERN_WARNING
		"fh_lock_parent: %s/%s parent changed or child unhashed\n",
		dchild->d_parent->d_name.name, dchild->d_name.name);

	fh_unlock(parent_fh);
	return nfserr_noent;
}


/*
 * Look up one component of a pathname.
 * N.B. After this call _both_ fhp and resfh need an fh_put
 *
 * If the lookup would cross a mountpoint, and the mounted filesystem
 * is exported to the client with NFSEXP_CROSSMNT, then the lookup is
 * accepted as it stands and the mounted directory is
 * returned. Otherwise the covered directory is returned.
 * NOTE: this mountpoint crossing is not supported properly by all
 *   clients and is explicitly disallowed for NFSv3
 *      NeilBrown <neilb@cse.unsw.edu.au>
 */
int
nfsd_lookup(struct svc_rqst *rqstp, struct svc_fh *fhp, const char *name,
					int len, struct svc_fh *resfh)
{
	struct svc_export	*exp;
	struct dentry		*dparent, *dchild;
	int			err;

	dprintk("nfsd: nfsd_lookup(fh %p, %s)\n", SVCFH_DENTRY(fhp), name);

	/* Obtain dentry and export. */
	err = fh_verify(rqstp, fhp, S_IFDIR, MAY_EXEC);
	if (err)
		goto out;

	dparent = fhp->fh_dentry;
	exp  = fhp->fh_export;

#if 0
	err = nfsd_permission(exp, dparent, MAY_EXEC);
	if (err)
		goto out;
#endif
	err = nfserr_acces;

	/* Lookup the name, but don't follow links */
	dchild = lookup_dentry(name, dget(dparent), 0);
	if (IS_ERR(dchild))
		goto out_nfserr;
	/*
	 * check if we have crossed a mount point ...
	 */
	if (dchild->d_sb != dparent->d_sb) {
		struct svc_export *exp2 = NULL;
		exp2 = exp_get(rqstp->rq_client,
			       dchild->d_inode->i_dev,
			       dchild->d_inode->i_ino);
		if (exp2 && EX_CROSSMNT(exp2))
			/* successfully crossed mount point */
			exp = exp2;
		else if (dchild->d_covers->d_sb == dparent->d_sb) {
			/* stay in the original filesystem */
			struct dentry *tdentry = dget(dchild->d_covers);
			dput(dchild);
			dchild = tdentry;
		} else {
			/* This cannot possibly happen */
			printk("nfsd_lookup: %s/%s impossible mount point!\n", dparent->d_name.name, dchild->d_name.name);
			dput(dchild);
			err = nfserr_acces;
			goto out;

		}
	}
	/*
	 * Note: we compose the file handle now, but as the
	 * dentry may be negative, it may need to be updated.
	 */
	fh_compose(resfh, exp, dchild);
	err = nfserr_noent;
	if (dchild->d_inode)
		err = 0;
out:
	return err;

out_nfserr:
#ifdef NFSV4
	err = nfsd4errno(-PTR_ERR(dchild));
#else /*NFSV4*/
	err = nfserrno(-PTR_ERR(dchild));
#endif /*NFSV4*/
	goto out;
}

/*
 * Set various file attributes.
 * N.B. After this call fhp needs an fh_put
 */
int
nfsd_setattr(struct svc_rqst *rqstp, struct svc_fh *fhp, struct iattr *iap)
{
	struct dentry	*dentry;
	struct inode	*inode;
	int		accmode = MAY_SATTR;
	int		ftype = 0;
	int		imode;
	int		err;
	kernel_cap_t	saved_cap = 0;

	nfsv4_printk (vfs, "setattr(): starting\n");

	if (iap->ia_valid & (ATTR_ATIME | ATTR_MTIME | ATTR_SIZE))
		accmode |= MAY_WRITE;
	if (iap->ia_valid & ATTR_SIZE)
		ftype = S_IFREG;

	/* Get inode */
	err = fh_verify(rqstp, fhp, ftype, accmode);
	if (err || !iap->ia_valid)
		goto out;

	dentry = fhp->fh_dentry;
	inode = dentry->d_inode;

	err = inode_change_ok(inode, iap);
	/* could be a "touch" (utimes) request where the user is not the owner but does
	 * have write permission. In this case the user should be allowed to set
	 * both times to the current time.  We could just assume any such SETATTR
	 * is intended to set the times to "now", but we do a couple of simple tests
	 * to increase our confidence.
	 */
#define BOTH_TIME_SET (ATTR_ATIME_SET | ATTR_MTIME_SET)
#define	MAX_TOUCH_TIME_ERROR (30*60)
	if (err
	    && (iap->ia_valid & BOTH_TIME_SET) == BOTH_TIME_SET
	    && iap->ia_mtime == iap->ia_ctime
	    ) {
	    /* looks good.  now just make sure time is in the right ballpark.
	     * solaris, at least, doesn't seem to care what the time request is
	     */
	    time_t delta = iap->ia_atime - CURRENT_TIME;
	    if (delta<0) delta = -delta;
	    if (delta < MAX_TOUCH_TIME_ERROR) {
		/* turn off ATTR_[AM]TIME_SET but leave ATTR_[AM]TIME
		 * this will cause notify_change to setthese times to "now"
		 */
		iap->ia_valid &= ~BOTH_TIME_SET;
		err = inode_change_ok(inode, iap);
		}
	}

	if (err)
		goto out_nfserr;

	/* The size case is special. It changes the file as well as the 
	* attributes.  
	*/
	if (iap->ia_valid & ATTR_SIZE) {
		if (!S_ISREG(inode->i_mode))
			printk("nfsd_setattr: size change??\n");
		if (iap->ia_size < inode->i_size) {
			err = nfsd_permission(fhp->fh_export, dentry, MAY_TRUNC);
			if (err)
				goto out;
		}
		err = get_write_access(inode);
		if (err)
			goto out_nfserr;

		DQUOT_INIT(inode);
	}

	imode = inode->i_mode;
	if (iap->ia_valid & ATTR_MODE) {
		iap->ia_mode &= S_IALLUGO;
		imode = iap->ia_mode |= (imode & ~S_IALLUGO);
	}

	/* Revoke setuid/setgid bit on chown/chgrp */
	if ((iap->ia_valid & ATTR_UID) && (imode & S_ISUID)
	 && iap->ia_uid != inode->i_uid) {
		iap->ia_valid |= ATTR_MODE;
		iap->ia_mode = imode &= ~S_ISUID;
	}
	if ((iap->ia_valid & ATTR_GID) && (imode & S_ISGID)
	 && iap->ia_gid != inode->i_gid) {
		iap->ia_valid |= ATTR_MODE;
		iap->ia_mode = imode &= ~S_ISGID;
	}

	/* Change the attributes. */


	iap->ia_valid |= ATTR_CTIME;
	if (current->fsuid != 0) {
		saved_cap = current->cap_effective;
		cap_clear(current->cap_effective);
	}
	if (iap->ia_valid & ATTR_SIZE) {
		fh_lock(fhp);
		err = notify_change(dentry, iap);
		if (!err) {
			vmtruncate(inode,iap->ia_size);		
			if (inode->i_op && inode->i_op->truncate)
				inode->i_op->truncate(inode);
		}
		fh_unlock(fhp);
		put_write_access(inode);
	}
	else
		err = notify_change(dentry, iap);
	if (current->fsuid != 0)
		current->cap_effective = saved_cap;
	if (err)
		goto out_nfserr;
	if (EX_ISSYNC(fhp->fh_export))
		write_inode_now(inode);
	err = 0;

	/* Don't unlock inode; the nfssvc_release functions are supposed
	 * to do this. */
out:
	nfsv4_printk (vfs, "setattr(): returning %d\n", err);
	return err;

out_nfserr:
#ifdef NFSV4
	err = nfsd4errno(-err);
#else /*NFSV4*/
	err = nfserrno(-err);
#endif /*NFSV4*/
	goto out;
}

#ifdef CONFIG_NFSD_V3
/*
 * Check server access rights to a file system object
 */
struct accessmap {
	u32		access;
	int		how;
};
static struct accessmap	nfs3_regaccess[] = {
    {	NFS3_ACCESS_READ,	MAY_READ			},
    {	NFS3_ACCESS_EXECUTE,	MAY_EXEC			},
    {	NFS3_ACCESS_MODIFY,	MAY_WRITE|MAY_TRUNC		},
    {	NFS3_ACCESS_EXTEND,	MAY_WRITE			},

    {	0,			0				}
};

static struct accessmap	nfs3_diraccess[] = {
    {	NFS3_ACCESS_READ,	MAY_READ			},
    {	NFS3_ACCESS_LOOKUP,	MAY_EXEC			},
    {	NFS3_ACCESS_MODIFY,	MAY_EXEC|MAY_WRITE|MAY_TRUNC	},
    {	NFS3_ACCESS_EXTEND,	MAY_EXEC|MAY_WRITE		},
    {	NFS3_ACCESS_DELETE,	MAY_REMOVE			},

    {	0,			0				}
};

static struct accessmap	nfs3_anyaccess[] = {
    /* XXX: should we try to cover read/write here for clients that
     * rely on us to do their access checking for special files? */

    {	0,			0				}
};

int
nfsd_access(struct svc_rqst *rqstp, struct svc_fh *fhp, u32 *access)
{
	struct accessmap	*map;
	struct svc_export	*export;
	struct dentry		*dentry;
	u32			query, result = 0;
	int			error;

	error = fh_verify(rqstp, fhp, 0, MAY_NOP);
	if (error)
		goto out;

	export = fhp->fh_export;
	dentry = fhp->fh_dentry;

	if (S_ISREG(dentry->d_inode->i_mode)) {
		map = nfs3_regaccess;
	} else if (S_ISDIR(dentry->d_inode->i_mode)) {
		map = nfs3_diraccess;
	} else {
		map = nfs3_anyaccess;
	}

	query = *access;
	while (map->access) {
		if (map->access & query) {
			error = nfsd_permission(export, dentry, (map->how | NO_OWNER_OVERRIDE));
			if (error == 0)
				result |= map->access;
			else if ((error == nfserr_perm) || (error == nfserr_acces)) {
				/*
				 *  This access type is denyed; but the 
				 *  access query itself succeeds.
				 */
				error = 0;
			} else {
				/*
				 *  Some fatal error.  Fail the query.
				 */
				goto out;
			}
		}
		map++;
	}
	*access = result;

out:
	return error;
}
#endif



/*
 * Open an existing file or directory.
 * The access argument indicates the type of open (read/write/lock)
 * N.B. After this call fhp needs an fh_put
 */
int
nfsd_open(struct svc_rqst *rqstp, struct svc_fh *fhp, int type,
			int access, struct file *filp)
{
	struct dentry	*dentry;
	struct inode	*inode;
	int		err = 0;

	err = fh_verify(rqstp, fhp, type, access);
	if (err)
		goto out;

	dentry = fhp->fh_dentry;
	inode = dentry->d_inode;

	/* Disallow access to files with the append-only bit set or
	 * with mandatory locking enabled
	 */
	err = nfserr_perm;
	if (IS_APPEND(inode) || IS_ISMNDLK(inode))
		goto out;
	if (!inode->i_op || !inode->i_op->default_file_ops)
		goto out;

	if ((access & MAY_WRITE) && (err = get_write_access(inode)) != 0)
		goto out_nfserr;

	memset(filp, 0, sizeof(*filp));
	filp->f_op    = inode->i_op->default_file_ops;
	filp->f_count = 1;
	filp->f_dentry = dentry;
	if (access & MAY_WRITE) {
		filp->f_flags = O_WRONLY;
		filp->f_mode  = FMODE_WRITE;
		DQUOT_INIT(inode);
	} else {
		filp->f_flags = O_RDONLY;
		filp->f_mode  = FMODE_READ;
	}

	err = 0;
	if (filp->f_op && filp->f_op->open) {
		err = filp->f_op->open(inode, filp);
		if (err) {
			if (access & MAY_WRITE)
				put_write_access(inode);

			/* I nearly added put_filp() call here, but this filp
			 * is really on callers stack frame. -DaveM
			 */
			filp->f_count--;
		}
	}
out_nfserr:
	if (err)
#ifdef NFSV4
		err = nfsd4errno(-err);
#else /*NFSV4*/
		err = nfserrno(-err);
#endif /*NFSV4*/
out:
	return err;
}

/*
 * Close a file.
 */
void
nfsd_close(struct file *filp)
{
	struct dentry	*dentry = filp->f_dentry;
	struct inode	*inode = dentry->d_inode;

	if (!inode->i_count)
		printk(KERN_WARNING "nfsd: inode count == 0!\n");
	if (!dentry->d_count)
		printk(KERN_WARNING "nfsd: wheee, %s/%s d_count == 0!\n",
			dentry->d_parent->d_name.name, dentry->d_name.name);
	if (filp->f_op && filp->f_op->release)
		filp->f_op->release(inode, filp);
	if (filp->f_mode & FMODE_WRITE) {
		put_write_access(inode);
	}
}

/*
 * Sync a file
 * As this calls fsync (not fdatasync) there is no need for a write_inode
 * after it.
 */
void
nfsd_sync(struct file *filp)
{
	dprintk("nfsd: sync file %s\n", filp->f_dentry->d_name.name);
	down(&filp->f_dentry->d_inode->i_sem);
	filp->f_op->fsync(filp, filp->f_dentry);
	up(&filp->f_dentry->d_inode->i_sem);
}

void
nfsd_sync_dir(struct dentry *dp)
{
	struct inode *inode = dp->d_inode;
	int (*fsync) (struct file *, struct dentry *);
	
	if (inode->i_op->default_file_ops
	    && (fsync = inode->i_op->default_file_ops->fsync)) {
		fsync(NULL, dp);
	}
}

/*
 * Obtain the readahead parameters for the file
 * specified by (dev, ino).
 */
static inline struct raparms *
nfsd_get_raparms(dev_t dev, ino_t ino)
{
	struct raparms	*ra, **rap, **frap = NULL;

	for (rap = &raparm_cache; (ra = *rap); rap = &ra->p_next) {
		if (ra->p_ino == ino && ra->p_dev == dev)
			goto found;
		if (ra->p_count == 0)
			frap = rap;
	}
	if (!frap)
		return NULL;
	rap = frap;
	ra = *frap;
	memset(ra, 0, sizeof(*ra));
	ra->p_dev = dev;
	ra->p_ino = ino;
found:
	if (rap != &raparm_cache) {
		*rap = ra->p_next;
		ra->p_next   = raparm_cache;
		raparm_cache = ra;
	}
	ra->p_count++;
	return ra;
}

/*
 * Read data from a file. count must contain the requested read count
 * on entry. On return, *count contains the number of bytes actually read.
 * N.B. After this call fhp needs an fh_put
 */
int
nfsd_read(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset,
          char *buf, unsigned long *count)
{
	struct raparms	*ra;
	mm_segment_t	oldfs;
	int		err;
	struct file	file;

	err = nfsd_open(rqstp, fhp, S_IFREG, MAY_READ, &file);
	if (err)
		goto out;
	err = nfserr_perm;
	if (!file.f_op->read)
		goto out_close;

	/* Get readahead parameters */
	ra = nfsd_get_raparms(fhp->fh_handle.fh_dev, fhp->fh_handle.fh_ino);
	if (ra) {
		file.f_reada = ra->p_reada;
		file.f_ramax = ra->p_ramax;
		file.f_raend = ra->p_raend;
		file.f_ralen = ra->p_ralen;
		file.f_rawin = ra->p_rawin;
	}
	file.f_pos = offset;

	oldfs = get_fs(); set_fs(KERNEL_DS);
	err = file.f_op->read(&file, buf, *count, &file.f_pos);
	set_fs(oldfs);

	/* Write back readahead params */
	if (ra != NULL) {
		dprintk("nfsd: raparms %ld %ld %ld %ld %ld\n",
			file.f_reada, file.f_ramax, file.f_raend,
			file.f_ralen, file.f_rawin);
		ra->p_reada = file.f_reada;
		ra->p_ramax = file.f_ramax;
		ra->p_raend = file.f_raend;
		ra->p_ralen = file.f_ralen;
		ra->p_rawin = file.f_rawin;
		ra->p_count -= 1;
	}

	if (err >= 0) {
		*count = err;
		err = 0;
	} else 
#ifdef NFSV4
		err = nfsd4errno(-err);
#else /*NFSV4*/
		err = nfserrno(-err);
#endif /*NFSV4*/
out_close:
	nfsd_close(&file);
out:
	return err;
}

/*
 * Write data to a file.
 * The stable flag requests synchronous writes.
 * N.B. After this call fhp needs an fh_put
 */
int
nfsd_write(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset,
				char *buf, unsigned long cnt, int stable)
{
	struct svc_export	*exp;
	struct file		file;
	struct dentry		*dentry;
	struct inode		*inode;
	mm_segment_t		oldfs;
	int			err = 0;
#ifdef CONFIG_QUOTA
	uid_t			saved_euid;
#endif

	if (!cnt)
		goto out;
	err = nfsd_open(rqstp, fhp, S_IFREG, MAY_WRITE, &file);
	if (err)
		goto out;
	err = nfserr_perm;
	if (!file.f_op->write)
		goto out_close;

	dentry = file.f_dentry;
	inode = dentry->d_inode;
	exp   = fhp->fh_export;

	/*
	 * Request sync writes if
	 *  -	the sync export option has been set, or
	 *  -	the client requested O_SYNC behavior (NFSv3 feature).
	 *  -   The file system doesn't support fsync().
	 * When gathered writes have been configured for this volume,
	 * flushing the data to disk is handled separately below.
	 */
#if defined(CONFIG_NFSD_V3) || defined(NFSV4)
	if (rqstp->rq_vers == 2)
		stable = EX_ISSYNC(exp);
	else if (file.f_op->fsync == 0)
	       stable = 1;
	if (stable && !EX_WGATHER(exp))
		file.f_flags |= O_SYNC;
#else
	if ((stable || (stable = EX_ISSYNC(exp))) && !EX_WGATHER(exp))
		file.f_flags |= O_SYNC;
#endif /* CONFIG_NFSD_V3 */

	fh_lock(fhp);			/* lock inode */
	file.f_pos = offset;		/* set write offset */

	/* Write the data. */
	oldfs = get_fs(); set_fs(KERNEL_DS);
#ifdef CONFIG_QUOTA
	/* This is for disk quota. */
	saved_euid = current->euid;
	current->euid = current->fsuid;
	err = file.f_op->write(&file, buf, cnt, &file.f_pos);
	current->euid = saved_euid;
#else
	err = file.f_op->write(&file, buf, cnt, &file.f_pos);
#endif
	set_fs(oldfs);

	/* clear setuid/setgid flag after write */
	if (err >= 0 && (inode->i_mode & (S_ISUID | S_ISGID))) {
		struct iattr	ia;
		kernel_cap_t	saved_cap = 0;

		ia.ia_valid = ATTR_MODE;
		ia.ia_mode  = inode->i_mode & ~(S_ISUID | S_ISGID);
		if (current->fsuid != 0) {
			saved_cap = current->cap_effective;
			cap_clear(current->cap_effective);
		}
		notify_change(dentry, &ia);
		if (current->fsuid != 0)
			current->cap_effective = saved_cap;
	}

	fh_unlock(fhp);			/* unlock inode */

	if (err >= 0 && stable) {
		static unsigned long	last_ino = 0;
		static kdev_t		last_dev = NODEV;

		/*
		 * Gathered writes: If another process is currently
		 * writing to the file, there's a high chance
		 * this is another nfsd (triggered by a bulk write
		 * from a client's biod). Rather than syncing the
		 * file with each write request, we sleep for 10 msec.
		 *
		 * I don't know if this roughly approximates
		 * C. Juszak's idea of gathered writes, but it's a
		 * nice and simple solution (IMHO), and it seems to
		 * work:-)
		 */
		if (EX_WGATHER(exp) && (inode->i_writecount > 1
		 || (last_ino == inode->i_ino && last_dev == inode->i_dev))) {
#if 0
			interruptible_sleep_on_timeout(&inode->i_wait, 10 * HZ / 1000);
#else
			dprintk("nfsd: write defer %d\n", current->pid);
/* FIXME: Olaf commented this out [gam3] */
			schedule_timeout((HZ+99)/100);
			dprintk("nfsd: write resume %d\n", current->pid);
#endif
		}

		if (inode->i_state & I_DIRTY) {
			dprintk("nfsd: write sync %d\n", current->pid);
			nfsd_sync(&file);
		}
		wake_up(&inode->i_wait);
		last_ino = inode->i_ino;
		last_dev = inode->i_dev;
	}

	dprintk("nfsd: write complete err=%d\n", err);
	if (err >= 0)
		err = 0;
	else 
#ifdef NFSV4
		err = nfsd4errno(-err);
#else /*NFSV4*/
		err = nfserrno(-err);
#endif /*NFSV4*/
out_close:
	nfsd_close(&file);
out:
	return err;
}


#ifdef CONFIG_NFSD_V3
/*
 * Commit all pendig writes to stable storage.
 * Strictly speaking, we could sync just indicated the file region here,
 * but there's currently no way we can ask the VFS to do so.
 *
 * We lock the file to make sure we return full WCC data to the client.
 */
int
nfsd_commit(struct svc_rqst *rqstp, struct svc_fh *fhp,
               off_t offset, unsigned long count)
{
	struct file	file;
	int		err;

	if ((err = nfsd_open(rqstp, fhp, S_IFREG, MAY_WRITE, &file)) != 0)
		return err;

	if (file.f_op && file.f_op->fsync) {
		nfsd_sync(&file);
	} else {
		err = nfserr_notsupp;
	}

	nfsd_close(&file);
	return err;
}
#endif /* CONFIG_NFSD_V3 */

#ifndef NFSV4
/*
 * Create a file (regular, directory, device, fifo); UNIX sockets 
 * not yet implemented.
 * If the response fh has been verified, the parent directory should
 * already be locked. Note that the parent directory is left locked.
 *
 * N.B. Every call to nfsd_create needs an fh_put for _both_ fhp and resfhp
 */
int
nfsd_create(struct svc_rqst *rqstp, struct svc_fh *fhp,
		char *fname, int flen, struct iattr *iap,
		int type, dev_t rdev, struct svc_fh *resfhp)
{
	struct dentry	*dentry, *dchild;
	struct inode	*dirp;
	nfsd_dirop_t	opfunc = NULL;
	int		err;

	err = nfserr_perm;
	if (!flen)
		goto out;
		iap->ia_mode = 0; 

	err = fh_verify(rqstp, fhp, S_IFDIR, MAY_CREATE);
	if (err)
		goto out;

	dentry = fhp->fh_dentry;
	dirp = dentry->d_inode;

	err = nfserr_notdir;
	if(!dirp->i_op || !dirp->i_op->lookup)
		goto out;
	/*
	 * Check whether the response file handle has been verified yet.
	 * If it has, the parent directory should already be locked.
	 */
	if (!resfhp->fh_dverified) {
		dchild = lookup_dentry(fname, dget(dentry), 0);
		err = PTR_ERR(dchild);
		if (IS_ERR(dchild))
			goto out_nfserr;
		fh_compose(resfhp, fhp->fh_export, dchild);
		/* Lock the parent and check for errors ... */
		err = fh_lock_parent(fhp, dchild);
		if (err)
			goto out;
	} else {
		dchild = resfhp->fh_dentry;
		if (!fhp->fh_locked) {
			/* not actually possible */
			printk(KERN_ERR
				"nfsd_create: parent %s/%s not locked!\n",
				dentry->d_parent->d_name.name,
				dentry->d_name.name);
			err = -EIO;
			goto out;
		}
	}
	/*
	 * Make sure the child dentry is still negative ...
	 */
	err = nfserr_exist;
	if (dchild->d_inode) {
		dprintk("nfsd_create: dentry %s/%s not negative!\n",
			dentry->d_name.name, dchild->d_name.name);
		goto out; 
	}

	/*
	 * Get the dir op function pointer.
	 */
	err = nfserr_perm;
	switch (type) {
	case S_IFREG:
		opfunc = (nfsd_dirop_t) dirp->i_op->create;
		break;
	case S_IFDIR:
		opfunc = (nfsd_dirop_t) dirp->i_op->mkdir;
		break;
	case S_IFCHR:
	case S_IFBLK:
		/* The client is _NOT_ required to do security enforcement */
		if(!capable(CAP_SYS_ADMIN)) {
			err = -EPERM;
			goto out;
		}
	case S_IFIFO:
	case S_IFSOCK:
		opfunc = dirp->i_op->mknod;
		break;
	default:
	        printk("nfsd: bad file type %o in nfsd_create\n", type);
		err = nfserr_inval;
	}
	if (!opfunc)
		goto out;

	if (!(iap->ia_valid & ATTR_MODE))
		iap->ia_mode = 0;
	iap->ia_mode = (iap->ia_mode & S_IALLUGO) | type;

	/*
	 * Call the dir op function to create the object.
	 */
	DQUOT_INIT(dirp);
	err = opfunc(dirp, dchild, iap->ia_mode, rdev);
	if (err < 0)
		goto out_nfserr;

	if (EX_ISSYNC(fhp->fh_export)) {
		nfsd_sync_dir(dentry);
		write_inode_now(dchild->d_inode);
	}

	/*
	 * Update the file handle to get the new inode info.
	 */
	fh_update(resfhp);

	/* Set file attributes. Mode has already been set and
	 * setting uid/gid works only for root. Irix appears to
	 * send along the gid when it tries to implement setgid
	 * directories via NFS.
	 */
	err = 0;
	if ((iap->ia_valid &= ~(ATTR_UID|ATTR_GID|ATTR_MODE)) != 0)
		err = nfsd_setattr(rqstp, resfhp, iap);
out:
	return err;

out_nfserr:
	err = nfserrno(-err);
	goto out;
}
#endif /*NFSV4*/

#ifdef CONFIG_NFSD_V3
/*
 * NFSv3 version of nfsd_create
 */
int
nfsd_create_v3(struct svc_rqst *rqstp, struct svc_fh *fhp,
    char *fname, int flen, struct iattr *iap,
    struct svc_fh *resfhp, int createmode, u32 *verifier)
{
	struct dentry *dentry, *dchild;
	struct inode  *dirp;
	int   err;

	err = nfserr_perm;
	if (!flen)
		goto out;
	if (!(iap->ia_valid & ATTR_MODE))
		iap->ia_mode = 0;
	err = fh_verify(rqstp, fhp, S_IFDIR, MAY_CREATE);
	if (err)
		goto out;

	dentry = fhp->fh_dentry;
	dirp = dentry->d_inode;

	/* Get all the sanity checks out of the way before
	* we lock the parent. */
	err = nfserr_notdir;
	if(!dirp->i_op || !dirp->i_op->lookup)
		goto out;
	err = nfserr_perm;
	if(!dirp->i_op->create)
		goto out;

	/*
	* Compose the response file handle.
	*/
	dchild = lookup_dentry(fname, dget(dentry), 0);
	err = PTR_ERR(dchild);
	if(IS_ERR(dchild))
		goto out_nfserr;
	fh_compose(resfhp, fhp->fh_export, dchild);

	/*
	* We must lock the directory before we check for the inode.
	*/
	err = fh_lock_parent(fhp, dchild);
	if (err)
		goto out;

	if (dchild->d_inode) {
		err = 0;

		if (resfhp->fh_handle.fh_ino == 0)
		/* inode might have been instantiated while we slept */
			fh_update(resfhp);

		switch (createmode) {
		case NFS3_CREATE_UNCHECKED:
			if (! S_ISREG(dchild->d_inode->i_mode))
				err = nfserr_exist;
			else {
			iap->ia_valid &= ATTR_SIZE;
			goto set_attr;
			}
		break;
		case NFS3_CREATE_EXCLUSIVE:
			if (   dchild->d_inode->i_mtime == verifier[0]
					&& dchild->d_inode->i_atime == verifier[1]
					&& dchild->d_inode->i_mode == S_IFREG
					&& dchild->d_inode->i_size == 0 )
		break;
		/* fallthru */
		case NFS3_CREATE_GUARDED:
			err = nfserr_exist;
		}
		goto out;
	}

	err = dirp->i_op->create(dirp, dchild, iap->ia_mode);
	if (err < 0)
		goto out_nfserr;

	if (EX_ISSYNC(fhp->fh_export)) {
		nfsd_sync_dir(dentry);
		/* setattr will sync the child (or not) */
	}

	/*
	* Update the filehandle to get the new inode info.
	*/
	fh_update(resfhp);
	err = 0;

	if (createmode == NFS3_CREATE_EXCLUSIVE) {
		/* Cram the verifier into atime/mtime */
		iap->ia_valid = ATTR_MTIME|ATTR_ATIME|ATTR_MTIME_SET|ATTR_ATIME_SET;
		iap->ia_mtime = verifier[0];
		iap->ia_atime = verifier[1];
	}

	/* Set file attributes. Mode has already been set and
	* setting uid/gid works only for root. Irix appears to
	* send along the gid when it tries to implement setgid
	* directories via NFS. Clear out all that cruft.
	*/
	set_attr:
		if ((iap->ia_valid &= ~(ATTR_UID|ATTR_GID|ATTR_MODE)) != 0)
			err = nfsd_setattr(rqstp, resfhp, iap);

	out:
		fh_unlock(fhp);
		return err;

	out_nfserr:
		err = nfserrno(-err);
		goto out;
}
#endif /* CONFIG_NFSD_V3 */

/*
 * Truncate a file.
 * The calling routines must make sure to update the ctime
 * field and call notify_change.
 *
 * XXX Nobody calls this thing? -DaveM
 * N.B. After this call fhp needs an fh_put
 */
int
nfsd_truncate(struct svc_rqst *rqstp, struct svc_fh *fhp, unsigned long size)
{
	struct dentry	*dentry;
	struct inode	*inode;
	struct iattr	newattrs;
	int		err;
	kernel_cap_t	saved_cap = 0;

	err = fh_verify(rqstp, fhp, S_IFREG, MAY_WRITE | MAY_TRUNC);
	if (err)
		goto out;

	dentry = fhp->fh_dentry;
	inode = dentry->d_inode;

	err = get_write_access(inode);
	if (err)
		goto out_nfserr;

	/* Things look sane, lock and do it. */
	fh_lock(fhp);
	DQUOT_INIT(inode);
	newattrs.ia_size = size;
	newattrs.ia_valid = ATTR_SIZE | ATTR_CTIME;
	if (current->fsuid != 0) {
		saved_cap = current->cap_effective;
		cap_clear(current->cap_effective);
	}
	err = notify_change(dentry, &newattrs);
	if (!err) {
		vmtruncate(inode, size);
		if (inode->i_op && inode->i_op->truncate)
			inode->i_op->truncate(inode);
	}
	if (current->fsuid != 0)
		current->cap_effective = saved_cap;
	put_write_access(inode);
	if (EX_ISSYNC(fhp->fh_export))
		nfsd_sync_dir(dentry);
	fh_unlock(fhp);
out_nfserr:
	if (err)
#ifdef NFSV4
		err = nfsd4errno(-err);
#else /*NFSV4*/
		err = nfserrno(-err);
#endif /*NFSV4*/
out:
	return err;
}

/*
 * Read a symlink. On entry, *lenp must contain the maximum path length that
 * fits into the buffer. On return, it contains the true length.
 * N.B. After this call fhp needs an fh_put
 */
int
nfsd_readlink(struct svc_rqst *rqstp, struct svc_fh *fhp, char *buf, int *lenp)
{
	struct dentry	*dentry;
	struct inode	*inode;
	mm_segment_t	oldfs;
	int		err;

	err = fh_verify(rqstp, fhp, S_IFLNK, MAY_READ);
	if (err)
		goto out;

	dentry = fhp->fh_dentry;
	inode = dentry->d_inode;

	err = nfserr_inval;
	if (!inode->i_op || !inode->i_op->readlink)
		goto out;

	UPDATE_ATIME(inode);
	/* N.B. Why does this call need a get_fs()??
	 * Remove the set_fs and watch the fireworks:-) --okir
	 */

	oldfs = get_fs(); set_fs(KERNEL_DS);
	err = inode->i_op->readlink(dentry, buf, *lenp);
	set_fs(oldfs);

	if (err < 0)
		goto out_nfserr;
	*lenp = err;
	err = 0;
out:
	return err;

out_nfserr:
#ifdef NFSV4
	err = nfsd4errno(-err);
#else /*NFSV4*/
	err = nfserrno(-err);
#endif /*NFSV4*/
	goto out;
}

/*
 * Create a symlink and look up its inode
 * N.B. After this call _both_ fhp and resfhp need an fh_put
 */
int
nfsd_symlink(struct svc_rqst *rqstp, struct svc_fh *fhp,
				char *fname, int flen,
				char *path,  int plen,
				struct svc_fh *resfhp,
				struct iattr *iap)
{
	struct dentry	*dentry, *dnew;
	struct inode	*dirp;
	int		err;

	err = nfserr_noent;
	if (!flen || !plen)
		goto out;

	err = fh_verify(rqstp, fhp, S_IFDIR, MAY_CREATE);
	if (err)
		goto out;
	dentry = fhp->fh_dentry;

	err = nfserr_perm;
	dirp = dentry->d_inode;
	if (!dirp->i_op	|| !dirp->i_op->symlink)
		goto out;

	dnew = lookup_dentry(fname, dget(dentry), 0);
	err = PTR_ERR(dnew);
	if (IS_ERR(dnew))
		goto out_nfserr;

	/*
	 * Lock the parent before checking for existence
	 */
	err = fh_lock_parent(fhp, dnew);
	if (err)
		goto out_compose;

	err = nfserr_exist;
	if (!dnew->d_inode) {
		DQUOT_INIT(dirp);
		err = dirp->i_op->symlink(dirp, dnew, path);
		if (!err) {
			if (EX_ISSYNC(fhp->fh_export))
				nfsd_sync_dir(dentry);
			if (iap) {
				iap->ia_valid &= ATTR_MODE /* ~(ATTR_MODE|ATTR_UID|ATTR_GID)*/;
				if (iap->ia_valid) {
					iap->ia_valid |= ATTR_CTIME;
					iap->ia_mode = (iap->ia_mode&S_IALLUGO)
						| S_IFLNK;
					err = notify_change(dnew, iap);
					if (!err && EX_ISSYNC(fhp->fh_export))
						write_inode_now(dentry->d_inode);
			       }
			}
		} else
#ifdef NFSV4
			err = nfsd4errno(-err);
#else /*NFSV4*/
			err = nfserrno(-err);
#endif /*NFSV4*/
	}
	fh_unlock(fhp);

	/* Compose the fh so the dentry will be freed ... */
out_compose:
	fh_compose(resfhp, fhp->fh_export, dnew);

out:
	return err;

out_nfserr:
#ifdef NFSV4
	err = nfsd4errno(-err);
#else /*NFSV4*/
	err = nfserrno(-err);
#endif /*NFSV4*/
	goto out;
}

/*
 * Create a hardlink
 * N.B. After this call _both_ ffhp and tfhp need an fh_put
 */
int
nfsd_link(struct svc_rqst *rqstp, struct svc_fh *ffhp,
				char *fname, int len, struct svc_fh *tfhp)
{
	struct dentry	*ddir, *dnew, *dold;
	struct inode	*dirp, *dest;
	int		err;

	err = fh_verify(rqstp, ffhp, S_IFDIR, MAY_CREATE);
	if (err)
		goto out;
	err = fh_verify(rqstp, tfhp, -S_IFDIR, MAY_NOP);
	if (err)
		goto out;

	err = nfserr_perm;
	if (!len)
		goto out;

	ddir = ffhp->fh_dentry;
	dirp = ddir->d_inode;

	dnew = lookup_dentry(fname, dget(ddir), 0);
	err = PTR_ERR(dnew);
	if (IS_ERR(dnew))
		goto out_nfserr;
	/*
	 * Lock the parent before checking for existence
	 */
	err = fh_lock_parent(ffhp, dnew);
	if (err)
		goto out_dput;

	err = nfserr_exist;
	if (dnew->d_inode)
		goto out_unlock;

	dold = tfhp->fh_dentry;
	dest = dold->d_inode;

	err = (rqstp->rq_vers == 2) ? nfserr_acces : nfserr_xdev;
	if (dirp->i_dev != dest->i_dev)
		goto out_unlock;

	err = nfserr_perm;
	if (IS_IMMUTABLE(dest) || IS_APPEND(dest))
		goto out_unlock;
	if (!dirp->i_op || !dirp->i_op->link)
		goto out_unlock;

	DQUOT_INIT(dirp);
	err = dirp->i_op->link(dold, dirp, dnew);
	if (!err) {
		if (EX_ISSYNC(ffhp->fh_export)) {
			nfsd_sync_dir(ddir);
			write_inode_now(dest);
		}
	} else
#ifdef NFSV4
	err = nfsd4errno(-err);
#else /*NFSV4*/
		err = nfserrno(-err);
#endif /*NFSV4*/

out_unlock:
	fh_unlock(ffhp);
out_dput:
	dput(dnew);
out:
	return err;

out_nfserr:
#ifdef NFSV4
	err = nfsd4errno(-err);
#else /*NFSV4*/
	err = nfserrno(-err);
#endif /*NFSV4*/
	goto out;
}

/*
 * This follows the model of double_lock() in the VFS.
 */
static inline void nfsd_double_down(struct semaphore *s1, struct semaphore *s2)
{
	if (s1 != s2) {
		if ((unsigned long) s1 < (unsigned long) s2) {
			struct semaphore *tmp = s1;
			s1 = s2;
			s2 = tmp;
		}
		down(s1);
	}
	down(s2);
}

static inline void nfsd_double_up(struct semaphore *s1, struct semaphore *s2)
{
	up(s1);
	if (s1 != s2)
		up(s2);
}

/*
 * Rename a file
 * N.B. After this call _both_ ffhp and tfhp need an fh_put
 */
int
nfsd_rename(struct svc_rqst *rqstp, struct svc_fh *ffhp, char *fname, int flen,
			    struct svc_fh *tfhp, char *tname, int tlen)
{
	struct dentry	*fdentry, *tdentry, *odentry, *ndentry;
	struct inode	*fdir, *tdir;
	int		err;

	err = fh_verify(rqstp, ffhp, S_IFDIR, MAY_REMOVE);
	if (err)
		goto out;
	err = fh_verify(rqstp, tfhp, S_IFDIR, MAY_CREATE);
	if (err)
		goto out;

	fdentry = ffhp->fh_dentry;
	fdir = fdentry->d_inode;

	tdentry = tfhp->fh_dentry;
	tdir = tdentry->d_inode;

	err = (rqstp->rq_vers == 2) ? nfserr_acces : nfserr_xdev;
	if (fdir->i_dev != tdir->i_dev)
		goto out;

	/* N.B. We shouldn't need this ... dentry layer handles it */
	err = nfserr_perm;
	if (!flen || (fname[0] == '.' && 
	    (flen == 1 || (flen == 2 && fname[1] == '.'))) ||
	    !tlen || (tname[0] == '.' && 
	    (tlen == 1 || (tlen == 2 && tname[1] == '.'))))
		goto out;

	odentry = lookup_dentry(fname, dget(fdentry), 0);
	err = PTR_ERR(odentry);
	if (IS_ERR(odentry))
		goto out_nfserr;

	err = -ENOENT;
	if (!odentry->d_inode)
		goto out_dput_old;

	ndentry = lookup_dentry(tname, dget(tdentry), 0);
	err = PTR_ERR(ndentry);
	if (IS_ERR(ndentry))
		goto out_dput_old;

	/*
	 * Lock the parent directories.
	 */
	nfsd_double_down(&tdir->i_sem, &fdir->i_sem);

#ifdef CONFIG_NFSD_V3
	/* Fill in the pre-op attr for the wcc data for both 
	 * tdir and fdir
	 */ 
	fill_pre_wcc(ffhp);
	fill_pre_wcc(tfhp);
#endif /* CONFIG_NFSD_V3 */

	err = -ENOENT;
	/* GAM3 check for parent changes after locking. */
	if (nfsd_check_parent(fdentry, odentry) &&
	    nfsd_check_parent(tdentry, ndentry)) {

		err = vfs_rename(fdir, odentry, tdir, ndentry);
		if (!err && EX_ISSYNC(tfhp->fh_export)) {
			nfsd_sync_dir(tdentry);
			nfsd_sync_dir(fdentry);
		}
	} else
		dprintk("nfsd: Caught race in nfsd_rename");

#ifdef CONFIG_NFSD_V3
        /* Fill in the post-op attr for the wcc data for both 
         * tdir and fdir
         */
	fill_post_wcc(ffhp);
	fill_post_wcc(tfhp);
#endif /* CONFIG_NFSD_V3 */
	nfsd_double_up(&tdir->i_sem, &fdir->i_sem);
	dput(ndentry);

out_dput_old:
	dput(odentry);
	if (err)
		goto out_nfserr;
out:
	return err;

out_nfserr:
#ifdef NFSV4
	err = nfsd4errno(-err);
#else /*NFSV4*/
	err = nfserrno(-err);
#endif /*NFSV4*/
	goto out;
}

/*
 * Unlink a file or directory
 * N.B. After this call fhp needs an fh_put
 */
int
nfsd_unlink(struct svc_rqst *rqstp, struct svc_fh *fhp, int type,
				char *fname, int flen)
{
	struct dentry	*dentry, *rdentry;
	struct inode	*dirp;
	int		err;

	/* N.B. We shouldn't need this test ... handled by dentry layer */
	err = nfserr_acces;
	if (!flen || isdotent(fname, flen))
		goto out;
	err = fh_verify(rqstp, fhp, S_IFDIR, MAY_REMOVE);
	if (err)
		goto out;

	dentry = fhp->fh_dentry;
	dirp = dentry->d_inode;

	rdentry = lookup_dentry(fname, dget(dentry), 0);
	err = PTR_ERR(rdentry);
	if (IS_ERR(rdentry))
		goto out_nfserr;

	if (!rdentry->d_inode) {
		dput(rdentry);
		err = nfserr_noent;
		goto out;
	}


	if (type != S_IFDIR) {
		/* It's UNLINK */

		err = fh_lock_parent(fhp, rdentry);
		if (err)
			goto out;

		err = vfs_unlink(dirp, rdentry);

		fh_unlock(fhp);

		dput(rdentry);

	} else {
		/* It's RMDIR */
		/* See comments in fs/namei.c:do_rmdir */

		rdentry->d_count++;
		nfsd_double_down(&dirp->i_sem, &rdentry->d_inode->i_sem);

#ifdef CONFIG_NFSD_V3
		fill_pre_wcc(fhp);
#else
		fhp->fh_locked = 1;
#endif /* CONFIG_NFSD_V3 */

		err = -ENOENT;
		if (nfsd_check_parent(dentry, rdentry))
			err = vfs_rmdir(dirp, rdentry);

		rdentry->d_count--;
#ifdef CONFIG_NFSD_V3
		fill_post_wcc(fhp);
#else
		fhp->fh_locked = 0;
#endif /* CONFIG_NFSD_V3 */
		nfsd_double_up(&dirp->i_sem, &rdentry->d_inode->i_sem);

		dput(rdentry);
	}

	if (err)
		goto out_nfserr;
	if (EX_ISSYNC(fhp->fh_export)) {
		down(&dentry->d_inode->i_sem);
		nfsd_sync_dir(dentry);
		up(&dentry->d_inode->i_sem);
	}
out:
	return err;

out_nfserr:
#ifdef NFSV4
	err = nfsd4errno(-err);
#else /*NFSV4*/
	err = nfserrno(-err);
#endif /*NFSV4*/
	goto out;
}

/*
 * Read entries from a directory.
 * The verifier is an NFSv3 thing we ignore for now.
 */
int
nfsd_readdir(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset, 
             encode_dent_fn func, u32 *buffer, int *countp, u32 *verf)
{
	struct inode	*inode;
	u32		*p;
	int		oldlen, eof, err;
	struct file	file;
	struct readdir_cd cd;

	err = 0;
	if (offset > ~(u32) 0)
		goto out;

	err = nfsd_open(rqstp, fhp, S_IFDIR, MAY_READ, &file);
	if (err)
		goto out;

	err = nfserr_notdir;
	if (!file.f_op->readdir)
		goto out_close;
	file.f_pos = offset;

	/* Set up the readdir context */
	memset(&cd, 0, sizeof(cd));
	cd.rqstp  = rqstp;
	cd.buffer = buffer;
	cd.buflen = *countp; /* count of words */
	cd.dirfh  = fhp;

	/*
	 * Read the directory entries. This silly loop is necessary because
	 * readdir() is not guaranteed to fill up the entire buffer, but
	 * may choose to do less.
	 */
	inode = file.f_dentry->d_inode;
	while (1) {
		oldlen = cd.buflen;

		/*
		dprintk("nfsd: f_op->readdir(%x/%ld @ %d) buflen = %d (%d)\n",
			file.f_inode->i_dev, file.f_inode->i_ino,
			(int) file.f_pos, (int) oldlen, (int) cd.buflen);
		 */
		down(&inode->i_sem);
		err = file.f_op->readdir(&file, &cd, (filldir_t) func);
		up(&inode->i_sem);
		if (err < 0)
			goto out_nfserr;
		if (oldlen == cd.buflen)
			break;
		if (cd.eob)
			break;
	}

	/* If we didn't fill the buffer completely, we're at EOF */
	eof = !cd.eob;

	if (cd.offset) {
#ifdef CONFIG_NFSD_V3
		if (rqstp->rq_vers == 3)
			(void)enc64(cd.offset, file.f_pos);
		else
#endif /* CONFIG_NFSD_V3 */
			*cd.offset = htonl(file.f_pos);
	}

	p = cd.buffer;
	*p++ = 0;			/* no more entries */
	*p++ = htonl(eof);		/* end of directory */
	*countp = (caddr_t) p - (caddr_t) buffer;

	dprintk("nfsd: readdir result %d bytes, eof %d offset %d\n",
				*countp, eof,
				cd.offset? ntohl(*cd.offset) : -1);
	err = 0;
out_close:
	nfsd_close(&file);
out:
	return err;

out_nfserr:
#ifdef NFSV4
	err = nfsd4errno(-err);
#else /*NFSV4*/
	err = nfserrno(-err);
#endif /*NFSV4*/
	goto out_close;
}

/*
 * Get file system stats
 * N.B. After this call fhp needs an fh_put
 */
int
nfsd_statfs(struct svc_rqst *rqstp, struct svc_fh *fhp, struct statfs *stat)
{
	struct dentry		*dentry;
	struct inode		*inode;
	struct super_block	*sb;
	mm_segment_t		oldfs;
	int			err;

	err = fh_verify(rqstp, fhp, 0, MAY_NOP);
	if (err)
		goto out;
	dentry = fhp->fh_dentry;
	inode = dentry->d_inode;

	err = nfserr_io;
	if (!(sb = inode->i_sb) || !sb->s_op->statfs)
		goto out;

	oldfs = get_fs();
	set_fs (KERNEL_DS);
	sb->s_op->statfs(sb, stat, sizeof(*stat));
	set_fs (oldfs);
	err = 0;

out:
	return err;
}

/*
 * Check for a user's access permissions to this inode.
 */
int
nfsd_permission(struct svc_export *exp, struct dentry *dentry, int acc)
{
	struct inode	*inode = dentry->d_inode;
	int		err;
	kernel_cap_t	saved_cap = 0;
	int		owneraccess;

	/*
	 *  Check if we are to use "owner may always access" semantics,
	 *  then clean out the flag bit which controls this. It might be
	 *  clearer to reverse the logic of this flag, but I didn't
	 *  want to change a lot of code in a stable kernel - dhiggen.
	 */

	if (acc & NO_OWNER_OVERRIDE) {
		owneraccess = 0;
		acc &= ~NO_OWNER_OVERRIDE;
	} else {
		owneraccess = 1;
	}

	if (acc == MAY_NOP)
		return 0;
#if 0
	dprintk("nfsd: permission 0x%x%s%s%s%s%s%s mode 0%o%s%s%s\n",
		acc,
		(acc & MAY_READ)?	" read"  : "",
		(acc & MAY_WRITE)?	" write" : "",
		(acc & MAY_EXEC)?	" exec"  : "",
		(acc & MAY_SATTR)?	" sattr" : "",
		(acc & MAY_TRUNC)?	" trunc" : "",
#ifndef NFSV4
		(acc & MAY_LOCK)?	" lock"  : "",
#endif /*NFSV4*/
		inode->i_mode,
		IS_IMMUTABLE(inode)?	" immut" : "",
		IS_APPEND(inode)?	" append" : "",
		IS_RDONLY(inode)?	" ro" : "");
	dprintk("      owner %d/%d user %d/%d\n",
		inode->i_uid, inode->i_gid, current->fsuid, current->fsgid);
#endif

	if (acc & (MAY_WRITE | MAY_SATTR | MAY_TRUNC)) {
		if (EX_RDONLY(exp) || IS_RDONLY(inode))
			return nfserr_rofs;
		if (/* (acc & MAY_WRITE) && */ IS_IMMUTABLE(inode))
			return nfserr_perm;
	}
	if ((acc & MAY_TRUNC) && IS_APPEND(inode))
		return nfserr_perm;

	if (acc & MAY_LOCK) {
		/* If we cannot rely on authentication in NLM requests,
		 * just allow locks, others require read permission
		 */
		if (exp->ex_flags & NFSEXP_NOAUTHNLM)
			return 0;
		else
			acc = MAY_READ;
	}
	/*
	 * The file owner always gets access permission (except in the 
	 * special case of a V3 ACCESS call, which is used for checking at 
	 * open() time). This is to make file access work even when the 
	 * client has done a fchmod(fd, 0).
	 *
	 * However, `cp foo bar' should fail nevertheless when bar is
	 * readonly. A sensible way to do this might be to reject all
	 * attempts to truncate a read-only file, because a creat() call
	 * always implies file truncation.
	 *  dhXXX we are not currently setting MAY_TRUNC from a SETATTR which
	 *  changes the size, so this check is not enforced.  It probably
	 *  should be?
	 */
	if (owneraccess && inode->i_uid == current->fsuid /* && !(acc & MAY_TRUNC) */)
		return 0;

	if (current->fsuid != 0) {
		saved_cap = current->cap_effective;
		cap_clear(current->cap_effective);
	}

	err = permission(inode, acc & (MAY_READ|MAY_WRITE|MAY_EXEC));

	/* Allow read access to binaries even when mode 111 */
	if (err == -EACCES && S_ISREG(inode->i_mode) && acc == MAY_READ)
		err = permission(inode, MAY_EXEC);

	if (current->fsuid != 0)
		current->cap_effective = saved_cap;

#ifdef NFSV4
	return err? nfsd4errno(-err) : 0;
#else /*NFSV4*/
	return err? nfserrno(-err) : 0;
#endif /*NFSV4*/
}

void
nfsd_racache_shutdown(void)
{
	if (!raparm_cache)
		return;
	dprintk("nfsd: freeing %d readahead buffers.\n", FILECACHE_MAX);
	kfree(raparml);
	nfsd_nservers = 0;
	raparm_cache = raparml = NULL;
}
/*
 * Initialize readahead param cache
 */
void
nfsd_racache_init(void)
{
	int	i;

	if (raparm_cache)
		return;
	raparml = kmalloc(sizeof(struct raparms) * FILECACHE_MAX, GFP_KERNEL);

	if (raparml != NULL) {
		dprintk("nfsd: allocating %d readahead buffers.\n",
			FILECACHE_MAX);
		memset(raparml, 0, sizeof(struct raparms) * FILECACHE_MAX);
		for (i = 0; i < FILECACHE_MAX - 1; i++) {
			raparml[i].p_next = raparml + i + 1;
		}
		raparm_cache = raparml;
	} else {
		printk(KERN_WARNING
		       "nfsd: Could not allocate memory read-ahead cache.\n");
		nfsd_nservers = 0;
	}
}
#ifdef NFSV4
/*
 * NFSV4 version of nfsd_create
 */
int
nfsd4_create(struct svc_rqst *rqstp, utf8string *utf8fname, struct iattr *iap, struct svc_fh *resfhp,int createmode, char *verifier)
{
	struct dentry	*dentry, *dchild;
	struct inode	*dirp;
	int		err;
	struct svc_fh *fhp = &CURRENT_FH;
	char *fname;

	nfsv4_printk(open, "nfsd4_create() starting\n");

	err = nfserr_perm;
	if (!utf8fname->utf8string_len)
		goto out;

	if (utf8fname->utf8string_len > NFS4_MAXNAMLEN)
		utf8fname->utf8string_len = NFS4_MAXNAMLEN;

	if (!(iap->ia_valid & ATTR_MODE))
		iap->ia_mode = 0;
	err = fh_verify(rqstp, fhp, S_IFDIR, MAY_CREATE);
	if (err)
		goto out;

	dentry = fhp->fh_dentry;
	dirp = dentry->d_inode;
	nfsv4_printk(open, "nfsd4_create(): fname=\"%.*s\", dname=\"%s\"\n",
		     (int) utf8fname->utf8string_len, utf8fname->utf8string_val,
		     fhp->fh_dentry->d_name.name);
	
	/* Get all the sanity checks out of the way before
	 * we lock the parent. */
	err = nfserr_notdir;
	if(!dirp->i_op || !dirp->i_op->lookup)
		goto out;
	err = nfserr_perm;
	if(!dirp->i_op->create)
		goto out;

	/*
	 * Compose the response file handle.
	 */
  /* utf8_2_ascii,kfree(fname):
  * The vfs layer assumes ascii strings. NFSv4 uses
  * UTF-8 encoding. Until the vfs layer decides otherwise,
  * linux NFSv4 servers deal only in ascii.
  */   
	utf8_2_ascii(utf8fname,&fname);
	dchild = lookup_dentry(fname, dget(dentry), 0);
	kfree(fname);
	err = PTR_ERR(dchild);
	if (IS_ERR(dchild)) {
		nfsv4_printk(open, "nfsd4_create(): bad child dentry\n");
		goto out_nfserr;
	}
	fh_compose(resfhp, fhp->fh_export, dchild);

	/*
	 * We must lock the directory before we check for the inode.
	 */
	err = fh_lock_parent(fhp, dchild);
	if (err)
		goto out;
	if (dchild->d_inode) {
		err = 0;

		if (resfhp->fh_handle.fh_ino == 0)
		     /* inode might have been instantiated while we slept */
		    fh_update(resfhp);
		
		switch (createmode) {
		case UNCHECKED4:
			if (! S_ISREG(dchild->d_inode->i_mode)) {
				nfsv4_printk(open, "nfsd4_create(): not a regular file\n");
				err = nfserr_exist;
			}
			else {
				iap->ia_valid &= ATTR_SIZE;
				goto set_attr;
			}
			break;
		case EXCLUSIVE4:
		/* not supported yet */
#if 0
			if (   dchild->d_inode->i_mtime == verifier[0]
			    && dchild->d_inode->i_atime == verifier[1]
			    && dchild->d_inode->i_mode == S_IFREG
			    && dchild->d_inode->i_size == 0 )
				break;
#endif /*0*/
			err = nfserr_notsupp;
			 /* fallthru */
		case GUARDED4:
			nfsv4_printk(open, "nfsd4_create(): file already exists\n");
			err = nfserr_exist;
		}
		goto out;
	}

	err = dirp->i_op->create(dirp, dchild, iap->ia_mode);
	if (err < 0) {
		nfsv4_printk(open, "nfsd4_create(): create op failed\n");
		goto out_nfserr;
	}

	if (EX_ISSYNC(fhp->fh_export)) {
		nfsd_sync_dir(dentry);
		/* setattr will sync the child (or not) */
	}

	/*
	 * Update the filehandle to get the new inode info.
	 */
	fh_update(resfhp);
	err = 0;

	/* not supported yet*/
	if (createmode == EXCLUSIVE4) {
		err = nfserr_notsupp;
#if 0
		/* Cram the verifier into atime/mtime */
		iap->ia_valid = ATTR_MTIME|ATTR_ATIME|ATTR_MTIME_SET|ATTR_ATIME_SET;
		iap->ia_mtime = verifier[0];
		iap->ia_atime = verifier[1];
#endif /*0*/
	}

	/* Set file attributes. Mode has already been set and
	 * setting uid/gid works only for root. Irix appears to
	 * send along the gid when it tries to implement setgid
	 * directories via NFS. Clear out all that cruft.
	 */
set_attr:
	if ((iap->ia_valid &= ~(ATTR_UID|ATTR_GID|ATTR_MODE)) != 0) {
 		err = nfsd_setattr(rqstp, resfhp, iap);
		if (err)
			nfsv4_printk(open, "nfsd4_create(): nfsd_setattr failed\n");
	}

out:
	fh_unlock(fhp);
	nfsv4_printk(vfs,"NFSDV4: nfsd4_create returning err %d\n",err);
 	return err;
 
out_nfserr:
	err = nfsd4errno(-err);
	goto out;
}
/* create root ("/") fh.
 * for now, don't deal with the case where
 * "/" is exported.
 */
int
nfsd4_putrootfh(struct svc_rqst *rqstp)
{
int err;
struct svc_fh fh;

	nfsv4_printk(vfs,"NFSDV4: nfsd4_putrootfh\n");

	err = exp_pseudorootfh(&fh);
	if(err != 0){
		nfsv4_printk(vfs,"  exp_pseudorootfh returns: %d\n",err);
		goto out;
	}

	fh_init(&CURRENT_FH);

	nfsv4_printk(vfs,
		"NFSDV4: nfsd4_putrootfh: fh dev/ino: %d/%d name:%s\n",
		SVCFH_DEV(&fh), SVCFH_INO(&fh),
		fh.fh_dentry->d_name.name);

	if(!err)
		fh_copy(&CURRENT_FH, &fh);

	nfsv4_printk(vfs,
		"NFSDV4: nfsd4_putrootfh: CURRENT_FH dev/ino: %d/%d name:%s\n",
		SVCFH_DEV(&CURRENT_FH), SVCFH_INO(&CURRENT_FH),
		CURRENT_FH.fh_dentry->d_name.name);


	nfsv4_printk(vfs,
		"NFSDV4 nfsd4_putrootfh CURRENT_FHl xdev %d xino %d\n",
		CURRENT_FH.fh_handle.fh_xdev,
		CURRENT_FH.fh_handle.fh_xino);

out:
	return err;
}


/* dparent - dentry of input fhp. dchild, result of lookup on name
*
* pseudo-fs:
* we do things a bit backwards - we get the inode first, then the 
* dentry. this is because if the pseudo_fs inode doesn't exist, we're
* hosed - it's the only place where the data is!
*/
int
nfsd4_lookup_pseudo(struct svc_rqst *rqstp,struct svc_fh *fhp,
      const char *name, struct svc_fh *resfh)
{
struct svc_export *exp;
struct dentry *dchild,*dparent;
struct inode *iparent,*ichild;
int     err = 0;

	nfsv4_printk(vfs,
		"NFSDV4: nfsd4_lookup_pseudo (fh %p %s)\n", 
		SVCFH_DENTRY(fhp),name); 

	/* in the pseudo_fs where inodes and dentries are kept,
	* we are guarenteed a valid fhp->fh_dentry
	* but call fh_verify for ext2 dentries */

	err = fh_verify(rqstp,fhp,S_IFDIR,MAY_EXEC);
	if (err){
		nfsv4_printk(vfs,"NFSDV4 nfsd4_lookup_pseudo: fh_verify failed\n");
		goto out;
	}                   
	print_pseudo_dentry(fhp->fh_dentry);
	dparent = fhp->fh_dentry;

	/* get export and dchild */
	if(is_pseudo(fhp->fh_handle.fh_dev)){

		/* special case for pseudo root */
		if((IS_ROOT(fhp->fh_dentry)) && (!strcmp(name,"/"))){
			fh_init(resfh);
			fh_copy(resfh,fhp);
			goto out;
		}

		nfsv4_printk(vfs,
			"NFSV4: nfsd_lookup_pseudo fhp->fh_handle.fh_ino... %u\n",
			SVCFH_INO(fhp));

		/* get export */
		exp = get_new_pseudo_root_exp();

		/* get parent inode */
		iparent = pseudo_iget(fhp->fh_handle.fh_ino);	
		if(!iparent){
			printk("NFSV4: nfsd_lookup_pseudo: can't find parent inode# %d\n",
				fhp->fh_handle.fh_ino);
			return nfserr_noent;	
		}
		/* get child inode */
		ichild = pseudo_lookup(iparent,name);
		if(IS_ERR(ichild)){
			printk("NFSV4: nfsd_lookup_pseudo: can't find child inode# %d\n",
				fhp->fh_handle.fh_ino);
			return nfserr_noent;	
		}
		/* and child dentry */
		dchild = find_pseudo_dentry(ichild->i_ino);
		if(IS_ERR(dchild)){
			nfsv4_printk(andros,"ANDROS: nfs4_lookup_pseudo can't find child dentry\n");
			err = PTR_ERR(dchild);
			goto out;
		}
		print_pseudo_dentry(dchild);

		/* check if we have crossed a mount point ...  */
		if (dchild->d_covers != dchild->d_mounts) {
		struct dentry *tdentry;
		struct dentry *tdentry2;

			printk("NFSV4: nfsd_lookup_pseudo: pseudo_fs crossed mountpoint parent name %s, child name %s\n",
				dparent->d_name.name, dchild->d_name.name);
			tdentry = dchild->d_covers;
			tdentry2 = dchild->d_mounts;
#if 0
			if (tdentry == dchild){
				nfsv4_printk(vfs,"NFSDV4 nfsd4_lookup_pseudo: tdentry == dchild\n");
				err = nfserr_acces;
				goto out;
			}
#endif /*0*/
			/* leave pseudo_fs dentry alone - no dput() please !*/
			dchild = dget(tdentry2);
			if(dchild->d_inode){
				nfsv4_printk(vfs,"NFSV4 nfsd4_lookup_pseudo:ext2 dentry mount point HAS inode\n");
exp_get_again:
				exp = exp_get(rqstp->rq_client,dchild->d_inode->i_dev,
					dchild->d_inode->i_ino);
				if(!exp){
					nfsv4_printk(vfs,"NFSV4 nfsd4_lookup_pseudo: no export for inode(%d/%ld)\n",
						dchild->d_inode->i_dev,dchild->d_inode->i_ino);
				/* dchild is mounted on pseudo fs so 
				* 1) check permissions - if user is permitted to cross into 
				* this ext2 fs then,  
				* 2) call exp_client_export  - if no error then
				* 3) call exp_get again and proceed
				*/
					if(!rqstp->rq_client){
					/* this will occur if the client has yet to call SETCLIENTID 
					* which is the only other place exp_newaddclient is called.
					*/
						err = exp_newaddclient(rqstp);	
						printk("NFSV4 exp_newaddclient returns %d\n",err);
						if(err){
							err = nfserr_stale;
							goto out;
						}
					}
					err = exp_client_export(rqstp->rq_client,dchild);
					if(!err){
						printk("NFSV4 exp_client_export return success, goto exp_get_again\n");
						goto exp_get_again;
					}
					printk("NFSV4 exp_client_export failed\n");
					err = nfserr_stale;
					goto out;
				}
			}
			else {
				printk("NFSV4 nfsd4_lookup_pseudo: no inode ext2 dentry mount point\n");
				err = nfserr_noent;
				goto out;
			}
		}
	} /* is_pseudo */
	else { /* is ext2 */
		exp  = fhp->fh_export;
		/* initially, we set ex_sec via exportfs nfsctl_export by using
		* the "10.10.10.10" client string. (ugh). this will change soon
		* to a new field in /etc/exports. 
		* we also have one gss_svc mode hardwired - RPC_GSS_SVC_NONE.
		* futhermore, we have yet to decide where the gss creds will be stored
		* we think in a slab cache managed by the gssd interface. so, we
		* just check the cred struct flavor. if it's gss -> RPC_GSS_SVC_NONE.
		*/
		if(exp->ex_sec == RPC_GSS_SVC_NONE){
printk("RPC_GSS_SVC_NONE directory boundry\n");
			if(rqstp->rq_cred.cr_flavor != RPC_AUTH_GSS){
printk("rqstp->rq_cred.cr_flavor  %d NOT CORRECT - should be %d\n",rqstp->rq_cred.cr_flavor,RPC_AUTH_GSS);
				err = nfserr_wrongsec;
				goto out;
			}
		}
		dchild = lookup_dentry(name, dget(dparent), 0);
		if (IS_ERR(dchild)){
			nfsv4_printk(vfs,"NFSDV4 nfsd4_lookup_pseudo: lookup_dentry failed\n");
			goto out_nfserr;
		}
	}

	fh_compose(resfh, exp, dchild);
	err = nfserr_noent;
	if (dchild->d_inode){
		nfsv4_printk(vfs,"NFSDV4 nfsd4_lookup_pseudo: dchild HAS inode!\n");
		err = 0;
		goto out;
	}
	if(err == nfserr_noent){
		nfsv4_printk(vfs,
			"NFSDV4 nfsd4_lookup_pseudo: dchild has no inode err: %d\n",
			nfserr_noent);
		goto out;
	}

out_nfserr:
	err = nfsd4errno(-PTR_ERR(dchild));
	goto out;   
out:
	nfsv4_printk(vfs,"NFSDV4 nfsd4_lookup_pseudo: returning. err %d\n",err);
	return err;
}


/* 
* starting with the CURRENT_FH, look up each name in 
* args->FVAL[i].utf8string_val, and set TEMP_FH to resfh.
* Upon last successful lookup, set CURRENT_FH to TEMP_FH.
*/

int
nfsd4_lookup(struct svc_rqst *rqstp, pathname4 p4)
{
int i,err = 0;
char buf[NFS3_MAXNAMLEN];  
struct svc_fh resfh; 			/* hold intermediate results */

	nfsv4_printk(vfs,
		"NFSDV4 nfsd4_lookup: filenames_len: %d\n",p4.pathname4_len);
	nfsv4_printk(vfs,
		"NFSDV4 nfsd4_lookup:CURRENT_FH.fh_dentry->d_name.name: %s\n", 
		CURRENT_FH.fh_dentry->d_name.name);

	if(!strcmp(CURRENT_FH.fh_dentry->d_name.name,"/")){
		nfsv4_printk(vfs,
			"NFSDV4 nfsd4_lookup CURRENT_FH is root ex_flags: 0x%x\n",
			CURRENT_FH.fh_export->ex_flags);	

		if(CURRENT_FH.fh_export->ex_flags & NFSEXP_PSEUDO_EXP){
			nfsv4_printk(vfs,
				"NFSDV4 nfsd4_lookup CURRENT_FH is root and not exported\n");
		}
	}
	/* leave CURRENT_FH alone until done. work with TEMP_FH */
	fh_init(&TEMP_FH);
	fh_copy(&TEMP_FH,&CURRENT_FH);
	nfsv4_printk(vfs,"NFSDV4: nfs4d_lookup: TEMP_FH dev/ino: %d/%d name:%s\n",
	SVCFH_DEV(&TEMP_FH), SVCFH_INO(&TEMP_FH),
	TEMP_FH.fh_dentry->d_name.name);

	if(p4.pathname4_len == 0 ){
		nfsv4_printk(vfs,"NFSDV4 nfsd4_lookup pathname4_len is zero\n");
		return nfserr_inval;
	}
	/* parse pathname */
	for(i=0;i<p4.pathname4_len;i++){
		if(p4.pathname4_val[i].utf8string_len == 0){
			nfsv4_printk(vfs,"NFSDV4 nfsd4_lookup pathname4_len[%d].utf8string_len is zero\n",i);
			return nfserr_inval;
		}
		memset(buf,0,NFS4_MAXNAMLEN);
		memcpy(buf,p4.pathname4_val[i].utf8string_val,p4.pathname4_val[i].utf8string_len);
		nfsv4_printk(vfs,"NFSDV4 nfsd4_lookup name: %s\n",buf);

		fh_init(&resfh);
		err = 0;
		err = nfsd4_lookup_pseudo(rqstp,&TEMP_FH, buf,&resfh);
		if(err) {
			nfsv4_printk(vfs,"NFSDV4 nfsd4_lookup_pseudo failed - err %d\n",ntohl(err));
			return err;
		}
		if(!(TEMP_FH.fh_dentry)){
			fh_put(&TEMP_FH);
			printk("NFSV4 dentry not pseudo\n");
		}
		if(!IS_PSEUDO(resfh.fh_dentry)){
			fh_put(&resfh);
			printk("NFSV4 dentry not pseudo\n");
		}
		/* set TEMP_FH */	
 		fh_copy(&TEMP_FH,&resfh);
		nfsv4_printk(vfs,"NFSDV4 nfs4d_lookup: TEMP_FH dev/ino: %d/%d name:%s\n",
			SVCFH_DEV(&TEMP_FH), SVCFH_INO(&TEMP_FH),
			TEMP_FH.fh_dentry->d_name.name);
	}	
	/*success. set CURRENT_FH
	 * TEMP_FH has correct dentry/inode, but wrong export 
	 * (pseudo_fs export...). 
	 */

	if(IS_PSEUDO(TEMP_FH.fh_dentry)){
		nfsv4_printk(vfs,"NFSV4 dentry is pseudo\n");
		print_pseudo_dentry(TEMP_FH.fh_dentry);
		fh_copy(&CURRENT_FH,&TEMP_FH);
		goto out;
	}

/* not for new pseudo.. */
		err = set_client_exp(rqstp);
out:
	return err;
}
/* 
 * nfs4_lookup has detected a mount request and has successfully set the
 * mount point dentry and inode in temp_fh. now copy temp_fh to current_fh
 * and set current_fh's export from the pseudo_fs to the correct client's 
 * cl_export[i].
 */
int
set_client_exp(struct svc_rqst *rqstp)
{
struct svc_export *exp;
u32 error = 0;
        
	nfsv4_printk(vfs,"NFSDV4 set_client_exp\n");
	error = nfserr_stale;
	nfsv4_printk(vfs,"NFSDV4 set_client_exp calling exp_get dev %d ino %d\n",TEMP_FH.fh_handle.fh_dev,TEMP_FH.fh_handle.fh_ino);
        exp = exp_get(rqstp->rq_client,
			u32_to_kdev_t(TEMP_FH.fh_handle.fh_dev),
			u32_to_ino_t(TEMP_FH.fh_handle.fh_ino));
	if(exp){
		nfsv4_printk(vfs,"NFSDV4 set_client_exp found pseudo client exp, pseudo? %d\n",(exp->ex_flags && NFSEXP_PSEUDO_EXP)?1:0);
		if(exp->ex_flags && NFSEXP_PSEUDO_EXP)
			exp->ex_flags &= NFSEXP_PSEUDO_EXP;
		nfsv4_printk(vfs,"NFSDV4 set_client_exp still pseudo? %d\n",(exp->ex_flags && NFSEXP_PSEUDO_EXP)?1:0);
		TEMP_FH.fh_handle.fh_xdev = kdev_t_to_u32(exp->ex_dev);
		TEMP_FH.fh_handle.fh_xino = ino_t_to_u32(exp->ex_ino);
		TEMP_FH.fh_export = exp;
		fh_copy(&CURRENT_FH,&TEMP_FH);
		return 0;
	}else{
        	exp = exp_get(rqstp->rq_client,
			u32_to_kdev_t(TEMP_FH.fh_handle.fh_xdev),
			u32_to_ino_t(TEMP_FH.fh_handle.fh_xino));
		if(exp){
			fh_copy(&CURRENT_FH,&TEMP_FH);
			nfsv4_printk(vfs,"NFSDV4 set_client_exp found client exp, pseudo? %d\n",(exp->ex_flags && NFSEXP_PSEUDO_EXP)?1:0);
	nfsv4_printk(vfs,"NFSDV4 set_client_exp CURRENT_FH dev %d ino %d xdev %d xino %d\n",
			CURRENT_FH.fh_handle.fh_dev,
			CURRENT_FH.fh_handle.fh_ino,
			CURRENT_FH.fh_handle.fh_xdev,
			CURRENT_FH.fh_handle.fh_xino);

		return 0;
		}
	}
	nfsv4_printk(vfs,"NFSDV4 set_client_exp no client exp\n");
	return error;
}

/* set the CURRENT_FH */
int
nfsd4_putfh(struct svc_rqst *rqstp, PUTFH4args *args)
{
int 	size;
u32	error = 0;

	nfsv4_printk(vfs,"NFSDV4 nfsd4_putfh first args fh then TEMP_FH fh\n");
	print_nfs_fh4(&args->object);
	fh_init(&TEMP_FH);
	size = args->object.nfs_fh4_len;
	if(size != sizeof(struct knfs_fh))
		return EIO;
	memcpy(&TEMP_FH.fh_handle,args->object.nfs_fh4_val,size);
	print_knfs_fh(&TEMP_FH.fh_handle);
	error = fh_verify(rqstp,&TEMP_FH,0,MAY_NOP);
	if(!error)
		fh_copy(&CURRENT_FH,&TEMP_FH);
	nfsv4_printk(vfs,"NFSDV4 nfsd4_putfh STATUS %d \n", error);
	return error;
}

/* return the CURRENT_FH */
int
nfsd4_getfh(struct svc_rqst *rqstp, GETFH4res *res)
{
int 	size = sizeof(struct knfs_fh);
nfs_fh4 *fh4 = &res->GETFH4res_u.resok4.object;

	nfsv4_printk(vfs,"NFSDV4 nfsd4_getfh - first print knfs_fh, then nfs_fh4\n");
	print_knfs_fh(&CURRENT_FH.fh_handle);
	fh4->nfs_fh4_len = size;
	fh4->nfs_fh4_val = kmalloc(NFS4_FHSIZE,GFP_KERNEL);
	if(fh4->nfs_fh4_val){
		memset(fh4->nfs_fh4_val,0,NFS4_FHSIZE);
		memcpy(res->GETFH4res_u.resok4.object.nfs_fh4_val,&CURRENT_FH.fh_handle,size);
		print_nfs_fh4(&res->GETFH4res_u.resok4.object);
		return 0;
	}
	else 
		return nfserr_resource;
}

/* set the SAVE_FH */
int
nfsd4_savefh(struct svc_rqst *rqstp)
{
u32     status = 0;

        nfsv4_printk(rename,"NFSDV4 nfsd4_savefh\n");
        fh_init(&SAVE_FH);
	memset(&SAVE_FH.fh_handle,0,sizeof(struct knfs_fh));
        memcpy(&SAVE_FH.fh_handle,&CURRENT_FH.fh_handle,sizeof(struct knfs_fh));
        print_knfs_fh(&SAVE_FH.fh_handle);
        status = fh_verify(rqstp,&SAVE_FH,0,MAY_NOP);
        if(!status)
                fh_copy(&SAVE_FH,&CURRENT_FH);
        nfsv4_printk(rename,"NFSDV4 nfsd4_savefh STATUS %d \n", status);
        return status;
}

int
nfsd4_restorefh (struct svc_rqst *rqstp)
{
u32     status = 0;

        nfsv4_printk (rename,"NFSDV4 nfsd4_restorefh\n");
        fh_init (&CURRENT_FH);
	memset (&CURRENT_FH.fh_handle, 0, sizeof(struct knfs_fh));
        memcpy (&CURRENT_FH.fh_handle, &SAVE_FH.fh_handle, sizeof(struct knfs_fh));
        print_knfs_fh (&CURRENT_FH.fh_handle);
        status = fh_verify (rqstp, &CURRENT_FH, 0, MAY_NOP);
        if(!status)
                fh_copy (&CURRENT_FH, &SAVE_FH);
        nfsv4_printk(rename,"NFSDV4 nfsd4_restorefh STATUS %d \n", status);
        return status;
}


/* return bitmap4 marked attributes of the CURRENT_FH 
 * resp space has been kmalloc'ed in nfsd4_proc_compound() 
 */
int
nfsd4_getattr(struct svc_rqst *rqstp, GETATTR4args *args, GETATTR4res *res)
{
int status,attrlen = res->GFATTR4.FATTRSLEN;

	status = get_attributes(rqstp,&args->attr_request,&res->GFATTR4,(struct inode *)NULL);
	/* 
	 * NOTE: need to check each possible error - when
	 * server supported attribute failed,return no
	 * attributes. for now, this occurs on all errors.....
	 */
	if (status != 0) {
		memset(res->GFATTR4.FATTRSVAL,0,attrlen);
		res->GFATTR4.FATTRSLEN = 0;
	}
	if(!(CURRENT_FH.fh_dentry->d_flags & DCACHE_NFSDV4_PSEUDO_FS)){
		nfsv4_printk(attributes,"NFSV4: nfsd4_getattr calling fh_put\n");
		fh_put(&CURRENT_FH);
	}
	return status;
}

/*
 * gather CURRENT_FH attributes indicated in bitmap4, fill fattr4
 * with attribute values. 
 */
int
get_attributes(struct svc_rqst *rqstp,bitmap4 *bm4,fattr4 *f4,struct inode *ino)
{
char 			*attrvalptr = (char *)f4->FATTRSVAL;
int			attrlen = f4->FATTRSLEN;
int 			fh_v_err = 0;
struct inode 		*inode = NULL;
int 			statfs_err = 0;
int 			i = 0,bit;
int 			status = 0;
struct super_block 	*sb;
mm_segment_t 		oldfs;
struct statfs 		stat;

	/* f4->FATTRSVAL is kmalloc'ed in nfs4_proc_compound() large
	 * enough to hold a maximum size successful return.
         * it's size in bytes is passed in f4->FATTRSLEN and is 
	 * used to check for buffer overflow.
	 * f4->FATTRSLEN is then zero'ed and will hold the number of bytes
         * used in the return. 
         */
	f4->FATTRSLEN = 0;

nfsv4_printk(vfs,"NFSDV4 get_attributes attrvalptr 0x%x, FATTRSLEN %d\n",(unsigned int)attrvalptr,attrlen);

	/* NOTE: could write macro to check if any bits in args
	 * require an fh_verify() or a statfs() call. if so, do it once.
         * for now, do it once anyway
	 */
	if(!ino){ /* use CURRENT_FH */
		nfsv4_printk(vfs, "get_attributes() called on CURRENT_FH %s, inode %ld\n",
			     CURRENT_FH.fh_dentry->d_name.name, 
			     CURRENT_FH.fh_dentry->d_inode->i_ino);
		fh_v_err = fh_verify(rqstp,&CURRENT_FH,0,MAY_NOP);
		inode = CURRENT_FH.fh_dentry->d_inode;
		nfsv4_printk(vfs,"NFSDV4 get_attributes using CURRENT_FH\n");
	}
	else { /* use passed in ino - see fill_entry */
		nfsv4_printk(vfs, "NFSDV4 get_attributes using inode * 0x%x, not CURRENT_FH\n",
			     (unsigned int)inode);
		inode = ino;
	}

	if (!inode) {
		printk("nfsd: NULL inode in %s:%d", __FILE__, __LINE__);
		statfs_err = nfserr_io;
	}

	/*
	 * if no inode, all fattr4 requests (bits in args->attr_request)
 	 * that need the inode will fail. since this server  
	 * claims to support said requests, the whole getattr call
	 * should fail upon one of these failed inode dependent requests.
	 * the same is true of the statfs dependent requests.
         */
	else {
	  if (!(sb = inode->i_sb) || !sb->s_op->statfs)
	    statfs_err = nfserr_io;
	  else {
	    oldfs = get_fs();    // should all of this stuff be in the else..else branch?
	    set_fs (KERNEL_DS);
	    statfs_err = sb->s_op->statfs(sb, &stat, sizeof(struct statfs));
	    if (statfs_err)
	      statfs_err = nfsd4errno(-statfs_err);
	    set_fs (oldfs);
	  }
	}

	nfsv4_printk(vfs,"NFSDV4: vfs.c get_attributes bitmap val[0] 0x%x val[1] 0x%x\n",bm4->BVAL[0], bm4->BVAL[1]);
	do{
		bit = get_bitmap4_bit(bm4,i);
		i++;
		if(!bit || bit == -1)
			continue;
	
		nfsv4_printk(attributes,
		  "NFSDV4: vfs get_attributes bit %d is set\n",i-1);
		switch(i-1){
			case FATTR4_SUPPORTED_ATTRS:
			{
			int i;
		/* this case can not fail*/
			    	nfsv4_printk(attributes,
				  "NFSDV4 vfs.c get_attributes: FATTR4_SUPPORTED_ATTRS\n");
			    	*(u_int *)attrvalptr = ATTR_BITMAP.BLEN;
			    	attrvalptr += sizeof(u_int);
			    	f4->FATTRSLEN += sizeof(u_int);

			    	for(i=0;i<ATTR_BITMAP.BLEN;i++){
						*(uint32_t *)attrvalptr = ATTR_BITMAP.BVAL[i];
						attrvalptr += sizeof(uint32_t);
						f4->FATTRSLEN += sizeof(uint32_t);
					}

					set_bitmap4_bit(FATTR4_SUPPORTED_ATTRS,&f4->attrmask);
			}
			break;
			case FATTR4_TYPE:
					nfsv4_printk(attributes,
				"NFSDV4 vfs.c get_attributes: FATTR4_TYPE\n");
		    		if(inode){
		       	 		nfsv4_printk(attributes,
							"NFSDV4 - nfsd4_getattr  FATTR4_TYPE = %d\n",
							nfs4_ftypes[(inode->i_mode & S_IFMT) >> 12]);

		    /* translate CURRENT_FH:inode->i_mode 
		     * into NFSv4 type 
		     */
						*(fattr4_type *)attrvalptr = (u32)nfs4_ftypes[(inode->i_mode & S_IFMT) >> 12];

		       		 set_bitmap4_bit(FATTR4_TYPE,&f4->attrmask);
		       		 f4->FATTRSLEN +=  sizeof(fattr4_type); 
		       		 attrvalptr += sizeof(fattr4_type);
		    		}
		    		else{ /* FAILURE */
		       	 		status = fh_v_err;
		       	 		nfsv4_printk(attributes,
							"NFSDV4 - nfsd4_getattr  FATTR4_TYPE FAILED. status %d\n",
							ntohl(status));
		        	goto errout;
					}
			break;
			case FATTR4_FH_EXPIRE_TYPE:
		    	{
		    	fattr4_fh_expire_type fhtype = 0;

		    		nfsv4_printk(attributes,"NFSDV4 vfs.c get_attributes: FATTR4_FH_EXPIRE_TYPE\n");
		    		fhtype |= FH4_PERSISTENT;
		    		*(fattr4_fh_expire_type *)attrvalptr = fhtype;
		    		set_bitmap4_bit(FATTR4_FH_EXPIRE_TYPE,&f4->attrmask);
		    		f4->FATTRSLEN +=  sizeof(fattr4_fh_expire_type); 
		    		attrvalptr += sizeof(fattr4_fh_expire_type);
		   	 }
			break;
			case FATTR4_CHANGE:
		    		nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_CHANGE\n");
		    		if(inode){
						nfsv4_printk(attributes,
			   				 "NFSDV4 - nfsd4_getattr  FATTR4_CHANGE = %d\n",
			 	  			 (int)inode->i_mtime);
						*(fattr4_change *)attrvalptr = (u64)inode->i_mtime;

						set_bitmap4_bit(FATTR4_CHANGE,&f4->attrmask);
						f4->FATTRSLEN +=  sizeof(fattr4_change);
						attrvalptr += sizeof(fattr4_change);
		   		 	}
		   		 	else{ /* FAILURE */
						status = fh_v_err;
						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes  FATTR4_CHANGE FAILED. status %d\n",
							ntohl(status));
						goto errout;
		   	 		}
			break;
			case FATTR4_SIZE:
		    	nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_SIZE\n");
		    	if(inode){
		      	nfsv4_printk(attributes,
							"NFSDV4 - nfsd4_getattr  FATTR4_SIZE = %ld\n",
							inode->i_size);
						if(S_ISLNK(inode->i_mode) && inode->i_size > NFS4_MAXPATHLEN) {
			   			*(fattr4_size *)attrvalptr = (u64)NFS4_MAXPATHLEN;
		    		}else {
			   				 *(fattr4_size *)attrvalptr = (u64)inode->i_size;
						}
						set_bitmap4_bit(FATTR4_SIZE,&f4->attrmask);
						f4->FATTRSLEN +=  sizeof(fattr4_size); 
						attrvalptr += sizeof(fattr4_size);
		   		}
		    	else{ /* FAILURE */
						status = fh_v_err;
						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes  FATTR4_SIZE FAILED. status %d\n",
							ntohl(status));
						goto errout;
		   		}
				break;
				case FATTR4_LINK_SUPPORT:
		   	 nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_LINK_SUPPORT = 1\n");
		   	 *(fattr4_link_support *)attrvalptr = (fattr4_link_support)1;
		   	 set_bitmap4_bit(FATTR4_LINK_SUPPORT,&f4->attrmask);
		   	 f4->FATTRSLEN +=  sizeof(fattr4_link_support); 
		   	 attrvalptr += sizeof(fattr4_link_support);
				break;
				case FATTR4_SYMLINK_SUPPORT:
		    	nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_SYMLINK_SUPPORT = 1\n");
		   	 *(fattr4_symlink_support *)attrvalptr = (fattr4_symlink_support)1;
		   	 set_bitmap4_bit(FATTR4_SYMLINK_SUPPORT,&f4->attrmask);
		   	 f4->FATTRSLEN +=  sizeof(fattr4_symlink_support); 
		   	 attrvalptr += sizeof(fattr4_symlink_support);
				break;
				case FATTR4_NAMED_ATTR:
			    nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_NAMED_ATTR\n");
				break;
				case FATTR4_FSID:
				{
				fattr4_fsid *ptr = (fattr4_fsid *)attrvalptr;

					nfsv4_printk(attributes,"NFSDV4 vfs.c get_attributes: FATTR4_FSID\n");
					if(inode){
		        nfsv4_printk(attributes,
							"NFSDV4 - nfsd4_getattr  FATTR4_FSID. Major = %d, Minor = %d\n",
							MAJOR(inode->i_dev),MINOR(inode->i_dev));

		        ptr->major = (u64)MAJOR(inode->i_dev);
		        ptr->minor = (u64)MINOR(inode->i_dev);

		        set_bitmap4_bit(FATTR4_FSID,&f4->attrmask);
    		        f4->FATTRSLEN +=  sizeof(fattr4_fsid); 
		        attrvalptr += sizeof(fattr4_fsid);
		    	}
		    	else{ /* FAILURE */
						status = fh_v_err;
						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes  FATTR4_FSID FAILED. status %d\n",
							ntohl(status));
						goto errout;
		 	  	}
				}
				break;
				case FATTR4_ACL:
		   	 nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_ACL\n");
				break;
				case FATTR4_ACLSUPPORT:
		   	 nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_ACLSUPPORT\n");
				break;
				case FATTR4_ARCHIVE:
			    nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_ARCHIVE\n");
				break;
				case FATTR4_CANSETTIME:
			    nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_CANSETTIME\n");
				break;
				case FATTR4_CASE_INSENSITIVE:
		   	 nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_CASE_INSENSITIVE\n");
				break;
				case FATTR4_CASE_PRESERVING:
			  	nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_PRESERVING\n");
				break;
				case FATTR4_CHOWN_RESTRICTED:
		    	nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_CHOWN_RESTRICTED\n");
				break;
				case FATTR4_FILEHANDLE:
		   	 nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_FILEHANDLE\n");
				break;
				case FATTR4_FILEID:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_FILEID\n");
					if(inode){
						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes  FATTR4_FILEID = %ld\n",inode->i_ino);
					*(fattr4_fileid *)attrvalptr = (u64)inode->i_ino;

					set_bitmap4_bit(FATTR4_FILEID,&f4->attrmask);
					f4->FATTRSLEN +=  sizeof(fattr4_fileid);
					attrvalptr += sizeof(fattr4_fileid);
		    }
		    else{ /* FAILURE */
					status = fh_v_err;
					nfsv4_printk(attributes,
						"NFSDV4 - get_attributes  FATTR4_FILEID FAILED. status %d\n",
					ntohl(status));
				goto errout;
		    }
				break;
				case FATTR4_FILES_AVAIL:
		   	 nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_FILE_AVAIL\n");
				break;
				case FATTR4_FILES_FREE:
				  if (statfs_err) {      // FAILURE
				    status = statfs_err;
				    goto errout;
				  }
				  set_bitmap4_bit (FATTR4_FILES_FREE, &f4->attrmask);
				  f4->FATTRSLEN += sizeof(fattr4_files_free);
				  *((fattr4_files_free *) attrvalptr) = stat.f_ffree;
				  attrvalptr += sizeof(fattr4_files_free);
				break;
				case FATTR4_FILES_TOTAL:
				  if (statfs_err) {      // FAILURE
				    status = statfs_err;
				    goto errout;
				  }
				  set_bitmap4_bit (FATTR4_FILES_TOTAL, &f4->attrmask);
				  f4->FATTRSLEN += sizeof(fattr4_files_total);
				  *((fattr4_files_total *) attrvalptr) = stat.f_files;
				  attrvalptr += sizeof(fattr4_files_total);
				break;
				case FATTR4_HIDDEN:
		    nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_FILE_HIDDEN\n");
				break;
				case FATTR4_HOMOGENEOUS:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_FILE_HOMOGENEOUS\n");
					*(fattr4_homogeneous *)attrvalptr = (fattr4_homogeneous)1;
		    	set_bitmap4_bit(FATTR4_HOMOGENEOUS,&f4->attrmask);
		    	attrvalptr += sizeof(fattr4_homogeneous);
		    	f4->FATTRSLEN += sizeof(fattr4_homogeneous);
				break;
				case FATTR4_MAXFILESIZE:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_MAXFILESIZE\n");
					*(fattr4_maxfilesize *)attrvalptr = (u64)(~(u32) 0);
					set_bitmap4_bit(FATTR4_MAXFILESIZE,&f4->attrmask);
					f4->FATTRSLEN +=  sizeof(fattr4_maxfilesize); 
					attrvalptr += sizeof(fattr4_maxfilesize);
				break;
				case FATTR4_MAXLINK:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_MAXLINK\n");
				break;
				case FATTR4_MAXNAME:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_MAXNAME\n");
					*(fattr4_maxname *)attrvalptr = (u32)NFS4_MAXNAMLEN;
					set_bitmap4_bit(FATTR4_MAXNAME,&f4->attrmask);
					f4->FATTRSLEN +=  sizeof(fattr4_maxname); 
					attrvalptr += sizeof(fattr4_maxname);
				break;
				case FATTR4_MAXREAD:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_MAXREAD\n");
					*(fattr4_maxread *)attrvalptr = (u64)NFSSVC_MAXBLKSIZE;
					set_bitmap4_bit(FATTR4_MAXREAD,&f4->attrmask);
					f4->FATTRSLEN +=  sizeof(fattr4_maxread); 
					attrvalptr += sizeof(fattr4_maxread);
				break;
				case FATTR4_MAXWRITE:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_MAXWRITE\n");
					*(fattr4_maxwrite *)attrvalptr = (u64)NFSSVC_MAXBLKSIZE;
					set_bitmap4_bit(FATTR4_MAXWRITE,&f4->attrmask);
					f4->FATTRSLEN +=  sizeof(fattr4_maxwrite); 
					attrvalptr += sizeof(fattr4_maxwrite);
				break;
				case FATTR4_MODE:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_MODE\n");
					if(inode){
						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes  FATTR4_MODE = %d\n",
						inode->i_mode);
						*(fattr4_mode *)attrvalptr = (u32)inode->i_mode;
						set_bitmap4_bit(FATTR4_MODE,&f4->attrmask);
						f4->FATTRSLEN +=  sizeof(fattr4_mode); /* increment attrvalprt */
						attrvalptr += sizeof(fattr4_mode);
					}
					else{ /* FAILURE */
						status = fh_v_err;
						nfsv4_printk(attributes,
						"NFSDV4 - get_attributes  FATTR4_MODE FAILED. status %d\n",
								ntohl(status));
					goto errout;
				}
				break;
				case FATTR4_NO_TRUNC:
					nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_NO_TRUNK\n");
				break;
				case FATTR4_NUMLINKS:
					nfsv4_printk(attributes,"NFSDV4 vfs.c get_attributes: FATTR4_NUMLINKS\n");
					if(inode){
						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes  FATTR4_NUMLINKS = %d\n",inode->i_nlink);
						*(fattr4_numlinks *)attrvalptr = (u32)inode->i_nlink;
						set_bitmap4_bit(FATTR4_NUMLINKS,&f4->attrmask);
						f4->FATTRSLEN +=  sizeof(fattr4_numlinks); 
						attrvalptr += sizeof(fattr4_numlinks);
					}
					else{ /* no inode, FAILURE */
						status = fh_v_err;
						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes  FATTR4_NUMLINKS FAILED. status %d\n",ntohl(status));
					goto errout;
				}
				break;
				case FATTR4_OWNER:
				{
				  int err;
				  utf8string s;
				  int i;

				  if (!inode) {
				    status = fh_v_err;
				    printk ("NFSDV4-get_attributes FATTR4_OWNER FAILED. status %d\n", ntohl(status));
				    goto errout;
				  }

				  nfsv4_printk(vfs, "get_attributes() calling get_name() with uid %d\n", inode->i_uid);
				  err = get_name (NUM2STR_UID, inode->i_uid, &s);
				  if (err) {
					  nfsv4_printk(vfs, "get_name() returned error\n");
					  status = -1;    // XXX: what is the right error to return?
					  goto errout;
				  }
				  nfsv4_printk(vfs, "get_name() returned %.*s\n",
					       (int) s.utf8string_len, s.utf8string_val);

				  *((u_int *) attrvalptr) = s.utf8string_len;
				  attrvalptr += sizeof(u_int);
				  *((char **) attrvalptr) = attrvalptr + sizeof(char *);
				  attrvalptr += sizeof(char *);
				  memcpy (attrvalptr, s.utf8string_val, s.utf8string_len);
				  attrvalptr += s.utf8string_len;
				  set_bitmap4_bit (FATTR4_OWNER, &f4->attrmask);
				  f4->FATTRSLEN += sizeof(fattr4_owner) + s.utf8string_len;
				  kfree (s.utf8string_val);
				}
				break;
				case FATTR4_OWNER_GROUP:
				{
				  int err;
				  utf8string s;
				  int i;

				  if (!inode) {
				    status = fh_v_err;
				    printk ("NFSDV4-get_attributes FATTR4_OWNER_GROUP FAILED. status %d\n", ntohl(status));
				    goto errout;
				  }

				  nfsv4_printk(vfs, "get_attributes() calling get_name() with gid %d\n", inode->i_gid);
				  err = get_name (NUM2STR_GID, inode->i_gid, &s);
				  if (err) {
				    nfsv4_printk(vfs, "get_name() returned error\n");
				    status = -1;    /* XXX: what is the right error to return? */
				    goto errout;
				  }
				  nfsv4_printk(vfs, "get_name() returned %.*s\n",
					       (int) s.utf8string_len, s.utf8string_val);

				  *((u_int *) attrvalptr) = s.utf8string_len;
				  attrvalptr += sizeof(u_int);
				  *((char **) attrvalptr) = attrvalptr + sizeof(char *);
				  attrvalptr += sizeof(char *);
				  memcpy (attrvalptr, s.utf8string_val, s.utf8string_len);
				  attrvalptr += s.utf8string_len;
				  set_bitmap4_bit (FATTR4_OWNER_GROUP, &f4->attrmask);
				  f4->FATTRSLEN += sizeof(fattr4_owner_group) + s.utf8string_len;
				  kfree (s.utf8string_val);
				}
				break;
				case FATTR4_QUOTA_HARD:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_QUOTA_HARD\n");
				break;
				case FATTR4_QUOTA_SOFT:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_QUOTA_SOFT\n");
				break;
				case FATTR4_QUOTA_USED:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_QUOTA_USED\n");
				break;
				case FATTR4_RAWDEV:
				{
				fattr4_rawdev *ptr = (fattr4_rawdev *)attrvalptr;
					nfsv4_printk(attributes,"NFSDV4 vfs.c get_attributes: FATTR4_RAWDEV\n");
					if(inode){

					nfsv4_printk(attributes,
						"NFSDV4 - get_attributes  FATTR4_RAWDEV. Major = %d, Minor = %d\n",
						(u32)MAJOR(inode->i_rdev),(u32)MINOR(inode->i_rdev));
					ptr->specdata1 = (u32)MAJOR(inode->i_rdev);
					ptr->specdata2 = (u32)MINOR(inode->i_rdev);

					set_bitmap4_bit(FATTR4_RAWDEV,&f4->attrmask);
					f4->FATTRSLEN +=  sizeof(fattr4_rawdev); 
					attrvalptr += sizeof(fattr4_rawdev);
					}
					else{ /* FAILURE */
					status = fh_v_err;
					nfsv4_printk(attributes,
						"NFSDV4 - get_attributes  FATTR4_RAWDEV FAILED. status %d\n",
						ntohl(status));
					goto errout;
					}
				}
				break;
				case FATTR4_SPACE_AVAIL:
				  if (statfs_err) {      // FAILURE
				    status = statfs_err;
				    goto errout;
				  }
				  set_bitmap4_bit (FATTR4_SPACE_AVAIL, &f4->attrmask);
				  f4->FATTRSLEN += sizeof(fattr4_space_avail);
				  *((fattr4_space_avail *) attrvalptr) = stat.f_bavail * stat.f_bsize;
				  attrvalptr += sizeof(fattr4_space_avail);
				break;
				case FATTR4_SPACE_FREE:
				  if (statfs_err) {      // FAILURE
				    status = statfs_err;
				    goto errout;
				  }
				  set_bitmap4_bit (FATTR4_SPACE_FREE, &f4->attrmask);
				  f4->FATTRSLEN += sizeof(fattr4_space_free);
				  *((fattr4_space_free *) attrvalptr) = stat.f_bfree * stat.f_bsize;
				  attrvalptr += sizeof(fattr4_space_free);
				break;
				case FATTR4_SPACE_TOTAL:
				  if (statfs_err) {      // FAILURE
				    status = statfs_err;
				    goto errout;
				  }
				  set_bitmap4_bit (FATTR4_SPACE_TOTAL, &f4->attrmask);
				  f4->FATTRSLEN += sizeof(fattr4_space_total);
				  *((fattr4_space_total *) attrvalptr) = stat.f_blocks * stat.f_bsize;
				  attrvalptr += sizeof(fattr4_space_total);
				break;
				case FATTR4_SPACE_USED:
					nfsv4_printk(attributes,"NFSDV4 vfs.c get_attributes: FATTR4_SPACE_USED\n");
					if(inode){
					nfsv4_printk(attributes,
						"NFSDV4 - get_attributes  FATTR4_SPACE_USED = %ld\n",
						inode->i_blksize * inode->i_blocks);
					*(fattr4_size *)attrvalptr = (u64)(inode->i_blksize * inode->i_blocks);
					set_bitmap4_bit(FATTR4_SPACE_USED,&f4->attrmask);
					f4->FATTRSLEN +=  sizeof(fattr4_space_used); 
					attrvalptr += sizeof(fattr4_space_used);
					}
					else{ /* FAILURE */
						status = fh_v_err;
						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes  FATTR4_SPACE_USED FAILED. status %d\n",
							ntohl(status));
						goto errout;
		 	   }
				break;
				case FATTR4_SYSTEM:
					nfsv4_printk(attributes,"NFSDV4 vfs.c get_attributes: FATTR4_SYSTEM\n");
				break;
				case FATTR4_TIME_ACCESS:
				{
				fattr4_time_access *ptr = (fattr4_time_access *)attrvalptr;

		   	 nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_TIME_ACCESS\n");
					if(inode){

						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes  FATTR4_TIME_ACCESS. seconds = %d \n",
							(int)inode->i_mtime);

						ptr->seconds = (u64)inode->i_atime;
						nfsv4_printk(attributes,
                 "NFSV4 get_attributes time_access seconds: [0x%lx , %08lx]\n",
    (unsigned long)(ptr->seconds >> 32),
    (unsigned long)(ptr->seconds & 0xFFFFFFFFUL));
						ptr->nseconds = (u32)0;

						set_bitmap4_bit(FATTR4_TIME_ACCESS,&f4->attrmask);
						f4->FATTRSLEN +=  sizeof(fattr4_time_access); 
						attrvalptr += sizeof(fattr4_time_access);
		  	  }
		    	else{ /* FAILURE */
						status = fh_v_err;
						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes FATTR4_TIME_ACCESS FAILED. status %d\n",
							ntohl(status));
						goto errout;
		 	 	  }
				}
				break;
				case FATTR4_TIME_ACCESS_SET:
		   	 nfsv4_printk(attributes,
					"NFSDV4 vfs.c get_attributes: FATTR4_TIME_ACCESS_SET\n");
				break;
				case FATTR4_TIME_BACKUP:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_TIME_BACKUP\n");
				break;
				case FATTR4_TIME_CREATE:
				{
				fattr4_time_create *ptr = (fattr4_time_create *)attrvalptr;

			    nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_TIME_CREATE\n");
			    if(inode){
						nfsv4_printk(attributes,"NFSDV4 - get_attributes  FATTR4_TIME_CREATE. seconds = %d \n",(int)inode->i_ctime);

						ptr->seconds = (u64)inode->i_ctime;
						nfsv4_printk(attributes,
                 "NFSV4 get_attributes time_create seconds: [0x%lx , %08lx]\n",
    (unsigned long)(ptr->seconds >> 32),
    (unsigned long)(ptr->seconds & 0xFFFFFFFFUL));
						ptr->nseconds = (u32)0;
	
						set_bitmap4_bit(FATTR4_TIME_CREATE,&f4->attrmask);
						f4->FATTRSLEN +=  sizeof(fattr4_time_create); 
						attrvalptr += sizeof(fattr4_time_create);
					}
					else{ /* FAILURE */
						status = fh_v_err;
						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes  FATTR4_TIME_CREATE FAILED. status %d\n",
							ntohl(status));
						goto errout;
		    	}
				}
				break;
				case FATTR4_TIME_DELTA:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_TIME_DELTA\n");
				break;
				case FATTR4_TIME_METADATA:
					nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_TIME_METADATA\n");
				break;
				case FATTR4_TIME_MODIFY:
				{
				fattr4_time_modify *ptr = (fattr4_time_modify *)attrvalptr;
		 			nfsv4_printk(attributes,"NFSDV4 vfs.c get_attributes: FATTR4_TIME_MODIFY\n");
					if(inode){
						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes  FATTR4_TIME_MODIFY. seconds = %d \n",
							(int)inode->i_mtime);
						ptr->seconds = (u64)inode->i_mtime;
						nfsv4_printk(attributes,
                 "NFSV4 get_attributes time_modify seconds: [0x%lx , %08lx]\n",
    (unsigned long)(ptr->seconds >> 32),
    (unsigned long)(ptr->seconds & 0xFFFFFFFFUL));

						ptr->nseconds = (u32)0;
						set_bitmap4_bit(FATTR4_TIME_MODIFY,&f4->attrmask);
						f4->FATTRSLEN +=  sizeof(fattr4_time_modify); 
						attrvalptr += sizeof(fattr4_time_modify);
		   		}
		   		else{ /* FAILURE */
						status = fh_v_err;
						nfsv4_printk(attributes,
							"NFSDV4 - get_attributes  FATTR4_TIME_MODIFY FAILED. status %d\n",
								ntohl(status));
						goto errout;
		    	}
				}
				break;
				case FATTR4_TIME_MODIFY_SET:
		 			nfsv4_printk(attributes,
						"NFSDV4 vfs.c get_attributes: FATTR4_TIME_MODIFY_SET\n");
				break;
				default:
				break;
		}
	}while(bit != -1);
	f4->FATTRSLEN = XDR_QUADLEN(f4->FATTRSLEN) << 2;
	nfsv4_printk(vfs,"NFSDV4 get_attributes returning bitmap [0] 0x%x [1] 0x%x, FATTRSLEN %d\n",f4->attrmask.BVAL[0],f4->attrmask.BVAL[1],f4->FATTRSLEN);
	return 0;
errout:
	return status;
}

/*
 * Read data from a file. count must contain the requested read count
 * on entry. On return, *count contains the number of bytes actually read.
 * N.B. After this call fhp needs an fh_put
 */
int
nfsd4_read(struct svc_rqst *rqstp,READ4args *args,READ4res *res) 
{
	struct raparms	*ra;
	mm_segment_t	oldfs;
	int		err;
	struct file	file;
	struct inode *inode = CURRENT_FH.fh_dentry->d_inode;

	nfsv4_printk(vfs,"NFSDV4 nfsd4_read\n");

	err = nfsd_open(rqstp, &CURRENT_FH, S_IFREG, MAY_READ, &file);
	if (err)
		goto out;
	nfsv4_printk(vfs,"NFSDV4 nfsd4_read 1\n");
	err = nfserr_perm;
	if (!file.f_op->read)
		goto out_close;
	nfsv4_printk(vfs,"NFSDV4 nfsd4_read 2\n");

	/* Get readahead parameters */
	ra = nfsd_get_raparms(CURRENT_FH.fh_handle.fh_dev, CURRENT_FH.fh_handle.fh_ino);
	nfsv4_printk(vfs,"NFSDV4 nfsd4_read 3\n");
	if (ra) {
		file.f_reada = ra->p_reada;
		file.f_ramax = ra->p_ramax;
		file.f_raend = ra->p_raend;
		file.f_ralen = ra->p_ralen;
		file.f_rawin = ra->p_rawin;
	}
	file.f_pos = args->offset;

	oldfs = get_fs(); set_fs(KERNEL_DS);
	err = file.f_op->read(&file, res->READVAL, args->count, &file.f_pos);
	set_fs(oldfs);
	nfsv4_printk(vfs,"NFSDV4 nfsd4_read 4\n");

	/* Write back readahead params */
	if (ra != NULL) {
		dprintk("nfsd: raparms %ld %ld %ld %ld %ld\n",
			file.f_reada, file.f_ramax, file.f_raend,
			file.f_ralen, file.f_rawin);
		ra->p_reada = file.f_reada;
		ra->p_ramax = file.f_ramax;
		ra->p_raend = file.f_raend;
		ra->p_ralen = file.f_ralen;
		ra->p_rawin = file.f_rawin;
		ra->p_count -= 1;
	}

	nfsv4_printk(vfs,"NFSDV4 nfsd4_read 5 err %d\n",err);
	if (err >= 0) {
		res->READLEN = err;
		nfsv4_printk(vfs,"NFSDV4 nfsd4_read READLEN: %d\n",res->READLEN);
		err = 0;
		res->READEOF = ((args->offset + res->READLEN) >= inode->i_size);
		nfsv4_printk(vfs,
			"NFSDV4 nfsd4_read args->offset %Ld inode->i_size %ld\n", args->offset,(off_t)inode->i_size);
		nfsv4_printk(vfs,"NFSDV4 nfsd4_read res->READEOF: %d\n",res->READEOF);
	} else 
		err = nfsd4errno(-err);
out_close:
	nfsd_close(&file);
out:
	nfsv4_printk(vfs,"NFSDV4 nfsd4_read END err = %d\n",err);
	return err;
}

/*
 * Write data to a file.
 * The stable flag requests synchronous writes.
 * N.B. After this call fhp needs an fh_put
 * nfsd4_write(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset,
 *				char *buf, unsigned long cnt, int stable)
 */
int
nfsd4_write(struct svc_rqst *rqstp,WRITE4args *args,WRITE4res *res) 
{
	struct svc_export	*exp;
	struct file		file;
	struct dentry		*dentry;
	struct inode		*inode;
	mm_segment_t		oldfs;
	int			err = 0;
#ifdef CONFIG_QUOTA
	uid_t			saved_euid;
#endif

	nfsv4_printk(vfs,"NFSDV4 nfsd4_write. WRITELEN %d\n",args->WRITELEN);

	if (!args->WRITELEN)
		goto out;
	err = nfsd_open(rqstp, &CURRENT_FH, S_IFREG, MAY_WRITE, &file);
	if (err)
		goto out;
	err = nfserr_perm;
	if (!file.f_op->write)
		goto out_close;

	dentry = file.f_dentry;
	inode = dentry->d_inode;
	exp   = CURRENT_FH.fh_export;

#if 0
	/* NFSV4 - look at this later */

	/*
	 * Request sync writes if
	 *  -	the sync export option has been set, or
	 *  -	the client requested O_SYNC behavior (NFSv3 feature).
	 *  -   The file system doesn't support fsync().
	 * When gathered writes have been configured for this volume,
	 * flushing the data to disk is handled separately below.
	 */
	if (rqstp->rq_vers == 2)
		stable = EX_ISSYNC(exp);
	else if (file.f_op->fsync == 0)
	       stable = 1;
	if (stable && !EX_WGATHER(exp))
		file.f_flags |= O_SYNC;
#endif /*0*/

	fh_lock(&CURRENT_FH);			/* lock inode */
	file.f_pos = args->offset;		/* set write offset */

	/* Write the data. */
	oldfs = get_fs(); set_fs(KERNEL_DS);
#ifdef CONFIG_QUOTA
	/* This is for disk quota. */
	saved_euid = current->euid;
	current->euid = current->fsuid;
	err = file.f_op->write(&file, args->WRITEVAL, args->WRITELEN, &file.f_pos);
	current->euid = saved_euid;
#else
	err = file.f_op->write(&file, args->WRITEVAL, args->WRITELEN, &file.f_pos);
#endif
	set_fs(oldfs);

	/* clear setuid/setgid flag after write */
	if (err >= 0 && (inode->i_mode & (S_ISUID | S_ISGID))) {
		struct iattr	ia;
		kernel_cap_t	saved_cap = 0;

		ia.ia_valid = ATTR_MODE;
		ia.ia_mode  = inode->i_mode & ~(S_ISUID | S_ISGID);
		if (current->fsuid != 0) {
			saved_cap = current->cap_effective;
			cap_clear(current->cap_effective);
		}
		notify_change(dentry, &ia);
		if (current->fsuid != 0)
			current->cap_effective = saved_cap;
	}

	fh_unlock(&CURRENT_FH);			/* unlock inode */

	if (err >= 0 && args->stable) {
		static unsigned long	last_ino = 0;
		static kdev_t		last_dev = NODEV;

		/*
		 * Gathered writes: If another process is currently
		 * writing to the file, there's a high chance
		 * this is another nfsd (triggered by a bulk write
		 * from a client's biod). Rather than syncing the
		 * file with each write request, we sleep for 10 msec.
		 *
		 * I don't know if this roughly approximates
		 * C. Juszak's idea of gathered writes, but it's a
		 * nice and simple solution (IMHO), and it seems to
		 * work:-)
		 */
		if (EX_WGATHER(exp) && (inode->i_writecount > 1
		 || (last_ino == inode->i_ino && last_dev == inode->i_dev))) {
#if 0
			interruptible_sleep_on_timeout(&inode->i_wait, 10 * HZ / 1000);
#else
			dprintk("nfsd: write defer %d\n", current->pid);
/* FIXME: Olaf commented this out [gam3] */
			schedule_timeout((HZ+99)/100);
			dprintk("nfsd: write resume %d\n", current->pid);
#endif
		}

		if (inode->i_state & I_DIRTY) {
			dprintk("nfsd: write sync %d\n", current->pid);
			nfsd_sync(&file);
		}
		wake_up(&inode->i_wait);
		last_ino = inode->i_ino;
		last_dev = inode->i_dev;
	}

	nfsv4_printk(vfs,"NFSDV4 nfsd4_write: write complete err=%d\n", err);
	if (err >= 0)
	{
		res->WRITERES.count = err;
		/* first pass - simplest values */
		res->WRITERES.committed = UNSTABLE4;
		set_verifier(res->WRITERES.writeverf,NULL);

		err = 0;
	}
	else 
		err = nfsd4errno(-err);
out_close:
	nfsd_close(&file);
out:
	return err;
}
/*
 * fill entry4 struct.
 * entry4 is a cookie (2) plus utf8string_len (1) 
 * plus variable value of utf8string_len
 * plus fattr4 
 * 	bitmap4 (1) + variable bitmap4_len
 *	attr_vals.attrlist_len (1) 
 *	plus variable value of attr_vals.attrlist_len
 * plus entry4* (size 1)
 *
 * Because XDR requires the entry4 *nextentry as a
 * boolean at the beginning of each entry4, fill_entry stuffs
 * cd->buffer as follows:
 * nextentry (1 if there is one)
 * cookie
 * name
 * attrs
 *
 * NOTE: this will turn into an encode function
 * where instead of kmalloc a resp->RESVAL[i].RESU.opreaddir.ENTRIES
 * in nfs4proc.c, i'll pass in the sunrpc return buf iovec
 * and stuff xdr_encoded values here instead of in nfs4xdr.c.
 */
#define NFS_ENTRY4_BAGGAGE      24
#define NFS_ENTRY4_HACK          3

int
fill_entry4(struct readdir_cd *cd, const char *name,
                        int namlen, off_t offset, ino_t ino)
{

        u32             *p = cd->buffer;
        int             elen;
	int		err = 0;
	int		i;
	struct fattr4	attrs;
	u_int 		*p2;
	struct inode	*inode;

	nfsv4_printk(vfs,"NFSDV4 VFS fill_entry name %s namlen %d offset %ld ino %ld buffer 0x%x\n", name,namlen,offset,ino,(unsigned int)cd->buffer);

#if 1
	if ((namlen == 1) && (name[0] == '.'))      // server does not return "." to client
	  return 0;
	if ((namlen == 2) && (name[0] == '.') && (name[1] == '.'))   // server does not return ".." to client
	  return 0;
#endif

	/* 
	 * here we stuff the current file.pos into the previous entry4 
	 * cookie field pointed to by cd->offset, overwriting the 
	 * NFS_OFFSETMAX value.
	 */
        if (cd->offset) {
		nfsv4_printk(vfs,"NFSDV4 fill_entry RESET cd->offset\n");
                *cd->offset = (u64) offset;
	}

        /* nfsd_readdir calls us with name == 0 when it wants us to
         * set the last offset entry. */
        if (name == 0){
		nfsv4_printk(vfs,"NFSDV4 fill_entry name = 0\n");
                return 0;
	}

        /* truncate filename if too long */
        if (namlen > NFS4_MAXNAMLEN)
                namlen = NFS4_MAXNAMLEN;

	/* Calculate largest size of entry4, see if it's too big for 
	 * whats left of the buffer. 
	 * could be less if get_attributes fails.
         */
        elen = NFS_ENTRY4_BAGGAGE + namlen + (4 * cd->bm4->BLEN) + get_fattr4_res_buf_size(cd->bm4) + NFS_ENTRY4_HACK;
        if ((cd->buflen - elen) < 0) {
                cd->eob = 1;
                return -EINVAL;
        }

	/* entry4 values */

		/* nextentry */
	*(u32 *)p = 1; 		/* 1 indicates more entry4s in p*/
	p += 1;


	/* cookie: where the READDIR should start
	* within the ext2 directory. This will get overwritten
	* the next time through the nfsd4_lookup while loop
	* with the file.pos of the next dirent.
	*/
        cd->offset = p;                 /* remember pointer */
	*(u64 *)p =  NFS_OFFSET_MAX;
	p += 2;

	*(u_int *)p = namlen;
	p += 1;
	for(i=0;i<namlen;i++){
		*(char *)p = name[i];
		(char *)p += sizeof(char);
	}
		/* attrs */
	/* give get_attributes a 'real' fattr... 
	 * and let it stuff p with its results.
	 *
	 * return bitmap4_len = args bitmap4_len...
	 */	
	*(u_int *)p = cd->bm4->BLEN;  
	p += 1;

	attrs.attrmask.bitmap4_len = cd->bm4->BLEN;
	attrs.attrmask.bitmap4_val = p;     /* save bitmap_val in fattr */
	memset (attrs.attrmask.bitmap4_val, 0, 4 * attrs.attrmask.bitmap4_len);
	p += attrs.attrmask.bitmap4_len;                 /* move past bitmap_val */

	p2 = (u_int *)p; /* remember attrlist_len */
	p += 1;		/* move past attrlist_len */

	attrs.attr_vals.attrlist4_val = (char *)p;
	if(CURRENT_FH.fh_dentry->d_flags & DCACHE_NFSDV4_PSEUDO_FS){
		nfsv4_printk(vfs,"NFSV4 fill_entry inode is PSEUDO\n");
		inode = pseudo_iget(ino);	
		err = get_attributes(cd->rqstp,cd->bm4,&attrs,inode);
	}else{
		inode = iget(CURRENT_FH.fh_dentry->d_sb,ino);	
		err = get_attributes(cd->rqstp,cd->bm4,&attrs,inode);
		iput(inode);
	}
	if(err)nfsv4_printk(vfs,"NFSDV4 fill_entry get_attributes err %d\n",err);

	*p2 = attrs.FATTRSLEN;
	(char *)p += attrs.FATTRSLEN;
	
/* for now assume get_attributes succeeds and buflen is accurate */
	cd->buflen -= ((char *)p - (char *)cd->buffer) + NFS_ENTRY4_HACK;

        cd->buffer = p;			/* point to remaining buffer */
	cd->nextentry = p;	/* remember pointer */
	nfsv4_printk(vfs,"NFSDV4 fill_entry4 returning\n");
        return 0;
}


/* 
 * Read entries from a directory, encode them into res->ENTRIES.
 * NOTE: therefore, the res->ENTRIES buffer does not need to be
 * touched in nfs4svc_encode_COMPOUND4res.
 * also encode eof (res->REOF). 
 * NOTE: need to code the verifier behaviour.
 */
int
nfsd4_readdir(struct svc_rqst *rqstp,READDIR4args *args, READDIR4res *res)
{
	struct inode	*inode;
	int		oldlen, eof, err;
	struct file	file;
	struct readdir_cd cd;
	struct svc_fh * fhp = &CURRENT_FH;

	nfsv4_printk(vfs,"NFSDV4: nfsd4_readdir\n");

	err = 0;
	if (args->cookie > ~(u32) 0)
		goto out;

	err = nfsd_open(rqstp, fhp, S_IFDIR, MAY_READ, &file);
	if (err){
		nfsv4_printk(vfs,"NFSDV4: nfsd4_readdir: nfsd_open failed. err:%d\n",err);
		goto out;
	}

	err = nfserr_notdir;
	if (!file.f_op->readdir){
		nfsv4_printk(vfs,"NFSDV4: nfsd4_readdir: !file.f_op->readdir\n");
		goto out_close;
	}
	file.f_pos = args->cookie;

	nfsv4_printk(vfs,"NFSDV4: nfsd4_readdir. args->cookie %ld args->maxcount %d\n",(long int)args->cookie, args->maxcount);

	/* Set up the readdir context. Note due to memset, cd.ofset = 0 */
	memset(&cd, 0, sizeof(cd));
	cd.rqstp  = rqstp;
	cd.buffer = cd.nextentry = (u32 *)res->ENTRIES;
	cd.buflen = args->maxcount; /* count of words */
/* do i need the fhp? */
	cd.dirfh  = fhp;
	cd.bm4 = &args->attr_request;

	/*
	 * Read the directory entries. */
	inode = file.f_dentry->d_inode;
	while (1) {
		oldlen = cd.buflen;

		nfsv4_printk(vfs,"NFSDV4 nfsd4_readdir:f_op->readdir(%x/%ld @ %d) buflen = %d (%d)\n",
			file.f_dentry->d_inode->i_dev, 
			file.f_dentry->d_inode->i_ino,
			(int) file.f_pos, (int) oldlen, (int) cd.buflen);
		down(&inode->i_sem);
		/* read directory */
		err = file.f_op->readdir(&file, &cd, (filldir_t)fill_entry4);
		nfsv4_printk(vfs,"NFSDV4 nfsd4_readdir after op->readdir err %d\n",err);
		up(&inode->i_sem);

		if (err < 0){
			nfsv4_printk(vfs,"NFSDV4: nfsd4_readdir: fop->readdir failed. err %d\n",err);
			goto out_nfserr;
		}

		if (oldlen == cd.buflen){
			nfsv4_printk(vfs,"NFSDV4: nfsd4_readdir: while loop oldlen = cd.buflen \n");
			*cd.nextentry = 0; /* set last entry4.nextentry to 
					    * to indicate ain't no more.
					    */
			break;
		}
		if (cd.eob){
			nfsv4_printk(vfs,"NFSDV4: nfsd4_readdir: while loop cd.eob\n");
			*cd.nextentry = 0;
			break;
		}
	}

	/* If we didn't fill the buffer completely, we're at EOF.
	 * this is signaled with a fill_entry() namelen = 0, and 
	 * cd->eob 0. cd->eob=1 means there was not enough buffer
	 * left, so continue reading with next readdir call.
	 */
	eof = !cd.eob;

	/* stuff the cookie field of the last entry4 */
	if (cd.offset){
		nfsv4_printk(andros,"ANDROS: last entry offset: (cookie) %Ld\n",(long long)file.f_pos);
		*cd.offset = (u64)file.f_pos;
	}

	/* here ends res->ENTRIES */
	res->REOF = eof;
	nfsv4_printk(vfs,"NFSDV4: readdir result eof %d offset %d\n",
				eof, cd.offset? ntohl(*cd.offset) : -1);
	err = 0;
out_close:
	nfsv4_printk(vfs,"NFSDV4 nfsd4_readdir at out_close\n");
	nfsd_close(&file);
	nfsv4_printk(vfs,"NFSDV4 nfsd4_readdir after out_close\n");
out:
	return err;

out_nfserr:
	nfsv4_printk(vfs,"NFSDV4 nfsd4_readdir at out_nfserr\n");
	err = nfsd4errno(-err);
	goto out_close;
}

/* use nfs v3 access logic code for now. e.g. on the wire nfs v4
 * vfs is v3
 *
 * Check server access rights to a file system object
 */
struct accessmap {
        u32             access;
        int             how;
};
static struct accessmap nfs4_regaccess[] = {
    {   ACCESS4_READ,       MAY_READ                        },
    {   ACCESS4_EXECUTE,    MAY_EXEC                        },
    {   ACCESS4_MODIFY,     MAY_WRITE|MAY_TRUNC             },
    {   ACCESS4_EXTEND,     MAY_WRITE                       },

    {   0,                      0                               }
};
static struct accessmap nfs4_diraccess[] = {
    {   ACCESS4_READ,       MAY_READ                        },
    {   ACCESS4_LOOKUP,     MAY_EXEC                        },
    {   ACCESS4_MODIFY,     MAY_EXEC|MAY_WRITE|MAY_TRUNC    },
    {   ACCESS4_EXTEND,     MAY_EXEC|MAY_WRITE              },

    {   0,                      0                               }
};

static struct accessmap nfs4_anyaccess[] = {
    /* XXX: should we try to cover read/write here for clients that
     * rely on us to do their access checking for special files? */

    {   0,                      0                               }
};

int
nfsd4_access(struct svc_rqst *rqstp, ACCESS4args *args, ACCESS4res *res)
{
	struct accessmap	*map;
	struct svc_export	*export;
	struct dentry		*dentry;
	u32			result = 0;
	int			error;
	u32			query = 0;

	nfsv4_printk(vfs,"NFSDV4 nfsd4_access \n");
	error = fh_verify(rqstp, &CURRENT_FH, 0, MAY_NOP);
	if (error){
		nfsv4_printk(vfs,"NFSDV4 nfsd4_access fh_verify error %d\n",error);
		goto out;
	}

	export = CURRENT_FH.fh_export;
	dentry = CURRENT_FH.fh_dentry;

	if (S_ISREG(dentry->d_inode->i_mode)) {
		nfsv4_printk(vfs,"NFSDV4 nfsd4_access map = regaccess\n");
		map = nfs4_regaccess;
	} else if (S_ISDIR(dentry->d_inode->i_mode)) {
		nfsv4_printk(vfs,"NFSDV4 nfsd4_access map = diraccess\n");
		map = nfs4_diraccess;
	} else {
		nfsv4_printk(vfs,"NFSDV4 nfsd4_access map = anyaccess\n");
		map = nfs4_anyaccess;
	}

/*map between v4 access flags and v3 access flags */
	res->SUPPORTED = 0;
	query = args->access;
	while (map->access) {
		res->SUPPORTED |= map->access;
		if (map->access & query) {
			error = nfsd_permission(export, dentry, map->how);
			nfsv4_printk(vfs,"NFSDV4 nfsd4_access nfsd_permission error %d\n",error);
			if (error == 0)
				 result |= map->access;
		}
		map++;
	}
	error = 0; /* nfsd_permission failing simply means don't return
		    *
                    */
        res->ACCESS = result;
        nfsv4_printk(vfs,"NFSDV4 nfsd4_access access: 0x%x\n",(unsigned int)result);

out:
        return error;
}

/* this function attempts to rename the source name (args->oldname) located in
 * the directory refrenced by SAVE_FH to the target name (args->newname)
 * located in the directory refrenced by CURRENT_FH
 */
int
nfsd4_rename(struct svc_rqst *rqstp,RENAME4args *args,RENAME4res *res)
{
struct dentry   *tar_dentry, *src_dentry, *old_dentry=NULL, *new_dentry;
struct inode    *tar_dir, *src_dir;
int             err;
char 						*temp_old = NULL, *temp_new = NULL; 				

	nfsv4_printk(rename, "NFSDv4: RENAME: vfs 'nfsd4_rename' function\n");
	print_utf8string(&args->oldname);
	print_utf8string(&args->newname);

	nfsv4_printk(rename, "NFSDv4: RENAME: vfs 1:\n");

	if (!args->OLDNAME_LEN || (args->OLDNAME_VAL[0] == '.' 
		&& (args->OLDNAME_LEN == 1 || (args->OLDNAME_LEN == 2 
		&& args->OLDNAME_VAL[1] == '.'))) 
		|| !args->NEWNAME_LEN || (args->NEWNAME_VAL[0] == '.' 
		&& (args->NEWNAME_LEN == 1 
		|| (args->NEWNAME_LEN == 2 && args->NEWNAME_VAL[1] == '.')))){
		err = -NFS4ERR_INVAL;
		goto out_nfserr;
	}

	err = fh_verify(rqstp, &SAVE_FH, S_IFDIR, MAY_REMOVE);
	if (err)
		goto out;
	err = fh_verify(rqstp, &CURRENT_FH, S_IFDIR, MAY_CREATE);
	if (err)
		goto out;

	/* grab the dentry and inode from SAVE_FH */
	src_dentry = SAVE_FH.fh_dentry;
	src_dir = src_dentry->d_inode;

	nfsv4_printk(rename, 
		"NFSDv4: RENAME: SAVE FH is: %s\n", src_dentry->d_name.name);

	/* grab the dentry and inode from CURRENT_FH */
	tar_dentry = CURRENT_FH.fh_dentry;
	tar_dir = tar_dentry->d_inode;

	/* check to see if file systems are different */
	if (src_dentry->d_sb != tar_dentry->d_sb) {
		nfsv4_printk(rename, 
			"NFDSDV4: RENAME: the two files reside on separate file systems, quitting with err NFS4ER_XDEV\n");
		err = -NFS4ERR_XDEV;
		goto out_dput_old;
	}

	nfsv4_printk(rename, 
		"NFSDv4: RENAME: CURRENT FH is: %s\n", tar_dentry->d_name.name);

	/* utf8_2_ascii,kfree():
	* The vfs layer assumes ascii strings. NFSv4 uses
	* UTF-8 encoding. Until the vfs layer decides otherwise,
	* linux NFSv4 servers deal only in ascii.
	*/
	utf8_2_ascii(&args->oldname,&temp_old);
	old_dentry = lookup_dentry(temp_old, dget(src_dentry), 0);
	kfree(temp_old);
	err = PTR_ERR(old_dentry);
	if (IS_ERR(old_dentry))
		goto out_nfserr;

	err = -ENOENT;
	if (!old_dentry->d_inode)
		goto out_dput_old;

	utf8_2_ascii(&args->newname,&temp_new);
	new_dentry = lookup_dentry(temp_new, dget(tar_dentry), 0);
	kfree(temp_new);
	err = PTR_ERR(new_dentry);
	if (IS_ERR(new_dentry))
		goto out_dput_old;

	/* check to see if the two files are the same (hard links of each other) */
	if (new_dentry->d_inode){
		if (old_dentry->d_inode){
			if (new_dentry->d_inode->i_ino == old_dentry->d_inode->i_ino){
				nfsv4_printk(rename, 
					"NFSDV4: RENAME: the two files are the same file %ld : %ld, exiting with success\n", 
					(long int)new_dentry->d_inode->i_ino, 
					(long int)old_dentry->d_inode->i_ino);
				err = -NFS4_OK;
				goto out;
			}
			}else 
				nfsv4_printk(rename, "NFSDV4: RENAME: old doesn't have an inode\n");
		}else
			nfsv4_printk(rename, "NFSDV4: RENAME: new doesn't have an inode\n");

		/*
		* Lock the parent directories.
		*/
		nfsd_double_down(&src_dir->i_sem, &tar_dir->i_sem);

		err = -ENOENT;
		/* GAM3 check for parent changes after locking. */
		if (nfsd_check_parent(src_dentry, old_dentry) &&
		nfsd_check_parent(tar_dentry, new_dentry)) {
		res->RENAME_RES_SRC.atomic = 1;
		res->RENAME_RES_TAR.atomic = 1;

		*(time_t *)&res->RENAME_RES_SRC.before = src_dir->i_mtime;
		*(time_t *)&res->RENAME_RES_TAR.before = tar_dir->i_mtime;

		err = vfs_rename(src_dir, old_dentry, tar_dir, new_dentry);

		*(time_t *)&res->RENAME_RES_SRC.after = src_dir->i_mtime;
		*(time_t *)&res->RENAME_RES_TAR.after = tar_dir->i_mtime;

		nfsv4_printk(rename, 
			"NFSDv4: RENAME: nfsd4_rename: err on vfs_rename %d\n", err);
		if (!err && EX_ISSYNC(CURRENT_FH.fh_export)) {
			nfsd_sync_dir(src_dentry);
			nfsd_sync_dir(tar_dentry);
		}
	} else
		dprintk("nfsd: Caught race in nfsd_rename");

	nfsd_double_up(&tar_dir->i_sem, &src_dir->i_sem);
	dput(new_dentry);

out_dput_old:
	nfsv4_printk(rename, "NFSDv4: RENAME: nfsd4_rename: out_dput_old\n");
	dput(old_dentry);
	if (err)
		goto out_nfserr;
out:
	nfsv4_printk(rename, "NFSDv4: RENAME: nfsd4_rename: out, err %d\n", err);
	return err;

out_nfserr:
	nfsv4_printk(rename, "NFSDv4: RENAME: nfsd4_rename: out_nfserr\n");
	err = nfsd4errno(-err);
	goto out;
}

/* Create a hardlink
 * This function sets a link from the original file (SAVE_FH) to
 * the new file (args->newname) that is residing in the directory (CURRENT_FH).
 */
int
nfsd4_link(struct svc_rqst *rqstp,LINK4args *args,LINK4res *res)
{
struct dentry *ddir, *dnew, *dold;
struct inode  *dirp, *dest;
int   err;
char * temp;

	nfsv4_printk(link, "NFSDV4: LINK: link new file: ");
	nfsv4_printk(link, "%s / ", CURRENT_FH.fh_dentry->d_name.name);
	print_utf8string(&args->newname);
	nfsv4_printk(link, " to old file %s \n", SAVE_FH.fh_dentry->d_name.name);

  if (args->newname.utf8string_len == 0){
    nfsv4_printk(link, "NFSDV4: LINK: target name has 0 length, failing\n");
    err = -NFS4ERR_INVAL;
    goto out_nfserr;
  }

	/* need error check here to make sure 'args->newname' obeys UTF-8 definition */


	err = fh_verify(rqstp, &CURRENT_FH, S_IFDIR, MAY_CREATE);
  if (err)
    goto out;
	err = fh_verify(rqstp, &SAVE_FH, -S_IFDIR, MAY_NOP);
  if (err)
    goto out;

  ddir = CURRENT_FH.fh_dentry;
  dirp = ddir->d_inode;

	/* utf8_2_ascii,kfree():
	* The vfs layer assumes ascii strings. NFSv4 uses
	* UTF-8 encoding. Until the vfs layer decides otherwise,
	* linux NFSv4 servers deal only in ascii.
	*/
	utf8_2_ascii(&args->newname,&temp);
  dnew = lookup_dentry(temp, dget(ddir), 0);
	kfree(temp);

  err = PTR_ERR(dnew);
  if (IS_ERR(dnew))
    goto out_nfserr;
  /*
   * Lock the parent before checking for existence
   */
	err = fh_lock_parent(&CURRENT_FH, dnew);
  if (err)
    goto out_dput;

	nfsv4_printk(link, "NFSDV4: locked CURRENT_FH\n");
	/*  err = nfserr_exist; */
	err = -NFS4ERR_EXIST;
  if (dnew->d_inode)
    goto out_unlock;

  dold = SAVE_FH.fh_dentry; 
  dest = dold->d_inode;

	/* this was removed in latest nfsv3 server.....  
	err = -NFS4ERR_ACCES;
	if (nfsd_iscovered(ddir, CURRENT_FH.fh_export))
    goto out_unlock;
	*/

	if (dirp->i_dev != dest->i_dev){
		nfsv4_printk(link, 
			"NFSDV4: LINK: nfsd4_link: files do not share same device\n");
		err = -NFS4ERR_XDEV;
		goto out_unlock;
	}

	/*  err = nfserr_perm; */
	err = -NFS4ERR_PERM;
  if (IS_IMMUTABLE(dest) || IS_APPEND(dest))
    goto out_unlock;
  if (!dirp->i_op || !dirp->i_op->link)
    goto out_unlock;

	res->LINK_RES_INFO.atomic = 1;
  DQUOT_INIT(dirp);
	*(time_t *)&res->LINK_RES_INFO.before = dirp->i_mtime;
  err = dirp->i_op->link(dold, dirp, dnew);
	*(time_t *)&res->LINK_RES_INFO.after = dirp->i_mtime;

  if (!err) {
		if (EX_ISSYNC(CURRENT_FH.fh_export)) {
			nfsv4_printk(link, "NFSDV4: LINK: write inodes now\n");
      nfsd_sync_dir(ddir);
      write_inode_now(dest);
    }
  } else
    err = nfsd4errno(-err);

out_unlock:
	err = nfsd4errno(-err);
	fh_unlock(&CURRENT_FH);
	nfsv4_printk(link, "NFSDV4: LINK: unlock CURRENT_FH\n");
out_dput:
  dput(dnew);
out:
	nfsv4_printk(link, "NFSDV4: LINK: out with err %d\n", err);
  return err;

out_nfserr:
  err = nfsd4errno(-err);
  goto out;
}

/*
 * Remove a file or directory
 * The file/directory is in the directory CURRENT_FH.
 * The file/directory is referenced by argp->target (a component4).
 */
int
nfsd4_remove(struct svc_rqst *rqstp, REMOVE4args * argp, REMOVE4res *resp)
{
  struct dentry *dentry, *rdentry;
  struct inode  *dirp;
  int   err, type;
	char name[NFS4_MAXNAMLEN];
	
  /* N.B. We shouldn't need this test ... handled by dentry layer */
  err = nfserr_inval;
  if (!argp->target.utf8string_len || isdotent(argp->target.utf8string_val, argp->target.utf8string_len) )
    goto out;

  err = fh_verify(rqstp, &CURRENT_FH, S_IFDIR, MAY_REMOVE);
  if (err)
    goto out;

  dentry = (&CURRENT_FH)->fh_dentry;
  dirp = dentry->d_inode;

	memset(name,0,NFS4_MAXNAMLEN);
	memcpy(name,argp->target.utf8string_val,argp->target.utf8string_len);

  rdentry = lookup_dentry(name, dget(dentry), 0);
  err = PTR_ERR(rdentry);
  if (IS_ERR(rdentry))
    goto out_nfserr;

  if (!rdentry->d_inode) {
    dput(rdentry);
    err = nfserr_noent;
    goto out;
  }

  if (S_ISDIR(rdentry->d_inode->i_mode) )
    type = S_IFDIR;
  else
    type = -S_IFDIR;

  if (type != S_IFDIR) {
    /* It's UNLINK */
    err = fh_lock_parent(&CURRENT_FH, rdentry);
    if (err)
      goto out;

    resp->REMOVE4res_u.resok4.cinfo.atomic = 1;

    *(time_t *)&resp->REMOVE4res_u.resok4.cinfo.before = dirp->i_mtime;
    err = vfs_unlink(dirp, rdentry);
    *(time_t *)&resp->REMOVE4res_u.resok4.cinfo.after = dirp->i_mtime;

    fh_unlock(&CURRENT_FH);

    dput(rdentry);

  } else {
    /* It's RMDIR */
    /* See comments in fs/namei.c:do_rmdir */

    rdentry->d_count++;
    nfsd_double_down(&dirp->i_sem, &rdentry->d_inode->i_sem);
    (&CURRENT_FH)->fh_locked = 1;

    err = -ENOENT;
    if (nfsd_check_parent(dentry, rdentry)){
      resp->REMOVE4res_u.resok4.cinfo.atomic = 1;
      *(time_t *)&resp->REMOVE4res_u.resok4.cinfo.before = dirp->i_mtime;
      err = vfs_rmdir(dirp, rdentry);
      *(time_t *)&resp->REMOVE4res_u.resok4.cinfo.after = dirp->i_mtime;
    }

    rdentry->d_count--;
    (&CURRENT_FH)->fh_locked = 0;
    nfsd_double_up(&dirp->i_sem, &rdentry->d_inode->i_sem);

    dput(rdentry);
  }

  if (err)
    goto out_nfserr;
  if (EX_ISSYNC( (&CURRENT_FH)->fh_export) ){
      down(&dentry->d_inode->i_sem);
      nfsd_sync_dir(dentry);
      up(&dentry->d_inode->i_sem);
  }
out:
  return err;

out_nfserr:
  err = nfsd4errno(-err);
  goto out;
}

/*
 * Create a symlink and look up its inode
 * N.B. After this call _both_ fhp and resfhp need an fh_put
 */
int
nfsd4_symlink (struct svc_rqst *rqstp, CREATE4args *argp, CREATE4res *resp, struct svc_fh *resfhp)
{
  struct dentry *dentry, *dnew;
  struct inode  *dirp;
  int   err;
  char *temp;
  char *path;
  struct svc_fh *fhp = &CURRENT_FH;

  err = nfserr_noent;
  if (!argp->objname.utf8string_len )
    goto out;

  err = fh_verify(rqstp, fhp, S_IFDIR, MAY_CREATE);
  if (err)
    goto out;
  dentry = fhp->fh_dentry;

  err = nfserr_perm;
  dirp = dentry->d_inode;
  if (!dirp->i_op || !dirp->i_op->symlink)
    goto out;

  utf8_2_ascii(&argp->objname, &temp);
  dnew = lookup_dentry(temp, dget(dentry), 0);
  err = PTR_ERR(dnew);
  if (IS_ERR(dnew))
    goto out_nfserr;

  /*
   * Lock the parent before checking for existence
   */
  err = fh_lock_parent(fhp, dnew);
  if (err)
    goto out_compose;

  err = nfserr_exist;
  if (!dnew->d_inode) {
    DQUOT_INIT(dirp);
    utf8_2_ascii(&argp->objtype.createtype4_u.linkdata, &path);
    resp->CREATEinfo.atomic = 1;
    *(time_t *)&resp->CREATEinfo.before = dirp->i_mtime;
    err = dirp->i_op->symlink(dirp, dnew, path);
    if (!err) {
      *(time_t *)&resp->CREATEinfo.after = dirp->i_mtime;
      if (EX_ISSYNC(fhp->fh_export))
        nfsd_sync_dir(dentry);
      err = nfsd4errno(-err);
    }
    fh_unlock(fhp);
  }

  /* Compose the fh so the dentry will be freed ... */
out_compose:
  fh_compose(resfhp, fhp->fh_export, dnew);

out:
  return err;

out_nfserr:
#ifdef NFSV4
  err = nfsd4errno(-err);
#else /*NFSV4*/
  err = nfserrno(-err);
#endif /*NFSV4*/
  goto out;
}

/*
 * Create a directory.
 *
 */
int
nfsd4_mkdir(struct svc_rqst *rqstp, CREATE4args *argp, CREATE4res *resp, 
	    struct svc_fh *resfhp)
{
	struct dentry *dentry, *dchild;
	struct inode *dirp;
	int err;
	struct iattr *iap;
	char *fname;
	dev_t rdev=0;          /* NOTE this needs to be set somewhere before using! */
	struct svc_fh *fhp = &CURRENT_FH;


	iap = (struct iattr *)kmalloc(sizeof(struct iattr), GFP_KERNEL);


	err = nfserr_perm;
	if (!argp->objname.utf8string_len)
			goto out;
	iap->ia_mode = 0;

	err = fh_verify(rqstp, fhp, S_IFDIR, MAY_CREATE);
	if (err)
			goto out;

	dentry = fhp->fh_dentry;
	dirp = dentry->d_inode;

	err = nfserr_notdir;
	if (!dirp->i_op || !dirp->i_op->lookup)
			goto out;

	if (!resfhp->fh_dverified) {
		utf8_2_ascii(&argp->objname, &fname);
		dchild = lookup_dentry(fname, dget(dentry), 0);
		kfree(fname);
		err = PTR_ERR(dchild);
		if (IS_ERR(dchild))
			goto out_nfserr;
		fh_compose(resfhp, fhp->fh_export, dchild);
    err = fh_lock_parent(fhp, dchild);
    if (err)
      goto out;
  } else {
    dchild = resfhp->fh_dentry;
    if (!fhp->fh_locked) {
      /* not actually possible */
      printk(KERN_ERR
        "nfsd_create: parent %s/%s not locked!\n",
        dentry->d_parent->d_name.name,
        dentry->d_name.name);
      err = -EIO;
      goto out;
    }
  }
  /*
   * Make sure the child dentry is still negative ...
   */
  err = nfserr_exist;
  if (dchild->d_inode) {
    dprintk("nfsd_create: dentry %s/%s not negative!\n",
      dentry->d_name.name, dchild->d_name.name);
    goto out;
  }
	err = nfserr_perm;
	switch (argp->objtype.type){
		case NF4DIR:
			DQUOT_INIT(dirp);
			resp->CREATEinfo.atomic = 1;
			*(time_t *)&resp->CREATEinfo.before = dirp->i_mtime;
			err = dirp->i_op->mkdir(dirp, dchild, iap->ia_mode);
			*(time_t *)&resp->CREATEinfo.after = dirp->i_mtime;
		break;
		case NF4CHR:
  	case NF4BLK:
			/* The client is _NOT_ required to do security enforcement */
    	if(!capable(CAP_SYS_ADMIN)) {
      	err = -EPERM;
      	goto out;
    	}
		break;
		case NF4FIFO:
  	case NF4SOCK:
			DQUOT_INIT(dirp);
    	err = dirp->i_op->mknod(dirp, dchild, iap->ia_mode, rdev);
    break;
  	default:
			printk("nfsd: bad file type %o in nfsd4_mkdir\n", argp->objtype.type);
			err = nfserr_inval;
			goto out;
		break;
	}

	if (err < 0)
		goto out_nfserr;


	if (EX_ISSYNC(fhp->fh_export)) {
		nfsd_sync_dir(dentry);
		write_inode_now(dchild->d_inode);
	}

	/*
   * Update the file handle to get the new inode info.
   */
  fh_update(resfhp);

	/* Set file attributes. Mode has already been set and
   * setting uid/gid works only for root. Irix appears to
   * send along the gid when it tries to implement setgid
   * directories via NFS.
   */
  err = 0;
/*
  if ((iap->ia_valid &= ~(ATTR_UID|ATTR_GID|ATTR_MODE)) != 0)
    err = nfsd_setattr(rqstp, resfhp, iap);
*/
out:
	fh_unlock(fhp);
	kfree(iap);
  return err;

out_nfserr:
  err = nfsd4errno(-err);
  goto out;
}

/*
 *
 * This function attempts to set attributes for the CFH as defined in argp.
 * The results of which attributes are set is returned in resp.
 * The resp bitmap is preset to 0.
 * Note: It is possible for this function to set some but not all requested
 * attributes.
 */

int nfsd4_setattr (struct svc_rqst *rqstp, SETATTR4args *argp, SETATTR4res *resp)
{
  struct iattr attr;
  fattr4_type dummy;
  int err;
  
  err = nfs4_set_iattr (&argp->obj_attributes, &attr, &dummy);
  if (err)
    return err;
  if (!argp->stateid)
    attr.ia_valid &= ~ATTR_SIZE;
  err = nfsd_setattr (rqstp, &CURRENT_FH, &attr);
  if (!err)
    memcpy (resp->attrsset.bitmap4_val, argp->obj_attributes.attrmask.bitmap4_val,
	    argp->obj_attributes.attrmask.bitmap4_len * sizeof(uint32_t));
  return err;
}

int
nfsd4_readlink (struct svc_rqst *rqstp, READLINK4res *resp)
{
	struct dentry	*dentry;
	struct inode	*inode;
	mm_segment_t	oldfs;
	int		err;
	struct svc_fh   *fhp = &CURRENT_FH;
	utf8string      *s;

	err = fh_verify(rqstp, fhp, S_IFLNK, MAY_READ);
	if (err)
		goto out;

	dentry = fhp->fh_dentry;
	inode = dentry->d_inode;

	err = nfserr_inval;
	if (!inode->i_op || !inode->i_op->readlink)
		goto out;

	UPDATE_ATIME(inode);
	/* N.B. Why does this call need a get_fs()??
	 * Remove the set_fs and watch the fireworks:-) --okir
	 */

	oldfs = get_fs(); set_fs(KERNEL_DS);
	s = &resp->READLINK4res_u.resok4.link;
	err = inode->i_op->readlink (dentry, s->utf8string_val, NFS4_MAXNAMLEN);
	set_fs(oldfs);

	if (err < 0)
		goto out_nfserr;
	s->utf8string_len = err;
	err = 0;
out:
	return err;

out_nfserr:
	err = nfsd4errno(-err);
	goto out;
}

/* 
 * nfsd4_Create() processes the CREATE compound call.  Unfortunately, a function named nfsd4_create() already
 * exists (it processes the OPEN compound call), hence the capitalization
 */
int
nfsd4_Create (struct svc_rqst *rqstp, CREATE4args *argp, CREATE4res *resp, struct svc_fh *newfhp)
{
  struct svc_fh  *fhp = &CURRENT_FH;
  struct dentry  *dentry, *dchild;
  struct inode   *dirp;
  utf8string     *name = &argp->objname;
  char           *temp;
  int            err;

  fh_init (newfhp);

  /* First verify the parent file handle */
  err = fh_verify (rqstp, fhp, S_IFDIR, MAY_CREATE);      // ??? - MAY_EXEC
  if (err)
    goto out; /* must fh_put fhp even on error */         // ??? - goto done

  dentry = fhp->fh_dentry;
  dirp = dentry->d_inode;

  err = nfserr_notdir;
  if (!dirp->i_op || !dirp->i_op->lookup)
    goto out;

  // check for write permission on the directory has been temporarily omitted!
  // (see nfsd_proc_create() in nfsproc.c)

  // check that name has positive length, truncate if too long
  err = nfserr_perm;
  if (name->utf8string_len <= 0)
    goto out;
  if (name->utf8string_len > NFS4_MAXNAMLEN)
    name->utf8string_len = NFS4_MAXNAMLEN;

  // get dentry for new object (should be negative), initialize fh
  utf8_2_ascii (name, &temp);
  dchild = lookup_dentry (name->utf8string_val, dget(dentry), 0);
  kfree (temp);
  err = PTR_ERR (dchild);
  if (IS_ERR(dchild))
    goto out_nfserr;
  fh_compose (newfhp, fhp->fh_export, dchild);

  // lock dir, check that child dentry is negative
  err = fh_lock_parent (fhp, dchild);
  if (err)
    goto out;       // out_nfserr?
  err = nfserr_exist;
  if (dchild->d_inode) {
    printk ("nfsd_create(): dentry %s/%s not negative!\n",
	    dentry->d_name.name, dchild->d_name.name);
    goto out;
  }

  DQUOT_INIT (dirp);
  resp->CREATEinfo.atomic = 1;
  *(time_t *)&resp->CREATEinfo.before = dirp->i_mtime;
  switch (argp->objtype.type) {
    // case NF4BLK:
    //   if(!capable(CAP_SYS_ADMIN)) {
    //     err = -EPERM;
    //     goto out;
    //   }
    //   break;
  case NF4DIR:        // should test (dirp->i_op->mkdir == NULL) ?
    err = dirp->i_op->mkdir(dirp, dchild, 0);            // what mode?
    break;
  case NF4LNK:        // should test (dirp->i_op->symlink == NULL) ?
    utf8_2_ascii (&argp->objtype.createtype4_u.linkdata, &temp);
    err = dirp->i_op->symlink(dirp, dchild, temp);
    kfree (temp);
    break;
    // case NF4SOCK:
    //   DQUOT_INIT(dirp);
    //   err = dirp->i_op->mknod(dirp, dchild, iap->ia_mode, rdev);
    //   break;
  default:     // many cases are not yet implemented
    printk ("nfsd_create(): unsupported file type %d\n", argp->objtype.type);
    err = nfserr_notsupp;
    goto out;
  }
  *(time_t *)&resp->CREATEinfo.after = dirp->i_mtime;

  if (err < 0)
    goto out_nfserr;

  if (EX_ISSYNC(fhp->fh_export)) {
    nfsd_sync_dir(dentry);
    write_inode_now(dchild->d_inode);
  }

  fh_update (newfhp);
  err = 0;

  // should we set some default file permissions here?

out:
  fh_unlock (fhp);
  return err;

out_nfserr:
  err = nfsd4errno (-err);
  goto out;
}

int
nfs4_open(struct svc_rqst *rqstp, OPEN4args *aptr, OPEN4res *rptr)
{
	struct inode *inode = CURRENT_FH.fh_dentry->d_inode; 
	struct nfs4_lock_owner *lowner = NULL;
	struct nfs4_file *file;
	struct dentry *dentry,*dchild = NULL;
	stateid4 stateid = 0;
	int status = 0;
	time_t now = CURRENT_TIME;
	struct svc_client *clp = exp_getclient(&rqstp->rq_addr, aptr->owner.clientid);

	nfsv4_printk(open, "nfs4_open(): called in dir %s\n", CURRENT_FH.fh_dentry->d_name.name);
	rptr->status = 0;
	if (inode) {                      /* directory inode exists */
		switch (aptr->claim.claim) {
		case CLAIM_NULL:
		{
			if (aptr->claim.CLAIM_FILE_LEN > 1) {
				/* save the CURRENT_FH in case open fails... */
				fh_init(&SAVE_FH);
				fh_copy(&SAVE_FH, &CURRENT_FH);
				if ((status = nfsd4_lookup(rqstp, aptr->claim.CLAIM_FILE))) {
					nfsv4_printk(open,"nfs4_open(): nfsd4_lookup failed\n");
					goto open_out;	
				}
			}
			switch (aptr->openhow.opentype) {
			case OPEN4_NOCREATE:
			{
				char *fname;
				struct svc_fh *filefh;

				/* check and perhaps renew lease, open file, 
				 * get lockowner, check for share lock conflicts, set CURRENT_FH
				 */

				nfsv4_printk(open, "nfs4_open() case NOCREATE calling check_and_set_lease\n");
				if ((status = check_and_set_lease(&rqstp->rq_addr, aptr->owner.clientid,now))) {
					nfsv4_printk(open, "nfs4_open(): check_and_set_lease failed\n");
					goto open_out;
				}
				/* directory (CURRENT_FH) has been verified in PUTFH 
				 * lookup file to be opened
				 */
				dentry = CURRENT_FH.fh_dentry;
				if (!inode->i_op || !inode->i_op->lookup) {
					status = nfserr_notdir;
					goto open_out;
				}
				utf8_2_ascii(&aptr->claim.CLAIM_FILE_VAL[aptr->claim.CLAIM_FILE_LEN-1], &fname);
				if (aptr->claim.CLAIM_FILE_LEN == 1) { 
					nfsv4_printk(open,"fname: %s dentry %p\n", fname, dentry);
					dchild = lookup_dentry(fname, dget(dentry), 0);
					kfree(fname); 
					if (IS_ERR(dchild)) {
						nfsv4_printk(open, "nfs4_open(): lookup_dentry failed\n");
						status = PTR_ERR(dchild);
						goto open_out;
					}
					filefh = fh_init(&TEMP_FH);
					fh_compose(filefh, CURRENT_FH.fh_export, dchild);		
				} else {
					filefh = fh_init(&TEMP_FH);
					fh_compose(filefh, CURRENT_FH.fh_export, dentry);		
				}

				/* find or create nfs4_file, and open the file
				 * perform share checks & if no conflicts, create a stateid.
				 * we hold the dentry use count while the file is open.
				 */
				nfsv4_printk(open, "nfs4_open() case OPEN_NOCREATE calling nfs4_get_file\n");
				if ((status = nfs4_get_file(rqstp, &file, filefh))) {
					nfsv4_printk(open, "nfs4_open(): nfs4_get_file failed\n");
					goto open_out;
				}
				
				/* check and perhaps set lease
				 * find or create the nfs4_lockowner, set current_lockowner
				 */
				nfsv4_printk(open, "nfs4_open() case OPEN_NOCREATE calling nfs4_get_lockowner\n");
				if ((status = nfs4_get_lockowner(rqstp, &aptr->owner,
								 aptr->seqid, &lowner,now))) {
					nfsv4_printk(open, "nfs4_open(): nfs4_get_lockowner failed\n");
					goto open_out;
				}
				nfsv4_printk(open, "nfs4_open() case OPEN_NOCREATE calling nfs4_test_lock\n");
				if ((status = nfs4_test_lock(lowner, file, aptr->share_access, aptr->share_deny,
							     &stateid))) {
					nfsv4_printk(open, "nfs4_open(): nfs4_test_lock failed\n");
					nfs4_file_close(&file->f_file);	
					goto open_out;
				}
				
				nfsv4_printk(open, "nfs4_open(): nfs4_test_lock returned stateid [0x%x, %08x]\n",
					     (__u32) (stateid >> 32), (__u32) (stateid & 0xFFFFFFFFUL));
				/* set CURRENT_FH */
				fh_copy(&CURRENT_FH,filefh);
			}
			break;
			case OPEN4_CREATE:
			{
				svc_fh *newfhp = NULL;
				struct fattr4 *fattr = &aptr->openhow.FLAG_HOW.CREATE_FATTR;
				struct iattr iap;
				fattr4_type f4type;

				nfsv4_printk(open, "nfs4_open() case OPEN_CREATE calling check_and_set_lease\n");
				if ((status = check_and_set_lease(&rqstp->rq_addr, aptr->owner.clientid, now))) {
					nfsv4_printk(open, "nfs4_open(): check_and_set_lease failed\n");
					goto open_out;
				}
				
				newfhp = fh_init(&TEMP_FH);
				if ((status = fh_verify(rqstp, &CURRENT_FH, S_IFDIR, MAY_CREATE))) {
					nfsv4_printk(open, "nfs4_open(): fh_verify failed\n");
					goto open_out;
				}
				      
				/* translate from fattr4 to iattr */
				if ((status = nfs4_set_iattr(fattr, &iap, &f4type))) {
					nfsv4_printk(open, "nfs4_open(): nfs4_set_iattr failed\n");
					goto open_out;
				}

				nfsv4_printk(open, "nfs4_open() case OPEN_CREATE calling nfsd4_create\n");
				/* Now create the file and set attributes */
				if ((status = nfsd4_create(rqstp, &aptr->claim.CLAIM_FILE_VAL[aptr->claim.CLAIM_FILE_LEN-1],
							   &iap, newfhp, aptr->openhow.FLAG_HOW_MODE,
							   (char *)NULL))) {
					nfsv4_printk(open, "nfs4_open(): nfsd4_create failed\n");
					goto open_out;
				}
				
				/* find or create nfs4_file, and open the file
				 * perform share checks & if no conflicts, create a stateid.
				 * we hold the dentry use count while the file is open.
				 */
				nfsv4_printk(open, "nfs4_open() case OPEN_CREATE calling nfs4_get_file\n");
				if ((status = nfs4_get_file(rqstp, &file, newfhp))) {
					nfsv4_printk(open, "nfs4_open(): nfs4_get_file failed\n");
					goto open_out;
				}
				
				/* find or create the nfs4_lockowner, set current_lockowner
				 */
				nfsv4_printk(open, "nfs4_open() case OPEN_CREATE calling nfs4_get_lockowner\n");
				if ((status = nfs4_get_lockowner(rqstp, &aptr->owner, aptr->seqid, &lowner,now))) {
					nfsv4_printk(open, "nfs4_open(): nfs4_get_lockowner failed\n");
					goto open_out;
				}
				nfsv4_printk(open, "nfs4_open() case OPEN_CREATE calling nfs4_test_lock\n");
				if ((status = nfs4_test_lock(lowner, file, aptr->share_access, aptr->share_deny,
							     &stateid))) {
					nfsv4_printk(open, "nfs4_open(): nfs4_test_lock failed\n");
					nfs4_file_close(&file->f_file);	
					goto open_out;
				}
				
				nfsv4_printk(open, "nfs4_open(): nfs4_test_lock returned stateid [0x%x, %08x]\n",
					     (__u32) (stateid >> 32), (__u32) (stateid & 0xFFFFFFFFUL));

				/* set CURRENT_FH */
				fh_copy(&CURRENT_FH, newfhp);
			}
			break;
			default:
				rptr->status = nfserr_inval;
				goto open_out;
			}

			/* decide what type of delegation to return, if any
			 * NOTE: before calling nfs4_issue_delegation(), CURRENT_FH must be set
			 * to the file being opened
			 */
			nfsv4_printk(open, "nfs4_open() case OPEN calling delegation functions\n");
			nfs4_recall_delegations(file, clp, aptr->share_access);
			nfs4_issue_delegation(&rptr->OPEN_RES.delegation, file, clp, aptr->share_access);


			rptr->OPEN_RES.stateid = stateid;
				
				/* changeid4's are time_modify value of the
				 * directory (CURRENT_FH) containing the open'ed object
				 * before and after the open - i.e. these
				 * values are only different when the 
				 * file to be opened is created.
				 * the atomic field is set to 1 to indicate 
				 * that the linux ext2 inode time_modify is atomic
				 * with respect to the link operation. 
				 */
			rptr->OPEN_RES.cinfo.atomic = 1;
				
			*(time_t *)&rptr->OPEN_RES.cinfo.before = inode->i_mtime;  
			*(time_t *)&rptr->OPEN_RES.cinfo.after = inode->i_mtime;  
				
				/* 
				 * if first OPEN4 for this nfs lockowner, ask for an OPEN_CONFIRM.
				 */
			rptr->OPEN_RES.rflags = 0;
			if (lowner->lo_state == OPEN_CONFIRM_REQ_NEGOTIATING) {
				rptr->OPEN_RES.rflags = OPEN4_RESULT_CONFIRM;
				/* we need the stateid in the OPEN_CONFIRM call, 
				 * else can't find the lockowner and check the seq#
				 */
				nfsv4_printk(open, "NFSV4: OPEN setting open_confirm rflags: %d\n",rptr->OPEN_RES.rflags);
				memcpy(rptr->OPEN_RES.open_confirm, &stateid, NFS4_CREATEVERFSIZE); 
			}
			
		}
		break;
		case CLAIM_PREVIOUS:
			rptr->status = nfserr_notsupp;
			break;
		case CLAIM_DELEGATE_CUR:
			rptr->status = nfserr_notsupp;
			break;
		case CLAIM_DELEGATE_PREV:
			rptr->status = nfserr_notsupp;
			break;
		default:
			rptr->status = nfserr_inval;
			break;
		}

	} /* if inode */
	else {
		status = nfserr_exist;
	}	
			  
open_out:
	nfsv4_printk(open, "nfs4_open(): returning status %d\n", ntohl(status));
	rptr->status = status;
	if (status && (aptr->claim.claim == CLAIM_NULL) && (aptr->claim.CLAIM_FILE_LEN > 1)) {
		/* nfsd4_lookup changed CURRENT_FH - put it back */
		nfsv4_printk(open,"NFSV4: nfsd4_lookup changed CURRENT_FH - put it back\n"); 
		fh_init(&CURRENT_FH);
		fh_copy(&CURRENT_FH, &SAVE_FH);
	}
	return status;
}

#endif /*NFSV4*/
