/*
 * include/linux/nfsd/nfsd4export.h
 * 
 * Public declarations for NFS exports. The definitions for the
 * syscall interface are in nfsctl.h
 * NFSV4 Implementation
 *      by Andy Adamson <andros@citi.umich.edu>
 *      Copyright (C) 1999 The Regents of the University of Michigan
 *
 * based on NFSv3 implementation by
 * Copyright (C) 1995-1997 Olaf Kirch <okir@monad.swb.de>
 *
 */

#ifndef NFSD4_EXPORT_H
#define NFSD4_EXPORT_H

#include <asm/types.h>
#ifdef __KERNEL__
# include <linux/types.h>
# include <linux/in.h>
#endif

/*
 * Important limits for the exports stuff.
 */
#define NFSCLNT_IDMAX		1024
#define NFSCLNT_ADDRMAX		16
#define NFSCLNT_KEYMAX		32

/*
 * Export flags.
 */
#define NFSEXP_READONLY		0x0001
#define NFSEXP_INSECURE_PORT	0x0002
#define NFSEXP_ROOTSQUASH	0x0004
#define NFSEXP_ALLSQUASH	0x0008
#define NFSEXP_ASYNC		0x0010
#define NFSEXP_GATHERED_WRITES	0x0020
#define NFSEXP_UIDMAP		0x0040
#define NFSEXP_KERBEROS		0x0080		/* not available */
#define NFSEXP_SUNSECURE	0x0100
#define NFSEXP_CROSSMNT		0x0200
#define NFSEXP_NOSUBTREECHECK	0x0400
#define	NFSEXP_NOAUTHNLM	0x0800		/* Don't authenticate NLM requests - just trust */
#define NFSEXP_PSEUDO_EXP 0x1000
#define NFSEXP_ALLFLAGS		0x1FFF

#define MAX_PSEUDO_EXP    32 /* abritrary temporary limit */ 


#ifdef __KERNEL__

/* The following are hashtable sizes and must be powers of 2 */
#define NFSCLNT_EXPMAX		16

/* unlike NFS v2/3 in NFSV4, there is a one to one correspondence between 
* svc_client and in_addr's
*/
/* 64 bit clientid consists of
* | 32 bits of nfsclient boot time |
* | 16 bits of hashindex into clnt_hash[] |
* | 16 bits of hash bucket generation num |
*/
#define GET_NFS4_CLIENTID_VERF(clid)  (clid & 0xFFFFFFFFUL)
#define GET_NFS4_CLIENTID_HINDEX(clid) (((clid >> 32) & 0xFFFF0000UL) >> 16)
#define GET_NFS4_CLIENTID_GEN(clid)   ((clid >> 32) & 0x0000FFFFUL)

/* values for clid_state */
#define CLID_NOSETCLIENTID 0x0 /* when SETCLIENTID has yet to be called */
#define CLID_NEGOTIATING	0x1 /* SETCLIENTID called and succeeded*/
#define CLID_CONFIRMED		0x2 /* SETCLIENTID_CONFIRM succeeded */ 

/* 
* Boot: no svc_client structs created. Pseudo fs can be 
* browsed by a client without server keeping any state.
*
* LOOKUP: when a user crosses from pseudo fs into an exported
* ext2, an svc_client struct is created, and all svc_client fields including
* adding the svc_export for the ext2 fs are set except the
* client id fields: nfsdclientid, clnt4, cl_ident.
*
* First OPEN: setclientid/setclientid_confirm negotiation
* sets the svc_client's client id fields. 
* 
* cl_ident is storage for clnt4.id.id_val.
*/
struct svc_client {
	struct svc_client *	cl_next;
	clientid4 nfsdclientid;
	nfs_client_id4 clnt4;
	char			cl_ident[NFSCLNT_IDMAX]; /* pointed to by clnt4.id.id_val*/
	int clid_state;
	verifier4	setclientid_confirm;         /* holds the verifier when 
						      * clid_state = CLID_NEGOTIATING 
						      */
	struct svc_export *	cl_export[NFSCLNT_EXPMAX]; /* ext2 exported fs's that
							    * have been successfully
							    * entered 
							    */  
	struct in_addr		cl_addr;
	uid_t cl_uid;  /* for now. uidmap is #if 0 in auth.h*/
	struct semaphore cl_sema;
	time_t	cl_lease; /* lease start */
	int cl_lease_len; /* in seconds */
	
	struct rpc_clnt	*cl_callback_clnt;		/* used for callbacks to the client */
};

struct svc_clnthash {
	struct svc_client *h_hash_head;
	int	clid_seq_num;	      /* bumped by successful SETCLIENTID_CONFIRM */ 
};
	
struct svc_export {
	struct svc_export *	ex_next;
	char			ex_path[NFS_MAXPATHLEN+1];
	rpc_gss_svc_t	ex_sec;
	struct svc_export *	ex_parent;
	struct svc_client *	ex_client; 
	int			ex_flags;
	struct dentry *		ex_dentry;
	kdev_t			ex_dev;
	ino_t			ex_ino;
	uid_t			ex_anon_uid;
	gid_t			ex_anon_gid;
};

#define EX_SECURE(exp)		(!((exp)->ex_flags & NFSEXP_INSECURE_PORT))
#define EX_ISSYNC(exp)		(!((exp)->ex_flags & NFSEXP_ASYNC))
#define EX_RDONLY(exp)		((exp)->ex_flags & NFSEXP_READONLY)
#define EX_CROSSMNT(exp)	((exp)->ex_flags & NFSEXP_CROSSMNT)
#define EX_SUNSECURE(exp)	((exp)->ex_flags & NFSEXP_SUNSECURE)
#define EX_WGATHER(exp)		((exp)->ex_flags & NFSEXP_GATHERED_WRITES)


/*
 * Function declarations
 */
void			nfsd_export_init(void);
void			nfsd_export_shutdown(void);
void			exp_readlock(void);
int			exp_writelock(void);
void			exp_unlock(void);
struct svc_client *	exp_getclient(struct sockaddr_in *sin,clientid4 clid);
void			exp_putclient(struct svc_client *clp);
struct svc_export *	exp_get(struct svc_client *clp, kdev_t dev, ino_t ino);
int			exp_rootfh(struct svc_client *, kdev_t, ino_t,
				   char *path, struct knfs_fh *);
int			nfserrno(int errno);
void			exp_nlmdetach(void);
#ifdef NFSV4
u32     exp_pseudorootfh(struct svc_fh *f);
struct svc_export * exp_get_pseudo(kdev_t dev, ino_t ino);
int       pseudo_root_exp_init(struct dentry *dent, char *name);
u32       create_fh(struct dentry *dent, struct svc_fh *f);
int 		nfsd_pseudo_fs_init(void);
struct svc_export * get_new_pseudo_root_exp(void);
struct svc_client *exp_getclientbyclientid(clientid4 clid);
#endif /*NFSV4*/


extern __inline__ int
exp_checkaddr(struct svc_client *clp, struct in_addr addr)
{
	if (clp->cl_addr.s_addr == addr.s_addr)
		return 1;
	return 0;
}

#endif /* __KERNEL__ */

#endif /* NFSD4_EXPORT_H */

