/*
 * linux/net/sunrpc/svcauth_gss.c
 *
 * RPCSEC_GSS server authentication.
 * 
 * Copyright (c) 2000 Dug Song <dugsong@monkey.org>
 * Copyright (c) 2000 The Regents of the University of Michigan
 *
 * $Id: svcauth_gss.c,v 1.6 2000/08/24 19:27:48 dugsong Exp $
 */

#include <linux/types.h>
#include <linux/sched.h>
#include <linux/sunrpc/types.h>
#include <linux/sunrpc/xdr.h>
#include <linux/sunrpc/svcauth.h>
#include <linux/sunrpc/svcsock.h>
#include <linux/sunrpc/auth_gss.h>

#ifdef RPC_DEBUG
# define RPCDBG_FACILITY	RPCDBG_AUTH
#endif

static int
xdr_argsize_check(struct svc_rqst *req, u32 *p)
{
	struct svc_buf *buf = &req->rq_argbuf;

	buf->len = p - buf->base;
	dprintk("RPC: argsize_check p %p - base %p = len %d <= buflen %d\n",
		p, buf->base, buf->len, buf->buflen);
	return (buf->len <= buf->buflen);
}

static int
xdr_ressize_check(struct svc_rqst *req, u32 *p)
{
	struct svc_buf *buf = &req->rq_resbuf;

	buf->len = p - buf->base;
	dprintk("RPC: ressize_check p %p base %p len %d\n",
		p, buf->base, buf->len);
	return (buf->len <= buf->buflen);
}
	
static u32 *
xdr_decode_rpc_gss_cred(struct svc_rqst *req, u32 *p, struct rpc_gss_cred *gc)
{
	gc->gc_v = (u32) ntohl(*p++);
	gc->gc_proc = (u32) ntohl(*p++);
	gc->gc_seq = (u32) ntohl(*p++);
	gc->gc_svc = (u32) ntohl(*p++);
	p = xdr_decode_netobj(p, &gc->gc_ctx);

	dprintk("RPC: xdr_decode_rpc_gss_cred() = "
		"(%d, %d, %d, %d, %p:%d)\n",
		gc->gc_v, gc->gc_proc, gc->gc_seq, gc->gc_svc,
		gc->gc_ctx.data, gc->gc_ctx.len);
	
	if (!xdr_argsize_check(req, p))
		return NULL;
	return p;
}

static u32 *
xdr_decode_gss_init_arg(struct svc_rqst *req, u32 *p,
			struct xdr_netobj *argp)
{
	p = xdr_decode_netobj(p, argp);
	
	dprintk("RPC: xdr_decode_gss_init_arg() = (%p:%d)\n",
		argp->data, argp->len);
	
	if (!xdr_argsize_check(req, p))
		return NULL;
	return p;
}

int
xdr_encode_rpc_gss_init_res(struct svc_rqst *req, u32 *p,
			    struct rpc_gss_init_res *resp)
{
	dprintk("RPC: xdr_encode_rpc_gss_init_res"
		"(%p:%d, %d, %d, %d, %p:%d)\n",
		resp->gr_ctx.data, resp->gr_ctx.len,
		resp->gr_major, resp->gr_minor, resp->gr_win,
		resp->gr_token.data, resp->gr_token.len);

	p = xdr_encode_netobj(p, &resp->gr_ctx);
	*p++ = (u32) htonl(resp->gr_major);
	*p++ = (u32) htonl(resp->gr_minor);
	*p++ = (u32) htonl(resp->gr_win);
	p = xdr_encode_netobj(p, &resp->gr_token);

	return xdr_ressize_check(req, p);
}

void
svcauth_gss(struct svc_rqst *rqstp, u32 *statp, u32 *authp)
{
	struct svc_buf	       *argp = &rqstp->rq_argbuf;
	struct svc_buf	       *resp = &rqstp->rq_resbuf;
	struct svc_cred	       *cred = &rqstp->rq_cred;
	u32		        uid, *bufp = argp->buf;

	struct rpc_gss_cred	gc;
	struct rpc_gss_init_res	gr;
	GSS_BUFFER_T		input_token, client_name;
	u32			ret_flags;

	if (argp->len < 6) {
		*statp = rpc_garbage_args;
		return;
	}
	bufp++;					/* XXX - skip opaque length */
	
	if ((bufp = xdr_decode_rpc_gss_cred(rqstp, bufp, &gc)) == NULL) {
		*statp = rpc_garbage_args;
		return;
	}
	bufp++;					/* XXX - skip verf flavor */
	bufp += XDR_QUADLEN(ntohl(*bufp)) + 1;	/* XXX - skip length + verf */

	argp->len -= (bufp - argp->buf);
	argp->buf = bufp;
	
	if (gc.gc_v != RPC_GSS_VERSION)
		goto rejcred;
	
	/* Switch on RPCSEC_GSS phase. */
	switch (gc.gc_proc) {

	case RPC_GSS_PROC_DATA:
		dprintk("RPC: RPCSEC_GSS_DATA\n");
		break;

	case RPC_GSS_PROC_DESTROY:
		dprintk("RPC: RPCSEC_GSS_DESTROY\n");
		/* XXX - whack context here! */
		break;

	case RPC_GSS_PROC_INIT:
	case RPC_GSS_PROC_CONTINUE_INIT:
		if (rqstp->rq_proc != 0) {	/* NULLPROC */
			dprintk("RPC: *INIT state but not NULLPROC: %d\n",
			       rqstp->rq_proc);
			goto rejcred;		/* XXX - ? */
		}
		if (gc.gc_proc == RPC_GSS_PROC_INIT) {
			dprintk("RPC: RPCSEC_GSS_INIT\n");
			if (gc.gc_ctx.len != 0) {
				dprintk("RPC: non-empty handle on initial context creation request\n");
				goto rejcred;	/* XXX */
			}
		}
		else {
			dprintk("RPC: RPCSEC_GSS_CONTINUE_INIT\n");
			if (gc.gc_ctx.len == 0) {
				dprintk("RPC: expected non-empty handle in continue\n");
				goto rejcred;	/* XXX */
			}
		}
		/* Decode arguments. */
		if (!xdr_decode_gss_init_arg(rqstp, bufp, &input_token)) {
			goto rejcred;	/* XXX */
		}
		if (gssd_accept_sec_context(&gr.gr_major,
					    &gr.gr_minor,
					    &gc.gc_ctx,
					    &input_token,
					    &client_name,
					    &gr.gr_token,
					    &ret_flags) < 0)
			goto rejcred;	/* XXX */
		
		if (gr.gr_major != GSS_S_COMPLETE &&
		    gr.gr_major != GSS_S_CONTINUE_NEEDED) {
			dprintk("RPC: accept_sec_context failed!\n");
			goto rejcred;	/* XXX */
		}
		/* Accepted security context. */
		dprintk("RPC: accepted connection, ctx.len=%d\n",
			gc.gc_ctx.len);
		*authp = rpc_auth_ok;
	
		/* Convert client name, fill in rpc_cred struct with uid */
		if (gssd_name_to_uid(&client_name, &uid) < 0) {
			*statp = rpc_system_err;
			return;
		}
		cred->cr_uid = (uid_t) uid;
		gr.gr_win = 10;
		gr.gr_ctx = gc.gc_ctx;
		
		if (gr.gr_major == GSS_S_COMPLETE) {
			/* Put GSS verifier */
			rqstp->rq_verfed = 1;
			svc_putlong(resp, RPC_AUTH_GSS);
			svc_putlong(resp, 0);
			/* XXX - should be checksum of win w/ QOP 0 */
		}
		else {
			/* Put NULL verifier */
			rqstp->rq_verfed = 1;
			svc_putlong(resp, RPC_AUTH_NULL);
			svc_putlong(resp, 0);
		}
		/*
		 * XXX - Save results to marshal later as NULLPROC results,
		 * faking the decode stage for our dispatcher below.
		 */
		*(struct rpc_gss_init_res *)rqstp->rq_resp = gr;
		return;
		
	default:
		dprintk("RPC: unknown RPCSEC_GSS phase %d\n", gc.gc_proc);
		goto rejcred;
		break;
	}
	/* Put NULL verifier */
	rqstp->rq_verfed = 1;
	svc_putlong(resp, RPC_AUTH_NULL);
	svc_putlong(resp, 0);
	return;
	
 rejcred:
	*authp = rpc_autherr_rejectedcred;
}

int
svcauth_gss_dispatch(struct svc_rqst *rqstp, u32 *statp)
{
	struct svc_procedure   *proc;
	kxdrproc_t		xdr;

	dprintk("RPC: svcauth_gss_dispatch: vers %d proc %d\n",
		rqstp->rq_vers, rqstp->rq_proc);
	proc = rqstp->rq_procinfo;

	/* XXX - NULLPROC decode stage handled above by svcauth_gss(). */
	if (rqstp->rq_proc == 0) {
		if (!xdr_encode_rpc_gss_init_res(rqstp, rqstp->rq_resbuf.buf,
						 (struct rpc_gss_init_res *)
						 rqstp->rq_resp))
			*statp = rpc_system_err;
		return 1;
	}
	/* Decode arguments. */
	xdr = proc->pc_decode;
	if (xdr && !xdr(rqstp, rqstp->rq_argbuf.buf, rqstp->rq_argp)) {
		dprintk("RPC: svcauth_gss_dispatch: can't decode arguments\n");
		*statp = rpc_garbage_args;
		return 1;
	}
	*statp = proc->pc_func(rqstp, rqstp->rq_argp, rqstp->rq_resp);

	/* Encode results. */
	xdr = proc->pc_encode;
	if (*statp == rpc_success &&
	    xdr && !xdr(rqstp, rqstp->rq_resbuf.buf, rqstp->rq_resp)) {
		dprintk("RPC: svcauth_gss_dispatch: failed to encode reply\n");
		*statp = rpc_system_err;
	}
	return 1;
};
